!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : parallel_mod
!> @brief   REIN parallel modules
!! @author  Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    NM: 2009/04/02
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module parallel_mod

  use common_mod
  use constants_mod
  use mpi

  implicit none
  private

  ! structures
  type, public :: s_spawn_kit
    character(icha_2h)              :: command
    integer                         :: inter_communicator
    integer                         :: root
  end type s_spawn_kit

  ! subroutines
  public  :: spawn_initialize
#ifndef NO_SPAWN
  public  :: loop_spawn_for_replica
  private :: spawn_run
#else
  public  :: loop_no_spawn_for_replica
#endif

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  spawn_initialize
  !> @brief      spawn_initialize
  !! @authors    NT, NM
  !! @date       2012/06/27 
  !! @param[in]    world    MPI information
  !! @param[in]    main     main information
  !! @param[inout] comm_md  md spawn information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine spawn_initialize(world, main, comm_md)

    ! formal arguments
    type(s_comm_x),                 intent(in)    :: world
    type(s_main_info),              intent(in)    :: main
    type(s_spawn_kit), allocatable, intent(inout) :: comm_md(:)


    allocate(comm_md(main%total_replicas))

    return

  end subroutine spawn_initialize


#ifndef NO_SPAWN

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  loop_spawn_for_replica
  !> @brief      loop_spawn_for_replica
  !! @authors    NT, NM
  !! @date       2012/06/27 
  !! @param[in]    world    MPI information
  !! @param[in]    main     main information
  !! @param[inout] software md spawn information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine loop_spawn_for_replica(world, main, software)

    ! formal arguments
    type(s_comm_x),    intent(in)    :: world
    type(s_main_info), intent(in)    :: main
    type(s_spawn_kit), intent(inout) :: software(:)

    ! local variables
    integer            :: i, ierr


    !do i = 1, main%total_replicas
    !  write(6,*)"chk====spawn parameter chk"
    !  write(6,*)"chk ==replica= ", i
    !  write(6,*)"chk command=",trim(software(i)%command)
    !  write(6,*)"chk root=",software(i)%root
    !  write(6,*)"chk======================"
    !end do

    do i = 1, main%total_replicas
      call MPI_BARRIER(world%comm, ierr)
      call spawn_run(world, software(i))
    end do

    return

  end subroutine loop_spawn_for_replica


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  spawn_run
  !> @brief      spawn_run
  !! @authors    NT, NM
  !! @date       2012/06/27 
  !! @param[in]    world    MPI information
  !! @param[inout] software md spawn information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine spawn_run(world, software)

    ! formal arguments
    type(s_comm_x),     intent(in)    :: world
    type(s_spawn_kit),  intent(inout) :: software

    ! local variables
    integer             :: ierr


    call mpi_comm_spawn(software%command,            &
                        MPI_ARGV_NULL,               &
                        Gnumber_of_cpus_for_exec,    &
                        MPI_INFO_NULL,               &
                        software%root,               &
                        world%comm,                  &
                        software%inter_communicator, &
                        MPI_ERRCODES_IGNORE,         &
                        ierr)

    if (ierr /= 0) then
      write(io_out,'(A,A,A,A,I4,A,I4,A,I4)') &
                 'Spawn_Run> ERROR:', &
                         '  cmd : ', TRIM(software%command), &
                         '  rank: ', world%rank, &
                         '  ncpu: ', Gnumber_of_cpus_for_exec, &
                         '  icom: ', software%inter_communicator
      stop
    end if

    return

  end subroutine spawn_run

#else

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  loop_no_spawn_for_replica
  !> @brief      loop_no_spawn_for_replica
  !! @authors    NT, NM
  !! @date       2012/10/11 
  !! @param[in]  ini  initial rank
  !! @param[in]  fin  final rank
  !! @param[in]  software md spawn information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine loop_no_spawn_for_replica(ini, fin, software)

    ! parameter
    integer, parameter :: RetryCount = 10

    ! formal arguments
    integer,           intent(in)    :: ini
    integer,           intent(in)    :: fin
    type(s_spawn_kit), intent(in)    :: software(:)

    ! local variables
    integer            :: i, j, system, iret


    if (ini /= fin) &
      write(6,*) 'Loop_No_Spawn_For_Replica> WARNING: multiple MD per REIN.'

    do i = ini, fin

      do j = 1, RetryCount

        iret = system("sh " // trim(software(i)%command))
        if (iret == 0) &
          exit
        write(6,*) '## Rep ', i, trim(software(i)%command), iret, ' Retry ', j
        call sleep(1)

      end do

    end do

    return

  end subroutine loop_no_spawn_for_replica

#endif

end module parallel_mod
