!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : setup_mod
!> @brief   setup condition
!! @author  Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    NM: 2009/10/18
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module setup_mod

  use common_mod
  use constants_mod
  use mpi

  implicit none
  private

  public :: setup

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  setup
  !> @brief      setup condition
  !! @authors    NT, NM
  !! @date       2009/10/18 (NM)
  !! @param[in]    world  MPI information
  !! @param[inout] main   main information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup(world, main)

    ! formal arguments
    type(s_comm_x),    intent(in)    :: world
    type(s_main_info), intent(inout) :: main

    ! local variables
    integer                      :: i, k, ist, ierr, nrep, nrep_real


    main%total_replicas          = 1
    main%num_of_values           = 0
    main%flag_velocity_rescaling = 0
    main%flag_has_harmonic       = 0
    main%exchange_axis_switch    = 1

    do i = 1, Gnumber_of_axis

      main%total_replicas = main%total_replicas * Gnumber_of_replicas(i)
      main%num_of_values  = main%num_of_values + Gaxis_num_of_param(i)

      select case (trim(adjustl(Gaxis_type_name(i))))

      case ('temperature')
        main%flag_velocity_rescaling = Gaxis_control_param(i)

      case ('harmonic')
        main%flag_has_harmonic = 1

      end select

    end do


    ! allocate memory
    !
    nrep = main%total_replicas

    allocate( &
         main%replica_rank        (1:nrep, 1:1),                   &
         main%replica_values      (1:nrep, 1:main%num_of_values),  &
         main%replica_on          (1:nrep, 1:3),                   &
         main%replica_ep          (1:nrep, 1:Gnumber_of_exchange), &
         main%axis_exchange_index (1:Gnumber_of_exchange),         &
         main%replica_tscale      (1:nrep, 1:1),                   &
         main%replica_no          (1:nrep, 1:3),                   &

         main%replica_energy      (1:nrep),                        &
         main%replica_potentiale  (1:nrep),                        &
         main%replica_distance    (1:nrep),                        &
         main%replica_extra_pote  (1:nrep),                        &
         main%replica_temperature (1:nrep),                        &
         main%replica_rep_temp    (1:nrep),                        &
         main%replica_kinetice    (1:nrep),                        &
         main%replica_prerank     (1:nrep, 1:1),                   &
         main%replica_param2rep   (1:nrep, 1:1),                   &
         main%replica_param_p     (1:nrep, 1:Gnumber_of_axis),     &
         main%exchange_switch     (1:Gnumber_of_axis))
      
    main%exchange_switch(:) = 0


    ! read input files
    !

    if (world%rank == 0) then

      write(io_out,*) 'READ INPUT FILES START =='
      
      write(io_out,*) '  file_rank      : ', trim(Ginput_replica_rank)
      open(tmp_io2, &
           file   = trim(adjustl(Ginput_replica_rank)), &
           status = 'old', &
           form   = 'formatted')
      do k = 1,nrep
        read(tmp_io2,*,iostat=ist) main%replica_rank(k,1)
      end do
      close(tmp_io2)


      write(io_out,*) '  file_value     : ',trim(Ginput_replica_values)
      open(tmp_io3, &
           file   = trim(adjustl(Ginput_replica_values)), &
           status = 'old', &
           form   = 'formatted')
      do k = 1, nrep
        read(tmp_io3,*,iostat=ist) &
             (main%replica_values(k,i), i=1, main%num_of_values, 1)
      end do
      close(tmp_io3)


      write(io_out,*) '  file_on        : ',trim(Ginput_replica_on)
      open(tmp_io4, &
           file   = trim(adjustl(Ginput_replica_on)), &
           status = 'old', &
           form='formatted')
      do k = 1, nrep
        read(tmp_io4,*,iostat=ist) &
             (main%replica_on(k,i),i=1,3,1)
      end do
      close(tmp_io4)


      write(io_out,*) '  file_ex list   : ', trim(Ginput_exchange_pattern)
      open(tmp_io5, &
           file   = trim(adjustl(Ginput_exchange_pattern)), &
           status = 'old', &
           form   = 'formatted')
      do k = 1, nrep
        read(tmp_io5,*,iostat=ist) &
             (main%replica_ep(k,i),i=1,Gnumber_of_exchange,1)
      end do
      close(tmp_io5)


      write(io_out,*) '  file_ex pattern: ', trim(Ginput_axis_exchange)
      open(tmp_io6, &
           file   = trim(adjustl(Ginput_axis_exchange)), &
           status = 'old', &
           form   = 'formatted')
      read(tmp_io6,*,iostat=ist) & 
           (main%axis_exchange_index(i), i=1, Gnumber_of_exchange, 1)
      close(tmp_io6)


      write(io_out,*) '  file_coef      : ', trim(Ginput_restart_tscale)
      open(tmp_io7, &
           file   = trim(adjustl(Ginput_restart_tscale)), &
           status = 'old', &
           form   ='formatted')
      do k = 1,nrep
        read(tmp_io7,*,iostat=ist) &
             main%replica_tscale(k,1)
      end do
      close(tmp_io7)

      write(io_out,*) ' '

    end if


    ! distribute to other rank
    !
    if (world%nprocs > 1) then

      if (world%rank == 0) &
        write(io_out,*) '  DISTRIBUTE INITIAL DATA on PREF START'

      call mpi_bcast(main%replica_rank(1,1), nrep*1, &
           MPI_INTEGER, 0, world%comm, ierr)

      call mpi_bcast(main%replica_values(1,1), nrep * main%num_of_values, &
           MPI_DOUBLE_PRECISION, 0, world%comm, ierr)

      call mpi_bcast(main%replica_on(1,1), nrep*3, &
           MPI_INTEGER, 0, world%comm, ierr)

      call mpi_bcast(main%replica_ep(1,1), nrep * Gnumber_of_exchange, &
           MPI_INTEGER, 0, world%comm, ierr)

      call mpi_bcast(main%axis_exchange_index(1), Gnumber_of_exchange, &
           MPI_INTEGER, 0, world%comm, ierr)

      call mpi_bcast(main%replica_tscale(1,1), nrep*1, &
           MPI_DOUBLE_PRECISION, 0, world%comm, ierr)

      if (world%rank == 0) then
        write(io_out,*) '  DISTRIBUTE INITIAL DATA on PREF END'
        write(io_out,*) ' '
      end if

    end if


    ! check actual number of replicas
    !
    nrep_real = 1
    do i = 1, nrep
      if (main%replica_on(i,1) == 1) then
        main%replica_no(nrep_real,1) = main%replica_on(i,1) 
        main%replica_no(nrep_real,2) = main%replica_on(i,2) 
        main%replica_no(nrep_real,3) = main%replica_on(i,3) 
        nrep_real = nrep_real + 1
      end if
      main%replica_param2rep(main%replica_rank(i,1),1) = i
    end do


    ! check multi-dimension
    !

    if (world%rank == 0) then

      write(io_out,'(" =DISPLAY======MULTI_DIMENSIONAL REPLICA CONDITIONS==")')

      do i = 1, Gnumber_of_axis

        write(io_out, '(" AXIS: "I4)') i

        write(io_out, '("   AXIS_INDEX:        ("I4")="I4)') &
             i, Gaxis_index(i)

        write(io_out, '("   AXIS_TYPE_NAME:    ("I4")="A14)') &
             i, trim(adjustl(Gaxis_type_name(i)))

        write(io_out, '("   NUMBER_OF_REPLICAS:("I4")="I4)') &
             i, Gnumber_of_replicas(i)

        write(io_out, '("   AXIS_CONTROL_PARAM:("I4")="I4)') &
             i, Gaxis_control_param(i)

        write(io_out, '("   AXIS_NUM_OF_PARAM: ("I4")="I4)') &
             i, Gaxis_num_of_param(i)

        do k = 1, Gaxis_num_of_param(i)
          write(io_out, '("   AXIS_PARAM:   ("I4","I4")="I4)') &
               k, i, Gaxis_index_coef(k,i)
        end do
      end do

      write(io_out, '("   EXCHANGE_INDEX="10I4)') &
           (main%axis_exchange_index(k), k=1, Gnumber_of_exchange, 1)

      write(io_out, '(" =END========MULTI_DIMENSIONAL REPLICA CONDITIONS==")')
      write(io_out,*) ' '

      write(io_out,*) 'READ INPUT FILES END =='
      write(io_out,*) ' '

    end if

    return

  end subroutine setup

end module setup_mod
