!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : marble_crd_rescaling_mod
!> @brief   REIN velocity rescaling for MARBLE crd file
!! @author  Naoyuki Miyashita (NM)
!! @date    NM: 2010/01/05
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module marble_crd_rescaling_mod

  use fileio_mod
  use constants_mod
  use common_mod

  implicit none
  private

  ! subroutines
  public  :: marble_crd_rescaling
  private :: tolower

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  marble_crd_rescaling
  !> @brief      rescale marble crd file
  !! @authors    NT, NM
  !! @date       2012/07/12 
  !! @param[in]  input   input file name
  !! @param[in]  output  output file name
  !! @param[in]  nline   number of lines
  !! @param[in]  coef    scaling coefficient
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine marble_crd_rescaling(input, output, nline, coef)

    ! parameters
    integer, parameter  :: dp = selected_real_kind(22)

    ! formal arguments
    character(icha_2h), intent(in)  :: input
    character(icha_2h), intent(in)  :: output
    integer,            intent(in)  :: nline
    double precision,   intent(in)  :: coef

    ! local variables
    double precision    :: dval(6)
    integer             :: io
    integer             :: i, ii, j, natom, nline_atom, nrigid, ntail
    integer             :: vel_start, vel_end

    real(kind=dp)       :: target_temperature
    integer             :: EX_System_T_flag
    integer             :: EX_System_P_flag
    integer             :: n_ex_system
    integer             :: n_chain_T

    character(icha_2h), allocatable :: lines(:)
    real(kind=dp),      allocatable :: eta(:,:),eta_v(:,:),Q(:,:)
    integer,            allocatable :: atom_ex_system(:,:)


    call get_io(io)

    ! read lines
    !

    allocate(lines(1:nline))

    open(io, &
         file   = adjustl(input), &
         status = 'old')
         
    do i = 1, nline
      read(io,'(a)') lines(i)
    end do

    close(io)


    ! rescaling velocities
    !

    read(lines(2),*) natom
    
    nline_atom = int(natom / 2) + mod(natom, 2)
    vel_start  = 2 + nline_atom + 1
    vel_end    = vel_start + nline_atom - 1

    do i = vel_start, vel_end - 1

      read(lines(i),*) (dval(j),j=1,6)
      dval = coef * dval
      write(lines(i),'(6F12.7)') (dval(j),j=1,6)

    end do

    if (mod(natom, 2) /= 0) then
      ntail = 3
    else
      ntail = 6
    end if

    read(lines(vel_end),*) (dval(j),j=1,ntail)
    dval = coef * dval
    write(lines(vel_end),'(6F12.7)') (dval(j),j=1,ntail)


    ! rescaling TPC
    !

    vel_start = 2 + nline_atom * 2 + 1

    if (len_trim(adjustl(lines(vel_start))) == 1) &
      vel_start = vel_start + 4

    if (vel_start > nline) then
      write(io_out,*) 'Marble_Crd_Rescaling> end file with velocity.'
      deallocate(lines)
      return
    end if

    i = vel_start
    read(lines(i),*) EX_System_T_flag, EX_System_P_flag

    ! rescaling temperature
    i = vel_start + 1
    read(lines(i),*) n_ex_system, n_chain_T, target_temperature

    target_temperature = coef * target_temperature

    write(lines(i), '(I0,X,I0,X,E21.15)') &
                     n_ex_system, n_chain_T, target_temperature
    call tolower(lines(i))

    ! rescaling Q(:,:)
    allocate(eta  (1:n_chain_T, 1:n_ex_system), &
             eta_v(1:n_chain_T, 1:n_ex_system), &
             Q    (1:n_chain_T, 1:n_ex_system), &
             atom_ex_system(1:3, 1:n_ex_system))

    do ii = 1, n_ex_system

      i = vel_start + 1 + ii + (ii - 1) * n_chain_T
      read(lines(i),*) (atom_ex_system(j,ii),j=1,2)

      do j = 1, n_chain_T

        i = vel_start + 1 + ii + (ii - 1) * n_chain_T + j
        read(lines(i),*) eta(j,ii), eta_v(j,ii), Q(j,ii)

        Q(j,ii) = coef * coef * Q(j,ii)
        write(lines(i),'(E21.15,X,E21.15,X,E21.15)') &
             eta(j,ii), eta_v(j,ii), Q(j,ii)

        call tolower(lines(i))

      end do

    end do

    deallocate(eta, eta_v, Q, atom_ex_system)

    select case(EX_System_P_flag)
    case (0)
      vel_end = i
    case (1)
      vel_end = i + 1
    case (2)
      vel_end = i + 4
    case default
      write(io_out,*) 'Marble_Crd_Rescaling> ERROR: Unknown EX_System_P_flag'
      stop
    end select


    ! rescaling rigid body
    !

    vel_start = vel_end + 1
    if (vel_start > nline) then
      write(io_out,*) 'Marble_Crd_Rescaling> end file with TPC.'
      deallocate(lines)
      return
    end if

    i = vel_start
    read(lines(i),*) nrigid

    do ii = 1, nrigid

      i = vel_start + 2 * ii
      read(lines(i),*) (dval(j),j=1,6)
      dval = coef * dval
      write(lines(i),'(6(E13.6,X))') (dval(j),j=1,6)
      call tolower(lines(i))

    end do


    ! write lines
    !

    open(io, &
         file   = adjustl(output), &
         status = 'replace')

    do i = 1, nline
      write(io,'(a)') trim(lines(i))
    end do

    close(io)

    deallocate(lines)

    call free_io(io)

    return

  end subroutine marble_crd_rescaling


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  tolower
  !> @brief      capital to small letter
  !! @authors    NT, NM
  !! @date       2012/07/12 
  !! @param[inout]  line  string line
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine tolower(line)

    ! formal arguments
    character(*), intent(inout) :: line

    ! local variables
    integer :: k


    do k = 1, len_trim(line)
      if (line(k:k) >= "A" .and. line(k:k) <= "Z" ) &
        line(k:k) = char(ichar(line(k:k)) + 32)
    end do

    return
    
  end subroutine tolower
   
end module marble_crd_rescaling_mod
