!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : marble_module_mod
!> @brief   REIN MARBLE modules
!! @author  Naoyuki Miyashita (NM)
!! @date    NM: 2010/07/09
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module marble_module_mod

  use marble_crd_rescaling_mod
  use fileio_mod
  use parallel_mod
  use common_mod
  use constants_mod
  use mpi

  implicit none
  private
   
  ! subroutines
  public  :: marble_make_input
  public  :: marble_rescaling
  public  :: marble_watchdog
  public  :: marble_get_output
  private :: get_io_name

contains
   
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  marble_make_input
  !> @brief      make MARBLE input files
  !! @authors    NT, NM
  !! @date       2012/07/02 
  !! @param[in]    world  MPI information
  !! @param[in]    main   main information
  !! @param[inout] marble marble local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine marble_make_input(world, main, marble)

    ! formal arguments
    type(s_comm_x),      intent(in)    :: world
    type(s_main_info),   intent(in)    :: main
    type(s_local_info),  intent(inout) :: marble


    call get_io_name(marble)

    if (main%i_steps == Gfirst_step) then

      call check_file_line(marble%base_file, &
                           marble%local_nline_inputfile)
      call check_file_line(marble%local_input_crd, &
                           marble%local_nline_crdfile)

    end if

    call create_local_input(                   &
                 marble%base_file,             &
                 marble%local_input_path,      &
                 marble%local_nline_inputfile, &
                 marble)

    return
    
  end subroutine marble_make_input
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  marble_rescaling
  !> @brief      rescaling MARBLE velocity/coordinate files
  !! @authors    NT, NM
  !! @date       2009/12/08 (NM)
  !! @param[in]    world    MPI information
  !! @param[in]    main     main information
  !! @param[inout] marble   marble local information
  !! @param[in]    ireplica replica number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine marble_rescaling(world, main, marble, ireplica)

#ifdef FUJI
    use service_routines, only: rename
#endif
#ifdef INTEL
    use iflport
    use iflposix
    use ifport
#endif

    ! formal arguments
    type(s_comm_x),      intent(in)    :: world
    type(s_main_info),   intent(in)    :: main
    type(s_local_info),  intent(inout) :: marble
    integer,             intent(in)    :: ireplica

    ! local variables
    double precision     :: Tpre, T, coef
    integer              :: it, istat
    character(icha_2h)   :: file_target, file_org
    

    it = main%flag_velocity_rescaling

    if (it == 0 .or. adjustl(Gvelocity_rescaling) .ne. "v_marble") &
      return
      
    if (main%i_steps == main%initial_steps) then
      coef = main%replica_tscale(ireplica, 1)
      T    = main%replica_values(main%replica_rank(ireplica, 1), it)
      Tpre = T/coef
    else 
      Tpre = main%replica_values(main%replica_prerank(ireplica, 1), it)
      T    = main%replica_values(main%replica_rank(ireplica, 1), it)
      coef = dsqrt(T / Tpre)
    end if

    if (dabs(coef-1.d0) >= err16) then 

      ! rename file
      !

      ! check file i/o complete
      call compare_file_line(marble%local_input_crd, &
                             marble%local_nline_crdfile)

      file_target = marble%local_input_crd
      file_org    = trim(file_target) // ".org"

      istat = rename(file_target, file_org)
      write(io_out,*) "Marble_Rescaling> " // &
                      " move : ", trim(file_target), &
                      " to ",     trim(file_org),    &
                      "  status:",  istat

      ! rescaling file
      !

      ! check file i/o complete
      call compare_file_line(file_org, marble%local_nline_crdfile)

      call marble_crd_rescaling(file_org, file_target, &
                                marble%local_nline_crdfile, coef)

      ! check file i/o complete (final)
      call compare_file_line(file_target, marble%local_nline_crdfile)

    end if

    return

  end subroutine marble_rescaling


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  marble_watchdog
  !> @brief      watchdog MARBLE process
  !! @authors    NT, NM
  !! @date       2009/12/08 (NM)
  !! @param[in]    world    MPI information
  !! @param[in]    main       main information
  !! @param[inout] marble   marble local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine marble_watchdog(world, main, marble)

    ! parameter
    integer, parameter   :: WatchInterval = 1    ! 1 second
    integer, parameter   :: MaxWatch      = 7200 ! 2 hour

    ! formal arguments
    type(s_comm_x),      intent(in)    :: world
    type(s_main_info),   intent(in)    :: main
    type(s_local_info),  intent(inout) :: marble

    ! local variables
    integer              :: i, found, nline
    character(icha_2h)   :: fileout, fileprop


    fileout = adjustl(marble%local_output_path)
    fileprop = adjustl(marble%local_output_prop)

    do i = 1, MaxWatch

      call check_file_tail(fileout, '     Elapsed Time:', 100, found)
      if (found == 0) &
      call check_file_tail(fileout, '     System Time:',  100, found)
      call check_file_line(fileprop, nline)

      if (found /= 0 .and. nline >= 2) &
        exit

      call sleep(WatchInterval)

    end do

    if (i >= MaxWatch) then
      write(io_out,'(a)') &
           'Marble_WatchDog> Over maximum watch. MARBLE is hunging-up?'
    end if

    return

  end subroutine marble_watchdog


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  marble_get_output
  !> @brief      get output for MARBLE
  !! @authors    NT, NM
  !! @date       2012/07/09 
  !! @param[in]    world    MPI information
  !! @param[inout] main     main information
  !! @param[in]    marble   marble local information
  !! @param[in]    ireplica replica number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine marble_get_output(world, main, marble, ireplica)

    ! formal arguments
    type(s_comm_x),      intent(in)    :: world
    type(s_main_info),   intent(inout) :: main
    type(s_local_info),  intent(in)    :: marble
    integer,             intent(in)    :: ireplica

    ! local variables
    double precision     :: dval, dvalv(20)
    integer              :: io_get, ierr, ival, nrank, i, l, dist_no
    character(icha_3h)   :: line
    character(icha_1h)   :: sval, phead(20)

    double precision, allocatable :: pval(:)


    ! output file
    !

    call get_io(io_get)

    open(io_get, &
         file   = adjustl(marble%local_output_path), &
         status = 'old',       &
         form   = 'formatted', &
         iostat = ierr)

    do while (.true.)

      read(io_get,'(a300)',iostat=ierr) line
      if (ierr /= 0) &
        exit

      if (line(1:4) == 'STEP') then

        nrank = marble%nrank

        read(line,*,err=20) &
             sval, ival, sval, dval, sval, sval, &
             main%replica_temperature(nrank)

        line(1:1) = ' '
        do while(line(1:1) /= '-')
          
          read(io_get,'(a300)',iostat=ierr) line
          if (ierr /= 0) &
            exit

          ! read TOTAL_E line ...
          if (line(1:7) == 'TOTAL_E') then

            l = LEN_TRIM(line) ! case of: "POTENTIAL-857617.96" -> no blank!
            do i = 1, l
              if (line(i:i) >= 'A' .and. line(i:i) <= 'Z' .or. &
                  line(i:i) == '_') &
                line(i:i) = ' '
            end do
            read(line,*,err=20)                  &
                 main%replica_energy(nrank),     &
                 main%replica_potentiale(nrank), &
                 main%replica_kinetice(nrank)

          ! read VDW line ...
          else if (line(1:4) == 'VDW ') then

            read(line,*,end=20) &
                 sval, dval, sval, dval, sval, main%replica_extra_pote(nrank)

          end if

        end do

      end if

    end do

    close(io_get)

    call free_io(io_get)


    ! prop file
    !
    call get_io(io_get)

    open(io_get, &
         file   = adjustl(marble%local_output_prop), &
         status = 'old',       &
         form   = 'formatted', &
         iostat = ierr)

    allocate(pval(marble%num_axis_type*2))

    ! header
    read(io_get,'(a300)',err=21) line

    phead(:) = ''
    dist_no  = 0
    read(line,*,end=10,err=21) (phead(i), i = 1,20)

10  do i = 1,20
      if (phead(i) == '') &
        exit
      if (phead(i)(2:9) == 'DISTANCE') then
        dist_no = i
        exit
      end if
    end do

    ! values
    if (dist_no /= 0) then

      do while (.true.)

        read(io_get,'(a300)',iostat=ierr) line
        if (ierr /= 0) &
             exit

        read(line,*) &
             (dvalv(i), i = 1, dist_no-1), &
             (pval (i), i = 1, marble%num_axis_type*2)

      end do

      do i = 1, marble%num_axis_type
        main%replica_param_p(marble%nrank, i) = pval(i*2-1)
      end do

    end if

    deallocate(pval)

    close(io_get)

    call free_io(io_get)

    return

20  write(io_out,'(a)') 'Marble_Get_Output> MARBLE output read error.'
    stop
21  write(io_out,'(a)') 'Marble_Get_Output> MARBLE prop file read error.'
    stop

  end subroutine marble_get_output


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  get_io_name
  !> @brief      get i/o name
  !! @authors    NT, NM
  !! @date       2012/07/09 
  !! @param[inout] marble   marble local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_io_name(marble)

    ! formal arguments
    type(s_local_info),   target, intent(inout) :: marble

    ! local variables
    character(icha_2h)    :: blank_cha


    ! file
    blank_cha = " "
    call make_filename(marble%local_file,           &
                       Gio_file_name,               &
                       blank_cha,                   &
                       marble%name_pref_rank,       &
                       marble%name_i_steps)

    ! prefile
    blank_cha = " "
    call make_filename(marble%local_pre_input_file, &
                       Gio_file_name,               &
                       blank_cha,                   &
                       marble%name_pref_rank,       &
                       marble%name_ipre_steps)

    ! file.inp
    blank_cha = ".inp"
    call make_filename(marble%local_input_file,     &
                       Gio_file_name,               &
                       blank_cha,                   &
                       marble%name_pref_rank,       &
                       marble%name_i_steps)

    ! file.out
    blank_cha = ".out"
    call make_filename(marble%local_output_file,    &
                       Gio_file_name,               &
                       blank_cha,                   &
                       marble%name_pref_rank,       &
                       marble%name_i_steps)

    ! file.sh
    blank_cha = ".sh"
    call make_filename(marble%local_command_file,   &
                       Gio_file_name,               &
                       blank_cha,                   &
                       marble%name_pref_rank,       &
                       marble%name_i_steps)


    ! dir/file
    call add_directory(marble%local_path,           &
                       marble%name_pref_rank,       &
                       Gio_file_name,               &
                       marble%local_file)

    ! dir/prefile
    call add_directory(marble%local_pre_input_path, &
                       marble%name_pref_rank,       &
                       Gio_file_name,               &
                       marble%local_pre_input_file)

    ! dir/file.inp
    call add_directory(marble%local_input_path,     &
                       marble%name_pref_rank,       &
                       Gio_file_name,               &
                       marble%local_input_file)

    ! dir/file.out
    call add_directory(marble%local_output_path,    &
                       marble%name_pref_rank,       &
                       Gio_file_name,               &
                       marble%local_output_file)

    marble%base_file = &
                      trim(adjustl(Ginitial_file_name)) // ".inp"

    marble%local_dir = &
                      trim(adjustl(Gio_file_name)) // &
                      trim(adjustl(marble%name_pref_rank))

    marble%local_command_path = &
                      trim(adjustl(marble%local_dir)) // "/" // &
                      trim(adjustl(marble%local_command_file))


    marble%local_command = &
                      trim(adjustl(Gparallel_module_md))     // " " // &
                      trim(adjustl(Gparallel_exec))          // " " // &
                      trim(adjustl(Goption_md))              // " " // &
                      trim(adjustl(marble%local_input_path)) // " " // &
                      trim(adjustl(Gextend_option_md))       // " > " // &
                      trim(adjustl(marble%local_output_path))

    marble%local_input_crd  = &
                      trim(adjustl(marble%local_dir)) // '/' //&
                      trim(adjustl(marble%local_pre_input_file)) // ".crd"
    marble%local_output_crd = &
                      trim(adjustl(marble%local_dir)) // '/' //&
                      trim(adjustl(marble%local_file)) // ".crd"
    marble%local_output_trj = &
                      trim(adjustl(marble%local_dir)) // '/' //&
                      trim(adjustl(marble%local_file)) // ".trj"
    marble%local_output_prop = &
                      trim(adjustl(marble%local_dir)) // '/' //&
                      trim(adjustl(marble%local_file)) // ".prop"

    !write(6,*) '## base_file             :',TRIM(marble%base_file)
    !write(6,*) '##'
    !write(6,*) '## local_dir             :',TRIM(marble%local_dir)
    !write(6,*) '## local_file            :',TRIM(marble%local_file)
    !write(6,*) '## local_path            :',TRIM(marble%local_path)
    !write(6,*) '## local_input_file      :',TRIM(marble%local_input_file)
    !write(6,*) '## local_input_path      :',TRIM(marble%local_input_path)
    !write(6,*) '## local_output_file     :',TRIM(marble%local_output_file)
    !write(6,*) '## local_output_path     :',TRIM(marble%local_output_path)
    !write(6,*) '## local_pre_input_file  :',TRIM(marble%local_pre_input_file)
    !write(6,*) '## local_pre_input_path  :',TRIM(marble%local_pre_input_path)
    !write(6,*) '## local_command_file    :',TRIM(marble%local_command_file)
    !write(6,*) '## local_command_path    :',TRIM(marble%local_command_path)
    !write(6,*) '##'
    !write(6,*) '## local_command         :',TRIM(marble%local_command)
    !write(6,*) '##'
    !write(6,*) '## local_input_crd       :',TRIM(marble%local_input_crd)
    !write(6,*) '## local_output_crd      :',TRIM(marble%local_output_crd)
    !write(6,*) '## local_output_trj      :',TRIM(marble%local_output_trj)
    !write(6,*) '## local_output_prop     :',TRIM(marble%local_output_prop)
    !write(6,*) '##'
    !write(6,*) '## name_pref_rank        :',TRIM(marble%name_pref_rank)
    !write(6,*) '## name_i_steps          :',TRIM(marble%name_i_steps)
    !write(6,*) '## name_ipre_steps       :',TRIM(marble%name_ipre_steps)

    return

  end subroutine get_io_name

end module marble_module_mod
