!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : namd_rst_rescaling_mod
!> @brief   REIN velocity rescaling for NAMD binary restart file
!! @author  Naoyuki Miyashita (NM)
!! @date    NM: 2010/01/05
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module namd_rst_rescaling_mod

  use fileio_mod
  use constants_mod
  use common_mod

  implicit none
  private

  ! subroutines
  public  :: namd_rst_rescaling
  private :: namd_rstvel_rescaling
  private :: namd_rstpdb_rescaling

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  namd_rst_rescaling
  !> @brief      rescale namd restart file
  !! @authors    NT, NM
  !! @date       2012/07/04 
  !! @param[in] coef   rescaling coefficient
  !! @param[in] input  input file name
  !! @param[in] output output file name
  !! @param[in] flag   flag for file type
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine namd_rst_rescaling(coef, input, output, flag)

    ! formal arguments
    double precision,   intent(in) :: coef
    character(icha_2h), intent(in) :: input
    character(icha_2h), intent(in) :: output
    character(icha_2h), intent(in) :: flag


    select case(trim(adjustl(flag)))

    case("bin") ! n, xyz
      call namd_rstvel_rescaling(coef, input, output)

    case("ascii") ! pdb
      call namd_rstpdb_rescaling(coef, input, output)
    end select

    return

  end subroutine namd_rst_rescaling


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  namd_rstvel_rescaling
  !> @brief      rescale namd velocity restart file
  !! @authors    NT, NM
  !! @date       2012/07/26 
  !! @param[in]  coef    rescaling coefficient
  !! @param[in]  inpfile input file name
  !! @param[in]  outfile output file name
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine namd_rstvel_rescaling(coef, inpfile, outfile)

    ! formal arguments
    double precision,   intent(in)  :: coef
    character(icha_2h), intent(in)  :: inpfile
    character(icha_2h), intent(in)  :: outfile

    ! local variables
    integer             :: io, natom
    double precision,   allocatable :: values(:,:), values2(:,:)


    call get_io(io)

#ifdef GFORT
    open(io, &
         file   = adjustl(inpfile), &
         status = 'old',    &
         form   = 'unformatted', &
         access = 'stream')
#else
    open(io, &
         file   = adjustl(inpfile), &
         status = 'old',    &
         form   = 'binary', &
         access = 'sequential')
#endif
    
    ! read number of atoms, values
    read(io) natom

    allocate(values (1:3,1:natom), &
             values2(1:3,1:natom))

    rewind(io)
    read(io) natom, values

    close(io)
    
    ! rescale values
    values2 = coef * values
    
    ! write values
#ifdef GFORT
    open(io, &
         file   = adjustl(outfile), &
         status = 'new',    &
         form   = 'unformatted', &
         access = 'stream')
#else
    open(io, &
         file   = adjustl(outfile), &
         status = 'new',    &
         form   = 'binary', &
         access = 'sequential')
#endif

    write(io) natom, values2

    deallocate(values, values2)

    close(io)
    call free_io(io)

    return

  end subroutine namd_rstvel_rescaling


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  namd_rstpdb_rescaling
  !> @brief      rescale namd pdb restart file
  !! @authors    NT, NM
  !! @date       2012/07/04 
  !! @note       for PDB format (NAMD)
  !! @param[in]  coef    rescaling coefficient
  !! @param[in]  inpfile input file name
  !! @param[in]  outfile output file name
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine namd_rstpdb_rescaling(coef, inpfile, outfile)

    ! formal arguments
    double precision,   intent(in) :: coef
    character(icha_2h), intent(in) :: inpfile
    character(icha_2h), intent(in) :: outfile

    ! local variables
    double precision    :: x, y, z, occupancy, temp_factor
    integer             :: io1, io2
    integer             :: ist,ierr
    integer             :: serial, res_seq
    character(80)       :: linex
    character(6)        :: rname
    character(4)        :: atom_name
    character(1)        :: altloc, chainid
    character(4)        :: res_name
    character(2)        :: element, charge
    character(10)       :: blanc
    character(3)        :: term


    call get_io(io1)
    call get_io(io2)

    open(io1, &
         file   = trim(adjustl(inpfile)), &
         status = "OLD",       &
         form   = "formatted", &
         iostat = ierr)
    open(io2, &
         file   = trim(adjustl(outfile)), &
         status = "NEW", &
         form   = "formatted", &
         iostat = ierr)

    aci: do 
      read(io1,'(a80)',iostat=ist) linex
      if (ist /= 0) &
        exit

      rname = linex(1:6)
      term  = linex(1:3)

      select case(rname)
      case("ATOM  ")
        read(linex(7:11),*) serial
        atom_name = linex(13:16)
        altloc    = linex(17:17)
        res_name  = linex(18:21)
        chainid   = linex(22:22)
        read(linex(23:27),*) res_seq
        read(linex(31:38),*) x; x = coef * x
        read(linex(39:46),*) y; y = coef * y
        read(linex(47:54),*) z; z = coef * z
        read(linex(55:60),*) occupancy
        read(linex(61:66),*) temp_factor
        blanc   = linex(67:76)
        element = linex(77:78)
        charge  = linex(79:80)

      case ("HETATM")
        read(linex(7:11),*) serial
        atom_name = linex(13:16)
        altloc    = linex(17:17)
        res_name  = linex(18:21)
        chainid   = linex(22:22)
        read(linex(23:27),*) res_seq
        read(linex(31:38),*) x; x = coef * x
        read(linex(39:46),*) y; y = coef * y
        read(linex(47:54),*) z; z = coef * z
        read(linex(55:60),*) occupancy
        read(linex(61:66),*) temp_factor
        blanc     = linex(67:76)
        element   = linex(77:78)
        charge    = linex(79:80)

      case default

      end select

      select case(term)
      case ("TER")
        write(io2,'(a3)') "TER"

      case ("END")
        write(io2,'(a3)') "END"
        exit

      case ("ATO","HET")

        if (res_seq >= 10000) then
          write(io2, &
               '(a6,i5,x,a4,a1,a4,a1,i5,3x,3f8.3,2f6.2,a10,a2,a2)')          &
               rname, serial, atom_name, altloc, res_name, chainid, res_seq, &
               x, y, z, occupancy, temp_factor, blanc, element, charge
        else
          write(io2, &
               '(a6,i5,x,a4,a1,a4,a1,i4,x,3x,3f8.3,2f6.2,a10,a2,a2)')        &
               rname, serial, atom_name, altloc, res_name, chainid, res_seq, &
               x, y, z, occupancy, temp_factor, blanc, element, charge
        end if

      case default
        write(io2,*)trim(adjustl(linex))

      end select

      if (trim(adjustl(rname)) .eq. "END") &
        exit

      if (ist /= 0) &
        exit

    end do aci

    write(6,*) 'chk velfile out=3=', term, rname, trim(adjustl(outfile)), ist

    close(io2)
    close(io1)

    call free_io(io2)
    call free_io(io1)

    return

  end subroutine namd_rstpdb_rescaling

end module namd_rst_rescaling_mod
