!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : driver_mod
!> @brief   REIN driver modules
!! @author  Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    NM: 2012/03/02
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module driver_mod

  use marble_module_mod
  use namd_module_mod
  use exchange_mod
  use fileio_mod
  use parallel_mod
  use common_mod
  use constants_mod
  use mpi
  !$ use omp_lib

  implicit none
  private

  ! subroutines
  public  :: driver
  private :: init_driver
  private :: md_make_input
  private :: md_rescaling
  private :: md_watchdog
  private :: md_get_output
  private :: dist_energy
  private :: rank_restart_file
  private :: tscale_restart_file
  private :: trajectories_out
  private :: update_conds

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  driver
  !> @brief      driver
  !! @authors    NT, NM
  !! @date       2012/03/13 
  !! @param[in]    world  MPI information
  !! @param[inout] main   main information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine driver(world, main)

    ! formal arguments
    type(s_comm_x),      target,     intent(in)    :: world
    type(s_main_info),   target,     intent(inout) :: main
    
    ! local variables
    double precision                 :: ratex, cl
    integer                          :: i, j, ierr
    integer                          :: ini, fin, bin
    integer                          :: pp0, rep_temp
    character(icha_2h)               :: flag2

    type(s_spawn_kit),   allocatable :: comm_md(:)
    type(s_local_info),  allocatable :: local(:)
    character(icha_2h),  allocatable :: command_tmp(:)
    character(icha_2h),  allocatable :: command_all(:)
    integer,             allocatable :: list_bin(:)
    integer,             allocatable :: list_ini(:)

    integer,             pointer     :: initial_steps, last_steps, i_steps


    ! timer
    double precision     :: Time_setup
    double precision     :: Time_rescaling
#ifndef NO_SPAWN
    double precision     :: Time_spawn
    double precision     :: Time_watchdog
#endif
    double precision     :: Time_MD
    double precision     :: Time_edist
    double precision     :: Time_exchange
    double precision     :: Time_output
    double precision     :: Time_GET
    double precision     :: Time_setup_ini,     Time_setup_fin
    double precision     :: Time_rescaling_ini, Time_rescaling_fin
#ifndef NO_SPAWN
    double precision     :: Time_spawn_ini,     Time_spawn_fin
    double precision     :: Time_watchdog_ini,  Time_watchdog_fin
#endif
    double precision     :: Time_MD_ini,        Time_MD_fin
    double precision     :: Time_GET_ini,       Time_GET_fin
    double precision     :: Time_edist_ini,     Time_edist_fin
    double precision     :: Time_exchange_ini,  Time_exchange_fin
    double precision     :: Time_output_ini,    Time_output_fin
    double precision     :: Time_Loop,          Time_Loop_woio


    initial_steps => main%initial_steps
    last_steps    => main%last_steps
    i_steps       => main%i_steps


    ! INITIAL SETTINGS =================================
    !
    call init_driver(world, main, local, comm_md)

    ini = main%master_nrank_array(world%rank + 1, 1)
    fin = main%master_nrank_array(world%rank + 1, 2)
    bin = main%master_nrank_array(world%rank + 1, 3)

    allocate(command_tmp(1:bin),                 &
             command_all(1:main%total_replicas), &
             list_bin(1:world%nprocs),           &
             list_ini(1:world%nprocs))

    list_bin(1) = main%master_nrank_array(1,3)*icha_2h
    list_ini(1) = 0

    do i = 2, world%nprocs
      list_bin(i) = main%master_nrank_array(i,3)*icha_2h
      list_ini(i) = list_ini(i-1) + list_bin(i-1)
    end do

    !$omp parallel private(i,ratex,cl)
    !$omp do
    do i = ini, fin, 1
      ratex = dble(i)/dble(main%total_replicas)
      cl = main%clockt * ratex + main%clockt2 * (1.d0 - ratex)
      call random_numb(cl)
    end do
    !$omp end do
    !$omp end parallel

    call MPI_BARRIER(world%comm, ierr)

    main%replica_energy     = 0.d0
    main%replica_kinetice   = 0.d0
    main%replica_potentiale = 0.d0


    !==== REIN LOOP ====================================
    !
    if (world%rank == 0) then
      write(io_out,*) 'REIN LOOP START =='
      write(io_out,*) ''
    end if

    do i_steps = initial_steps, last_steps

      ! DISPLAY HEADER ====
      !
      if (i_steps - initial_steps + 1 <= Gequiv_steps) then

        flag2 = "equil"
        main%i_equiv = main%i_equiv + 1

        if (world%rank == 0) then
          write(io_out,*) &
               '  ======================================Umbrella sampling'
          write(io_out,'(A,I7,A,I7)') &
               '   REIN_EQUI_RUN:', main%i_equiv, ' / TOTAL: ', i_steps
          write(io_out,*) ' '
        end if

      else

        flag2 = "prod"
        main%i_product = main%i_product + 1

        if (world%rank == 0) then
          write(io_out,*) &
               '  ======================================Replica-exchange'
          write(io_out,'(A,I7,A,I7)') &
               '   REIN_PRODUCT_RUN:', main%i_product, ' / TOTAL: ', i_steps
          write(io_out,*) ' '
        end if

      end if


      ! SETTING ====
      !
      Time_setup_ini = mpi_wtime()

      call update_conds(main, local)

      call md_make_input(world, main, local, comm_md)

      do i = 1, bin
        j = ini + i -1
        command_tmp(i) = comm_md(j)%command
      end do

      call MPI_ALLGATHERV( &
           command_tmp(1), list_bin(world%rank+1), MPI_CHARACTER, &
           command_all(1), list_bin, list_ini,     MPI_CHARACTER, &
           world%comm, ierr)

      do i = 1, main%total_replicas
        comm_md(i)%command = command_all(i)
      end do

      Time_setup_fin     = mpi_wtime()


      ! RESCALING ====
      !
      Time_rescaling_ini = mpi_wtime()

      if (flag2 .eq. "prod") then
        call md_rescaling(world, main, local)
      end if

      call MPI_BARRIER(world%comm, ierr)

      Time_rescaling_fin = mpi_wtime()



      ! MD run ====
      !

      if (world%rank == 0) &
        write(io_out,*) '  MD START'

#ifdef FUJI
      call fipp_start()
#endif

#ifndef NO_SPAWN

      Time_spawn_ini = mpi_wtime()
      call loop_spawn_for_replica(world, main, comm_md)
      Time_spawn_fin = mpi_wtime()

#else

      Time_MD_ini    = mpi_wtime()
      call loop_no_spawn_for_replica(ini, fin, comm_md)
      time_MD_fin    = mpi_wtime()

#endif

      call MPI_BARRIER(world%comm, ierr)


      ! Watchdog ====
      !
#ifndef NO_SPAWN

      Time_watchdog_ini = mpi_wtime()
      call md_watchdog(world, main, local)
      call MPI_BARRIER(world%comm, ierr)
      Time_watchdog_fin = mpi_wtime()

#endif

      if (world%rank == 0) then
        write(io_out,*) '  MD END'
        write(io_out,*) ''
      end if


      ! GET OUTPUT ====
      !

      Time_GET_ini      = mpi_wtime()

      call md_get_output(world, main, local)

      Time_GET_fin   = mpi_wtime()


      ! DISTRIBUTE ENERGY ====
      !
      Time_edist_ini = mpi_wtime()

      if (world%nprocs > 1) then
        call dist_energy(world,        &
                         main,         &
                         local(1)%num_axis_type)
      else

        !$omp parallel private(i,pp0,rep_temp)
        !$omp do
        do i = 1, main%total_replicas
          if (main%flag_velocity_rescaling /= 0 ) then
            pp0      = main%replica_rank(i,1) ! rep->param
            rep_temp = main%replica_values(pp0, main%flag_velocity_rescaling)
            main%replica_rep_temp(i) = rep_temp
          else
            main%replica_rep_temp(i) = Gbase_temperature
          end if
        end do
        !$omp end do
        !$omp end parallel
      end if


      ! OUTPUT RANK, ENERGY ====
      !

      if (world%rank == 0) &
        call trajectories_out(main, local(1)%num_axis_type)

      Time_edist_fin    = mpi_wtime()


      ! EXCHANGE ====
      !

      Time_exchange_ini = mpi_wtime()

      if (trim(adjustl(flag2)) .eq. "prod") &
        call exchange(world, main)

#ifdef FUJI
      call fipp_stop()
#endif

      Time_exchange_fin = mpi_wtime()


      ! CREATE RESTART FILES ====
      !

      Time_output_ini   = mpi_wtime()

      if (world%rank == 0) then

        call rank_restart_file(main)   !RESTART RANK FILE
        call tscale_restart_file(main) !RESTART TSCALE FILE

      end if

#ifndef NO_SPAWN
      do i = ini, fin
        call MPI_COMM_FREE(comm_md(i)%inter_communicator, ierr)
      end do
#endif

      Time_output_fin  = mpi_wtime()


      ! DISPLAY TIME 
      !
      Time_setup     = Time_setup_fin     - Time_setup_ini
      Time_rescaling = Time_rescaling_fin - Time_rescaling_ini
#ifndef NO_SPAWN
      Time_spawn     = Time_spawn_fin     - Time_spawn_ini
      Time_watchdog  = Time_watchdog_fin  - Time_watchdog_ini
      Time_MD        = Time_spawn         + Time_watchdog
#else
      Time_MD        = Time_MD_fin        - Time_MD_ini
#endif
      Time_edist     = Time_edist_fin     - Time_edist_ini
      Time_exchange  = Time_exchange_fin  - Time_exchange_ini
      Time_output    = Time_output_fin    - Time_output_ini
      Time_GET       = Time_GET_fin       - Time_GET_ini

      Time_Loop      = Time_setup     + &
                       Time_rescaling + &
                       Time_MD        + &
                       Time_edist     + &
                       Time_exchange  + &
                       Time_output
      Time_Loop_woio = Time_setup     + &
                       Time_rescaling + &
                       Time_MD        + &
                       Time_edist     + &
                       Time_exchange

      if (world%rank == 0) then
        write(io_out,'(A)')      '   PROCESS TIME (RANK 0)'
        write(io_out,'(A,F8.3)') '    Time_setup      : ', Time_setup
        write(io_out,'(A,F8.3)') '    Time_rescaling  : ', Time_rescaling
#ifndef NO_SPAWN
        write(io_out,'(A,F8.3)') '    Time_spawn      : ', Time_spawn
        write(io_out,'(A,F8.3)') '    Time_watchdog   : ', Time_watchdog
#endif
        write(io_out,'(A,F8.3)') '    Time_edist      : ', Time_edist
        write(io_out,'(A,F8.3)') '    Time_exchange   : ', Time_exchange
        write(io_out,'(A,F8.3)') '    Time_output     : ', Time_output
        write(io_out,'(A,F8.3)') '    Time_MD         : ', Time_MD
        write(io_out,'(A,F8.3)') '    Time_Loop       : ', Time_Loop
        write(io_out,'(A,F8.3)') '    TimeLoop_w/o IO : ', Time_Loop_woio
        write(io_out,*) ' '
        write(io_out,*) ' '
      end if

    end do  !BIG LOOP

    deallocate(command_tmp, command_all, list_bin, list_ini)

    return

  end subroutine driver


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  init_driver
  !> @brief      initialize driver
  !! @authors    NT, NM
  !! @date       2012/03/02 
  !! @param[in]    world    MPI information
  !! @param[inout] main     main information
  !! @param[inout] local    md local information
  !! @param[inout] comm_md  md spawn information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine init_driver(world, main, local, comm_md)

    ! formal arguments
    type(s_comm_x),                   intent(in)    :: world
    type(s_main_info),                intent(inout) :: main
    type(s_local_info),  allocatable, intent(inout) :: local(:)
    type(s_spawn_kit),   allocatable, intent(inout) :: comm_md(:)

    ! local variables
    integer                           :: ierr, i, j


    if (world%rank == 0) then
      write(io_out,*) 'INITIALIZE DRIVER START =='
      write(io_out,*) ' '
    end if

    call spawn_initialize(world, main, comm_md)

    !-- master node MPI deconpose----
    !
    allocate(main%master_nrank_array(1:world%nprocs,1:3))

    if (world%rank == 0) then
      call get_mpi_loop(main%total_replicas, &
                        world%nprocs,        &
                        main%master_nrank_array)

      write(io_out,*) '  MASTER NODE MPI DECOMPOSE>'
      do i = 1, world%nprocs
        write(io_out,'(A,I4,A,I5,A,I5,A,I5)') &
                     '     PROC:', i, &
                     '   ini:', main%master_nrank_array(i,1), &
                     '   fin:', main%master_nrank_array(i,2), &
                     '   bin:', main%master_nrank_array(i,3)
        if (main%master_nrank_array(i,3) == 0) then
          write(io_out,*) '  Init_Driver> ERROR: bin == 0'
          stop
        end if
      end do
    end if

    call MPI_BARRIER(world%comm, ierr)
    call MPI_BCAST  (main%master_nrank_array, 3*world%nprocs, &
                     MPI_INTEGER, 0, world%comm, ierr)

    ! ---ALLOCATION of LOCAL_INFO
    allocate(local(1:main%total_replicas))
    do i = 1, main%total_replicas
      allocate(local(i)%condition_values(1:main%num_of_values))
    end do

    !$omp parallel private(i,j)
    !$omp do
    do i = 1, main%total_replicas

      ! --- REPLICA_LOCAL
      do j = 1, main%num_of_values
        local(i)%condition_values(j) = &
             main%replica_values(main%replica_rank(i,1), j)
      end do

      ! --- LOCAL_INFO
      call trans_int2cha(i, local(i)%name_pref_rank)
      local(i)%nrank                  = i

      !--- FOR SPAWN
      comm_md(i)%root = 0
      comm_md(i)%inter_communicator = 0

      if (main%flag_velocity_rescaling == 0) then
        main%replica_rep_temp(i) = Gbase_temperature
        local(i)%num_axis_type   = Gnumber_of_axis
      else
        local(i)%num_axis_type   = Gnumber_of_axis - 1      
      end if
    end do
    !$omp end do
    !$omp end parallel


    ! INITIAL SETTINGS =================================
    !
    main%initial_steps   = 0 + Gfirst_step
    main%last_steps      = main%initial_steps + Gequiv_steps + Gproduct_steps-1
    main%i_steps         = 1
    main%i_equiv         = 0
    main%i_product       = 0
    main%clockt          = 0.1840198
    main%clockt2         = 0.4204822
    main%replica_prerank = main%replica_rank  ! reset replica_prerank

    !PRINT HEADER =====================================
    !
    if (world%rank == 0) then
      write(io_out,*) ''
      write(io_out,*) '  REIN_INITIAL:', main%initial_steps
      write(io_out,*) '  REIN_FINAL  :', main%last_steps
      write(io_out,*) '  REIN_EQUIV  :', Gequiv_steps
      write(io_out,*) '  REIN_PROD   :', Gproduct_steps
      write(io_out,*) ''
      write(io_out,*)  'INITIALIZE DRIVER END =='
      write(io_out,*) ''
    end if

    return

  end subroutine init_driver


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  md_make_input
  !> @brief      create files every steps
  !! @authors    NT, NM
  !! @date       2012/03/06 
  !! @param[in]    world    MPI information
  !! @param[in]    main     main information
  !! @param[inout] local    md local information
  !! @param[inout] comm_md  md spawn information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine md_make_input(world, main, local, comm_md)
#ifdef FUJI
    use service_routines, only:chmod
#endif
#ifdef INTEL
    use IFPORT
#endif

    ! formal arguments
    type(s_comm_x),     intent(in)    :: world
    type(s_main_info),  intent(in)    :: main
    type(s_local_info), intent(inout) :: local(:)
    type(s_spawn_kit),  intent(inout) :: comm_md(:)

    ! local variables
    integer             :: i, io, ich, i_replica
    integer             :: ini, bin
    character(icha_2h)  :: lcld
    character(icha_8)   :: n1, n2


    if (world%rank == 0) &
      write(io_out,*) '  MAKE INPUT FILE START'

    ini = main%master_nrank_array(world%rank + 1, 1)
    bin = main%master_nrank_array(world%rank + 1, 3)

    call trans_int2cha(main%i_steps,     n1)
    call trans_int2cha(main%i_steps - 1, n2)

    !CREATE INPUT FILE =====
    !

    !$omp parallel private(i_replica,i,lcld,io,ich)
    !$omp do

    do i_replica = 1, bin

      i = i_replica + ini - 1
      local(i)%name_i_steps    = n1
      local(i)%name_ipre_steps = n2

      select case(Gprogram_name)

      case ('namd')

        call namd_make_input(world,     &
                             main,      &
                             local(i))

      case ('marble')

        call marble_make_input(world,    &
                               main,     &
                               local(i))

      end select

      !for SPAWN
      lcld = trim(adjustl(local(i)%local_command_path))
      comm_md(i)%command = lcld

      call get_io(io)

      open(io, file = trim(adjustl(lcld)))
      write(io,'(a)') "#!/bin/sh"
      write(io,'(a)') trim(adjustl(local(i)%local_command))
      close(io)

      call free_io(io)

      ich = chmod(trim(adjustl(lcld)), "a+x")
      if (ich /= 0) then
        write(io_out,*) '  Md_Make_Input> ERROR: chmod was failed.'//trim(lcld)
        stop
      end if

      call compare_file_line(adjustl(lcld), 2)

    end do

    !$omp end do
    !$omp end parallel

    if (world%rank == 0) then
      write(io_out,*) '  MAKE INPUT FILE END'
      write(io_out,*) ''
    end if

    return

  end subroutine md_make_input


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  md_rescaling
  !> @brief      rescaling for each MD
  !! @authors    NT, NM
  !! @date       2012/03/06 
  !! @param[in]    world  MPI information
  !! @param[in]    main   main information
  !! @param[inout] local  md local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine md_rescaling(world, main, local)

    ! formal arguments
    type(s_comm_x),     intent(in)    :: world
    type(s_main_info),  intent(in)    :: main
    type(s_local_info), intent(inout) :: local(:)

    ! local variables
    integer             :: i, ini, fin


    if (world%rank == 0) &
      write(io_out,*) '  MD RESCALING START'

    ini = main%master_nrank_array(world%rank + 1, 1)
    fin = main%master_nrank_array(world%rank + 1, 2)

    !$omp parallel private(i)
    !$omp do

    do i = ini, fin

      select case(Gprogram_name)

      case ('namd')

        call namd_rescaling(world,    &
                            main,     &
                            local(i), &
                            i)

      case ('marble')

        call marble_rescaling(world,    &
                              main,     &
                              local(i), &
                              i)

      end select

    end do

    !$omp end do
    !$omp end parallel

    if (world%rank == 0) then
      write(io_out,*) '  MD RESCALING END'
      write(io_out,*) ''
    end if

    return

  end subroutine md_rescaling


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  md_watchdog
  !> @brief      watchdog for each MD
  !! @authors    NT, NM
  !! @date       2012/03/10 
  !! @param[in]    world  MPI information
  !! @param[in]    main   main information
  !! @param[inout] local  md local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine md_watchdog(world, main, local)


    ! formal arguments
    type(s_comm_x),     intent(in)    :: world
    type(s_main_info),  intent(in)    :: main
    type(s_local_info), intent(inout) :: local(:)

    ! local variables
    integer             :: i, ini, fin


    ini = main%master_nrank_array(world%rank + 1, 1)
    fin = main%master_nrank_array(world%rank + 1, 2)

    !$omp parallel private(i)
    !$omp do

    do i = ini, fin

      select case(Gprogram_name)

      case ('namd')

        call namd_watchdog(world, main, local(i))

      case ('marble')

        call marble_watchdog(world, main, local(i))

      end select


    end do

    !$omp end do
    !$omp end parallel

    return

  end subroutine md_watchdog


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  md_get_output
  !> @brief      get output for each MD
  !! @authors    NT, NM
  !! @date       2012/03/09 
  !! @param[in]    world  MPI information
  !! @param[inout] main   main information
  !! @param[in]    local  md local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine md_get_output(world, main, local)

    ! formal arguments
    type(s_comm_x),     intent(in)    :: world
    type(s_main_info),  intent(inout) :: main
    type(s_local_info), intent(in)    :: local(:)

    ! local variables
    integer             :: i, ini, fin


    ini = main%master_nrank_array(world%rank + 1, 1)
    fin = main%master_nrank_array(world%rank + 1, 2)

    !$omp parallel private(i)
    !$omp do

    do i = ini, fin

      select case(Gprogram_name)

      case ('namd')

        call namd_get_output  (main,     &
                               local(i), &
                               i)

      case ('marble')

        call marble_get_output(world,    &
                               main,     &
                               local(i), &
                               i)

      end select

    end do

    !$omp end do
    !$omp end parallel

    return

  end subroutine md_get_output


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  dist_energy
  !> @brief      distribute energy
  !! @authors    NT, NM
  !! @date       2012/03/02 
  !! @note       Hybrid : omp / mpi
  !! @param[in]    world  MPI information
  !! @param[inout] main   main information
  !! @param[in]    naxis  number of axis type
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dist_energy(world, main, naxis)

    ! formal arguments
    type(s_comm_x),               intent(in)    :: world
    type(s_main_info), target,    intent(inout) :: main
    integer,                      intent(in)    :: naxis

    ! local variables
    double precision              :: rep_temp
    integer                       :: pp0
    integer                       :: i, ii, j, ierr
    integer                       :: max_x, bin_x, ip, c1, c3
    integer                       :: ini, bin

    double precision, allocatable :: r0_positions(:)
    double precision, allocatable :: r_positions(:)
    integer,          allocatable :: inbin(:), list_bin(:), list_nbi(:)

    integer,          pointer     :: arrayx(:,:)


    arrayx => main%master_nrank_array

    ini = main%master_nrank_array(world%rank + 1, 1)
    bin = main%master_nrank_array(world%rank + 1, 3)

    c1 = 6 + naxis
    max_x = c1 * main%total_replicas
    bin_x = c1 * bin

    allocate(r0_positions(1:max_x),    &
             r_positions(1:bin_x),     &
             inbin(1:world%nprocs),    &
             list_bin(1:world%nprocs), &
             list_nbi(1:world%nprocs))

    inbin(1)    = c1*arrayx(1,3)
    list_nbi(1) = 1
    list_bin(1) = 0

    do i = 2, world%nprocs
      inbin(i)    = c1*arrayx(i,3)
      list_bin(i) = list_bin(i-1) + inbin(i-1)   ! buffer
      list_nbi(i) = list_nbi(i-1) + arrayx(i-1,3)
    end do

    !$omp parallel private(i,j,ii,pp0,ip,rep_temp)
    !$omp do

    do i = 1, bin
      ii = ini + i - 1

      if (main%flag_velocity_rescaling /= 0) then
        pp0      = main%replica_rank(ii,1) ! rep->param
        rep_temp = main%replica_values(pp0, main%flag_velocity_rescaling)
        main%replica_rep_temp(ii) = rep_temp
      else
        main%replica_rep_temp(ii) = Gbase_temperature
      end if

      ip = (i-1)*c1
      r_positions(1+ip) = main%replica_energy     (ii)
      r_positions(2+ip) = main%replica_kinetice   (ii)
      r_positions(3+ip) = main%replica_potentiale (ii)
      r_positions(4+ip) = main%replica_temperature(ii)
      r_positions(5+ip) = main%replica_extra_pote (ii)
      r_positions(6+ip) = main%replica_rep_temp   (ii)

      if (naxis /= 0) then 
        do j = 1,naxis
          r_positions(6+j+ip) = main%replica_param_p(ii,j)
        end do

#ifdef DEBUG
        write(io_out,*) "#E# chk flag vv=", &
             i,world%rank, main%flag_velocity_rescaling
        write(io_out,*) "#E# chk=", ii, " rank=", world%rank
        write(io_out,*)"      tene=", r_positions(1+ip)
        write(io_out,*)"      kine=", r_positions(2+ip)
        write(io_out,*)"      pote=", r_positions(3+ip)
        write(io_out,*)"      temp=", r_positions(4+ip)
        write(io_out,*)"      exte=", r_positions(5+ip)
        write(io_out,*)"      rept=", r_positions(6+ip)
        write(io_out,*)"      dist=", r_positions(7+ip)
#endif

      end if
    end do

    !$omp end do
    !$omp end parallel

    ! allgather
    call MPI_ALLGATHERV(r_positions(1), bin_x,           MPI_DOUBLE_PRECISION, &
                        r0_positions(1),inbin, list_bin, MPI_DOUBLE_PRECISION, &
                        world%comm, ierr)

    ! reduce
    !   r0_position(max_x),  max_x = (6+naxis)*main%total_replicas
    !c1 = 6 + naxis

    !$omp parallel private(i,j,ii,ip,c3)
    !$omp do

    do i = 1, main%total_replicas
      ii = ini + i - 1
      ip = (i-1)*c1
      c3 = ip 

#ifdef DEBUG
      if (world%rank == 0) then
        write(io_out,*)"#P# chk=", i, " rank=", world%rank, " c3=",c3
        write(io_out,*)"     tene=", r0_positions(1+c3)
        write(io_out,*)"     kine=", r0_positions(2+c3)
        write(io_out,*)"     pote=", r0_positions(3+c3)
        write(io_out,*)"     temp=", r0_positions(4+c3)
        write(io_out,*)"     exte=", r0_positions(5+c3)
        write(io_out,*)"     rept=", r0_positions(6+c3)
        if (naxis/=0) then
          write(io_out,*)"     dist=", r0_positions(7+c3)
        end if
      end if
#endif

      main%replica_energy     (i) = r0_positions (1+c3)
      main%replica_kinetice   (i) = r0_positions (2+c3)
      main%replica_potentiale (i) = r0_positions (3+c3)
      main%replica_temperature(i) = r0_positions (4+c3)
      main%replica_extra_pote (i) = r0_positions (5+c3)
      main%replica_rep_temp   (i) = r0_positions (6+c3)
      if(naxis/=0)then
        do j = 1,naxis,1
          main%replica_param_p(i,j) = r0_positions(6+j+c3)
        end do
      end if
    end do

    !$omp end do
    !$omp end parallel


    main%replica_prerank = main%replica_rank

#ifdef DEBUG
    if (world%rank == 0) then

      if (naxis /= 0) then
        do i = 1, main%total_replicas
          write(io_out,'("P=",1I3,100D16.8)') i, &
               main%replica_energy(i),       &
               main%replica_kinetice(i),     &
               main%replica_potentiale(i),   &
               main%replica_temperature(i),  &
               (main%replica_param_p(i,j),j=1,naxis,1), &
               main%replica_extra_pote(i),   &
               main%replica_rep_temp(i)
        end do
      else
        do i = 1, main%total_replicas
          write(io_out,'(1I3,100D16.8)') i,      &
               main%replica_energy(i),       &
               main%replica_kinetice(i),     &
               main%replica_potentiale(i),   &
               main%replica_temperature(i),  &
               main%replica_extra_pote(i),   &
               main%replica_rep_temp(i)
        end do
      end if
    end if
#endif

    deallocate(r_positions, r0_positions, inbin, list_bin, list_nbi)

    return

  end subroutine dist_energy


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  rank_restart_file
  !> @brief      rank restart file
  !! @authors    NT, NM
  !! @date       2012/03/02 
  !! @param[in] main   main information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine rank_restart_file(main)

    ! formal arguments
    type(s_main_info), intent(in) :: main

    ! local variables
    double precision   :: next_rep_temp
    integer            :: io, k, kk, p_num0, ist


    call get_io(io)

    open(io, &
         file     = trim(adjustl(Goutput_restart_rank)), &
         position = 'rewind',  &
         status   = 'replace', &
         form     = 'formatted')

    do k = 1, main%total_replicas
      p_num0        = main%replica_rank(k,1)
      next_rep_temp = main%replica_values(p_num0,2)
      kk = main%replica_no(k,2)
      write(io,*,iostat=ist) main%replica_rank(k,1)
    end do

    close(io)

    call free_io(io)

    return
  
  end subroutine rank_restart_file


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  tscale_restart_file
  !> @brief      tscale restart file
  !! @authors    NT, NM
  !! @date       2012/03/02 
  !! @param[in] main   main information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine tscale_restart_file(main)

    ! formal arguments
    type(s_main_info), intent(inout) :: main

    ! local variables
    double precision   :: Tpre, T, coef
    integer            :: io, i


    call get_io(io)

    open(io, &
         file     = trim(adjustl(Goutput_restart_tscale)), &
         position = 'rewind',  &
         status   = 'replace', &
         form     = 'formatted')

    do i = 1, main%total_replicas
      Tpre = main%replica_values(main%replica_prerank(i,1),1)
      T    = main%replica_values(main%replica_rank(i,1),1)
      coef = dsqrt( T / Tpre )
      main%replica_tscale(i,1) = coef
      write(io,*) main%replica_tscale(i,1)
    end do

    close(io)

    call free_io(io)

    return

  end subroutine tscale_restart_file


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  trajectories_out
  !> @brief      trajectories out
  !! @authors    NT, NM
  !! @date       2012/03/02 
  !! @note       io1: rank,  io2: energy
  !! @param[in] main   main information
  !! @param[in] naxis  number of axis type
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine trajectories_out(main, naxis)

    ! formal arguments
    type(s_main_info), intent(in) :: main
    integer,           intent(in) :: naxis

    ! local variables
    double precision   :: next_rep_temp
    integer            :: io1, io2
    integer            :: k, kk, p_num0, ist, j


    call get_io(io1)

    open(io1, &
         file     = trim(adjustl(Goutput_replica_rank)), &
         position = 'append',  &
         status   = 'unknown', &
         form     = 'formatted')
    write(io1,*) '#========= EXCHANGE_STEP =', main%i_steps

    if (naxis /= 0) then

      do k = 1, main%total_replicas
        kk     = main%replica_no(k,2)
        p_num0 = main%replica_rank(kk,1) ! rep->param

        next_rep_temp = main%replica_values(p_num0,1)

! rep. cond. Temp. Param....  TotE KineE PotE 
        write(io1,'(2I8,100F21.5)',iostat=ist)       &
             k,                                      &
             main%replica_rank(kk,1),                &
             main%replica_temperature(kk),           &
            (main%replica_param_p(kk,j),j=1, naxis), &
             main%replica_energy(kk),                &
             main%replica_kinetice(kk),              &
             main%replica_potentiale(kk),            &
             main%replica_extra_pote(kk)
      end do

    else

! rep. cond. Temp. TotE KineE PotE 
      do k = 1, main%total_replicas
        kk     = main%replica_no(k,2)
        p_num0 = main%replica_rank(kk,1) ! rep->param

        next_rep_temp = main%replica_values(p_num0,1)

        write(io1,'(2I8,100F21.5)',iostat=ist)  &
             k,                                 &
             main%replica_rank(kk,1),           &
             main%replica_temperature(kk),      &
             main%replica_energy(kk),           &
             main%replica_kinetice(kk),         &
             main%replica_potentiale(kk)
      end do

    end if

    close(io1)

    call free_io(io1)

    return

  end subroutine trajectories_out


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  update_conds
  !> @brief      update condition values
  !! @authors    NT, NM
  !! @date       2012/03/02 
  !! @param[in]    main   main information
  !! @param[inout] local  md local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_conds(main, local)

    ! formal arguments
    type(s_main_info),  intent(in)    :: main
    type(s_local_info), intent(inout) :: local(:)

    ! local variables
    integer             :: i, j


    !$omp parallel private(i,j)
    !$omp do
    do i = 1, main%total_replicas
      do j = 1, main%num_of_values
        local(i)%condition_values(j) = &
             main%replica_values(main%replica_rank(i,1), j)
      end do
    end do
    !$omp end do
    !$omp end parallel

    return

  end subroutine update_conds

end module driver_mod
