!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : exchange_mod
!> @brief   REIN exchange modules
!! @author  Naoyuki Miyashita (NM)
!! @date    NM: 2012/03/12
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module exchange_mod

  use parallel_mod
  use common_mod
  use constants_mod
  use mpi
  !$ use omp_lib

  implicit none
  private

  ! structures
  type s_exchange
    double precision   :: beta1
    double precision   :: beta2
    double precision   :: T1
    double precision   :: T2
    double precision   :: deltaE 
    double precision   :: deltaE_potential
    double precision   :: deltaE_harmonic
    double precision   :: deltaE_pv
    double precision   :: E1
    double precision   :: E2
    double precision   :: E3
    double precision   :: E4
    integer            :: par_rank1
    integer            :: par_rank2
    integer            :: rep_rank1
    integer            :: rep_rank2
    integer            :: activ_axis_index
    integer            :: i_exchange_cycle
    integer            :: activ_control_index
    character(icha_2h) :: atname
    logical            :: exchange_go
  end type s_exchange

  ! subroutines
  public  :: exchange
  private :: par_rep_ranks
  private :: conv_energy
  private :: exchange_data
  private :: prepare_energy
  private :: deltaE_temperature
  private :: deltaE_harmonic

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  exchange
  !> @brief      exchange replica
  !! @authors    NT, NM
  !! @date       2010/07/20 (NM)
  !! @param[in]    world  MPI information
  !! @param[inout] main   main information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine exchange(world, main)

    ! formal arguments
    type(s_comm_x),             intent(in)    :: world
    type(s_main_info), target,  intent(inout) :: main

    ! local variables
    integer                     :: i, ii, ierr, i_tmp

    type(s_exchange),  target,  allocatable   :: ex(:)
    integer,                    allocatable   :: list_position(:)
    integer,                    allocatable   :: list_bin(:)
    logical,                    allocatable   :: local_exchange_go(:)
    logical,                    allocatable   :: all_exchange_go(:)

    integer,                    pointer       :: eas
    integer,                    pointer       :: i_steps
    integer,                    pointer       :: totrep
    integer,                    pointer       :: aa_idx(:), ac_idx(:)
    integer,                    pointer       :: ini,fin,bin
    integer,                    pointer       :: arrayx(:,:)


    aa_idx  => ex%activ_axis_index
    i_steps => main%i_steps
    totrep  => main%total_replicas
    eas     => main%exchange_axis_switch
    ac_idx  => ex%activ_control_index
    arrayx  => main%master_nrank_array

    ini     => main%master_nrank_array(world%rank + 1, 1)
    fin     => main%master_nrank_array(world%rank + 1, 2)
    bin     => main%master_nrank_array(world%rank + 1, 3)

    if (world%rank == 0) then
      write(io_out,*) '  EXCHANGE SECTION START'
    end if

    allocate(ex(totrep), aa_idx(totrep), ac_idx(totrep))

    eas = mod(i_steps, Gnumber_of_exchange)
    if (eas == 0) &
      eas = Gnumber_of_exchange  ! 1,2,3,4

    if (world%rank == 0) then
      write(io_out,'(A,I4,A,I4)') &
           '    EXCHANGE AXIS SWITCH: ', eas, &
           '  / # OF EXCHANGE: ', Gnumber_of_exchange
    end if

    !
    !$omp parallel private(i,i_tmp)
    !$omp do

    do i = 1, totrep

      aa_idx(i) = main%axis_exchange_index(eas)  ! 1,2,1,2
      i_tmp = aa_idx(i)
      ac_idx(i) = Gaxis_control_param(i_tmp)
      ex(i)%activ_axis_index    = i_tmp
      ex(i)%activ_control_index = ac_idx(i)

      ex(i)%i_exchange_cycle = eas              ! 1,2,3,4
      ex(i)%atname           = Gaxis_type_name(i_tmp)
      ex(i)%rep_rank1        = i

      call par_rep_ranks(main, ex(i))
      ex(i)%exchange_go      = .false.

    end do

    !$omp end do
    !$omp end parallel
    !


    if (world%nprocs > 1) then

      allocate(list_position(world%nprocs), list_bin(world%nprocs))
      allocate(all_exchange_go(totrep), local_exchange_go(bin))

      list_position(1) = 0
      list_bin(1)      = arrayx(1,3)

      do i = 2, world%nprocs
        list_bin(i)      = arrayx(i,3)
        list_position(i) = list_position(i-1) + list_bin(i-1)
      end do

    end if


    !
    !$omp parallel private(i)
    !$omp do

    do ii = 1, bin
      i = ini + ii - 1

      if (ex(i)%par_rank2 /= 0 ) then
        call prepare_energy(main, ex(i)) ! calculation of deltaE
        call conv_energy(main, ex(i), i) ! estimation of exp(-deltaE)
      end if

      if (world%nprocs > 1) then
        local_exchange_go(ii) = ex(i)%exchange_go
      end if

    end do

    !$omp end do
    !$omp end parallel
    !


    if (world%nprocs > 1) then

      call MPI_ALLGATHERV( &
           local_exchange_go, bin,                    MPI_LOGICAL, &
           all_exchange_go,  list_bin, list_position, MPI_LOGICAL, &
           world%comm, ierr)

      do i = 1,totrep
        ex(i)%exchange_go = all_exchange_go(i)
        call random_number(main%clockt) ! proceed rand for random with openmp
      end do

      deallocate(all_exchange_go, local_exchange_go)
      deallocate(list_position, list_bin)

    end if


    ! Communications (allgatherv):
    !             INT: ex% rep_rank1, rep_rank2, par_rank1, par_rank2
    !         LOGICAL: ex%exchange_go
    !
    ! exchange data
    call exchange_data(world, main, ex)

    deallocate(ex,aa_idx,ac_idx)

    if (world%rank == 0) then
      write(io_out,*) '  EXCHANGE SECTION END'
      write(io_out,*) ''
    end if

    return

  end subroutine exchange
     

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  par_rep_ranks
  !> @brief      find par_rank1, par_rank2, rep_rank1, rep_rank2
  !! @authors    NT, NM
  !! @date       2010/07/21 (NM)
  !! @param[in]    main   main information
  !! @param[inout] ex     exchange information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine par_rep_ranks(main, ex)

    ! formal arguments
    type(s_main_info),         intent(in)    :: main
    type(s_exchange),  target, intent(inout) :: ex

    ! local variables
    integer,           pointer :: rep_rank1, par_rank1
    integer,           pointer :: rep_rank2, par_rank2, iec


    rep_rank1 => ex%rep_rank1
    par_rank1 => ex%par_rank1
    rep_rank2 => ex%rep_rank2
    par_rank2 => ex%par_rank2
    iec       => ex%i_exchange_cycle

    par_rank1 = main%replica_rank(rep_rank1,1)
    par_rank2 = main%replica_ep(par_rank1,iec)

    if (par_rank2 == 0) then
      rep_rank2 = 0
    else
      rep_rank2 = main%replica_param2rep(par_rank2,1)
    end if

    return

  end subroutine par_rep_ranks
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  conv_energy
  !> @brief      estimation of exchange exp(-deltaE)
  !! @authors    NT, NM
  !! @date       2010/07/21 (NM)
  !! @param[inout] main   main information
  !! @param[inout] ex     exchange information
  !! @param[in]    i_rank rank number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine conv_energy(main, ex, i_rank)

    ! formal arguments
    type(s_main_info),  intent(inout) :: main
    type(s_exchange),   intent(inout) :: ex
    integer,            intent(in)    :: i_rank

    ! local variables
    double precision    :: mdelta, bb, cc
    integer             :: i
    character(icha_2h)  :: line_fmt


    if (ex%deltaE <= 0.d0) then
      ex%exchange_go = .true.
      bb = 1.d0
      cc = 1.d0
    else
      do i = 1, i_rank   ! for random with openmp
        call random_number(main%clockt) 
      end do

      mdelta = -1.d0 * ex%deltaE
      bb = exp(mdelta)
      cc = main%clockt
      if (cc <= bb ) &
        ex%exchange_go = .true.

    end if

#ifdef DEBUG
    if (ex%exchange_go) then
      line_fmt = &
    '("#(r,p)(1)=(",I3,",",I3,",",F16.10,")<=>(2)=(",I3,",",I3,",",F16.10, &
    "):",/,"SUCCESS !!: dE=",D16.10," exp(-d)=",D16.10" >= ",D16.10)'
    else
      line_fmt = &
    '("#(r,p)(1)=(",I3,",",I3,",",F16.10,")<=>(2)=(",I3,",",I3,",",F16.10, &
    "):",/,"NOT EXCN!!: dE=",D16.10," exp(-d)=",D16.10" <  ",D16.10)'
    end if
    
    write(io_out, fmt=line_fmt) &
         ex%rep_rank1,ex%par_rank1,ex%T1,ex%rep_rank2,ex%par_rank2,ex%T2, &
         ex%deltaE, bb, cc
#endif

    return

  end subroutine conv_energy
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  exchange_data
  !> @brief      exchange of data
  !! @authors    NT, NM
  !! @date       2010/07/21 (NM)
  !! @param[inout] main  main information
  !! @param[inout] ex    exchange information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine exchange_data(world, main, ex)

    ! formal arguments
    type(s_comm_x),                intent(in)    :: world
    type(s_main_info), target,     intent(inout) :: main
    type(s_exchange),  target,     intent(inout) :: ex(:)

    ! local variables
    integer,           allocatable :: comm_rank_data(:)
    integer,           allocatable :: param_list1(:)
    integer,           allocatable :: param_list2(:)
    integer,           allocatable :: comm_rank_data_sml(:)
    integer                        :: i,ichk
    integer                        :: totrep4

    integer,           pointer     :: totrep


    totrep    => main%total_replicas
    totrep4   = totrep*4

    allocate(comm_rank_data(1:totrep4), &
             comm_rank_data_sml(1:4),   &
             param_list1(1:totrep),     &
             param_list2(1:totrep))

    param_list1 = 0; param_list2 = 0; comm_rank_data = 0

    if (world%rank == 0) then
      write(io_out,*) '   EXCHANGE(PRE)'
      write(io_out,'(4X,I5,I3"RANKX"100I4)') &
           main%i_steps, main%total_replicas, &
           (main%replica_rank(i,1), i=1, totrep, 1)
    end if

    !$omp parallel private(i,ichk)
    !$omp do
    do i = 1, totrep
      if (ex(i)%exchange_go)then
        comm_rank_data(4*i-3) = ex(i)%rep_rank1
        comm_rank_data(4*i-2) = ex(i)%par_rank2
        comm_rank_data(4*i-1) = ex(i)%rep_rank2
        comm_rank_data(4*i)   = ex(i)%par_rank1
        ichk = 1
      else
        comm_rank_data(4*i-3) = ex(i)%rep_rank1
        comm_rank_data(4*i-2) = ex(i)%par_rank1
        comm_rank_data(4*i-1) = ex(i)%rep_rank2
        comm_rank_data(4*i)   = ex(i)%par_rank2
        ichk = 0
      end if
    end do
    !$omp end do
    !$omp end parallel


    !$omp parallel private(i)
    !$omp do
    do i = 1,totrep
      param_list1(comm_rank_data(4*i-3)) = comm_rank_data(4*i-2)
      if (comm_rank_data(4*i-1) /= 0 ) &
        param_list2(comm_rank_data(4*i-1)) = comm_rank_data(4*i)
    end do
    !$omp end do
    !$omp end parallel


    ! !$omp parallel private(i)
    ! !$omp do
    do i = 1, totrep,1
      if (param_list2(i) /= 0) param_list1(i) = param_list2(i)
    end do
    ! !omp end do 
    ! !omp end parallel


    if (world%rank == 0) then
      write(io_out,*) '   EXCHANGE(AFTER)'
      write(io_out,'(4X,I5,I3"CHKXX"100I4)') &
           main%i_steps, main%total_replicas, &
           (param_list1(i), i=1, totrep, 1)
    end if

    !NEW RANK CREATE
    !$omp parallel private(i)
    !$omp do
    do i = 1, totrep
      main%replica_rank(i,1) = param_list1(i)
      main%replica_param2rep(param_list1(i),1) = i
    end do
    !$omp end do
    !$omp end parallel
    !NEW RANK CREATE

    deallocate(comm_rank_data,     &
               comm_rank_data_sml, &
               param_list1,        &
               param_list2)

    return

  end subroutine exchange_data
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  prepare_energy
  !> @brief      Please make energy setup for replica axis
  !! @authors    NT, NM
  !! @date       2010/07/20 (NM)
  !! @param[in]    main  main information
  !! @param[inout] ex    exchange information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine prepare_energy(main, ex)

    ! formal arguments
    type(s_main_info), target, intent(in)    :: main
    type(s_exchange),  target, intent(inout) :: ex


#ifdef DEBUG
    write(io_out,*) '   PREPARE_ENERGY: ', trim(adjustl(ex%atname))
#endif

    select case(trim(adjustl(ex%atname)))

    case("temperature")
      call deltaE_temperature(main, ex)
      !ex%deltaE = ex%deltaE_potential

    case("harmonic")
      call deltaE_harmonic(main, ex)
      !ex%deltaE = ex%deltaE_harmonic

    case("quad")
      !call deltaE_quad()

    case("pressure")
      !call deltaE_pressure()

    case("volume")
      !call deltaE_volume()

    case default

    end select

    return

  end subroutine prepare_energy
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  deltaE_temperature
  !> @brief      Please make delta E for temperature exchange
  !! @authors    NT, NM
  !! @date       2010/07/20 (NM)
  !! @param[in]    main  main information
  !! @param[inout] ex    exchange information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine deltaE_temperature(main, ex)

    ! formal arguments
    type(s_main_info), target, intent(in)    :: main
    type(s_exchange),  target, intent(inout) :: ex


    ex%T1    = main%replica_rep_temp(ex%par_rank1)
    ex%beta1 = kboltz * ex%T1
    ex%beta1 = 1.d0/ex%beta1

    ex%T2    = main%replica_rep_temp(ex%par_rank2)
    ex%beta2 = kboltz * ex%T2
    ex%beta2 = 1.d0/ex%beta2

    ex%E1    = main%replica_potentiale(ex%rep_rank1)
    ex%E2    = main%replica_potentiale(ex%rep_rank2)

    ! (bm - bn) * ( E(qj) - E(qi) )
    ex%deltaE_potential = (ex%beta1 - ex%beta2)*(ex%E2 - ex%E1)

#ifdef DEBUG
    write(io_out,'("#(r,p)(1)=(",I3,",",I3,",",F16.10,")<=>'//&
                         '(2)=(",I3,",",I3,",",F16.10,"):",/,'//&
                 '" E1(",D16.3,")<=>E2(",D16.3,")=deltaE(",D16.3,")")') &
                 ex%rep_rank1, ex%par_rank1, ex%T1, &
                 ex%rep_rank2, ex%par_rank2, ex%T2, &
                 ex%E1, ex%E2, ex%deltaE_potential
#endif

    ex%deltaE = ex%deltaE_potential

    return

  end subroutine deltaE_temperature
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  deltaE_harmonic
  !> @brief      Please make delta E for harmonic exchange
  !! @authors    NT, NM
  !! @date       2010/07/20 (NM)
  !! @param[in]    main  main information
  !! @param[inout] ex    exchange information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine deltaE_harmonic(main, ex)

    ! formal arguments
    type(s_main_info), target, intent(in)    :: main
    type(s_exchange),  target, intent(inout) :: ex

    ! local variables
    double precision           :: d11, d12, d21, d22
    integer                    :: inum, id1, id2
    double precision           :: kk1, kk2, diff
    integer                    :: iadd
    integer,           pointer :: aa_idx, ac_idx
    double precision,  pointer :: r10, r20, k1, k2
    double precision           :: r1, r2, max, min


    aa_idx     => ex%activ_axis_index
    ac_idx     => ex%activ_control_index
    min  = Gperiodic_values(aa_idx, 1)
    max  = Gperiodic_values(aa_idx, 2)

#ifdef DEBUG
    write(6,*) "#deltaE_harmonic: aa_idx, ac_idx=>", aa_idx, ac_idx
#endif    

    ! CHECK. IT WILL BE MODIFIED.
    inum = Gaxis_num_of_param(aa_idx)
    id1  = Gaxis_index_coef(1, aa_idx)
    id2  = Gaxis_index_coef(2, aa_idx)

#ifdef DEBUG
    write(6,*) "#deltaE_harmonic: inum, id1, id2=>", inum, id1, id2
#endif

    !r1  => main%replica_distance(ex%rep_rank1)
    !r2  => main%replica_distance(ex%rep_rank2)
    if (main%flag_velocity_rescaling /= 0) then
      iadd = aa_idx - 1
    else
      iadd = aa_idx
    end if

#ifdef DEBUG
    write(6,*)"XXXchk iadd=", iadd, aa_idx
#endif

    r1  = main%replica_param_p(ex%rep_rank1,iadd) ! check dimension
    r2  = main%replica_param_p(ex%rep_rank2,iadd) ! check dimension
    !r10 => main%replica_values(ex%par_rank1,id2)
    !r20 => main%replica_values(ex%par_rank2,id2)
    !k1  => main%replica_values(ex%par_rank1,id1)
    !k2  => main%replica_values(ex%par_rank2,id1)
    r10 => main%replica_values(ex%par_rank1,id1)
    r20 => main%replica_values(ex%par_rank2,id1)
    k1  => main%replica_values(ex%par_rank1,id2)
    k2  => main%replica_values(ex%par_rank2,id2)

#ifdef DEBUG
    !write(6,*) "chk values=",main%replica_values
    write(6,*) "rank=",ex%par_rank1,ex%par_rank2
    write(6,*) "#deltaE_harmonic: r1,r2,r10,r20,k1,k2=>", & 
         r1,r2,r10,r20,k1,k2
    !write(6,*) "#XXXchk acci= ",r1,r2," == ", &
    !  main%replica_distance(ex%rep_rank1),main%replica_distance(ex%rep_rank2)
#endif

    if (Gperiodic(aa_idx) == "on") then
      call periodic_inbox(r1,min,max)
      call periodic_inbox(r2,min,max)
    end if

    ex%T1    = main%replica_rep_temp(ex%par_rank1)
    ex%beta1 = kboltz * ex%T1
    ex%beta1 = 1.d0/ex%beta1

#ifdef DEBUG
    write(6,*) "#deltaE_harmonic: T1, beta1, beta1=>", &
         ex%T1,ex%beta1,ex%beta1
#endif

    ex%T2    = main%replica_rep_temp(ex%par_rank2)
    ex%beta2 = kboltz * ex%T2
    ex%beta2 = 1.d0/ex%beta2

#ifdef DEBUG
    write(6,*) "#deltaE_harmonic: T2, beta2, beta2=>", ex%T2, ex%beta2
#endif

    kk1 = 0.5d0 * k1 * ex%beta1
    kk2 = 0.5d0 * k2 * ex%beta2

#ifdef DEBUG
    write(6,*) 'chk=k1,k2 = ', k1, k2, id1, id2
    write(6,*) 'chk r 1  r2 ', ex%rep_rank1, ex%rep_rank2, r1, r2, r10, r20
#endif

    d11 = r1 - r10
    d12 = r1 - r20
    d21 = r2 - r10
    d22 = r2 - r20

#ifdef DEBUG
    write(6,*) "chk=dd11,12,21,22: ", d11, d12, d21, d22
    write(6,*) "chkminmax = ", min, max
#endif

    if (trim(adjustl(Gperiodic(aa_idx))) == "on") then
      write(6,*) "chk periodic=", trim(adjustl(Gperiodic(aa_idx))), aa_idx
      call periodic_inbox(d11,min,max)
      call periodic_inbox(d12,min,max)
      call periodic_inbox(d21,min,max)
      call periodic_inbox(d22,min,max)
    end if

#ifdef DEBUG
    write(6,*) "chk modi=dd11,12,21,22: ", d11, d12, d21, d22
#endif     

    ex%E1 = kk1 * (d11 * d11 - d21 * d21) ! bm ( Um(qi) - Um(qj) )
    ex%E2 = kk2 * (d12 * d12 - d22 * d22) ! bn ( Un(qi) - Un(qj) )

    ! deltaE_harmonic = bm Um(qj) + bn Un(qi) - bm Um(qi) -bn Un(qj)
    ex%deltaE_harmonic = ex%E2 - ex%E1

    ex%deltaE_potential = 0.d0
    diff = dabs(ex%T1-ex%T2)
    if (diff >= err10) then
      ex%E3 = (main%replica_potentiale(ex%rep_rank1) - 0.5d0*k1*d11*d11)
      ex%E4 = (main%replica_potentiale(ex%rep_rank2) - 0.5d0*k2*d22*d22)
      ! deltaE_others = (bm - bn) * ( E(qj) - E(qi) )
      ex%deltaE_potential = (ex%beta1 - ex%beta2)*(ex%E4 - ex%E3)
    end if

#ifdef DEBUG
    write(io_out, FMT='("#(r,p)(1)=(",I3,",",I3,",",F16.10,")<=>'//   &
                              '(2)=(",I3,",",I3,",",F16.10,"):",/,'// &
        '" E1(",F16.10,")<=>E2(",F16.10,")=deltaE(",F16.10,"):",/,'// &
        '" R1(",F16.10,",",F16.10,")<=>R2(",F16.10,",",F16.10,")")')  &
        ex%rep_rank1, ex%par_rank1, ex%T1, &
        ex%rep_rank2, ex%par_rank2, ex%T2, &
        ex%E1, ex%E2, ex%deltaE_harmonic,  &
        r1, r10, r2, r20
#endif

    ex%deltaE = ex%deltaE_potential + ex%deltaE_harmonic

    return

  end subroutine deltaE_harmonic
  
end module exchange_mod
