!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   bb_ctrl_mod
!> @brief   read control parameters
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/10/02
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

module bb_ctrl_mod

  use fileio_control_mod
  use program_mod
  use messages_mod

  implicit none
  private

  ! structures

  ! section : <main>
  type, public :: s_main_info

    character(200)               :: directory             = '.'
    character(100)               :: rein_input            = 'rein.inp'
    character(20)                :: parallel              = 'mpiexec'
    character(200)               :: template_dir          = ''
    character(20)                :: chain_job             = 'no'
    character(200)               :: batch_builder_input   = 'INPBB'
    character(200)               :: batch_builder_restart = ''
    character(200)               :: batch_builder_exe     = './batch_builder'
    character(200)               :: skip_modify_input     = 'no'
    character(200)               :: spawn                 = 'yes'
    character(200)               :: openmp                = 'no'
    character(200)               :: gpgpu                 = 'no'
    character(200)               :: hybrid_md             = 'no'

  end type s_main_info

  ! section : K
  type, public :: s_batch_info
    character(100)               :: batchfile_name        = 'rein.sh'
    
    character(10)                :: job_scheduler         = 'K'
    character(100)               :: queue_name            = 'small'

    integer                      :: num_core_cpu          = 8
    integer                      :: num_threads_md        = 1
    integer                      :: num_mpiproc_md        = 8
    integer                      :: threads_1             = 7

    integer                      :: num_mpiproc_rein      = 1

    character(10)                :: direct_describe       = 'no'
    character(20)                :: node                  = '9'
    character(20)                :: process               = '9'
    character(20)                :: mpi_static_node_shape = '1'
    character(20)                :: mpi_static_node_proc  = '8'
    integer                      :: num_mpi_static_node_shape = 1
    integer                      :: num_mpi_static_node_proc  = 8

    character(100)               :: stage_out_dir         = './STAGE_OUT'
    character(20)                :: cpu_time              = '01:00:00'
    character(100)               :: machine_file          = ''

  end type s_batch_info

  type, public :: s_remd_info

    character(10)                :: replica_exchange      = "yes"
    integer                      :: num_of_exchange_steps = 1

    integer                      :: first_step            = 1

    integer                      :: num_of_product_steps  = 1
    integer                      :: num_of_equiv_steps    = 1

  end type s_remd_info

  type, public :: s_ctrl_data

    type(s_main_info)            :: main_info
    type(s_batch_info)           :: batch_info
    type(s_remd_info)            :: remd_info

  end type s_ctrl_data

  ! subroutines
  public  :: control
  private :: read_main
  private :: read_batch
  private :: read_remd

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  control
  !> @brief      a driver subroutine for reading control parmeters
  !! @authors    NT,NM
  !! @param[in]  filename    filename of control file
  !! @param[out] ctrl_data   control parameters in the program
  !! @date       2012/10/01 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine control(filename, ctrl_data)

    ! formal arguments
    character(*),      intent(in)    :: filename
    type(s_ctrl_data), intent(inout) :: ctrl_data

    ! local variables
    integer                          :: handle

    ! open control file
    !
    call open_ctrlfile(filename, handle)

    if (handle == 0) then
      call error_msg('Bb_Ctrl> File IO Error')
    end if

    ! read section 
    !
    call read_main(handle, ctrl_data%main_info)

    call read_batch(handle, ctrl_data%batch_info)

    call read_remd(handle, ctrl_data%remd_info)

    return

  end subroutine control


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_main
  !> @brief      read control parameters in <main> section
  !! @authors    NM, NT
  !! @date       2012/10/02 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_main(handle, main_info)

    character(*), parameter :: Section = ''

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_main_info),      intent(inout) :: main_info

    character(80)             :: arg1

    
    ! get input filename
    call getarg(1, arg1)
    main_info%batch_builder_input = TRIM ( ADJUSTL( arg1 ) )

    ! read control parameters
    !
    call read_ctrlfile_string(handle, Section, &
         'directory', main_info%directory)
    call read_ctrlfile_string(handle, Section, &
         'rein_input', main_info%rein_input)
    call read_ctrlfile_string(handle, Section, &
         'parallel', main_info%parallel)
    call read_ctrlfile_string(handle, Section, &
         'template_dir', main_info%template_dir)
    call read_ctrlfile_string(handle, Section, &
         'chain_job', main_info%chain_job)
    call read_ctrlfile_string(handle, Section, &
         'batch_builder_input', main_info%batch_builder_input)
    call read_ctrlfile_string(handle, Section, &
         'batch_builder_exe', main_info%batch_builder_exe)

    call read_ctrlfile_string(handle, Section, &
         'skip_modify_input', main_info%skip_modify_input)

    call read_ctrlfile_string(handle, Section, &
         'spawn', main_info%spawn)

    if (main_info%template_dir == '') then
      main_info%template_dir = TRIM(get_program_path())//'/templates'
    end if

    call read_ctrlfile_string(handle, Section, &
         'openmp_rein', main_info%openmp)

    call read_ctrlfile_string(handle, Section, &
         'gpgpu', main_info%gpgpu)

    call read_ctrlfile_string(handle, Section, &
         'hybrid_md', main_info%hybrid_md)

    ! restart batch_builder:  ./batch_builder bb.inp restart

    if ( TRIM( main_info%chain_job ) .eq. 'no' ) then
      main_info%batch_builder_restart = ''
    else
      main_info%batch_builder_restart =                   &
        TRIM( main_info%batch_builder_exe ) // ' '     // &
        TRIM( main_info%batch_builder_input ) //          &
        ' restart'
    end if

    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Main> Parameters of Main'
    write(MsgOut,'(A25,A)') '             directory = ', &
         TRIM(main_info%directory)
    write(MsgOut,'(A25,A)') '       rein input file = ', &
         TRIM(main_info%rein_input)
    write(MsgOut,'(A25,A)') '              parallel = ', &
         TRIM(main_info%parallel)
    write(MsgOut,'(A25,A)') '          template dir = ', &
         TRIM(main_info%template_dir)
    write(MsgOut,'(A25,A)') '          chain_job    = ', &
         TRIM(main_info%chain_job)
    if ( TRIM(main_info%chain_job) .eq. 'yes' ) then
      write(MsgOut,'(A25,A)') 'batch_builder_restart  = ', &
           TRIM(main_info%batch_builder_restart)
    end if
    write(MsgOut,'(A25,A)') '      skip modify input = ', &
         TRIM(main_info%skip_modify_input)
    write(MsgOut,'(A25,A)') '                 spawn  = ', &
         TRIM(main_info%spawn)
    write(MsgOut,*) ''

    return

  end subroutine read_main


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_batch
  !> @brief      read control parameters in BATCH section
  !! @authors    NM, NT
  !! @date       2012/10/02 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_batch(handle, batch_info)

    character(*), parameter :: Section = 'BATCH'

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_batch_info),     intent(inout) :: batch_info


    ! read control parameters
    !
    call read_ctrlfile_string(handle, Section, &
         'batchfile_name', batch_info%batchfile_name)
    call read_ctrlfile_string(handle, Section, &
         'job_scheduler', batch_info%job_scheduler)
    call read_ctrlfile_string(handle, Section, &
         'queue_name', batch_info%queue_name)

    call read_ctrlfile_integer(handle, Section, &
         'num_core_cpu', batch_info%num_core_cpu)
    call read_ctrlfile_integer(handle, Section, &
         'num_threads_md', batch_info%num_threads_md)
    call read_ctrlfile_integer(handle, Section, &
         'num_mpiproc_md', batch_info%num_mpiproc_md)

    call read_ctrlfile_integer(handle, Section, &
         'num_mpiproc_rein', batch_info%num_mpiproc_rein)

    call read_ctrlfile_string(handle, Section, &
         'cpu_time', batch_info%cpu_time)
    call read_ctrlfile_string(handle, Section, &
         'stage_out_dir', batch_info%stage_out_dir)
    if (batch_info%stage_out_dir == '') &
      batch_info%stage_out_dir = '.'

    call read_ctrlfile_string(handle, Section, &
         'machine_file', batch_info%machine_file)

    ! direct discribation
    call read_ctrlfile_string(handle, Section, &
         'direct_describe', batch_info%direct_describe)

    if ( trim ( batch_info%direct_describe ) .eq. 'yes' ) then

      call read_ctrlfile_string(handle, Section, &
           'node', batch_info%node)
      call read_ctrlfile_string(handle, Section, &
           'process', batch_info%process)
      call read_ctrlfile_string(handle, Section, &
           'mpi_static_node_shape', batch_info%mpi_static_node_shape)
      call read_ctrlfile_string(handle, Section, &
           'mpi_static_node_proc', batch_info%mpi_static_node_proc)
    else
      call read_ctrlfile_integer(handle, Section, &
           'mpi_static_node_shape', batch_info%num_mpi_static_node_shape)
    end if



    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_BATCH> Parameters of BATCH'
    write(MsgOut,'(A25,A)') '        job_scheduler  = ', &
         TRIM(batch_info%job_scheduler)
    write(MsgOut,'(A25,A)') '            queue name = ', &
         TRIM(batch_info%queue_name)
    write(MsgOut,'(A25,I0)')'    number of core cpu = ', &
         batch_info%num_core_cpu
    write(MsgOut,'(A25,I0)')'  number of threads md = ', &
         batch_info%num_threads_md
    write(MsgOut,'(A25,I0)')'  number of mpiproc md = ', &
         batch_info%num_mpiproc_md
    write(MsgOut,'(A25,I0)')'number of mpiproc rein = ', &
         batch_info%num_mpiproc_rein
    write(MsgOut,'(A25,A)') '         stage out dir = ', &
         TRIM(batch_info%stage_out_dir)
    write(MsgOut,'(A25,A)') '              cpu time = ', &
         TRIM(batch_info%cpu_time)
    write(MsgOut,'(A25,A)') '          machine file = ', &
         TRIM(batch_info%machine_file)
    write(MsgOut,*) ''


    return

  end subroutine read_batch


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_remd
  !> @brief      read control parameters in REMD section
  !! @authors    NM, NT
  !! @date       2012/10/02 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_remd(handle, remd_info)

    character(*), parameter :: Section = 'REMD'

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_remd_info),      intent(inout) :: remd_info

    character(10)                :: replica_exchange      = "yes"
    integer                      :: num_of_exchange_steps = 10

    ! read control parameters
    !

    call read_ctrlfile_string(handle, Section, &
         'replica_exchange', remd_info%replica_exchange)
    call read_ctrlfile_integer(handle, Section, &
         'num_of_exchange_steps', remd_info%num_of_exchange_steps)
    call read_ctrlfile_integer(handle, Section, &
         'first_step', remd_info%first_step)
    call read_ctrlfile_integer(handle, Section, &
         'num_of_product_steps', remd_info%num_of_product_steps)
    call read_ctrlfile_integer(handle, Section, &
         'num_of_equiv_steps', remd_info%num_of_equiv_steps)


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_REMD> Parameters of REMD'
    write(MsgOut,'(A25,A)') '      replica exchange = ', &
         TRIM(remd_info%replica_exchange)
    write(MsgOut,'(A25,I0)')' num_of_exchange_steps = ', &
         remd_info%num_of_exchange_steps
    write(MsgOut,'(A25,I0)')'            first_step = ', &
         remd_info%first_step
    write(MsgOut,'(A25,I0)')'  num_of_product_steps = ', &
         remd_info%num_of_product_steps
    write(MsgOut,'(A25,I0)')'    num_of_equiv_steps = ', &
         remd_info%num_of_equiv_steps
    write(MsgOut,*) ''

    return

  end subroutine read_remd

end module bb_ctrl_mod
