!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   bb_reininp_mod
!> @brief   parse REIN input file
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/05/31 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module bb_reininp_mod

  use bb_ctrl_mod
  use fileio_control_mod
  use fileio_mod
  use messages_mod

  implicit none
  private

  type, public :: s_rein_inp
    character(100) :: inp_files(1:6)     = ''
    character(100) :: out_files(1:3)     = ''
    character(20)  :: program_name       = ''
    character(20)  :: parallel_exec      = ''
    character(20)  :: initial_file_name  = ''
    character(20)  :: io_file_name       = ''
    integer        :: number_of_replicas = 1
    integer        :: first_step         = 0
    integer        :: equiv_steps        = 0
    integer        :: product_steps      = 0
    integer        :: number_of_cpus_md  = 8
  end type s_rein_inp

  type :: s_kv
    character(40)  :: key
    character(200) :: value
  end type s_kv

  integer, public  :: OutFileRestartRank   = 1
  integer, public  :: OutFileRestartTscale = 2
  integer, public  :: OutFileReplicaRank   = 3

  ! subroutines
  public  :: reininp

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  reininp
  !> @brief      parse REIN input file
  !! @authors    NM, NT
  !! @param[in]  ctrl_data   control parameters in the program
  !! @param[out] rein_inp    information of REIN input
  !! @date       2012/05/31 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reininp(ctrl_data, rein_inp)

    ! formal arguments
    type(s_ctrl_data), intent(in)    :: ctrl_data
    type(s_rein_inp),  intent(inout) :: rein_inp

    ! local variables
    type(s_kv),        pointer       :: kv(:)=>null()
    integer                          :: i, j, nrep
    character(200)                   :: file

    ! parameter
    character(*),      parameter     :: InpFiles(1:6) = &
         (/'input_replica_rank    ', &
           'input_replica_values  ', &
           'input_replica_on      ', &
           'input_exchange_pattern', &
           'input_axis_exchange   ', &
           'input_restart_tscale  '/)

    character(*),      parameter     :: OutFiles(1:3) = &
         (/'output_restart_rank  ', &
           'output_restart_tscale', &
           'output_replica_rank  ' /)

    ! read REIN input file Key-value
    !
    file = TRIM(ctrl_data%main_info%directory) // '/' // &
           TRIM(ctrl_data%main_info%rein_input)

    call read_key_value(file, kv)

    
    ! check REIN input 
    !
    rein_inp%number_of_replicas = 1

    do i = 1, SIZE(kv)

      ! check for input files
      do j = 1, SIZE(InpFiles)
        if (kv(i)%key == InpFiles(j)) then
          rein_inp%inp_files(j) = kv(i)%value
        end if
      end do

      ! check for output files
      do j = 1, SIZE(OutFiles)
        if (kv(i)%key == OutFiles(j)) then
          rein_inp%out_files(j) = kv(i)%value
        end if
      end do

      ! check for program_name
      if (kv(i)%key == 'program_name') then
        rein_inp%program_name = kv(i)%value
        call tolower(rein_inp%program_name)
      end if

      ! check for parallel_exec
      if (kv(i)%key == 'parallel_exec') then
        rein_inp%parallel_exec = kv(i)%value
      end if

      ! check initial file name
      if (kv(i)%key == 'initial_file_name') then
        rein_inp%initial_file_name = kv(i)%value
      end if

      ! check for io_file_name
      if (kv(i)%key == 'io_file_name') then
        rein_inp%io_file_name = kv(i)%value
      end if

      ! check for number_of_replicas
      if (kv(i)%key == 'number_of_replicas') then
        read(kv(i)%value,*,err=900) nrep
        rein_inp%number_of_replicas = rein_inp%number_of_replicas * nrep
      end if

      ! check for first_step
      if (kv(i)%key == 'first_step') then
        read(kv(i)%value,*,err=900) rein_inp%first_step
      end if

      ! check for equiv_steps
      if (kv(i)%key == 'equiv_steps') then
        read(kv(i)%value,*,err=900) rein_inp%equiv_steps
      end if
        
      ! check for product_steps
      if (kv(i)%key == 'product_steps') then
        read(kv(i)%value,*,err=900) rein_inp%product_steps
      end if

      ! check for number_of_cpus_for_exec
      if (kv(i)%key == 'number_of_cpus_for_exec') then
        read(kv(i)%value,*,err=900) rein_inp%number_of_cpus_md
      end if

    end do


    ! check variables
    do i = 1, SIZE(InpFiles)
      if (rein_inp%inp_files(i) == '') &
        call error_msg('Reininp> input file "'//TRIM(InpFiles(i))//&
        '" is not defined.')
    end do
    do i = 1, SIZE(OutFiles)
      if (rein_inp%out_files(i) == '') &
        call error_msg('Reininp> output file "'//TRIM(OutFiles(i))//&
        ' "is not defined.')
    end do
    if (rein_inp%program_name == '') &
      call error_msg('Reininp> program name is not defined.')
    if (rein_inp%parallel_exec == '') &
      call error_msg('Reininp> parallel exec is not defined.')
    if (rein_inp%initial_file_name == '') &
      call error_msg('Reininp> initial file name is not defined.')
    if (rein_inp%io_file_name == '') &
      call error_msg('Reininp> I/O file name is not defined.')
    if (rein_inp%number_of_replicas <= 1) &
      call error_msg('Reininp> # of replicas is not defined.')
    if (rein_inp%first_step == 0) &
    !  call error_msg('Reininp> first step is not defined.')


    write(MsgOut,'(A)') 'Reininp> REIN input file information.'
    do i = 1, SIZE(InpFiles)
      write(MsgOut,'(A22,A3,A)') &
         TRIM(InpFiles(i)), ' = ', TRIM(rein_inp%inp_files(i))
    end do
    do i = 1, SIZE(OutFiles)
      write(MsgOut,'(A22,A3,A)') &
         TRIM(OutFiles(i)), ' = ', TRIM(rein_inp%out_files(i))
    end do
    write(MsgOut,'(A25,A)')  '          program name = ', &
         TRIM(rein_inp%program_name)
    write(MsgOut,'(A25,A)')  '         parallel exec = ', &
         TRIM(rein_inp%parallel_exec)
    write(MsgOut,'(A25,A)')  '     initial file name = ', &
         TRIM(rein_inp%initial_file_name)
    write(MsgOut,'(A25,A)')  '         I/O file name = ', &
         TRIM(rein_inp%io_file_name)
    write(MsgOut,'(A25,I0)') '   # of total replicas = ', &
         rein_inp%number_of_replicas
    write(MsgOut,'(A25,I0)') '            first step = ', &
         rein_inp%first_step
    write(MsgOut,'(A25,I0)') '           equiv steps = ', &
         rein_inp%equiv_steps
    write(MsgOut,'(A25,I0)') '         product steps = ', &
         rein_inp%product_steps
    write(MsgOut,'(A25,I0)') 'number of cpus for exe = ', &
         rein_inp%number_of_cpus_md
    write(MsgOut,'(A)') ' '


    ! deallocate memory
    !
    deallocate(kv)

    return
  
900 call error_msg('Reininp> number_of_replicas value is BAD format')

  end subroutine reininp


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_key_value
  !> @brief      read REIN input file key-and-value
  !! @authors    NT
  !! @date       2012/05/31 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_key_value(file, kv)

    ! formal arguments
    character(*),              intent(in)    :: file
    type(s_kv),                pointer       :: kv(:)

    ! local variables
    integer                    :: unit_no, nline, i
    character(200)             :: line, str
    character(1)               :: c


    call open_file(unit_no, file, IOFileInput)

    ! pre read for allocation of array
    !
    nline = 0
    do while(.true.)
      read(unit_no,'(A)',end=10) line
      nline = nline + 1
    end do
10  continue

    REWIND(unit_no)

    ! allcate
    !
    allocate(kv(nline))

    ! read key-and-value
    !
    nline = 0
    do while(.true.)
      read(unit_no,'(A)',end=20) line
      
      c = adjustl(line)

      if ((TRIM(ADJUSTL(c)) == '#') .or. &
          (TRIM(ADJUSTL(c)) == '/') .or. &
          (TRIM(ADJUSTL(c)) == '!') .or. &
          (TRIM(ADJUSTL(c)) == '<') .or. &
          (TRIM(ADJUSTL(c)) == '>') .or. &
          (TRIM(ADJUSTL(c)) == '')) &
        cycle

      if (TRIM(ADJUSTL(c)) == '&') then
        read(line,*) c, str
        kv(nline)%value = TRIM(ADJUSTL(kv(nline)%value)) // &
                          TRIM(ADJUSTL(str))
      else
        nline = nline + 1
        read(line,*) kv(nline)%key, c, kv(nline)%value
      end if
    end do
20  continue

    call close_file(unit_no)

    ! check key-and-value
    !
    !DEBUG do i = 1, nline
    !DEBUG   write(MsgOut,'(A40, A)') kv(i)%key, TRIM(kv(i)%value)
    !DEBUG end do
    !DEBUG write(MsgOut,*) ' '

    return

  end subroutine read_key_value

end module bb_reininp_mod
