!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   bb_reininp_modify_mod
!> @brief   setup for modification of rein.inp
!! @authors Naoyuki Miyashita (NM)
!! @date    2012/10/04
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module bb_reininp_modify_mod

  use bb_reininp_mod
  use bb_ctrl_mod
  use replace_string_mod
  use fileio_control_mod
  use fileio_mod
  use messages_mod

  implicit none
  private

  ! parameters
  character(*), parameter :: RstBatFile  = '___restart.sh'

  ! subroutines
  public  :: reininp_modify

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  rein_inp_modify
  !> @brief      setup for modification of rein.inp
  !! @authors    NM
  !! @param[inout]  ctrl_data   control parameters in the program
  !! @param[in]  rein_inp    information of REIN input
  !! @date       2012/10/04 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reininp_modify(ctrl_data, rein_inp)

    ! formal arguments
    type(s_ctrl_data), intent(inout) :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp

    ! local variables
    character(100)     :: machine


    if (ctrl_data%main_info%skip_modify_input .eq. 'yes') &
      return

    ! difinite equiv and product steps
    if ( TRIM(ctrl_data%remd_info%replica_exchange) == "yes" ) then
      ctrl_data%remd_info%num_of_equiv_steps   = 0
      ctrl_data%remd_info%num_of_product_steps =    &
                        ctrl_data%remd_info%num_of_exchange_steps
    else
      ctrl_data%remd_info%num_of_equiv_steps   =    &
                        ctrl_data%remd_info%num_of_exchange_steps
      ctrl_data%remd_info%num_of_product_steps = 0
    end if

    ! process code of depends MD program
    if (ctrl_data%main_info%spawn .eq. 'yes') then

      if (ctrl_data%batch_info%num_threads_md >= 1 .and. &
          rein_inp%program_name .eq. "namd") &
        ctrl_data%batch_info%threads_1 = ctrl_data%batch_info%num_threads_md - 1

    else

      ctrl_data%batch_info%threads_1 = ctrl_data%batch_info%num_threads_md 

    end if

    call modify_inp(ctrl_data, rein_inp)

    return

  end subroutine reininp_modify

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  modify_inp
  !> @brief      build K batch script file
  !! @authors    NM
  !! @param[in]  ctrl_data   control parameters in the program
  !! @param[in]  rein_inp    information of REIN input
  !! @date       2012/10/04 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine modify_inp(ctrl_data, rein_inp)

    ! formal arguments
    type(s_ctrl_data), intent(in)    :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp

    ! local variables
    integer            :: bat_out, tmp_in, kw_index, i, iret, last_step
    character(200)     :: line
    character(200)     :: keyword, val, val2, val3, bat_file

    integer            :: system


    ! create batch file for preparing restart
    !
    bat_file = TRIM(ctrl_data%main_info%directory)//'/'//RstBatFile
    call open_file(bat_out, bat_file, IOFileOutputReplace)
    call open_file(tmp_in, TRIM(ctrl_data%main_info%template_dir)//&
                   '/t_reininp_modify', IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
      
      do while (check_replace(line, kw_index, keyword))
        
        select case(keyword)
          
        case ('MAIN_DIR')
          call replace(line, kw_index, keyword, ctrl_data%main_info%directory)

        case ('REIN_INPUT')
          call replace(line, kw_index, keyword, ctrl_data%main_info%rein_input)

        case ('FIRST_STEP')
          write(val,'(I0)') &
               rein_inp%first_step
          call replace(line, kw_index, keyword, val)

        case ('EQUIV_STEP')
          write(val,'(I0)') &
               ctrl_data%remd_info%num_of_equiv_steps
          call replace(line, kw_index, keyword, val)

        case ('PRODUCT_STEP')
          write(val,'(I0)') &
               ctrl_data%remd_info%num_of_product_steps
          call replace(line, kw_index, keyword, val)

        case ('NUM_MPIPROC_MD')
          write(val,'(I0)') &
               ctrl_data%batch_info%num_mpiproc_md
          call replace(line, kw_index, keyword, val)

        case ('PARALLEL_MOD')
          val = ''
          if ( TRIM (rein_inp%program_name) .eq. 'namd' ) then
            if ( TRIM (ctrl_data%main_info%spawn) .ne. 'yes' ) &
              val = ".\/charmrun"
          end if
          call replace(line, kw_index, keyword, val)

        case ('OPTION_1')
          val = ''
          if ( TRIM(rein_inp%program_name) .eq. 'namd') then
            if (ctrl_data%main_info%spawn .eq. 'yes') then
              write(val2,'(I0)') ctrl_data%batch_info%threads_1
              val = "+ppn " // TRIM(val2)
              if (ctrl_data%batch_info%num_threads_md == 1) &
                val = ""
            else
              write(val2,'(I0)') ctrl_data%batch_info%num_mpiproc_md
              write(val3,'(I0)') ctrl_data%batch_info%threads_1
              if (TRIM(ctrl_data%main_info%gpgpu ) .eq. 'yes') then
                val = "+idlepoll +p" // TRIM(val2)
              else
                write(val2,'(I0)') &
                 ctrl_data%batch_info%num_mpiproc_md
                val = "+p" // TRIM(val2) // " +ppn "//TRIM(val3)
              end if
              !if (ctrl_data%batch_info%num_threads_md == 1) &
              !  val = ""
            end if
          end if
          call replace(line, kw_index, keyword, val)

        case ('EXTEND_OP')
          val = ''
          call replace(line, kw_index, keyword, val)

        case default
          write(MsgOut,*) "Modify_Inp> Unknown template keyword : "//&
               TRIM(keyword)
          exit

        end select

      end do

      write(bat_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)
    call close_file(bat_out)


    ! run batch file for preparing restart
    !
    iret = system('sh '//TRIM(bat_file))
    if (iret /= 0) &
       call error_msg('Modify_Inp> Batch-file was failed.')

    iret = system('rm '//TRIM(bat_file))

    return

  end subroutine modify_inp

end module bb_reininp_modify_mod
