!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   bb_restart_mod
!> @brief   show usage of batch builder
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/08/02
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

module bb_restart_mod

  use bb_restart_num_mod
  use bb_reininp_mod
  use bb_ctrl_mod
  use replace_string_mod
  use fileio_control_mod
  use fileio_mod
  use messages_mod

  implicit none
  private

  ! parameters
  character(*), parameter :: RstBatFile  = '___restart.sh'

  ! subroutines
  public  :: restart
  private :: restart_main
  private :: get_last_step

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  restart
  !> @brief      setup for REIN restart
  !! @authors    NM, NT
  !! @param[inout]  ctrl_data   control parameters in the program
  !! @param[in]  rein_inp    information of REIN input
  !! @param[in]  rst_step    restart step number
  !! @date       2012/08/02 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine restart(ctrl_data, rein_inp, rst_step)

    ! formal arguments
    type(s_ctrl_data), intent(inout) :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp
    integer,           intent(in)    :: rst_step


    ! difinite equiv and product steps

    if ( TRIM( ctrl_data%remd_info%replica_exchange ) == "yes" ) then
      ctrl_data%remd_info%num_of_equiv_steps   = 0
      ctrl_data%remd_info%num_of_product_steps =    &
                        ctrl_data%remd_info%num_of_exchange_steps
    else
      ctrl_data%remd_info%num_of_equiv_steps   =    &
                        ctrl_data%remd_info%num_of_exchange_steps
      ctrl_data%remd_info%num_of_product_steps = 0
    end if


    call restart_main(ctrl_data, rein_inp, rst_step == 0)

    if (rst_step == 0) &
      return

    call restart_num(ctrl_data, rein_inp, rst_step)

    return

  end subroutine restart


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  restart_main
  !> @brief      build batch script file
  !! @authors    NM, NT
  !! @param[in]  ctrl_data   control parameters in the program
  !! @param[in]  rein_inp    information of REIN input
  !! @param[in]  show_msg    show messge and exit on not-restarable
  !! @date       2012/08/02 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine restart_main(ctrl_data, rein_inp, show_msg)

    ! formal arguments
    type(s_ctrl_data), intent(in)    :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp
    logical,           intent(in)    :: show_msg

    ! local variables
    integer            :: bat_out, tmp_in, kw_index, i, iret, last_step
    character(200)     :: line, job_scheduler
    character(200)     :: keyword, val, bat_file

    integer            :: system


    job_scheduler = ctrl_data%batch_info%job_scheduler
    call tolower(job_scheduler)


    ! check for restartable
    !
    last_step = get_last_step(ctrl_data, rein_inp)
    if (last_step == 0) then
      if (show_msg .eqv. .true.) then
        call error_msg('Restart_Main> output files are not found.'//&
                       ' Is the current job finish?')
      else
        return
      end if
    end if


    ! create batch file for preparing restart
    !
    bat_file = TRIM(ctrl_data%main_info%directory)//'/'//RstBatFile
    call open_file(bat_out, bat_file, IOFileOutputReplace)
    call open_file(tmp_in, &
         TRIM(ctrl_data%main_info%template_dir)//'/t_restart_'// &
         TRIM(job_scheduler), IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
      
      do while (check_replace(line, kw_index, keyword))
        
        select case(keyword)
          
        case ('MAIN_DIR')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%directory)

        case ('RESULT_DIR')
          write(val,'(A,I7.7,A,I7.7)') &
               'results/result', &
               rein_inp%first_step, '_', last_step
          call replace(line, kw_index, keyword, val)
               
        case ('REIN_INPUT')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%rein_input)

        case ('REIN_INPFILES')
          write(val,*) (TRIM(rein_inp%inp_files(i))//' ',&
               i=1,SIZE(rein_inp%inp_files))
          call replace(line, kw_index, keyword, val)

        case ('REIN_OUTFILES')
          write(val,*) (TRIM(rein_inp%out_files(i))//' ',&
               i=1,SIZE(rein_inp%out_files))
          call replace(line, kw_index, keyword, val)

        case ('REIN_INPFILE1')
          call replace(line, kw_index, keyword, &
                       rein_inp%inp_files(1))

        case ('REIN_INPFILE6')
          call replace(line, kw_index, keyword, &
                       rein_inp%inp_files(6))

        case ('REIN_OUTFILE1')
          call replace(line, kw_index, keyword, &
                       rein_inp%out_files(1))

        case ('REIN_OUTFILE2')
          call replace(line, kw_index, keyword, &
                       rein_inp%out_files(2))

        case ('NUM_REPLICAS')
          write(val,'(I0)') rein_inp%number_of_replicas
          call replace(line, kw_index, keyword, val)

        case ('REP_NAME')
          call replace(line, kw_index, keyword, &
                       rein_inp%io_file_name)

        case ('FIRST_STEP')
          write(val,'(I0)') &
               rein_inp%first_step
          call replace(line, kw_index, keyword, val)

        case ('LAST_STEP')
          write(val,'(I0)') last_step
          call replace(line, kw_index, keyword, val)

        case ('EQUIV_STEP')
          write(val,'(I0)') &
               ctrl_data%remd_info%num_of_equiv_steps
          call replace(line, kw_index, keyword, val)

        case ('PRODUCT_STEP')
          write(val,'(I0)') &
               ctrl_data%remd_info%num_of_product_steps
          call replace(line, kw_index, keyword, val)

        case ('NUM_MPIPROC_MD')
          write(val,'(I0)') &
               ctrl_data%batch_info%num_mpiproc_md
          call replace(line, kw_index, keyword, val)

        case ('STAGE_OUT_DIR')
          call replace(line, kw_index, keyword, &
                       ctrl_data%batch_info%stage_out_dir)

        case ('BATCHFILE_NAME')
          call replace(line, kw_index, keyword, &
                       ctrl_data%batch_info%batchfile_name)

        case default
          write(MsgOut,*) "Restart_Main> Unknown template keyword : "//&
               TRIM(keyword)
          exit

        end select

      end do

      write(bat_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)
    call close_file(bat_out)


    ! run batch file for preparing restart
    !
    iret = system('sh '//TRIM(bat_file))
    if (iret /= 0) &
       call error_msg('Restart_main> Batch-file restart was failed.')

    iret = system('rm '//TRIM(bat_file))

    return

  end subroutine restart_main


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function    get_last_step
  !> @brief      get last step from rank file
  !! @authors    NM, NT
  !! @return     integer last step number
  !! @date       2012/07/13 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function get_last_step(ctrl_data, rein_inp)

    ! return value
    integer :: get_last_step

    ! formal arguments
    type(s_ctrl_data), intent(in) :: ctrl_data
    type(s_rein_inp),  intent(in) :: rein_inp

    ! local variables
    integer            :: chk_in, i
    character(200)     :: chk_file, line

    integer            :: system


    get_last_step = 0

    ! check file exist
    chk_file    = TRIM(ctrl_data%main_info%directory)     //'/'// &
                  TRIM(ctrl_data%batch_info%stage_out_dir)//'/'// &
                  rein_inp%out_files(OutFileReplicaRank)

    i = system('test -e '//TRIM(chk_file))
    if (i /= 0) then
      
      chk_file  = TRIM(ctrl_data%main_info%directory)     //'/'// &
                  rein_inp%out_files(OutFileReplicaRank)

      i = system('test -e '//TRIM(chk_file))
      if (i /= 0) &
        return 

    end if

    ! check file contents
    call open_file(chk_in, chk_file, IOFileInput)

    do while (.true.)
      read(chk_in,'(A)',end=10) line
      i = SCAN(line, "=", .true.)
      if (i == 0) &
          cycle
      read(line(i+1:),*,err=90) get_last_step
    end do

10  call close_file(chk_in)

    return

90  call error_msg('Get_Last_Step> ERROR: Replica rank file is BAD format.')

  end function get_last_step


end module bb_restart_mod
