!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   bb_restart_num_mod
!> @brief   setup restart with step number
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/08/02
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module bb_restart_num_mod

  use bb_reininp_mod
  use bb_ctrl_mod
  use replace_string_mod
  use fileio_mod
  use messages_mod

  implicit none
  private

  ! parameters
  character(*), parameter :: RstBatFile  = '___restart.sh'

  ! subroutines
  public  :: restart_num

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  restart_num
  !> @brief      build GE batch script file
  !! @authors    NM, NT
  !! @param[in]  ctrl_data   control parameters in the program
  !! @param[in]  rein_inp    information of REIN input
  !! @param[in]  rst_step    restart step number
  !! @date       2012/08/02 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine restart_num(ctrl_data, rein_inp, rst_step)

    ! formal arguments
    type(s_ctrl_data), intent(in) :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp
    integer,           intent(in)    :: rst_step

    ! local variables
    integer            :: bat_out, tmp_in, kw_index, i, iret, iret2
    character(200)     :: line
    character(200)     :: keyword, val, bat_file

    integer            :: system


    ! create batch file for preparing restart
    !
    bat_file = TRIM(ctrl_data%main_info%directory)//'/'//RstBatFile
    call open_file(bat_out, bat_file, IOFileOutputReplace)
    call open_file(tmp_in, TRIM(ctrl_data%main_info%template_dir)//&
                   '/t_restart_num', IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
      
      do while (check_replace(line, kw_index, keyword))
        
        select case(keyword)

        case ('MAIN_DIR')
          call replace(line, kw_index, keyword, ctrl_data%main_info%directory)

        case ('RESULT_ROOT')
          call replace(line, kw_index, keyword, 'results')

        case ('REIN_INPUT')
          call replace(line, kw_index, keyword, ctrl_data%main_info%rein_input)

        case ('REIN_INPFILE1')
          call replace(line, kw_index, keyword, rein_inp%inp_files(1))

        case ('REIN_INPFILE6')
          call replace(line, kw_index, keyword, rein_inp%inp_files(6))

        case ('NUM_REPLICAS')
          write(val,'(I0)') rein_inp%number_of_replicas
          call replace(line, kw_index, keyword, val)

        case ('REP_NAME')
          call replace(line, kw_index, keyword, rein_inp%io_file_name)

        case ('EQUIV_STEP')
          write(val,'(I0)') &
               ctrl_data%remd_info%num_of_equiv_steps
          call replace(line, kw_index, keyword, val)

        case ('PRODUCT_STEP')
          write(val,'(I0)') &
               ctrl_data%remd_info%num_of_product_steps
          call replace(line, kw_index, keyword, val)

        case ('NUM_MPIPROC_MD')
          write(val,'(I0)') &
               ctrl_data%batch_info%num_mpiproc_md
          call replace(line, kw_index, keyword, val)

        case ('RESULT_DIR')
          write(val,'(A,I7.7,A)') &
               'results/result', rst_step, '_*[0-9]'
          call replace(line, kw_index, keyword, val)

        case ('STEP_NO')
          write(val,'(I0)') rst_step
          call replace(line, kw_index, keyword, val)

        case ('BATCHFILE_NAME')
          call replace(line, kw_index, keyword, ctrl_data%batch_info%batchfile_name)

        case default
          write(MsgOut,*) "Restart_num> Unknown template keyword : "//&
               TRIM(keyword)
          exit

        end select

      end do

      write(bat_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)
    call close_file(bat_out)


    ! run batch file for preparing restart
    !
    iret = system('sh '//TRIM(bat_file))
    iret2 = system('rm '//TRIM(bat_file))

    if (iret /= 0) &
       call error_msg('Restart_num> Batch-file restart was failed.')

    return

  end subroutine restart_num

end module bb_restart_num_mod
