!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   ib_build_file_d_mod
!> @brief   build REIN input files
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/06/11
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module ib_build_file_d_mod

  use ib_setup_mod
  use ib_ctrl_mod
  use fileio_mod
  use messages_mod

  implicit none
  private
  
  public  :: build_file_d
  private :: build_file_ex
  private :: build_file_value
  private :: build_file_on
  private :: build_file_rank
  private :: build_file_coef
  private :: build_file_axis_ex


  character(100),   parameter :: FileEx     = "file_ex.d"
  character(100),   parameter :: FileValue  = "file_value.d"
  character(100),   parameter :: FileOn     = "file_on.d"
  character(100),   parameter :: FileRank   = "file_rank.d"
  character(100),   parameter :: FileCoef   = "file_coef.d"
  character(100),   parameter :: FileAxisEx = "file_axis_ex.d"

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_file_d
  !> @brief      build file_*.d files
  !! @authors    NT, NM
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_file_d(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in)    :: ctrl_data
    type(s_build_info), intent(in)    :: build_info


    write(MsgOut,'(A)') 'Build_File_D>  Build file_*.d files.'

    call build_file_ex     (ctrl_data, build_info)
    call build_file_value  (ctrl_data, build_info)
    call build_file_on     (ctrl_data, build_info)
    call build_file_rank   (ctrl_data, build_info)
    call build_file_coef   (ctrl_data, build_info)
    call build_file_axis_ex(ctrl_data, build_info)

    write(MsgOut,'(A)') ' '

    return
  
  end subroutine build_file_d


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_file_ex
  !> @brief      build file_ex.d file
  !! @authors    NT,NM
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_file_ex(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in)    :: ctrl_data
    type(s_build_info), intent(in)    :: build_info

    ! local variables
    integer,            allocatable   :: ex(:,:), base(:,:), mm(:), ib(:)
    integer             :: naxis, nrep, nex, nrep_axis
    integer             :: file_out, i, j, jdm, idm
    character(200)      :: filename


    filename = TRIM(ctrl_data%main_info%directory)//'/'//FileEx

    if (check_file(filename)) then
      write(MsgOut,*) &
           ' WARNING: file is already exist. skip.. [', TRIM(filename), ']'
      return
    end if

    naxis = SIZE(build_info%axis_info)
    nrep  = build_info%replicas
    nex   = naxis * 2

    allocate(ex(nrep, nex), base(nrep, nex))
    allocate(mm(naxis), ib(naxis))

    mm(:) = 1
    
    do i = 1, nrep

      nrep_axis = build_info%axis_info(1)%replicas

      mm(1) = MOD(i, nrep_axis)

      if (mm(1) == 0) &
        mm(1) = nrep_axis

      ib(1) = MOD(mm(1), 2)
      if (ib(1) == 1) then
        base(i,1) = i + 1
        base(i,2) = 0
      else
        base(i,1) = 0
        base(i,2) = i + 1
      end if

      if (mm(1) == nrep_axis) then
        base(i,1) = 0
        base(i,2) = 0
      end if

      idm = 1
      do j = 2, naxis

        idm = idm * build_info%axis_info(j - 1)%replicas

        ib(j) = MOD(mm(j), 2)
        if (ib(j) == 1) then
          base(i, 2*j - 1) = i + idm
          base(i, 2*j)     = 0
        else
          base(i, 2*j - 1) = 0
          base(i, 2*j)     = i + idm
        end if
        !
        if (mm(j) == build_info%axis_info(j)%replicas) then
          base(i, 2*j - 1) = 0
          base(i, 2*j    ) = 0
        end if

        jdm = MOD(i, idm)
        if (jdm==0) &
          jdm=idm

        if (jdm == idm) &
          mm(j) = mm(j)+1

        if (mm(j) > build_info%axis_info(j)%replicas) then
          mm(j) = 1
        end if
      end do

      ex = base

    end do

    call open_file(file_out, filename, IOFileOutputReplace)

    do i = 1, nrep
      write(file_out,*) (ex(i,j), j = 1, nex, 2), &
                        (ex(i,j), j = 2, nex, 2)
    end do


    call close_file(file_out)

    deallocate(mm, ib)
    deallocate(ex, base)

    return
  
  end subroutine build_file_ex


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_file_value
  !> @brief      build file_value.d file
  !! @authors    NT,NM
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_file_value(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in)    :: ctrl_data
    type(s_build_info), intent(in)    :: build_info

    ! local variables
    double precision,   allocatable   :: v(:)
    integer             :: file_out, i, j, npval, axis_no
    character(200)      :: filename


    filename = TRIM(ctrl_data%main_info%directory)//'/'//FileValue

    if (check_file(filename)) then
      write(MsgOut,*) &
           ' WARNING: file is already exist. skip.. [', TRIM(filename), ']'
      return
    end if

    call open_file(file_out, filename, IOFileOutputReplace)

    npval = SIZE(build_info%pval_info)

    allocate(v(npval))

    do i = 1, build_info%replicas
      do j = 1, npval
        axis_no = build_info%pval_info(j)%axis_no
        if (axis_no == 0) then
          v(j) = build_info%pval_info(j)%value
        else
          v(j) = build_info%axis_info(axis_no)%values(i)
        end if
      end do
      write(file_out,'(20F12.5)') (v(j),j=1,npval)
    end do

    deallocate(v)

    call close_file(file_out)

    return
  
  end subroutine build_file_value


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_file_on
  !> @brief      build file_on.d file
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_file_on(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in)    :: ctrl_data
    type(s_build_info), intent(in)    :: build_info

    ! local variables
    integer             :: i, file_out, nrep
    character(200)      :: filename


    filename = TRIM(ctrl_data%main_info%directory)//'/'//FileOn

    if (check_file(filename)) then
      write(MsgOut,*) &
           ' WARNING: file is already exist. skip.. [', TRIM(filename), ']'
      return
    end if

    call open_file(file_out, filename, IOFileOutputReplace)

    nrep = build_info%replicas

    do i = 1, nrep
      write(file_out,*) 1, i, i
    end do

    call close_file(file_out)

    return
  
  end subroutine build_file_on


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_file_rank
  !> @brief      build file_rank.d file
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_file_rank(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in)    :: ctrl_data
    type(s_build_info), intent(in)    :: build_info

    ! local variables
    integer             :: i, file_out, nrep
    character(200)      :: filename


    filename = TRIM(ctrl_data%main_info%directory)//'/'//FileRank

    if (check_file(filename)) then
      write(MsgOut,*) &
           ' WARNING: file is already exist. skip.. [', TRIM(filename), ']'
      return
    end if

    call open_file(file_out, filename, IOFileOutputReplace)

    nrep = build_info%replicas

    do i = 1, nrep
      write(file_out,*) i
    end do

    call close_file(file_out)

    return
  
  end subroutine build_file_rank


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_file_coef
  !> @brief      build file_coef.d file
  !! @authors    NM, NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_file_coef(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in)    :: ctrl_data
    type(s_build_info), intent(in)    :: build_info

    ! local variables
    integer             :: i, file_out, nrep
    character(200)      :: filename


    filename = TRIM(ctrl_data%main_info%directory)//'/'//FileCoef

    if (check_file(filename)) then
      write(MsgOut,*) &
           ' WARNING: file is already exist. skip.. [', TRIM(filename), ']'
      return
    end if

    call open_file(file_out, filename, IOFileOutputReplace)

    nrep = build_info%replicas

    do i = 1, nrep
      select case (build_info%axis_info(1)%type)

      case (AxisTypeTemperature)
        write(file_out,*) build_info%axis_info(1)%values(i) / &
                          ctrl_data%md_info%md_temperature

      case default
        write(file_out,*) 1.d0

      end select

    end do

    call close_file(file_out)

    return
  
  end subroutine build_file_coef


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_file_axis_ex
  !> @brief      build file_axis_ex.d file
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_file_axis_ex(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in)    :: ctrl_data
    type(s_build_info), intent(in)    :: build_info

    ! local variables
    integer             :: i, j, file_out, naxis
    character(200)      :: filename


    filename = TRIM(ctrl_data%main_info%directory)//'/'//FileAxisEx

    if (check_file(filename)) then
      write(MsgOut,*) &
           ' WARNING: file is already exist. skip.. [', TRIM(filename), ']'
      return
    end if

    call open_file(file_out, filename, IOFileOutputReplace)

    naxis = SIZE(build_info%axis_info)

    write(file_out,'(20(I4))')((i,i = 1,naxis), j = 1,2)

    call close_file(file_out)

    return
  
  end subroutine build_file_axis_ex

end module ib_build_file_d_mod
