!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   ib_build_namd_mod
!> @brief   build REIN input files
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/06/11 
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

module ib_build_mdinp_mod

  use ib_setup_mod
  use ib_ctrl_mod
  use replace_string_mod
  use fileio_mod
  use messages_mod

  implicit none
  private
  
  public  :: build_mdinp
  private :: build_namd_inp
  private :: build_namd_colvar
  private :: build_namd_colvars
  private :: build_namd_harmonics
  private :: build_marble_inp
  private :: build_marble_groups
  private :: build_marble_group_h
  private :: build_marble_atom_h
  private :: make_marble_group_info

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_mdinp
  !> @brief      build MD input files
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/08 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_mdinp(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in) :: ctrl_data
    type(s_build_info), intent(in) :: build_info


    write(MsgOut,'(A,A)') 'Build_Mdinp>   Build MD input files. type : ', &
         TRIM(ctrl_data%md_info%md_program)

    if (ctrl_data%md_info%md_program == 'namd') then

      call build_namd_inp(ctrl_data, build_info)
      call build_namd_colvar(ctrl_data, build_info)

    else if (ctrl_data%md_info%md_program == 'marble') then

      call build_marble_inp(ctrl_data, build_info)

    end if

    write(MsgOut,'(A)') ' '

    return
  
  end subroutine build_mdinp


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_namd_inp
  !> @brief      build NAMD input files
  !! @authors    NM,NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/08 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_namd_inp(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in) :: ctrl_data
    type(s_build_info), intent(in) :: build_info

    ! local variables
    integer             :: tmp_in, inp_out, i
    integer             :: kw_index
    character(200)      :: tmp_file, inp_file
    character(200)      :: line, keyword, val


    inp_file = TRIM(ctrl_data%main_info%directory)//'/'//&
               TRIM(ctrl_data%md_info%init_md_file)//'.inp'
    tmp_file = TRIM(ctrl_data%main_info%template_dir)//'/t_namd.inp'

    call open_file(inp_out, inp_file, IOFileOutputReplace)
    call open_file(tmp_in,  tmp_file, IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)

        case ('FINALSTEP')
          write(val,'(I0)') ctrl_data%md_info%md_steps
          call replace(line, kw_index, keyword, val)

        case ('OUTSTEP')
          write(val,'(I0)') ctrl_data%md_info%md_steps / 10
          call replace(line, kw_index, keyword, val)

        case ('NAME')
          write(val,'(A,A)') '../', ctrl_data%md_info%init_md_file
          call replace(line, kw_index, keyword, val)

        case ('SET_PARAM1_SETTINGS')
          do i = 1, SIZE(build_info%axis_info)
            if (build_info%axis_info(i)%type == AxisTypeTemperature) &
              exit
          end do

          if (i <= SIZE(build_info%axis_info)) then
            write(inp_out, '(A)') 'rein_control: 1 1 "set param1 "'
          else
            write(inp_out, '(A)') 'rein_baseT: "set param1 "'
          end if

          line = '<'

        case ('PME_SETTINGS')
          write(inp_out, '(A,I0)') &
               'PMEGridSizeX         ', ctrl_data%md_info%pme_grid_size_x
          write(inp_out, '(A,I0)') &
               'PMEGridSizeY         ', ctrl_data%md_info%pme_grid_size_y
          write(inp_out, '(A,I0)') &
               'PMEGridSizeZ         ', ctrl_data%md_info%pme_grid_size_z

          line = '<'

        case ('COLVAR_SETTINGS')
          do i = 1, SIZE(build_info%axis_info)
            if (build_info%axis_info(i)%type == AxisTypeHarmonic) then
              write(inp_out, '(A)') ' '
              write(inp_out, '(A)') 'colvars on'
              write(inp_out, '(A)') 'rein_colvar: "colvarsConfig "'
              exit
            end if
          end do

          line = '<'

        case default
          write(MsgOut,*) 'Build_Namd_Inp> Unknown template keyword : '//&
               TRIM(keyword)
          exit

        end select

      end do

      if (line /= '<') &
        write(inp_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)
    call close_file(inp_out)

    return
  
  end subroutine build_namd_inp


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_namd_colvar
  !> @brief      build NAMD colvar input files
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/08 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_namd_colvar(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in) :: ctrl_data
    type(s_build_info), intent(in) :: build_info

    ! local variables
    integer             :: tmp_in, col_out, i
    integer             :: kw_index
    character(200)      :: tmp_file, col_file
    character(200)      :: line, keyword, val


    ! check colvar needed or not
    do i = 1, SIZE(build_info%axis_info)
      if (build_info%axis_info(i)%harmo_type /= HarmoTypeNone) &
        exit
    end do
    if (i == SIZE(build_info%axis_info) + 1) &
      return

    col_file = TRIM(ctrl_data%main_info%directory)//'/'//&
               TRIM(ctrl_data%md_info%init_md_file)//'.colvar'
    tmp_file = TRIM(ctrl_data%main_info%template_dir)//'/t_namd.colvar'

    call open_file(col_out, col_file, IOFileOutputReplace)
    call open_file(tmp_in,  tmp_file, IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)

        case ('COLVAR_TRJ_FREQ')
          write(val,'(I0)') ctrl_data%md_info%md_steps
          call replace(line, kw_index, keyword, val)

        case ('COLVARS')
          do i = 1, SIZE(build_info%axis_info)
            if (build_info%axis_info(i)%type == AxisTypeTemperature) &
              cycle
            call build_namd_colvars(ctrl_data,               &
                                    build_info%pval_info,    &
                                    build_info%axis_info(i), &
                                    col_out)
          end do
          line = '<'

        case ('HARMONICS')
          do i = 1, SIZE(build_info%axis_info)
            if (build_info%axis_info(i)%type == AxisTypeTemperature) &
              cycle
            call build_namd_harmonics(ctrl_data,             &
                                    build_info%pval_info,    &
                                    build_info%axis_info(i), &
                                    col_out)
          end do
          line = '<'

        case default
          write(MsgOut,*) 'Build_Namd_Colvar> Unknown template keyword : '//&
               TRIM(keyword)
          exit

        end select

      end do

      if (line /= '<') &
        write(col_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)
    call close_file(col_out)

    return

  end subroutine build_namd_colvar


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_namd_colvars
  !> @brief      build NAMD colvar file colvar blocks
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  pval_info  param values information
  !! @param[in]  axis_info  axis information
  !! @param[in]  col_out    colvar file unit no
  !! @date       2012/06/08 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_namd_colvars(ctrl_data, pval_info, axis_info, col_out)

    ! formal arguments
    type(s_ctrl_data), intent(in) :: ctrl_data
    type(s_pval_info), pointer    :: pval_info(:)
    type(s_axis_info), intent(in) :: axis_info
    integer,           intent(in) :: col_out

    ! local variables
    integer            :: tmp_in
    integer            :: kw_index
    character(200)     :: dir, name, line, keyword, val


    if (axis_info%harmo_item1(1) == '' .or. &
        axis_info%harmo_item2(1) == '' .or. &
        axis_info%harmo_item3(1) == '') then
      call error_msg('Build_Namd_Colvars> ERROR: Bad harmonic setting. ')
    end if

    select case(axis_info%harmo_type)
    case (HarmoTypeNone)
      call error_msg('Build_Namd_Colvars> ERROR: unexpected. ')
    case (HarmoTypeDistance)
      name = 'dis'
    case (HarmoTypeAngle)
      name = 'ang'
    case (HarmoTypeDihedral)
      name = 'dih'
    end select

    dir = ctrl_data%main_info%template_dir

    call open_file(tmp_in, &
                   TRIM(dir)//'/t_namd_colvar_'//TRIM(name), &
                   IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)

        case ('NAME')
          write(val,'(A,I0)') TRIM(name), axis_info%index
          call replace(line, kw_index, keyword, val)

        case ('RESI_1')
          call replace(line, kw_index, keyword, axis_info%harmo_item1(1))

        case ('RESI_2')
          call replace(line, kw_index, keyword, axis_info%harmo_item1(2))

        case ('RESI_3')
          call replace(line, kw_index, keyword, axis_info%harmo_item1(3))

        case ('RESI_4')
          call replace(line, kw_index, keyword, axis_info%harmo_item1(4))

        case ('ATOM_1')
          call replace(line, kw_index, keyword, axis_info%harmo_item2(1))

        case ('ATOM_2')
          call replace(line, kw_index, keyword, axis_info%harmo_item2(2))

        case ('ATOM_3')
          call replace(line, kw_index, keyword, axis_info%harmo_item2(3))

        case ('ATOM_4')
          call replace(line, kw_index, keyword, axis_info%harmo_item2(4))

        case ('ATOM_NO_1')
          call replace(line, kw_index, keyword, axis_info%harmo_item3(1))

        case ('ATOM_NO_2')
          call replace(line, kw_index, keyword, axis_info%harmo_item3(2))

        case ('ATOM_NO_3')
          call replace(line, kw_index, keyword, axis_info%harmo_item3(3))

        case ('ATOM_NO_4')
          call replace(line, kw_index, keyword, axis_info%harmo_item3(4))

        case default
          write(MsgOut,*) 'Build_Namd_Colvars> Unknown template keyword : '//&
               TRIM(keyword)
          exit

        end select

      end do

      if (line /= '<') &
        write(col_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)

    return

  end subroutine build_namd_colvars


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_namd_harmonics
  !> @brief      build NAMD colvar file harmonics blocks
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  pval_info  param values information
  !! @param[in]  axis_info  axis information
  !! @param[in]  col_out    colvar file unit no
  !! @date       2012/06/08 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_namd_harmonics(ctrl_data, pval_info, axis_info, col_out)

    ! formal arguments
    type(s_ctrl_data), intent(in) :: ctrl_data
    type(s_pval_info), pointer    :: pval_info(:)
    type(s_axis_info), intent(in) :: axis_info
    integer,           intent(in) :: col_out

    ! local variables
    integer            :: tmp_in, i
    integer            :: kw_index
    character(200)     :: dir, name, line, keyword, val


    dir = ctrl_data%main_info%template_dir

    select case(axis_info%harmo_type)
    case (HarmoTypeNone)
      call error_msg('Build_Namd_Harmonics> ERROR: unexpected. ')
    case (HarmoTypeDistance)
      name = 'dis'
    case (HarmoTypeAngle)
      name = 'ang'
    case (HarmoTypeDihedral)
      name = 'dih'
    end select

    call open_file(tmp_in, &
                   TRIM(dir)//'/t_namd_harmonic',&
                   IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)

        case ('NAME')
          write(val,'(A,A,I0)') 'harmo_', TRIM(name), axis_info%index
          call replace(line, kw_index, keyword, val)

        case ('COLVAR_NAME')
          write(val,'(A,I0)') TRIM(name), axis_info%index
          call replace(line, kw_index, keyword, val)

        case ('AXIS_NO')
          write(val,'(I0)') axis_info%index
          call replace(line, kw_index, keyword, val)

        case ('COL_NO')
          do i = 1, SIZE(pval_info)
            if (axis_info%index == pval_info(i)%axis_no) &
              exit
          end do
          if (i <= SIZE(pval_info)) then
            write(val,'(I0)') i
          else
            write(val,'(I0)') 0
          end if
          call replace(line, kw_index, keyword, val)

        case ('COL_NO+1')
          do i = 1, SIZE(pval_info)
            if (axis_info%index == pval_info(i)%axis_no) &
              exit
          end do
          if (i <= SIZE(pval_info)) then
            write(val,'(I0)') i+1
          else
            write(val,'(I0)') 1
          end if
          call replace(line, kw_index, keyword, val)

        case default
          write(MsgOut,*) &
               'Build_Namd_Harmonics> Unknown template keyword : '//&
               TRIM(keyword)
          exit

        end select

      end do

      if (line /= '<') &
        write(col_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)

    return

  end subroutine build_namd_harmonics


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_marble_inp
  !> @brief      build MARBLE input files
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/20 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_marble_inp(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),   intent(in) :: ctrl_data
    type(s_build_info),  intent(in) :: build_info

    ! local variables
    integer              :: tmp_in, inp_out, i, ngroup
    integer              :: kw_index
    character(200)       :: tmp_file, inp_file
    character(200)       :: line, keyword, val, group

    integer,             pointer :: group_no(:,:)
    character(200),      pointer :: groups(:)


    inp_file = TRIM(ctrl_data%main_info%directory)//'/'//&
               TRIM(ctrl_data%md_info%init_md_file)//'.inp'
    tmp_file = TRIM(ctrl_data%main_info%template_dir)//'/t_marble.inp'

    call open_file(inp_out, inp_file, IOFileOutputReplace)
    call open_file(tmp_in,  tmp_file, IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)

        case ('FINALSTEP')
          write(val,'(I0)') ctrl_data%md_info%md_steps
          call replace(line, kw_index, keyword, val)

        case ('OUTSTEP')
          write(val,'(I0)') ctrl_data%md_info%md_steps / 10
          call replace(line, kw_index, keyword, val)

        case ('MDATFILE')
          write(val,'(A,A,A)') &
               './', TRIM(ctrl_data%md_info%init_md_file), '.mdat'
          call replace(line, kw_index, keyword, val)

        case ('TEMPERATURE_SETTINGS')
          do i = 1, SIZE(build_info%axis_info)
            if (build_info%axis_info(i)%type == AxisTypeTemperature) &
              exit
          end do

          if (i <= SIZE(build_info%axis_info)) then
            write(inp_out, '(A)') 'rein_control: 1 1 "  temperature = "'
          else
            write(inp_out, '(A)') 'rein_baseT: "  temperature = "'
          end if

          line = '<'

        case ('HARMONIC_SETTINGS')

          do i = 1, SIZE(build_info%axis_info)
            if (build_info%axis_info(i)%type == AxisTypeHarmonic) &
              exit
          end do

          if (i > SIZE(build_info%axis_info)) then
            line = '<'
            exit
          end if

          ! build group information
          call make_marble_group_info(build_info, groups, group_no, ngroup)


          ! group section
          if (ngroup > 0) then

            write(inp_out, '(A)') ' '
            write(inp_out, '(A)') '[group]'

            do i = 1, ngroup

              if (groups(i) == '') &
                exit
          
              call build_marble_groups(ctrl_data, i-1, groups(i), inp_out)

            end do

          end if


          ! restraint section
          write(inp_out, '(A)') ' '
          write(inp_out, '(A)') '[restraint]'

          do i = 1, SIZE(build_info%axis_info)

            if (build_info%axis_info(i)%type /= AxisTypeHarmonic) &
              cycle

            if (build_info%axis_info(i)%harmo_item3(1) /= '') then

              call build_marble_group_h(ctrl_data,               &
                                        build_info%pval_info,    &
                                        build_info%axis_info(i), &
                                        group_no(:,i), inp_out)
            else

              call build_marble_atom_h (ctrl_data,               &
                                        build_info%pval_info,    &
                                        build_info%axis_info(i), &
                                        inp_out)
            end if

          end do

          deallocate(group_no, groups)

          line = '<'

        case ('PME_SETTINGS')

          write(inp_out, '(A)') '[ewald]'
          write(inp_out, '(A,I0,1X,I0,1X,I0)') &
               '  grid = ', &
               ctrl_data%md_info%pme_grid_size_x, &
               ctrl_data%md_info%pme_grid_size_y, &
               ctrl_data%md_info%pme_grid_size_z

          line = '<'

        case default
          write(MsgOut,*) 'Build_Marble_Inp> Unknown template keyword : '//&
               TRIM(keyword)
          exit

        end select

      end do

      if (line /= '<') &
        write(inp_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)
    call close_file(inp_out)

    return
  
  end subroutine build_marble_inp


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_marble_groups
  !> @brief      build MARBLE group section
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  group_no   group number
  !! @param[in]  group_name group name
  !! @param[in]  mdinp_out  md input file unit no
  !! @date       2012/07/20 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_marble_groups(ctrl_data, group_no, group_name, mdinp_out)

    ! formal arguments
    type(s_ctrl_data),   intent(in) :: ctrl_data
    integer,             intent(in) :: group_no
    character(*),        intent(in) :: group_name
    integer,             intent(in) :: mdinp_out

    ! local variables
    integer            :: tmp_in, i
    integer            :: kw_index
    character(200)     :: dir, line, keyword, val


    dir = ctrl_data%main_info%template_dir

    call open_file(tmp_in, &
                   TRIM(dir)//'/t_marble_group',&
                   IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)

        case ('GROUP_NO')
          write(val,'(I0)') group_no
          call replace(line, kw_index, keyword, val)

        case ('GROUP_NAME')
          call replace(line, kw_index, keyword, group_name)

        end select

      end do

      write(mdinp_out,'(A)') TRIM(line)

    end do

10  close(tmp_in)

    return

  end subroutine build_marble_groups


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_marble_group_h
  !> @brief      build MARBLE restraint section
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  pval_info  parameter values information
  !! @param[in]  axis_info  axis information
  !! @param[in]  group_no   group numbers
  !! @param[in]  mdinp_out  md input file unit no
  !! @date       2012/07/20 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_marble_group_h(ctrl_data, &
                                  pval_info, axis_info, group_no, mdinp_out)

    ! formal arguments
    type(s_ctrl_data),   intent(in) :: ctrl_data
    type(s_pval_info),   pointer    :: pval_info(:)
    type(s_axis_info),   intent(in) :: axis_info
    integer,             intent(in) :: group_no(:)
    integer,             intent(in) :: mdinp_out

    ! local variables
    integer            :: tmp_in, i, ndim
    integer            :: kw_index
    character(200)     :: dir, line, keyword, val, method


    select case(axis_info%harmo_type)

    case (HarmoTypeDistance)
      method = 'group_distance_harmonic'
      ndim   = 2
    case (HarmoTypeAngle)
      call error_msg('Build_Marble_Group_H> ERROR : '// &
                          'Group angle harmonic is not supported.')
      return
    case (HarmoTypeDihedral)
      call error_msg('Build_Marble_Group_H> ERROR : '// &
                          'Group dihedral harmonic is not supported.')
    end select

    dir = ctrl_data%main_info%template_dir

    call open_file(tmp_in, &
                   TRIM(dir)//'/t_marble_group_h',&
                   IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)

        case ('METHOD')
          call replace(line, kw_index, keyword, method)

        case ('GROUPS')
          write(val,'(4(1X,I0))') (group_no(i)-1,i=1,ndim)
          call replace(line, kw_index, keyword, val)

        case ('AXIS_NO')
          write(val,'(I0)') axis_info%index
          call replace(line, kw_index, keyword, val)

        case ('COL_NO')
          do i = 1, SIZE(pval_info)
            if (axis_info%index == pval_info(i)%axis_no) &
              exit
          end do
          if (i <= SIZE(pval_info)) then
            write(val,'(I0)') i
          else
            write(val,'(I0)') 0
          end if
          call replace(line, kw_index, keyword, val)

        case ('COL_NO+1')
          do i = 1, SIZE(pval_info)
            if (axis_info%index == pval_info(i)%axis_no) &
              exit
          end do
          if (i <= SIZE(pval_info)) then
            write(val,'(I0)') i+1
          else
            write(val,'(I0)') 1
          end if
          call replace(line, kw_index, keyword, val)

        end select

      end do

      write(mdinp_out,'(A)') TRIM(line)

    end do

10  close(tmp_in)

    return

  end subroutine build_marble_group_h


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_marble_atom_h
  !> @brief      build MARBLE restraint section
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  pval_info  parameter values information
  !! @param[in]  axis_info  axis information
  !! @param[in]  mdinp_out  md input file unit no
  !! @date       2012/08/01 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_marble_atom_h(ctrl_data, pval_info, axis_info, mdinp_out)

    ! formal arguments
    type(s_ctrl_data),   intent(in) :: ctrl_data
    type(s_pval_info),   pointer    :: pval_info(:)
    type(s_axis_info),   intent(in) :: axis_info
    integer,             intent(in) :: mdinp_out

    ! local variables
    integer            :: tmp_in, i
    integer            :: kw_index
    character(200)     :: dir, line, keyword, val, method


    select case(axis_info%harmo_type)

    case (HarmoTypeDistance)
      method = 'distance_harmonic'
    case (HarmoTypeDihedral)
      method = 'torsion_harmonic'
    case (HarmoTypeAngle)
      call error_msg('Build_marble_atom_h> ERROR : '// &
                     'Angle harmonic is not supported.')
    end select

    dir = ctrl_data%main_info%template_dir

    call open_file(tmp_in, &
                   TRIM(dir)//'/t_marble_atom_h',&
                   IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)

        case ('METHOD')
          call replace(line, kw_index, keyword, method)

        case ('ATOMS')
          write(val,'(4(A,1X))') &
               (TRIM(axis_info%harmo_item1(i)),i=1,SIZE(axis_info%harmo_item1))
          call replace(line, kw_index, keyword, val)

        case ('AXIS_NO')
          write(val,'(I0)') axis_info%index
          call replace(line, kw_index, keyword, val)

        case ('COL_NO')
          do i = 1, SIZE(pval_info)
            if (axis_info%index == pval_info(i)%axis_no) &
              exit
          end do
          if (i <= SIZE(pval_info)) then
            write(val,'(I0)') i
          else
            write(val,'(I0)') 0
          end if
          call replace(line, kw_index, keyword, val)

        case ('COL_NO+1')
          do i = 1, SIZE(pval_info)
            if (axis_info%index == pval_info(i)%axis_no) &
              exit
          end do
          if (i <= SIZE(pval_info)) then
            write(val,'(I0)') i+1
          else
            write(val,'(I0)') 1
          end if
          call replace(line, kw_index, keyword, val)

        end select

      end do

      write(mdinp_out,'(A)') TRIM(line)

    end do

10  close(tmp_in)

    return

  end subroutine build_marble_atom_h


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  make_marble_group_info
  !> @brief      make MARBLE group information
  !! @authors    NT
  !! @param[in]  build_info build information
  !! @param[in]  groups     group names
  !! @param[in]  group_no   group numbers
  !! @param[in]  ngroup     number of make groups
  !! @date       2012/08/01 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_marble_group_info(build_info, groups, group_no, ngroup)

    ! formal arguments
    type(s_build_info),  intent(in)    :: build_info
    character(*),        pointer       :: groups(:)
    integer,             pointer       :: group_no(:,:)
    integer,             intent(inout) :: ngroup

    ! local variables
    integer              :: i, j, jj
    character(200)       :: group


    allocate(groups  (   SIZE(build_info%axis_info)*4), &
             group_no(4, SIZE(build_info%axis_infO)))
    groups(:) = ''
    group_no(:,:) = 0

    ! group section
    do i = 1, SIZE(build_info%axis_info)

      if (build_info%axis_info(i)%type /= AxisTypeHarmonic .or. &
          build_info%axis_info(i)%harmo_item3(1) == '') &
           cycle

      do j = 1, SIZE(build_info%axis_info(i)%harmo_item1)

        group = TRIM(build_info%axis_info(i)%harmo_item1(j))//' '//&
                TRIM(build_info%axis_info(i)%harmo_item2(j))//' '//&
                TRIM(build_info%axis_info(i)%harmo_item3(j))

        do jj = 1, SIZE(groups)
          if (groups(jj) == '') then

            groups(jj) = group
            exit

          else if (groups(jj) == group) then
            exit

          end if

        end do

        group_no(j, i) = jj

      end do

    end do

    ngroup = 0

    do i = 1, SIZE(groups)
      if (groups(i) == '') &
        exit
      ngroup = ngroup + 1
    end do

    return

  end subroutine make_marble_group_info

end module ib_build_mdinp_mod
