!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   ib_build_reininp_mod
!> @brief   build REIN input files
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/06/11 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module ib_build_reininp_mod

  use ib_setup_mod
  use ib_ctrl_mod
  use replace_string_mod
  use fileio_control_mod
  use fileio_mod
  use messages_mod

  implicit none
  private
  
  public  :: build_reininp
  private :: build_reininp_axes


contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_reininp
  !> @brief      build REIN input file
  !! @authors    NM, NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/08 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_reininp(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in) :: ctrl_data
    type(s_build_info), intent(in) :: build_info

    ! local variables
    integer             :: tmp_in, inp_out, i
    integer             :: kw_index
    character(200)      :: tmp_file, inp_file, str
    character(200)      :: line, keyword, val


    write(MsgOut,'(A)') 'Build_Reininp> Build REIN input file.'

    inp_file = TRIM(ctrl_data%main_info%directory)//'/'//&
                    ctrl_data%main_info%rein_input
    tmp_file = TRIM(ctrl_data%main_info%template_dir)//&
                    '/t_rein.inp'

    call open_file(inp_out, inp_file, IOFileOutputReplace)
    call open_file(tmp_in,  tmp_file, IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
    
      do while (check_replace(line, kw_index, keyword))

        select case(keyword)


        case ('PARALLEL_MODULE_MD')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%parallel_module_md)

        case ('OPTION_MD')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%option_md)

        case ('EXTEND_OPTION_MD')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%extend_option_md)

        case ('MD_PROGRAM')
          call replace(line, kw_index, keyword, &
                       ctrl_data%md_info%md_program)

        case ('MD_EXE_FILE')
          call replace(line, kw_index, keyword, &
                       ctrl_data%md_info%md_exe_file)

        case ('VELOCITY_RESCALING')
          str = ctrl_data%md_info%md_program
          call tolower(str)
          if (str == 'namd') then
            call replace(line, kw_index, keyword, 'v_namd_bin')
          else if (str == 'marble') then
            call replace(line, kw_index, keyword, 'v_marble')
          end if
          
        case ('MD_TEMPERATURE')
          write(val,'(F0.3)') ctrl_data%md_info%md_temperature
          call replace(line, kw_index, keyword, val)

        case ('FIRST_STEP')
          write(val,'(I0)') ctrl_data%remd_info%first_step
          call replace(line, kw_index, keyword, val)

        case ('EQUIV_STEPS')
          write(val,'(I0)') ctrl_data%remd_info%equiv_steps
          call replace(line, kw_index, keyword, val)

        case ('PRODUCT_STEPS')
          write(val,'(I0)') ctrl_data%remd_info%product_steps
          call replace(line, kw_index, keyword, val)

        case ('INIT_MD_FILE')
          call replace(line, kw_index, keyword, ctrl_data%md_info%init_md_file)

        case ('NUM_AXIS')
          write(val,'(I0)') SIZE(build_info%axis_info)
          call replace(line, kw_index, keyword, val)

        case ('NUM_EXCHANGE')
          write(val,'(I0)') SIZE(build_info%axis_info) * 2
          call replace(line, kw_index, keyword, val)

        case ('AXES')
          call build_reininp_axes(ctrl_data, build_info, inp_out)
          line = '<'

        case ('CORE_FOR_MD')
          write(val,'(I0)') ctrl_data%md_info%core_for_md
          call replace(line, kw_index, keyword, val)

        case default
          write(MsgOut,*) 'Build_Reininp> Unknown template keyword : '//&
               TRIM(keyword)
          exit

        end select

      end do

      if (line /= '<') &
        write(inp_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)
    call close_file(inp_out)

    write(MsgOut,'(A)') ' '

    return
  
  end subroutine build_reininp


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  build_reininp_axes
  !> @brief      build REIN input file
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[in]  build_info build information
  !! @date       2012/06/08 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_reininp_axes(ctrl_data, build_info, inp_out)

    ! formal arguments
    type(s_ctrl_data),  intent(in) :: ctrl_data
    type(s_build_info), intent(in) :: build_info
    integer,            intent(in) :: inp_out

    ! local variables
    integer             :: tmp_in, i, j, jj, k, kw_index
    character(200)      :: tmp_file
    character(200)      :: line, keyword, val

    ! parameters
    character(*), parameter :: AxisTypeStr(2) = (/'temperature', &
                                                  'harmonic   '/)
    integer,      parameter :: AxisNParam(2)  = (/1,2/)


    tmp_file = TRIM(ctrl_data%main_info%template_dir)//&
                    '/t_rein_axis.inp'

    call open_file(tmp_in,  tmp_file, IOFileInput)

    do i = 1, SIZE(build_info%axis_info)

      do while (.true.)

        read(tmp_in,'(A)',end=10) line
    
        do while(check_replace(line, kw_index, keyword))

          select case(keyword)

          case ('AXIS_NO')
            write(val,'(I0)') i
            call replace(line, kw_index, keyword, val)

          case ('AXIS_TYPE')
            call replace(line, kw_index, keyword, &
                 AxisTypeStr(build_info%axis_info(i)%type))

          case ('NUM_REPLICAS')
            write(val,'(I0)') build_info%axis_info(i)%replicas
            call replace(line, kw_index, keyword, val)

          case ('AXIS_CTRL_PARAM')
            do j = 1, SIZE(build_info%pval_info)
              if (i == build_info%pval_info(j)%axis_no) &
                exit
            end do
            write(val,'(I0)') j
            call replace(line, kw_index, keyword, val)

          case ('AXIS_NUM_OF_PARAM')
            write(val,'(I0)') AxisNParam(build_info%axis_info(i)%type)
            call replace(line, kw_index, keyword, val)

          case ('AXIS_PARAMS')
            do j = 1, SIZE(build_info%pval_info)
              if (i == build_info%pval_info(j)%axis_no) &
                exit
            end do

            jj = j + AxisNParam(build_info%axis_info(i)%type) - 1
            do k = j, jj
              write(inp_out,'(A,I0)') 'axis_param              = ', k
            end do
            line = '<'

          case ('PERIODIC')
            if (build_info%axis_info(i)%periodic .eqv. .true.) then
              write(inp_out,'(A)')  'periodic                = "on"'
            else
              write(inp_out,'(A)')  'periodic                = "off"'
            end if
            write(inp_out,'(A,F0.3)') 'periodic_value_min      = ', &
                 build_info%axis_info(i)%value_lo
            write(inp_out,'(A,F0.3)') 'periodic_value_max      = ', &
                 build_info%axis_info(i)%value_hi
            line = '<'

          case default
            write(MsgOut,*) 'Build_Reininp> Unknown template keyword : '//&
                 TRIM(keyword)
            exit

          end select

        end do

        if (line /= '<') &
          write(inp_out,'(A)') TRIM(line)

      end do

10    REWIND(tmp_in)

    end do

    call close_file(tmp_in)

    return

  end subroutine build_reininp_axes

end module ib_build_reininp_mod
