!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   ib_copy_file_mod
!> @brief   copy file operations for MD input
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/06/11 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module ib_copy_file_mod

  use ib_setup_mod
  use ib_ctrl_mod
  use replace_string_mod
  use fileio_control_mod
  use fileio_mod
  use messages_mod

  implicit none
  private
  
  character(*),   parameter :: CopyBatFile = '___copy.sh'

  public  :: check_files
  public  :: copy_files

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  check_files
  !> @brief      check files existence in working directory
  !! @authors    NM, NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @date       2012/06/11 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_files(ctrl_data)

    ! formal arguments
    type(s_ctrl_data),  intent(in) :: ctrl_data

    ! local variables
    integer             :: handle, nfile, kw_index
    character(200)      :: tmp_file, dir
    character(100)      :: md_program, key, val
    character(30)       :: valL, valR
    logical             :: abort

    integer             :: system, iret

    
    dir = ctrl_data%main_info%directory

    md_program = ctrl_data%md_info%md_program
    call tolower(md_program)
    write(MsgOut,'(A)') 'Check_Files> check files for : '//TRIM(md_program)


    tmp_file = TRIM(ctrl_data%main_info%template_dir)// &
               '/t_'//TRIM(md_program)//'_copyfiles'

    call open_ctrlfile(tmp_file, handle)

    nfile = 1
    abort = .false.

    do while (.true.)
      val = ''
      write(key, fmt='(a4,i3.3)') 'file', nfile
      call read_ctrlfile_string(handle, 'CHECK_FILES', key, val)
      if (val == '') then
        exit
      end if

      valL = ''
      valR = ''
      read(val,*,end=10) valL, valR
10    call tolower(valR)

      do while (check_replace(valL, kw_index, key))
        select case(key)
          
        case ('MD_EXE_FILE')
          call replace(valL, kw_index, key, &
                      ctrl_data%md_info%md_exe_file)
        case ('INIT_MD_FILE')
          call replace(valL, kw_index, key, &
                       ctrl_data%md_info%init_md_file)
        case default
          write(MsgOut,*) "Check_Files> Unknown template keyword : "//&
               TRIM(key)
          exit

        end select
      end do

      if (valL /= '') then
        iret = system('test -f '//TRIM(dir)//'/'//valL)
        if (iret == 0) then
          write(MsgOut,'(A,A,A)') '   ', valL, ': found.'
        else
          write(MsgOut,'(A,A,A)') '   ', valL, ': not found.'
          if (valR == 'required') &
               abort = .true.
        end if
      end if

      nfile = nfile + 1
    end do

    call close_ctrlfile(handle)

    write(MsgOut,'(A)') ' '

    if (abort .eqv. .true.) then
      write(MsgOut,'(A)') 'Check_Files> ERROR: cannot continue processing.'
      write(MsgOut,'(A)') ' '
      stop
    end if

    return

  end subroutine check_files


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  copy_files
  !> @brief      copy working directory files to replica directory
  !! @authors    NM, NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @date       2012/06/11 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine copy_files(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in) :: ctrl_data
    type(s_build_info), intent(in) :: build_info

    ! local variables
    integer             :: bat_out, tmp_in, kw_index
    character(200)      :: tmp_file, bat_file
    character(200)      :: line, keyword, val
    character(100)      :: md_program

    integer             :: system, iret


    ! create batch file for preparing restart
    !
    md_program = ctrl_data%md_info%md_program
    call tolower(md_program)
    write(MsgOut,'(A)') 'Copy_Files> copy files for : '//TRIM(md_program)


    bat_file = TRIM(ctrl_data%main_info%directory)//'/'//CopyBatFile

    tmp_file = TRIM(ctrl_data%main_info%template_dir)// &
               '/t_'//TRIM(md_program)//'_copyfiles'

    call open_file(bat_out, bat_file, IOFileOutputReplace)
    call open_file(tmp_in, tmp_file, IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
      
      do while (check_replace(line, kw_index, keyword))
        
        select case(keyword)
          
        case ('MAIN_DIR')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%directory)

        case ('INIT_MD_FILE')
          call replace(line, kw_index, keyword, &
                       ctrl_data%md_info%init_md_file)

        case ('NUM_REPLICAS')
          write(val,'(I0)') build_info%replicas
          call replace(line, kw_index, keyword, val)

        case ('FIRST_STEP-1')
          write(val,'(I0)') ctrl_data%remd_info%first_step - 1
          call replace(line, kw_index, keyword, val)

        case default
          write(MsgOut,*) "Copy_Files> Unknown template keyword : "//&
               TRIM(keyword)
          exit

        end select

      end do

      if (line(:10) == '#ENDSCRIPT') &
        goto 10

      write(bat_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)
    call close_file(bat_out)


    ! run batch file for preparing restart
    !
    iret = system('sh '//TRIM(bat_file))
    if (iret /= 0) &
       call error_msg('Copy_Files> Batch-file restart was failed.')

    iret = system('rm '//TRIM(bat_file))

    write(MsgOut,'(A)') ' '

    return

  end subroutine copy_files

end module ib_copy_file_mod
