!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   ib_ctrl_mod
!> @brief   read control parameters
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/05/30 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module ib_ctrl_mod

  use fileio_control_mod
  use program_mod
  use messages_mod
  use ib_box2pme

  implicit none
  private

  ! structures

  ! section : <main>
  type, public :: s_main_info
    character(200)               :: directory             = './'
    character(100)               :: rein_input            = 'rein.inp'
    character(200)               :: template_dir          = ''
    character(100)               :: spawn                 = 'yes'
    character(100)               :: parallel_module_md    = ''
    character(100)               :: option_md             = ''
    character(100)               :: extend_option_md       = ''
  end type s_main_info

  ! section : REMD
  type, public :: s_remd_info
    integer                      :: first_step            = 1
    integer                      :: equiv_steps           = 0
    integer                      :: product_steps         = 0
    character(20),       pointer :: axis(:)               => null()
    integer,             pointer :: num_replicas(:)       => null()
    character(50),       pointer :: range(:)              => null()
    character(20),       pointer :: division(:)           => null()
    character(100),      pointer :: distance(:)           => null()
    character(100),      pointer :: angle(:)              => null()
    character(100),      pointer :: dihedral(:)           => null()
    double precision,    pointer :: spring_const(:)       => null()
  end type s_remd_info

  ! section : MD
  type, public :: s_md_info
    character(50)                :: init_md_file          = 'initial'
    integer                      :: core_for_md           = 8
    character(50)                :: md_program            = 'namd'
    character(50)                :: md_exe_file           = './namd2'
    double precision             :: md_temperature        = 300.0
    integer                      :: md_steps              = 1000
    character(50)                :: pme_grid              = 'no'
    integer                      :: pme_grid_size_x       = 24
    integer                      :: pme_grid_size_y       = 24
    integer                      :: pme_grid_size_z       = 24
    double precision             :: system_size_x         = 23.5
    double precision             :: system_size_y         = 23.5
    double precision             :: system_size_z         = 23.5
  end type s_md_info

  type, public :: s_ctrl_data
    type(s_main_info)            :: main_info
    type(s_remd_info)            :: remd_info
    type(s_md_info)              :: md_info
  end type s_ctrl_data

  ! subroutines
  public  :: control
  private :: read_main
  private :: read_remd
  private :: read_md

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  control
  !> @brief      a driver subroutine for reading control parmeters
  !! @authors    NM,NT
  !! @param[in]  filename    filename of control file
  !! @param[out] ctrl_data   control parameters in the program
  !! @date       2012/06/06 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine control(filename, ctrl_data)

    ! formal arguments
    character(*),      intent(in)    :: filename
    type(s_ctrl_data), intent(inout) :: ctrl_data

    ! local variables
    integer                          :: handle

    ! open control file
    !
    call open_ctrlfile(filename, handle)

    if (handle == 0) then
      call error_msg('Ib_ctrl> File IO Error')
    end if

    ! read section : <main>
    !
    call read_main(handle, ctrl_data%main_info)

    ! read section : REMD
    !
    call read_remd(handle, ctrl_data%remd_info)

    ! read section : MD
    !
    call read_md(handle, ctrl_data%md_info)

    return

  end subroutine control


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_main
  !> @brief      read control parameters in <main> section
  !! @authors    NT
  !! @date       2012/06/06 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_main(handle, main_info)

    character(*), parameter :: Section = ''

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_main_info),      intent(inout) :: main_info


    ! read control parameters
    !
    call read_ctrlfile_string(handle, Section, &
         'directory', main_info%directory)
    call read_ctrlfile_string(handle, Section, &
         'rein_input', main_info%rein_input)
    call read_ctrlfile_string(handle, Section, &
         'template_dir', main_info%template_dir)

    call read_ctrlfile_string(handle, Section, &
         'spawn', main_info%spawn)

    if ( TRIM(main_info%spawn) .eq. 'yes' ) then

      main_info%parallel_module_md = ''
      main_info%option_md          = ''
      main_info%extend_option_md   = ''

    else

      main_info%parallel_module_md = './charmrun'
      main_info%option_md          = '+p8'
      call read_ctrlfile_string(handle, Section, &
         'parallel_module_md', main_info%parallel_module_md)
      call read_ctrlfile_string(handle, Section, &
         'option_md', main_info%option_md)
      call read_ctrlfile_string(handle, Section, &
         'extend_option_md', main_info%extend_option_md)

    end if

    if (main_info%template_dir == '') then
      main_info%template_dir = TRIM(get_program_path())//'/templates'
    end if


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Main> Parameters of Main'
    write(MsgOut,'(A25,A)') '             directory = ', &
         TRIM(main_info%directory)
    write(MsgOut,'(A25,A)') '            rein input = ', &
         TRIM(main_info%rein_input)
    write(MsgOut,'(A25,A)') '          template dir = ', &
         TRIM(main_info%template_dir)
    write(MsgOut,*) ''

    return

  end subroutine read_main


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_remd
  !> @brief      read control parameters in REMD section
  !! @authors    NM,NT
  !! @date       2012/07/23 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_remd(handle, remd_info)

    character(*), parameter :: Section = 'REMD'

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_remd_info),      intent(inout) :: remd_info

    ! local variables
    integer                 :: i, naxis
    character(50)           :: val
    character(20)           :: key


    ! read control parameters
    !
    call read_ctrlfile_integer(handle, Section, &
         'first_step', remd_info%first_step)
    call read_ctrlfile_integer(handle, Section, &
         'equiv_steps', remd_info%equiv_steps)
    call read_ctrlfile_integer(handle, Section, &
         'product_steps', remd_info%product_steps)

    ! check readable axis count
    naxis = 0
    do while (.true.)
      val = ''
      write(key, fmt='(a4,i3.3)') 'axis', naxis + 1
      call read_ctrlfile_string(handle, Section, key, val)
      if (val == '') then
        exit
      end if
      naxis = naxis + 1
    end do

    if (naxis == 0) &
      call error_msg('Read_Remd> There is no one axis definition.')

    allocate(remd_info%axis(naxis),     &
             remd_info%num_replicas(naxis), &
             remd_info%range(naxis),    &
             remd_info%division(naxis), &
             remd_info%distance(naxis), &
             remd_info%angle(naxis),    &
             remd_info%dihedral(naxis), &
             remd_info%spring_const(naxis))
    remd_info%num_replicas(:)     = 0
    remd_info%range(:)        = ''
    remd_info%division(:)     = ''
    remd_info%distance(:)     = ''
    remd_info%angle(:)        = ''
    remd_info%dihedral(:)     = ''
    remd_info%spring_const(:) = 0.0

    do i = 1, naxis

      write(key,fmt='(a4,i3.3)') "axis", i
      call read_ctrlfile_string(handle, Section, key, &
                                remd_info%axis(i))

      write(key,fmt='(a12,i3.3)') "num_replicas", i
      call read_ctrlfile_integer(handle, Section, key, &
                                 remd_info%num_replicas(i))

      write(key,fmt='(a5,i3.3)') "range", i
      call read_ctrlfile_string(handle, Section, key, &
                                remd_info%range(i))

      write(key,fmt='(a8,i3.3)') "division", i
      call read_ctrlfile_string(handle, Section, key, &
                                remd_info%division(i))

      write(key,fmt='(a8,i3.3)') "distance", i
      call read_ctrlfile_string(handle, Section, key, &
                                remd_info%distance(i))

      write(key,fmt='(a5,i3.3)') "angle", i
      call read_ctrlfile_string(handle, Section, key, &
                                remd_info%angle(i))

      write(key,fmt='(a8,i3.3)') "dihedral", i
      call read_ctrlfile_string(handle, Section, key, &
                                remd_info%dihedral(i))

      write(key,fmt='(a12,i3.3)') "spring_const", i
      call read_ctrlfile_real(handle, Section, key, &
                              remd_info%spring_const(i))

    end do


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Remd> Parameters of REIN'
    write(MsgOut,'(A25,I0)')'            first step = ', &
         remd_info%first_step
    write(MsgOut,'(A25,I0)')'           equiv steps = ', &
         remd_info%equiv_steps
    write(MsgOut,'(A25,I0)')'         product steps = ', &
         remd_info%product_steps
    write(MsgOut,'(A20,I0)') '        # of axis : ', naxis
    do i = 1, naxis
      write(MsgOut,'(A19,I3.3,A3,A)') &
           '            axis : ', i, ' = ', TRIM(remd_info%axis(i))
      write(MsgOut,'(A25,I0)') &
           '          num_replicas = ', remd_info%num_replicas(i)
      write(MsgOut,'(A25,A)') &
           '                 range = ', TRIM(remd_info%range(i))
      write(MsgOut,'(A25,A)') &
           '         division type = ', TRIM(remd_info%division(i))
      if (remd_info%distance(i) /= '') &
        write(MsgOut,'(A25,A)') &
           '              distance = ', TRIM(remd_info%distance(i))
      if (remd_info%angle(i) /= '') &
        write(MsgOut,'(A25,A)') &
           '                 angle = ', TRIM(remd_info%angle(i))
      if (remd_info%dihedral(i) /= '') &
        write(MsgOut,'(A25,A)') &
           '              dihedral = ', TRIM(remd_info%dihedral(i))

      if (remd_info%distance(i) /= '' .or. &
          remd_info%angle(i)    /= '' .or. &
          remd_info%dihedral(i) /= '') &
        write(MsgOut,'(A25,F0.2)') &
           '          spring const = ', remd_info%spring_const(i)

    end do
    write(MsgOut,'(A)') ' '

    return

  end subroutine read_remd


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_md
  !> @brief      read control parameters in MD section
  !! @authors    NM,NT
  !! @date       2012/06/06 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_md(handle, md_info)

    character(*), parameter :: Section = 'MD'

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_md_info),        intent(inout) :: md_info


    ! read control parameters
    !
    call read_ctrlfile_integer(handle, Section, &
         'num_mpiproc_md', md_info%core_for_md)
    call read_ctrlfile_string(handle, Section, &
         'md_program', md_info%md_program)
    call read_ctrlfile_real(handle, Section, &
         'md_temperature', md_info%md_temperature)
    call read_ctrlfile_string(handle, Section, &
         'md_exe_file', md_info%md_exe_file)
    call read_ctrlfile_integer(handle, Section, &
         'md_steps', md_info%md_steps)


    call read_ctrlfile_string(handle, Section, &
         'pme_grid', md_info%pme_grid)

    if ( TRIM(md_info%pme_grid) .eq. 'yes' ) then

      call read_ctrlfile_integer(handle, Section, &
         'pme_grid_size_x', md_info%pme_grid_size_x)
      call read_ctrlfile_integer(handle, Section, &
         'pme_grid_size_y', md_info%pme_grid_size_y)
      call read_ctrlfile_integer(handle, Section, &
         'pme_grid_size_z', md_info%pme_grid_size_z)

      md_info%system_size_x = md_info%pme_grid_size_x
      md_info%system_size_y = md_info%pme_grid_size_y
      md_info%system_size_z = md_info%pme_grid_size_z

    else

      call read_ctrlfile_real(handle, Section, &
         'system_size_x', md_info%system_size_x)
      call read_ctrlfile_real(handle, Section, &
         'system_size_y', md_info%system_size_y)
      call read_ctrlfile_real(handle, Section, &
         'system_size_z', md_info%system_size_z)

      call box2pme(md_info%system_size_x, md_info%pme_grid_size_x)
      call box2pme(md_info%system_size_y, md_info%pme_grid_size_y)
      call box2pme(md_info%system_size_z, md_info%pme_grid_size_z)

    end if


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_MD> Parameters of MD'
    write(MsgOut,'(A25,I0)')'           core for MD = ', &
         md_info%core_for_md
    write(MsgOut,'(A25,A)') '           MD programs = ', &
         TRIM(md_info%md_program)
    write(MsgOut,'(A25,A)') '           MD exe file = ', &
         TRIM(md_info%md_exe_file)
    write(MsgOut,'(A25,F0.2)')'      MD temperature = ', &
         md_info%md_temperature
    write(MsgOut,'(A25,I0)')'              MD steps = ', &
         md_info%md_steps
    write(MsgOut,'(A25,A)') '              PME grid = ', &
         md_info%pme_grid
    write(MsgOut,'(A25,I0)')'       PME grid size X = ', &
         md_info%pme_grid_size_x
    write(MsgOut,'(A25,I0)')'       PME grid size Y = ', &
         md_info%pme_grid_size_y
    write(MsgOut,'(A25,I0)')'       PME grid size Z = ', &
         md_info%pme_grid_size_z
    write(MsgOut,*) ' '

    return

  end subroutine read_md

end module ib_ctrl_mod
