!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   ib_setup_mod
!> @brief   show usage of batch builder
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/06/14
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

module ib_setup_mod

  use ib_ctrl_mod
  use fileio_control_mod
  use messages_mod

  implicit none
  private

  integer, public, parameter   :: AxisTypeTemperature = 1
  integer, public, parameter   :: AxisTypeHarmonic    = 2

  integer, public, parameter   :: HarmoTypeNone       = 0
  integer, public, parameter   :: HarmoTypeDistance   = 1
  integer, public, parameter   :: HarmoTypeAngle      = 2
  integer, public, parameter   :: HarmoTypeDihedral   = 3


  ! param value information
  !
  type, public :: s_pval_info
    integer                    :: axis_no          = 0
    double precision           :: value            = 0.0
  end type s_pval_info


  ! axis information
  !
  type, public :: s_axis_info
    integer                    :: index            = 0
    integer                    :: type             = 0
    integer                    :: replicas         = 0
    logical                    :: periodic         = .false.
    double precision           :: value_lo         = 0.0
    double precision           :: value_hi         = 0.0
    double precision, pointer  :: values(:)        => null()
    integer                    :: harmo_type       = 0
    character(10),    pointer  :: harmo_item1(:)   => null()
    character(10),    pointer  :: harmo_item2(:)   => null()
    character(10),    pointer  :: harmo_item3(:)   => null()
  end type s_axis_info


  ! build information
  !
  type, public :: s_build_info
    integer                    :: replicas         = 0
    type(s_pval_info), pointer :: pval_info(:)     => null()
    type(s_axis_info), pointer :: axis_info(:)     => null()
  end type s_build_info


  ! subroutines
  public  :: setup
  private :: setup_pval
  private :: setup_axis
  private :: setup_harmonic

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  setup
  !> @brief      setup input builder build information
  !! @authors    NT
  !! @param[in]  ctrl_data  control parameters in the program
  !! @param[out] build_info build information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup(ctrl_data, build_info)

    ! formal arguments
    type(s_ctrl_data),  intent(in)    :: ctrl_data
    type(s_build_info), intent(inout) :: build_info

    ! local variables
    integer             :: i, naxis, nrep, nrep_acc


    naxis    = SIZE(ctrl_data%remd_info%axis)

    ! setup total replicas
    nrep     = 1
    nrep_acc = 1

    do i = 1, naxis
      nrep = nrep * ctrl_data%remd_info%num_replicas(i)
    end do

    build_info%replicas = nrep
    write(MsgOut,'(A,I0)') 'Setup> # of total replicas: ', build_info%replicas


    ! setup axis
    allocate(build_info%axis_info(naxis))

    do i = 1, naxis

      call setup_axis(ctrl_data%remd_info, &
                      i,        &
                      nrep,     &
                      nrep_acc, &
                      build_info%axis_info(i))

    end do

    ! setup param values
    call setup_pval(ctrl_data%remd_info, build_info)

    write(MsgOut,'(A)') ' '

    return

  end subroutine setup


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  setup_pval
  !> @brief      setup param values information
  !! @authors    NT
  !! @param[in]  remd_info  REIN information in control parameters
  !! @param[out] build_info build information
  !! @date       2012/06/14 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_pval(remd_info, build_info)

    ! formal arguments
    type(s_remd_info),  intent(in)    :: remd_info
    type(s_build_info), intent(inout) :: build_info

    ! local variables
    double precision    :: value
    integer             :: naxis, npval, i, axis_no
    character(100)      :: str


    naxis = SIZE(build_info%axis_info)

    ! check for # of parameter values
    npval = 0
    do i = 1, naxis
      select case(build_info%axis_info(i)%type)

      case (AxisTypeTemperature)
        npval = npval + 1

      case (AxisTypeHarmonic)
        npval = npval + 2

      end select
    end do

    ! allocate memory
    allocate(build_info%pval_info(npval))

    ! setup values
    npval = 1
    do i = 1, naxis
      select case(build_info%axis_info(i)%type)
        
      case (AxisTypeTemperature)

        build_info%pval_info(npval)%axis_no = i
        build_info%pval_info(npval)%value   = 0.0
        npval = npval + 1

      case (AxisTypeHarmonic)

        build_info%pval_info(npval  )%axis_no = i
        build_info%pval_info(npval  )%value   = 0.0
        build_info%pval_info(npval+1)%axis_no = 0
        build_info%pval_info(npval+1)%value   = remd_info%spring_const(i)
        npval = npval + 2

      end select
    end do

    npval = npval - 1
    write(MsgOut,'(A,I0)') 'Setup> # of param values : ', npval
    do i = 1, npval
      axis_no = build_info%pval_info(i)%axis_no
      value   = build_info%pval_info(i)%value
      if (axis_no == 0) then
        write(MsgOut,'(A,I3,A,F8.3)') '      param : ', i, ' = Fixed:', value
      else
        write(MsgOut,'(A,I3,A,I8)')   '      param : ', i, ' = Axis :', axis_no
        if (axis_no > SIZE(build_info%axis_info)) then
          write(str,'(A,I0,A)') &
               'Setup_Pval> Error: axis',axis_no,' is undefined.'
          call error_msg(str)
        end if
      end if
    end do

    return

  end subroutine setup_pval


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  setup_axis
  !> @brief      setup axis information
  !! @authors    NT
  !! @param[in]  remd_info  REIN information in control parameters
  !! @param[in]  iaxis      index of axis
  !! @param[in]  nrep       number of total replicas
  !! @@aram[in]  nrep_acc   accumlation of number of replicas
  !! @param[out] axis_info  axis information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_axis(remd_info, iaxis, nrep, nrep_acc, axis_info)

    ! formal arguments
    type(s_remd_info),  intent(in)    :: remd_info
    integer,            intent(in)    :: iaxis
    integer,            intent(in)    :: nrep
    integer,            intent(inout) :: nrep_acc
    type(s_axis_info),  intent(inout) :: axis_info

    ! local variables
    double precision    :: val_bin, val_range, val_lo, val_hi
    integer             :: j, jj, ival, nrep_axis
    character(20)       :: str


    ! axis index
    axis_info%index = iaxis

    ! axis type
    str = remd_info%axis(iaxis)
    call toupper(str)

    select case(str)
    case ('TEMPERATURE')
      axis_info%type = AxisTypeTemperature

    case ('HARMONIC')
      axis_info%type = AxisTypeHarmonic

      if (remd_info%distance(iaxis) == '' .and. &
          remd_info%angle(iaxis)    == '' .and. &
          remd_info%dihedral(iaxis) == '') &
        call error_msg('Setup> Error: There is no harmonic targets. '// &
        '(distance/angle/dihedral)')

    case default
      call error_msg('Setup> Error: Unknown axis type : '//TRIM(str))

    end select

    ! replicas
    nrep_axis = remd_info%num_replicas(iaxis)
    axis_info%replicas = nrep_axis
    if (nrep_axis <= 1) &
      call error_msg('Setup> Error: # of replica < 2')

    ! periodic
    str = remd_info%division(iaxis)
    call toupper(str)

    if (str == 'CYCLE') then
      axis_info%periodic = .true.
    else
      axis_info%periodic = .false.
    end if

    ! value low, high
    read(remd_info%range(iaxis),*,err=900,end=900) val_lo, val_hi
    axis_info%value_lo = val_lo
    axis_info%value_hi = val_hi

    val_range = val_hi - val_lo

    ! values
    allocate(axis_info%values(nrep))

    select case(str)
    case ('EXP')
      val_bin = (DLOG(val_hi/val_lo)) / DBLE(nrep_axis - 1)
    case ('EXPHF')
      val_bin = (DLOG(val_range + 1.d0))**2 / DBLE(nrep_axis - 1)
    case ('EQUIV')
      val_bin = val_range / DBLE(nrep_axis - 1)
    case ('CYCLE')
      val_bin = val_range / DBLE(nrep_axis)
    end select

    jj = 1
    do j = 1, nrep

      ival = MOD(jj, nrep_axis)
      if (ival == 0) &
        ival = nrep_axis

      if (ival == 1) then
        axis_info%values(j) = val_lo
        if (MOD(j, nrep_acc) == 0) &
          jj = jj + 1
        cycle
      end if

      select case(str)
      case ('EXP')
        axis_info%values(j) = &
             axis_info%values(ival - 1) * DEXP(val_bin)
      case ('EXPHF')
        axis_info%values(j) = &
             val_lo + DEXP((DBLE(ival - 1) * val_bin)**0.5) - 1.d0
      case ('EQUIV')
        axis_info%values(j) = &
             val_lo + (ival - 1) * val_bin
      case ('CYCLE')
        axis_info%values(j) = &
             val_lo + (ival - 1) * val_bin
      end select

      if (MOD(j, nrep_acc) == 0) &
        jj = jj + 1

    end do

    nrep_acc = nrep_acc * nrep_axis

    ! harmonic type
    if (remd_info%distance(iaxis) /= '') then
      axis_info%harmo_type = HarmoTypeDistance
      call setup_harmonic(remd_info%distance(iaxis), axis_info)

    else if (remd_info%angle(iaxis) /= '') then
      axis_info%harmo_type = HarmoTypeAngle
      call setup_harmonic(remd_info%angle(iaxis), axis_info)

    elseif (remd_info%dihedral(iaxis) /= '') then
      axis_info%harmo_type = HarmoTypeDihedral
      call setup_harmonic(remd_info%dihedral(iaxis), axis_info)

    else
      axis_info%harmo_type = HarmoTypeNone

    end if

    return

900 call error_msg('Setup_Axis> Error. axis value range '//&
         TRIM(remd_info%range(iaxis)))

  end subroutine setup_axis


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  setup_harmonic
  !> @brief      setup harmonic information
  !! @authors    NT
  !! @param[in]  harmo_info harmonic information
  !! @param[out] axis_info  axis information
  !! @date       2012/06/07 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_harmonic(harmo_info, axis_info)

    ! formal arguments
    character(*),       intent(in)    :: harmo_info
    type(s_axis_info),  intent(inout) :: axis_info

    ! local variables
    integer             :: i
    character(100)      :: str


    select case(axis_info%harmo_type)

    case (HarmoTypeDistance)
      allocate(axis_info%harmo_item1(2),    &
               axis_info%harmo_item2(2),    &
               axis_info%harmo_item3(2))

    case (HarmoTypeAngle)
      allocate(axis_info%harmo_item1(3),    &
               axis_info%harmo_item2(3),    &
               axis_info%harmo_item3(3))

    case (HarmoTypeDihedral)
      allocate(axis_info%harmo_item1(4),    &
               axis_info%harmo_item2(4),    &
               axis_info%harmo_item3(4))

    end select

    str = harmo_info

    do i = 1, LEN_TRIM(str)
      if (str(i:i) == ':') &
        str(i:i) = ' '
    end do

    read(str,*,err=10,end=10) &
                (axis_info%harmo_item1(i), &
                 axis_info%harmo_item2(i), &
                 axis_info%harmo_item3(i), i=1,SIZE(axis_info%harmo_item1))

    return

10  axis_info%harmo_item3(:) = ''
    read(str,*,err=90,end=90) &
                (axis_info%harmo_item1(i), i=1,SIZE(axis_info%harmo_item1))

    return

90  call error_msg('Setup_Harmonic> Error: Harmonic '//TRIM(harmo_info))

  end subroutine setup_harmonic

end module ib_setup_mod
