!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   fileio_mod
!> @brief   utilities for file I/O
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM), Yuji Sugita (YS)
!! @date    2012/05/24 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module fileio_mod

  use messages_mod

  implicit none
  private

  ! constants 
  integer, public, parameter :: InvalidUnitNo = 1
  integer, public, parameter :: FirstUnitNo   = 7
  integer, public, parameter :: MaxUnitCount  = 200

  integer, public, parameter :: IOFileInput  = 1
  integer, public, parameter :: IOFileOutputNew = 2
  integer, public, parameter :: IOFileOutputReplace = 3

  ! variables
  logical                    :: io_initial    = .false.
  logical                    :: io_used(MaxUnitCount)
  integer                    :: io_unit_no(MaxUnitCount)

  ! subroutines and functions
  public   :: open_file
  public   :: open_binary_file
  public   :: close_file
  public   :: check_file

  public   :: get_unit_no
  public   :: free_unit_no

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  open_file
  !> @brief      open a text file
  !! @authors    NT, YS
  !! @param[out] unit_no, unit number of a text file [int]
  !! @param[in]  filename file name of a text file [int]
  !! @param[in]  in_out flag for the file (IOFileInput/IOFileOutput) [int]
  !! @date       2012/05/24 (NT)
  !! @note       get_unit_no is used 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_file(unit_no, filename, in_out)

    ! formal arguments
    integer,      intent(out) :: unit_no
    character(*), intent(in)  :: filename
    integer,      intent(in)  :: in_out

    unit_no = get_unit_no()
    if (unit_no == InvalidUnitNo) goto 900

    if (in_out == IOFileInput) then
      open(unit_no, file=filename, status='old', form='formatted', err=900)
    else if (in_out == IOFileOutputNew) then
      open(unit_no, file=filename, status='new', form='formatted', err=900)
    else if (in_out == IOFileOutputReplace) then
      open(unit_no, file=filename, status='replace', form='formatted', err=900)
    endif

    return

900 write(ErrOut,'(A)') 'Open_File> ', TRIM(filename)
    call error_msg('Open_file> I/O Error.')

  end subroutine open_file

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  open_binary_file
  !> @brief      open a binary file
  !! @authors    NT, YS
  !! @param[out] unit_no, unit number of a file [int]
  !! @param[in]  filename file name of a file [int]
  !! @param[in]  in_out flag for the file (IOFileInput/IOFileOutput) [int]
  !! @date       2012/05/24 (NT)
  !! @note       get_unit_no is used
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_binary_file(unit_no, filename, in_out)

    ! formal arguments
    integer,      intent(out) :: unit_no
    character(*), intent(in)  :: filename
    integer,      intent(in)  :: in_out

    unit_no = get_unit_no()
    if (unit_no == InvalidUnitNo) goto 900

    if (in_out == IOFileInput) then
      open(unit_no, file=filename, status='old', form='unformatted', err=900)
    else if (in_out == IOFileOutputNew) then
      open(unit_no, file=filename, status='new', form='unformatted', err=900)
    else if (in_out == IOFileOutputReplace) then
      open(unit_no, file=filename, status='replace', form='unformatted', err=900)
    endif

    return

900 write(ErrOut,'(A)') 'Open_Binary_File> ', TRIM(filename)
    call error_msg('Open_Binary_file> I/O Error.')

  end subroutine open_binary_file

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  close_file
  !> @brief      close an input/output file
  !! @authors    NT, YS
  !! @param[in]  unit_no, unit number of a file [int]
  !! @date       2012/05/24 (NT)
  !! @note       free_unit_no is used
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_file(unit_no)

    ! formal arguments
    integer,      intent(in) :: unit_no

    close(unit_no)
    call free_unit_no(unit_no)

    return

  end subroutine close_file

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function  get_unit_no
  !> @brief    create and get unit number of a file
  !! @authors  NT, YS
  !! @return   unit_no of a file [int]
  !! @date     2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function get_unit_no()

    ! return value
    integer :: get_unit_no

    ! local variables
    integer :: i

    ! initialize table for unit no
    !
    if (.not. io_initial) then
      io_initial = .true.
      do i = 1, MaxUnitCount
        io_used(i)    = .false.
        io_unit_no(i) = i + FirstUnitNo - 1
      end do
    end if

    ! get unit no
    !
    do i = 1, MaxUnitCount
      if (.not. io_used(i)) then
        io_used(i)  = .true.
        get_unit_no = io_unit_no(i)
        return
      end if 
    end do

    get_unit_no = InvalidUnitNo

  end function get_unit_no

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine free_unit_no
  !> @brief     free unit number of a file
  !! @authors   NT, YS
  !! @param[in] unit_no unit number of an input/output file [int]
  !! @date      2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine free_unit_no(unit_no)

    ! formal arguments
    integer, intent(in) :: unit_no

    ! local variables
    integer             :: i

    do i = 1, MaxUnitCount
      if (unit_no == io_unit_no(i)) then
        io_used(i) = .false.
        return
      end if
    end do

    return

  end subroutine free_unit_no

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function    check_file
  !> @brief      check for file existence
  !! @authors    NT
  !! @param[in]  filename file name of a text file [int]
  !! @date       2012/08/01 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function check_file(filename)

    ! return
    logical       :: check_file

    ! formal arguments
    character(*), intent(in)  :: filename


    check_file = .false.
    inquire(file = filename, exist = check_file)

    return

  end function check_file

end module fileio_mod
