!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : fileio_control
!> @brief   utilities for reading control parameters from file
!! @author  Norio Takase (NT), Naoyuki Miyashita (NM), Yuji Sugita (YS)
!! @date    2012/05/24
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module fileio_control_mod

  use messages_mod

  implicit none
  private

  integer,    private, parameter :: MaxHandle = 10
  integer(8), private            :: g_handle(MaxHandle) = 0

  ! private subroutine
  !
  public  :: open_ctrlfile
  public  :: close_ctrlfile
  public  :: read_ctrlfile_string
  public  :: read_ctrlfile_integer
  public  :: read_ctrlfile_real
  public  :: read_ctrlfile_logical
  public  :: read_ctrlfile_type
  private :: search_string
  public  :: tolower
  public  :: toupper

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  open_ctrlfile
  !> @brief      open control file
  !! @authors    NT, YS
  !! @param[in]  filename : file name
  !! @param[out] handle   : unit number
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_ctrlfile(filename, handle)

    ! formal arguments
    character(*),  intent(in) :: filename
    integer,       intent(out):: handle

    ! local variables
    character(100)            :: str
    integer                   :: il, i
    integer(8)                :: c_handle

    str = filename
    il = LEN_TRIM(str)
    str(il+1:il+2) = CHAR(0)
    
    handle = 0
    
    call copenctrlfile(str, c_handle)
    if (c_handle == 0) return
    
    do i = 1, MaxHandle
       if (g_handle(i) == 0) then
          g_handle(i) = c_handle
          handle = i
          return
       end if
    end do
    
    return
  
  end subroutine open_ctrlfile

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  close_ctrlfile
  !> @brief      close control file
  !! @authors    NT, YS
  !! @param[in]  filename : unit number
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_ctrlfile(handle)

    ! formal arguments
    integer, intent(in) :: handle

    ! local variables
    integer(8)          :: c_handle

    if (handle > MaxHandle) return

    c_handle = g_handle(handle)

    call cclosectrlfile(c_handle)

    g_handle(handle) = 0

    return

  end subroutine close_ctrlfile

  !======1=========2=========3=========4=========5=========6=========7=========8
  ! 
  !  Subroutine  read_ctrlfile_string
  !> @brief      read character value from control file
  !! @authors    NT, YS
  !! @param[in]  handle  : unit number
  !! @param[in]  section :
  !! @param[in]  key     :
  !! @param[out] value   :
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrlfile_string(handle, section, key, value)

    ! formal arguments
    integer,      intent(in)    :: handle
    character(*), intent(in)    :: section
    character(*), intent(in)    :: key
    character(*), intent(inout) :: value
    
    ! local variables
    character(100)              :: csection, ckey, cdef_value
    integer                     :: idx, ii, ll
    integer(8)                  :: c_handle


    if (handle > MaxHandle) return

    c_handle   = g_handle(handle)
    csection   = section
    ckey       = key
    cdef_value = value

    idx = LEN_TRIM(csection)
    csection(idx+1:) = CHAR(0)

    idx = LEN_TRIM(ckey)
    ckey(idx+1:) = CHAR(0)

    idx = LEN_TRIM(cdef_value)
    cdef_value(idx+1:) = CHAR(0)

    ll = LEN(value)

    call creadctrlfilestring(c_handle, csection, ckey, cdef_value, value, ll)

    idx = INDEX(value, CHAR(0))
    do ii = idx, LEN(value)
      value(ii:) = ' '
    end do

    return

  end subroutine read_ctrlfile_string

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_ctrlfile_integer
  !> @brief      read integer value from control file
  !! @authors    NT, YS
  !! @param[in]  handle  : unit number
  !! @param[in]  section :
  !! @param[in]  key     :
  !! @param[out] value   :
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrlfile_integer(handle, section, key, value)

    ! formal arguments
    integer,      intent(in)    :: handle
    character(*), intent(in)    :: section
    character(*), intent(in)    :: key
    integer,      intent(inout) :: value

    ! local variables
    character(100)              :: csection, ckey
    integer                     :: idx
    integer(8)                  :: c_handle


    if (handle > MaxHandle) return

    c_handle = g_handle(handle)
    csection = section
    ckey = key

    idx = LEN_TRIM(csection)
    csection(idx+1:) = CHAR(0)

    idx = LEN_TRIM(ckey)
    ckey(idx+1:) = CHAR(0)

    call creadctrlfileinteger(c_handle, csection, ckey, value, value)

    return

  end subroutine read_ctrlfile_integer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_ctrlfile_real
  !> @brief      read real value from control file
  !! @authors    NT, YS
  !! @param[in]  handle  : unit number
  !! @param[in]  section :
  !! @param[in]  key     :
  !! @param[out] value   :
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrlfile_real(handle, section, key, value)

    ! formal arguments
    integer,          intent(in)    :: handle
    character(*),     intent(in)    :: section
    character(*),     intent(in)    :: key
    double precision, intent(inout) :: value

    ! local variables
    character(100) :: csection, ckey
    integer        :: idx
    integer(8)     :: c_handle


    if (handle > MaxHandle) return

    c_handle = g_handle(handle)
    csection = section
    ckey = key

    idx = LEN_TRIM(csection)
    csection(idx+1:) = CHAR(0)

    idx = LEN_TRIM(ckey)
    ckey(idx+1:) = CHAR(0)

    call creadctrlfilereal(c_handle, csection, ckey, value, value)

    return

  end subroutine read_ctrlfile_real

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_ctrlfile_logical
  !> @brief      read logical value from control file
  !! @authors    NT, YS
  !! @param[in]  handle  : unit number
  !! @param[in]  section :
  !! @param[in]  key     :
  !! @param[out] value   :
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrlfile_logical(handle, section, key, value)

    ! formal arguments
    integer,      intent(in)    :: handle
    character(*), intent(in)    :: section
    character(*), intent(in)    :: key
    logical,      intent(inout) :: value

    ! local variables
    character(100)              :: csection, ckey, cvalue, cdef_value
    integer                     :: idx, ii, ll
    integer(8)                  :: c_handle


    if (handle > MaxHandle) return

    c_handle = g_handle(handle)
    csection = section
    ckey = key
    if (value .eqv. .true.) then
       cdef_value = 'yes'
    else
       cdef_value = 'no'
    end if

    idx = LEN_TRIM(csection)
    csection(idx+1:) = CHAR(0)

    idx = LEN_TRIM(ckey)
    ckey(idx+1:) = CHAR(0)

    idx = LEN_TRIM(cdef_value)
    cdef_value(idx+1:) = CHAR(0)

    ll = LEN(cvalue)

    call creadctrlfilestring(c_handle, csection, ckey, cdef_value, cvalue, ll)

    idx = INDEX(cvalue, CHAR(0))
    do ii = idx, LEN(cvalue)
       cvalue(ii:) = ' '
    end do

    call tolower(cvalue)
    if (cvalue == 'yes') then
       value = .true.
    else
       value = .false.
    end if
    
    return

  end subroutine read_ctrlfile_logical

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_ctrlfile_type
  !> @brief      read type(integer) value from control file
  !! @authors    NT, YS
  !! @param[in]  handle  : unit number
  !! @param[in]  section :
  !! @param[in]  key     :
  !! @param[out] value   :
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrlfile_type(handle, section, key, value, typeStrs)

    ! formal arguments
    integer,      intent(in)    :: handle
    character(*), intent(in)    :: section
    character(*), intent(in)    :: key
    integer,      intent(inout) :: value
    character(*), intent(in)    :: typeStrs(:)

    ! local variables
    character(100)              :: csection, ckey, cvalue, cdef_value
    integer                     :: idx, ii, ll
    integer(8)                  :: c_handle


    if (handle > MaxHandle) return
    c_handle = g_handle(handle)

    if (value <= 0 .or. value > SIZE(typeStrs)) then
      call error_msg('read_ctrlfile_type> value is out of bounds.')
    end if

    csection = section
    ckey = key
    cdef_value = typeStrs(value)

    idx = LEN_TRIM(csection)
    csection(idx+1:) = CHAR(0)

    idx = LEN_TRIM(ckey)
    ckey(idx+1:) = CHAR(0)

    idx = LEN_TRIM(cdef_value)
    cdef_value(idx+1:) = CHAR(0)

    ll = LEN(cvalue)

    call creadctrlfilestring(c_handle, csection, ckey, cdef_value, cvalue, ll)

    idx = INDEX(cvalue, CHAR(0))
    do ii = idx, LEN(cvalue)
       cvalue(ii:) = ' '
    end do

    call search_string(typeStrs, cvalue, value)

    return

  end subroutine read_ctrlfile_type

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  search_string
  !> @brief      search string in line
  !! @authors    NT, YS
  !! @param[in]  strs  :
  !! @param[in]  str   :
  !! @param[out] index :
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
    
  subroutine search_string(strs, str, index)

    ! formal arguments
    character(*),     intent(in) :: strs(:)
    character(*),     intent(in) :: str
    integer,          intent(out):: index

    ! local variables
    character(40)                :: cstr1, cstr2
    integer                      :: i

    cstr1 = str
    call tolower(cstr1)

    do i = 1, SIZE(strs)

      cstr2 = strs(i)
      call tolower(cstr2)

      if (cstr1 == cstr2) then
        index = i
        return 
      end if

    end do

    call error_msg('search_string> Unsupported String : '//TRIM(str))

    return

  end subroutine search_string

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  tolower
  !> @brief      change capital letters into small letters
  !! @authors    NT, YS
  !! @param[out] str :
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine tolower(str)

    ! formal arguments
    character(*),    intent(inout) :: str

    ! local variables
    integer                        :: i

    do i = 1, LEN(str)
      if (str(i:i) >= 'A' .and. str(i:i) <= 'Z') then
        str(i:i) = CHAR(ICHAR(str(i:i)) + 32)
      end if
    end do
       
    return

  end subroutine tolower

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  toupper
  !> @brief      change small letters into capital letters
  !! @authors    NT, YS
  !! @param[out] str :
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine toupper(str)

    ! formal arguments
    character(*),    intent(inout) :: str

    ! local variables
    integer                        :: i

    do i = 1, LEN(str)
      if (str(i:i) >= 'a' .and. str(i:i) <= 'z') then
        str(i:i) = CHAR(ICHAR(str(i:i)) - 32)
      end if
    end do
       
    return

  end subroutine toupper

end module fileio_control_mod
