!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   fileio_psf_mod
!> @brief   read CHARMM PSF file
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM), Yuji Sugita (YS)
!! @date    2012/05/24 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module fileio_psf_mod

  use messages_mod
  use fileio_mod

  implicit none
  private

  ! structures
  type, public :: s_psf_hdr_info
    logical          :: ipsf, iext, icmap, icheq
    character(60)    :: fmt00, fmt01, fmt02
    character(60)    :: fmt03, fmt04, fmt05, fmt06
  end type s_psf_hdr_info

  type, public :: s_psf

    integer          :: type

    integer          :: num_atoms
    integer          :: num_residues
    integer          :: num_molecules
    integer          :: num_segments
    integer          :: num_bonds
    integer          :: num_angles
    integer          :: num_dihedrals
    integer          :: num_impropers
    integer          :: num_HB_donors
    integer          :: num_HB_acceptors
    integer          :: num_NB_exclusions
    integer          :: num_groups
    integer          :: num_cross_terms
    double precision :: total_charge

    ! ntitle

    ! atom
    integer,          pointer :: atom_no(:)         => null()
    character(4),     pointer :: segment_name(:)    => null()
    integer,          pointer :: segment_no(:)      => null()
    integer,          pointer :: residue_no(:)      => null()
    integer,          pointer :: residue_c_no(:)    => null()
    character(4),     pointer :: residue_name(:)    => null()
    character(4),     pointer :: atom_name(:)       => null()
    character(4),     pointer :: atom_cls_name(:)   => null()
    integer,          pointer :: atom_cls_no(:)     => null()
    double precision, pointer :: charge(:)          => null()
    double precision, pointer :: mass(:)            => null()
    integer,          pointer :: imove(:)           => null()

    integer,          pointer :: atom_connect_no(:) => null()
    integer,          pointer :: atom_connect(:,:)  => null()
    integer,          pointer :: molecule_no(:)     => null()

    ! bonds
    integer,          pointer :: bond_list(:,:)     => null()

    ! angles
    integer,          pointer :: angl_list(:,:)     => null()

    ! dihedrals
    integer,          pointer :: dihe_list(:,:)     => null()

    ! impropers
    integer,          pointer :: impr_list(:,:)     => null()

    ! ndonors
    integer,          pointer :: donr_list(:,:)     => null()

    ! acceptors
    integer,          pointer :: acce_list(:,:)     => null()

    ! nb
    integer,          pointer :: nb_list(:)         => null()

    ! grps
    integer,          pointer :: grp_list(:,:)      => null()

    ! cmaps
    integer,          pointer :: cmap_list(:,:)     => null()

    ! numlp

  end type s_psf

  ! variables
  integer, public, parameter :: PsfTypeCHARMM = 1
  integer, public, parameter :: PsfTypeXPLOR  = 2

  ! allocatable variables
  integer, public, parameter :: PsfAtom = 1
  integer, public, parameter :: PsfBond = 2
  integer, public, parameter :: PsfAngl = 3
  integer, public, parameter :: PsfDihe = 4
  integer, public, parameter :: PsfImpr = 5
  integer, public, parameter :: PsfDonr = 6
  integer, public, parameter :: PsfAcce = 7
  integer, public, parameter :: PsfNb   = 8
  integer, public, parameter :: PsfGrp  = 9
  integer, public, parameter :: PsfCmap = 10

  ! subroutines
  public  :: init_psf
  public  :: alloc_psf
  public  :: dealloc_psf
  public  :: dealloc_psf_all
  public  :: input_psf
  public  :: output_psf
  private :: read_psf
  private :: write_psf
  private :: check_psf_info

  private :: read_psf_header
  private :: read_psf_atom
  private :: read_psf_bond
  private :: read_psf_angl
  private :: read_psf_dihe
  private :: read_psf_impr
  private :: read_psf_donr
  private :: read_psf_acce
  private :: read_psf_nb
  private :: read_psf_grp
  private :: read_psf_cmap

  private :: write_psf_header
  private :: write_psf_atom
  private :: write_psf_bond
  private :: write_psf_angl
  private :: write_psf_dihe
  private :: write_psf_impr
  private :: write_psf_donr
  private :: write_psf_acce
  private :: write_psf_nb
  private :: write_psf_grp
  private :: write_psf_cmap

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  init_psf
  !> @brief      initialize PSF information
  !! @authors    NT, YS
  !! @param[in]  psf : structure of PSF information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine init_psf(psf)

    ! formal arguments
    type(s_psf),          intent(inout) :: psf

    psf%num_atoms         = 0 
    psf%num_residues      = 0
    psf%num_molecules     = 0
    psf%num_segments      = 0
    psf%num_bonds         = 0 
    psf%num_angles        = 0
    psf%num_dihedrals     = 0
    psf%num_impropers     = 0
    psf%num_HB_donors     = 0
    psf%num_HB_acceptors  = 0
    psf%num_NB_exclusions = 0
    psf%num_groups        = 0
    psf%num_cross_terms   = 0
    psf%total_charge      = 0.0d0

    return

  end subroutine init_psf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  alloc_psf
  !> @brief      allocate PSF information
  !! @authors    NT, YS
  !! @param[in]  psf      : structure of PSF information
  !! @param[in]  variable : selected variable
  !! @param[in]  var_size : size of the selected variable
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
 
  subroutine alloc_psf(psf, variable, var_size)

    ! formal arguments
    type(s_psf),          intent(inout) :: psf
    integer,              intent(in)    :: variable
    integer,              intent(in)    :: var_size

    ! local variables
    integer                             :: alloc_stat
    integer                             :: dealloc_stat


    ! initialize
    !
    alloc_stat   = 0
    dealloc_stat = 0

    ! allocate selected variables

    select case (variable)

    case(PsfAtom)

       if (associated(psf%atom_no)) then
          if (SIZE(psf%atom_no) == var_size) return
          deallocate(psf%atom_no,         &
                     psf%segment_name,    &
                     psf%segment_no,      &
                     psf%residue_no,      &
                     psf%residue_c_no,    &
                     psf%residue_name,    &
                     psf%atom_name,       &
                     psf%atom_cls_name,   &
                     psf%atom_cls_no,     &
                     psf%charge,          &
                     psf%mass,            &
                     psf%imove,           &
                     psf%atom_connect_no, &
                     psf%atom_connect,    &
                     psf%molecule_no,     &
                     stat = dealloc_stat)
       end if

       allocate(psf%atom_no(var_size),         &
                psf%segment_name(var_size),    &
                psf%segment_no(var_size),      &
                psf%residue_no(var_size),      &
                psf%residue_c_no(var_size),    &
                psf%residue_name(var_size),    &
                psf%atom_name(var_size),       &
                psf%atom_cls_name(var_size),   &
                psf%atom_cls_no(var_size),     &
                psf%charge(var_size),          &
                psf%mass(var_size),            &
                psf%imove(var_size),           &
                psf%atom_connect_no(var_size), &
                psf%atom_connect(4, var_size), &
                psf%molecule_no(var_size),     &
                stat = alloc_stat)

       psf%segment_name(:) = ''
       psf%residue_name(:) = ''
       psf%atom_name(:) = ''
       psf%atom_cls_name(:) = ''

    case (PsfBond)

       if (associated(psf%bond_list)) then
          if (SIZE(psf%bond_list(1,:)) == var_size) &
             return
          deallocate(psf%bond_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%bond_list(2, var_size), &
                stat = alloc_stat)

    case (PsfAngl)

       if (associated(psf%angl_list)) then
          if (SIZE(psf%angl_list(1,:)) == var_size) &
             return
          deallocate(psf%angl_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%angl_list(3, var_size), &
                stat = alloc_stat)

    case (PsfDihe)

       if (associated(psf%Dihe_list)) then
          if (SIZE(psf%dihe_list(1,:)) == var_size) &
             return
          deallocate(psf%dihe_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%dihe_list(4, var_size), &
                stat = alloc_stat)

    case (PsfImpr)

       if (associated(psf%impr_list)) then
          if (SIZE(psf%impr_list(1,:)) == var_size) &
             return
          deallocate(psf%impr_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%impr_list(4, var_size), &
                stat = alloc_stat)

    case (PsfDonr)

       if (associated(psf%donr_list)) then
          if (SIZE(psf%donr_list(1,:)) == var_size) &
             return
          deallocate(psf%donr_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%donr_list(2, var_size), &
                stat = alloc_stat)

    case (PsfAcce)

       if (associated(psf%acce_list)) then
          if (SIZE(psf%acce_list(1,:)) == var_size) &
             return
          deallocate(psf%acce_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%acce_list(2, var_size), &
                stat = alloc_stat)

    case (PsfNb)

       if (associated(psf%nb_list)) then
          if (SIZE(psf%nb_list) == var_size) &
             return
          deallocate(psf%nb_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%nb_list(var_size), &
                stat = alloc_stat)

    case (PsfGrp)

       if (associated(psf%grp_list)) then
          if (SIZE(psf%grp_list(1,:)) == var_size) &
             return
          deallocate(psf%grp_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%grp_list(3, var_size), &
                stat = alloc_stat)

    case (PsfCmap)

       if (associated(psf%cmap_list)) then
          if (SIZE(psf%cmap_list) == var_size) &
             return
          deallocate(psf%cmap_list, &
                     stat = dealloc_stat)
       end if

       allocate(psf%cmap_list(8, var_size), &
                stat = alloc_stat)

    case default

       call error_msg('Alloc_Psf> bad variable')

    end select

!    write(MsgOut,*) 'DEBUG::s_psf::alloc_psf::  VAR(',variable,') '//&
!         'SIZ(', var_size, ')'

    if (alloc_stat /= 0)   call error_msg_alloc
    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine alloc_psf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  dealloc_psf
  !> @brief      deallocate PSF information
  !! @authors    NT, YS
  !! @param[in]  psf      : structure of PSF information
  !! @param[in]  variable : selected variable
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_psf(psf, variable)

    ! formal arguments
    type(s_psf),          intent(inout) :: psf
    integer,              intent(in)    :: variable

    ! local variables
    integer                             :: dealloc_stat
   

    ! initialize
    !
    dealloc_stat = 0

    ! deallocate selected variable
    !
    select case (variable)

    case(PsfAtom)

       if (associated(psf%atom_no)) then
          deallocate(psf%atom_no,         &
                     psf%segment_name,    &
                     psf%segment_no,      &
                     psf%residue_no,      &
                     psf%residue_c_no,    &
                     psf%residue_name,    &
                     psf%atom_name,       &
                     psf%atom_cls_name,   &
                     psf%atom_cls_no,     &
                     psf%charge,          &
                     psf%mass,            &
                     psf%imove,           &
                     psf%atom_connect_no, &
                     psf%atom_connect,    &
                     psf%molecule_no,     &
                     stat = dealloc_stat)
          psf%atom_no         => null()
          psf%segment_name    => null()
          psf%segment_no      => null()
          psf%residue_no      => null()
          psf%residue_c_no    => null()
          psf%residue_name    => null()
          psf%atom_name       => null()
          psf%atom_cls_name   => null()
          psf%atom_cls_no     => null()
          psf%charge          => null()
          psf%mass            => null()
          psf%imove           => null()
          psf%atom_connect_no => null()
          psf%atom_connect    => null()
          psf%molecule_no     => null()
       end if

    case (PsfBond)

       if (associated(psf%bond_list)) then
          deallocate(psf%bond_list, &
                     stat = dealloc_stat)
          psf%bond_list => null()
       end if

    case (PsfAngl)

       if (associated(psf%angl_list)) then
          deallocate(psf%angl_list, &
                     stat = dealloc_stat)
          psf%angl_list => null()
       end if

    case (PsfDihe)

       if (associated(psf%dihe_list)) then
          deallocate(psf%dihe_list, &
                     stat = dealloc_stat)
          psf%dihe_list => null()
       end if


    case (PsfImpr)

       if (associated(psf%impr_list)) then
          deallocate(psf%impr_list, &
                     stat = dealloc_stat)
          psf%impr_list => null()
       end if


    case (PsfDonr)

       if (associated(psf%donr_list)) then
          deallocate(psf%donr_list, &
                     stat = dealloc_stat)
          psf%donr_list => null()
       end if


    case (PsfAcce)

       if (associated(psf%acce_list)) then
          deallocate(psf%acce_list, &
                     stat = dealloc_stat)
          psf%acce_list => null()
       end if


    case (PsfNb)

       if (associated(psf%nb_list)) then
          deallocate(psf%nb_list, &
                     stat = dealloc_stat)
          psf%nb_list => null()
       end if


    case (PsfGrp)

       if (associated(psf%grp_list)) then
          deallocate(psf%grp_list, &
                     stat = dealloc_stat)
          psf%grp_list => null()
       end if


    case (PsfCmap)

       if (associated(psf%cmap_list)) then
          deallocate(psf%cmap_list, &
                     stat = dealloc_stat)
          psf%cmap_list => null()
       end if

    case default

       call error_msg('Dealloc_Psf> bad variable')

    end select

    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine dealloc_psf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  dealloc_psf_all
  !> @brief      deallocate PSF information
  !! @authors    NT, YS
  !! @param[in]  psf : structure of PSF information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_psf_all(psf)

    ! formal arguments
    type(s_psf),          intent(inout) :: psf


    call dealloc_psf(psf, PsfAtom)
    call dealloc_psf(psf, PsfBond)
    call dealloc_psf(psf, PsfAngl)
    call dealloc_psf(psf, PsfDihe)
    call dealloc_psf(psf, PsfImpr)
    call dealloc_psf(psf, PsfDonr)
    call dealloc_psf(psf, PsfAcce)
    call dealloc_psf(psf, PsfNb  )
    call dealloc_psf(psf, PsfGrp )
    call dealloc_psf(psf, PsfCmap)

    return

  end subroutine dealloc_psf_all

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  input_psf
  !> @brief      a driver subroutine for reading PSF file
  !! @authors    NT, YS
  !! @param[in]  psf_filename : filename of PSF file
  !! @param[in]  psf          : structure of PSF information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_psf(psf_filename, psf)

    ! formal arguments
    character(*), intent(in)    :: psf_filename
    type(s_psf),  intent(inout) :: psf
    
    ! local variables
    integer                     :: file

    ! open PSF file
    !
    call open_file(file, psf_filename, IOFileInput)

    ! read coordinate data from PSF file
    !
    call read_psf(file, psf)

    ! close PSF file
    !
    call close_file(file)

    return

  end subroutine input_psf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  output_psf
  !> @brief      a driver subroutine for writing PSF file
  !! @authors    NT, YS
  !! @param[in]  psf_filename : filename of PSF file
  !! @param[in]  psf          : structure of PSF information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_psf(psf_filename, psf)

    ! formal arguments
    character(*), intent(in)    :: psf_filename
    type(s_psf),  intent(in)    :: psf
    
    ! local variables
    integer                     :: file

    ! open PSF file
    !
    call open_file(file, psf_filename, IOFileOutputNew)

    ! write coordinate data from PSF file
    !
    call write_psf(file, psf)

    ! close PSF file
    !
    call close_file(file)

    return

  end subroutine output_psf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf
  !> @brief      read data from PSF file
  !! @authors    NT, YS
  !! @param[in]  file : unit number of PSF file
  !! @param[in]  psf  : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf(file, psf)
       
    ! formal arugments
    integer,             intent(in)   :: file
    type(s_psf),         intent(out)  :: psf
 
    ! local variables
    type(s_psf_hdr_info) :: hdr_info
    integer              :: n1, n2
    character(100)       :: key
    logical              :: can_read


    ! deallocate old data
    !
    call dealloc_psf_all(psf)
    call init_psf(psf)

    ! read header section
    !
    call read_psf_header(file, hdr_info)

    do while(.true.)

       can_read = .false.
       read(file, hdr_info%fmt06,err=10,end=10) n1, n2, key
       can_read = .true.

10     if (.not. can_read) then
          BACKSPACE(file)
          read(file, hdr_info%fmt01,err=100,end=100) n1, key
       end if

       if (key(1:5) == 'NATOM') then

          ! read atom section
          !
          call read_psf_atom(file, n1, hdr_info, psf)

       else if (key(1:5) == 'NBOND') then
 
          ! read bond section
          !
          call read_psf_bond(file, n1, hdr_info, psf)
 
       else if (key(1:6) == 'NTHETA') then
 
          ! read angle section
          !
          call read_psf_angl(file, n1, hdr_info, psf)
 
       else if (key(1:4) == 'NPHI') then
 
          ! read dihedral section
          !
          call read_psf_dihe(file, n1, hdr_info, psf)
 
       else if (key(1:6) =='NIMPHI') then
 
          ! read improper section
          !
          call read_psf_impr(file, n1, hdr_info, psf)
 
       else if (key(1:4) == 'NDON') then
 
          ! read donor
          !
          call read_psf_donr(file, n1, hdr_info, psf)
 
       else if (key(1:4) == 'NACC') then
 
          ! read acceptor
          !
          call read_psf_acce(file, n1, hdr_info, psf)
 
       else if (key(1:3) == 'NNB') then
 
          ! read nb
          !
          call read_psf_nb(file, n1, hdr_info, psf)
 
       else if (key(1:4) == 'NGRP') then
 
          ! read grp
          !
          call read_psf_grp(file, n1, hdr_info, psf)
 
       else if (key(1:7) == 'NCRTERM') then
 
          ! read cmap section
          !
          call read_psf_cmap(file, n1, hdr_info, psf)

       end if

    end do
100 continue

    ! check PSF information [charge, residues, 
    !
    call check_psf_info(psf)

    ! write summary of PSF information
    !
    write(MsgOut,'(A)') 'Read_Psf> Summary of PSF file'
    if (psf%type == PsfTypeXPLOR) then
      write(MsgOut,'(A)') '  psftype         =      xplor'
    else
      write(MsgOut,'(A)') '  psftype         =     charmm'
    endif

    write(MsgOut,'(A20,I10,A20,I10)')                  &
         '  num_atoms       = ', psf%num_atoms,        &
         '  num_residues    = ', psf%num_residues
    write(MsgOut,'(A20,I10,A20,I10)')                  &
         '  num_molecules   = ', psf%num_molecules,    &
         '  num_segments    = ', psf%num_segments
    write(MsgOut,'(A20,I10,A20,I10)')                  &
         '  num_bonds       = ', psf%num_bonds,        &
         '  num_angles      = ', psf%num_angles
    write(MsgOut,'(A20,I10,A20,I10)')                  &
         '  num_dihedrals   = ', psf%num_dihedrals,    &
         '  num_impropers   = ', psf%num_impropers
    write(MsgOut,'(A20,I10,A20,I10)')                  &
         '  num_cmap_terms  = ', psf%num_cross_terms,  &
         '  num_HB_donors   = ', psf%num_HB_donors
    write(MsgOut,'(A20,I10,A20,I10)')                  &
         '  num_HB_acceptors= ', psf%num_HB_acceptors, &
         '  num_NB_exclusion= ', psf%num_NB_exclusions
    write(MsgOut,'(A20,I10,A20,F10.3)')                &
         '  num_groups      = ', psf%num_groups,       &
         '  total_charge    = ', psf%total_charge
    write(MsgOut,'(A)') ' '

    return

  end subroutine read_psf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf
  !> @brief      write data to PSF file
  !! @authors    NT, YS
  !! @param[in]  file : unit number of PSF file
  !! @param[in]  psf  : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf(file, psf)

    ! formal arguments
    integer,             intent(in) :: file
    type(s_psf),         intent(in) :: psf
 
    ! local variables
    type(s_psf_hdr_info) :: hdr_info


    ! write summary of PSF information
    !
    if (psf%type == PsfTypeXPLOR) then
      write(MsgOut,*) 'Write_Psf> TYPE = XPLOR'
    else
      write(MsgOut,*) 'Write_Psf> TYPE = CHARMM'
    endif

    write(MsgOut,*) 'Write_Psf> Number of Atoms         = ', psf%num_atoms
    write(MsgOut,*) 'Write_Psf> Number of Residues      = ', psf%num_residues
    write(MsgOut,*) 'Write_Psf> Number of Molecules     = ', psf%num_molecules
    write(MsgOut,*) 'Write_Psf> Number of Segments      = ', psf%num_segments
    write(MsgOut,*) 'Write_Psf> Number of Bonds         = ', psf%num_bonds
    write(MsgOut,*) 'Write_Psf> Number of Angles        = ', psf%num_angles
    write(MsgOut,*) 'Write_Psf> Number of Dihedrals     = ', psf%num_dihedrals
    write(MsgOut,*) 'Write_Psf> Number of Impropers     = ', psf%num_impropers
    write(MsgOut,*) 'Write_Psf> Number of Cross-terms   = ', psf%num_cross_terms
    write(MsgOut,*) 'Write_Psf> Number of HB acceptors  = ', psf%num_HB_acceptors
    write(MsgOut,*) 'Write_Psf> Number of HB donors     = ', psf%num_HB_donors
    write(MsgOut,*) 'Write_Psf> Number of NB exclusions = ', psf%num_NB_exclusions
    write(MsgOut,*) 'Write_Psf> Number of Groups        = ', psf%num_groups
    write(MsgOut,*) 'Write_Psf> Total charge            = ', psf%total_charge
    write(MsgOut,*) ' '

    !  write header section
    !
    call write_psf_header(file, hdr_info)

    !  write atom section
    !
    call write_psf_atom(file, hdr_info, psf)

    !  write bond section
    !
    call write_psf_bond(file, hdr_info, psf)

    !  write angle section
    !
    call write_psf_angl(file, hdr_info, psf)

    !  write dihedral section
    !
    call write_psf_dihe(file, hdr_info, psf)

    !  write improper section
    !
    call write_psf_impr(file, hdr_info, psf)

    !  write donor section
    !
    call write_psf_donr(file, hdr_info, psf)

    !  write acceptor section
    !
    call write_psf_acce(file, hdr_info, psf)

    !  write nb section
    !
    call write_psf_nb(file, hdr_info, psf)

    !  write grp section
    !
    call write_psf_grp(file, hdr_info, psf)

    !  write cmap section
    !
    call write_psf_cmap(file, hdr_info, psf)

    return

  end subroutine write_psf

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  check_psf_info
  !> @brief      check PSF
  !! @authors    NT, YS
  !! @param[in]  psf  : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_psf_info(psf)

    ! formal arguments
    type(s_psf), intent(inout) :: psf

    ! local variables
    integer          :: i, j, k
    integer          :: resno_pre, resno_next
    integer          :: i1, i2
    integer          :: mi, mj, mk, nmol
    character(4)     :: segment_pre, segment_next
    character(4)     :: residue_pre, residue_next
    character(4)     :: cls1, cls2
    double precision :: ttl_charge
 
    ! total charge
    !
    ttl_charge = 0.0d0
    do i = 1, psf%num_atoms
      ttl_charge = ttl_charge + psf%charge(i)
    end do
    psf%total_charge = ttl_charge

    ! count the number of segments
    !
    do i = 1, psf%num_atoms
      if (i == 1) then
        psf%segment_no(1) = 1
        segment_pre = psf%segment_name(1)
      else
        if (psf%segment_name(i) /= segment_pre) then
          psf%segment_no(i) = psf%segment_no(i-1) + 1
          segment_pre       = psf%segment_name(i)
        else
          psf%segment_no(i) = psf%segment_no(i-1)
        end if
      end if
    end do
    psf%num_segments = psf%segment_no(psf%num_atoms)

    ! count the number of residues
    !
    do i = 1, psf%num_atoms
      if (i == 1) then
        segment_pre = psf%segment_name(i)
        residue_pre = psf%residue_name(i)
        resno_pre   = psf%residue_no(i)
        psf%residue_c_no(i) = 1
      else
        segment_pre  = psf%segment_name(i-1)
        residue_pre  = psf%residue_name(i-1)
        resno_pre    = psf%residue_no(i-1)
        segment_next = psf%segment_name(i)
        residue_next = psf%residue_name(i)
        resno_next   = psf%residue_no(i)
        
        if ((segment_pre /= segment_next) .or. &
            (residue_pre /= residue_next) .or. &
            (resno_pre   /= resno_next)) then
          psf%residue_c_no(i) = psf%residue_c_no(i-1) + 1
        else
          psf%residue_c_no(i) = psf%residue_c_no(i-1)
        end if
      end if
    end do
    psf%num_residues = psf%residue_c_no(psf%num_atoms)

    ! check connectivities
    !
    psf%atom_connect_no(:) = 0
    
    do i = 1, psf%num_bonds
      i1   = psf%bond_list(1,i)
      i2   = psf%bond_list(2,i)
      cls1 = psf%atom_cls_name(i1)
      cls2 = psf%atom_cls_name(i2)
      
      if (((cls1(1:1) == "H") .or. (cls1(1:2) == " H")) .and. &
          ((cls2(1:1) == "H") .or. (cls2(1:2) == " H")))      &
        cycle
      
      psf%atom_connect_no(i1) = psf%atom_connect_no(i1) + 1
      psf%atom_connect_no(i2) = psf%atom_connect_no(i2) + 1
      
      psf%atom_connect(psf%atom_connect_no(i1), i1) = i2
      psf%atom_connect(psf%atom_connect_no(i2), i2) = i1
    end do

    ! count the number of molecules 
    ! this algorithm is taken from tinker (molecule.f)
    !
    nmol = 0
    psf%molecule_no(:) = 0

    do i = 1, psf%num_atoms
      if (psf%molecule_no(i) == 0) then
        nmol = nmol + 1
        psf%molecule_no(i) = nmol
      end if
      mi = psf%molecule_no(i)
      do i2 = 1, psf%atom_connect_no(i)
        j = psf%atom_connect(i2,i)
        mj = psf%molecule_no(j)
        if (mj == 0) then
          psf%molecule_no(j) = mi
        else if (mi < mj) then
          nmol = nmol - 1
          do k = 1, psf%num_atoms
            mk = psf%molecule_no(k)
            if (mk == mj) then
              psf%molecule_no(k) = mi
            else if (mk > mj) then
              psf%molecule_no(k) = mk - 1
            end if
          end do
        else if (mi > mj) then
          nmol = nmol - 1
          do k = 1, psf%num_atoms
            mk = psf%molecule_no(k)
            if (mk == mi) then
              psf%molecule_no(k) = mj
            else if (mk > mi) then
              psf%molecule_no(k) = mk - 1
            end if
          end do
          mi = mj
        end if
      end do
    end do

    psf%num_molecules = psf%molecule_no(psf%num_atoms)

    return

  end subroutine check_psf_info

  !**********************************************************************

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_header
  !> @brief      read header information of PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_header(file, hdr_info)

    ! parameter
    integer,                 parameter     :: MAXROW = 80

    ! formal arguments
    integer,                 intent(in)    :: file
    type(s_psf_hdr_info),    intent(out)   :: hdr_info

    ! local variables
    character(100)           :: line, title
    integer                  :: i, ntitle


    read(file,'(a80)') line

    hdr_info%ipsf  = .false.
    hdr_info%iext  = .false.
    hdr_info%icmap = .false.
    hdr_info%icheq = .false.

    do i = 1, MAXROW
       if (line(i:i+3) == 'PSF ') hdr_info%ipsf  = .true.
       if (line(i:i+3) == 'EXT ') hdr_info%iext  = .true.
       if (line(i:i+3) == 'CMAP') hdr_info%icmap = .true.
       if (line(i:i+3) == 'CHEQ') hdr_info%icheq = .true.
    end do

    if (.not. hdr_info%ipsf) then
       call error_msg('Read_Psf_Header> ERROR: Format is not correct')

    else if (hdr_info%iext) then
     !  write(MsgOut,*) '  Read_Psf_Header> SORRY: EXT is not supported'

    else if (hdr_info%icmap) then
     !  write(MsgOut,*) '  Read_Psf_Header> SORRY: CMAP is not supported'
     !  stop

    else if (hdr_info%icheq) then
       write(MsgOut,*) '  Read_Psf_Header> SORRY: CHEQ is not supported'
    end if

    if (hdr_info%iext) then
       ! EXTENDED format (# of atoms exceeds 100,000)
       hdr_info%fmt00 = '(/I10)'
       hdr_info%fmt01 = '(I10,2X,A20)'
       hdr_info%fmt02 = '(I10,1X,A8,1X,A8,1X,A8,1X,A8,1X,I4,1X,2G14.6,I8)'
       hdr_info%fmt03 = '(I10,1X,A8,1X,A8,1X,A8,1X,A8,1X,A4,1X,2G14.6,I8)'
       hdr_info%fmt04 = '(8I10)'
       hdr_info%fmt05 = '(9I10)'
       hdr_info%fmt06 = '(2I10,2X,A20)'
    else
       hdr_info%fmt00 = '(/I8)'
       hdr_info%fmt01 = '(I8,2X,A20)'
       hdr_info%fmt02 = '(I8,1X,A4,1X,A4,1X,A4,1X,A4,1X,I4,1X,2G14.6,I8)'
       hdr_info%fmt03 = '(I8,1X,A4,1X,A4,1X,A4,1X,A4,1X,A4,1X,2G14.6,I8)'
       hdr_info%fmt04 = '(8I8)'
       hdr_info%fmt05 = '(9I8)'
       hdr_info%fmt06 = '(2I8,2X,A20)'
    end if

    read(file, hdr_info%fmt00) ntitle
    do i = 1, ntitle
       read(file, '(a80)') title
    end do

    return

  end subroutine read_psf_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_atom
  !> @brief      read atom information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  natom    : total number of atoms
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_atom(file, natom, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: natom
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    character(8)             :: cstr, cres_no, cseg_nm, cres_nm, catm_nm
    integer                  :: i

      
    ! allocate Atom variables
    !
    call alloc_psf(psf, PsfAtom, natom)

    ! Check psf file type
    !
    if (hdr_info%iext) then
       read(file, '(47X,A4)') cstr
    else
       read(file, '(29X,A4)') cstr
    end if
    if (cstr(1:1) >= 'A' .and. cstr(1:1) <= 'Z' .or. &
        cstr(1:1) >= 'a' .and. cstr(1:1) <= 'z') then
       psf%type = PsfTypeXPLOR
    else
       psf%type = PsfTypeCHARMM
    end if
    BACKSPACE(file)

    ! Read atom
    !
    psf%num_atoms = natom

    if (psf%type == PsfTypeCHARMM) then

       do i = 1, natom
          read(file, hdr_info%fmt02)              &
                            psf%atom_no(i),       &
                            cseg_nm,              &
                            cres_no,              &
                            cres_nm,              &
                            catm_nm,              &
                            psf%atom_cls_no(i),   &
                            psf%charge(i),        &
                            psf%mass(i),          &
                            psf%imove(i)
          read(cres_no, '(i4)') psf%residue_no(i)
          psf%segment_name(i) = cseg_nm(1:4)
          psf%residue_name(i) = cres_nm(1:4)
          psf%atom_name(i)    = catm_nm(1:4)
       end do

    else  ! PsfTypeXPLOR

       do i = 1, natom
          read(file, hdr_info%fmt03)              &
                            psf%atom_no(i),       &
                            psf%segment_name(i),  &
                            cres_no,              &
                            psf%residue_name(i),  &
                            psf%atom_name(i),     &
                            psf%atom_cls_name(i), &
                            psf%charge(i),        &
                            psf%mass(i),          &
                            psf%imove(i)
          read(cres_no, '(i4)') psf%residue_no(i)
       end do

    end if

    return

  end subroutine read_psf_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_bond
  !> @brief      read bond information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  nbnd     : total number of bonds
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_bond(file, nbnd, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: nbnd
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    integer                  :: i


    ! Allocate Bond variables
    !
    call alloc_psf(psf, PsfBond, nbnd)


    ! Read Bond
    !
    psf%num_bonds = nbnd

    read(file, hdr_info%fmt04) (psf%bond_list(1, i), &
                                psf%bond_list(2, i), &
                                i = 1, nbnd)

    return

  end subroutine read_psf_bond

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_angl
  !> @brief      read angle information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  nang     : total number of angles
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_angl(file, nang, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: nang
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    integer                  :: i


    ! Allocate Angle variables
    !
    call alloc_psf(psf, PsfAngl, nang)


    ! Read Angl
    !
    psf%num_angles = nang

    read(file, hdr_info%fmt05) (psf%angl_list(1, i), &
                                psf%angl_list(2, i), &
                                psf%angl_list(3, i), &
                                i = 1, nang)
    
    return

  end subroutine read_psf_angl

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_dihe
  !> @brief      read dihedral angle information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  ndih     : total number of dihedral angles
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_dihe(file, ndih, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: ndih
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    integer                  :: i


    ! Allocate Dihedral variables
    !
    call alloc_psf(psf, PsfDihe, ndih)


    ! read Dihedral
    !
    psf%num_dihedrals = ndih

    read(file, hdr_info%fmt04) (psf%dihe_list(1, i), &
                                psf%dihe_list(2, i), &
                                psf%dihe_list(3, i), &
                                psf%dihe_list(4, i), &
                                i = 1, ndih)
                               

    return

  end subroutine read_psf_dihe

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_impr
  !> @brief      read improper angle information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  nimp     : total number of improper angles
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_impr(file, nimp, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: nimp
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    integer                  :: i


    ! Allocate Improper Dihedral variables
    !
    call alloc_psf(psf, PsfImpr, nimp)


    ! read Improper
    !
    psf%num_impropers = nimp

    read(file, hdr_info%fmt04) (psf%impr_list(1, i), &
                                psf%impr_list(2, i), &
                                psf%impr_list(3, i), &
                                psf%impr_list(4, i), &
                                i = 1, nimp)

    return

  end subroutine read_psf_impr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_donr
  !> @brief      read donor information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  ndonr    : total number of donors
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_donr(file, ndonr, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: ndonr
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    integer                  :: i


    ! Allocate Donor variables
    !
    call alloc_psf(psf, PsfDonr, ndonr)


    ! read donor
    !
    psf%num_HB_donors = ndonr

    read(file, hdr_info%fmt04) (psf%donr_list(1, i), &
                                psf%donr_list(2, i), &
                                i = 1, ndonr)

    return

  end subroutine read_psf_donr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_acce
  !> @brief      read acceptor information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  nacce    : total number of acceptors
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_acce(file, nacce, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: nacce
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    integer                  :: i


    ! Allocate Acceptor variables
    !
    call alloc_psf(psf, PsfAcce, nacce)


    ! read acceptor
    !
    psf%num_HB_acceptors = nacce

    read(file, hdr_info%fmt04) (psf%acce_list(1, i), &
                                psf%acce_list(2, i), &
                                i = 1, nacce)

    return

  end subroutine read_psf_acce

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_nb
  !> @brief      read nb information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  nnb      : total number of nbs
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_nb(file, nnb, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: nnb
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    integer                  :: i


    ! Allocate NB variables
    !
    call alloc_psf(psf, PsfNb, psf%num_atoms)


    ! read nb
    !
    psf%num_NB_exclusions = nnb

    read(file, hdr_info%fmt04) (psf%nb_list(i), &
                                i = 1, psf%num_atoms)

    return

  end subroutine read_psf_nb

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_grp
  !> @brief      read group information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  ngrp     : total number of groups
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_grp(file, ngrp, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: ngrp
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! integer
    integer                  :: i


    ! Allocate Grp variables
    !
    call alloc_psf(psf, PsfGrp, ngrp)


    ! read nb
    !
    psf%num_groups = ngrp

    read(file, hdr_info%fmt05) (psf%grp_list(1, i), &
                                psf%grp_list(2, i), &
                                psf%grp_list(3, i), &
                                i = 1, ngrp)

    return

  end subroutine read_psf_grp

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_psf_cmap
  !> @brief      read cmap information from PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  ncmap    : total number of cmaps
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_psf_cmap(file, ncmap, hdr_info, psf)

    ! formal arguments
    integer,                 intent(in)    :: file
    integer,                 intent(in)    :: ncmap
    type(s_psf_hdr_info),    intent(in)    :: hdr_info
    type(s_psf),             intent(inout) :: psf

    ! local variables
    integer                  :: i


    !  Allocate cross-term variables
    !
    call alloc_psf(psf, PsfCmap, ncmap)


    ! read cross-terms
    !
    psf%num_cross_terms = ncmap

    read(file, hdr_info%fmt04) (psf%cmap_list(1, i), &
                                psf%cmap_list(2, i), &
                                psf%cmap_list(3, i), &
                                psf%cmap_list(4, i), &
                                psf%cmap_list(5, i), &
                                psf%cmap_list(6, i), &
                                psf%cmap_list(7, i), &
                                psf%cmap_list(8, i), &
                                i = 1, ncmap)

    return

  end subroutine read_psf_cmap

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_header
  !> @brief      write header information of PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_header(file, hdr_info)

    ! parameter
    integer,              parameter   :: MAXROW = 80

    ! formal arguments
    integer,              intent(in)  :: file
    type(s_psf_hdr_info), intent(out) :: hdr_info

    ! local variables
    character(8)          date
    character(10)         time
    character(5)          zone
    integer               date_time(8)


    write(file,'(a)') 'PSF'

    hdr_info%ipsf  = .true.  ! now unselectable
    hdr_info%iext  = .false. ! now unselectable
    hdr_info%icmap = .false. ! now unselectable
    hdr_info%icheq = .false. ! now unselectable

    hdr_info%fmt00 = '(/I8)'
    hdr_info%fmt01 = '(/I8,1X,"!",A)'
    hdr_info%fmt02 = '(I8,1X,A4,1X,A4,1X,A4,1X,A4,1X,I4,1X,2G14.6,I8)'
    hdr_info%fmt03 = '(I8,1X,A4,1X,A4,1X,A4,1X,A4,1X,A4,1X,2G14.6,I8)'
    hdr_info%fmt04 = '(8I8)'
    hdr_info%fmt05 = '(9I8)'
    hdr_info%fmt06 = '(/2I8,1X,"!",A)'

    write(file,hdr_info%fmt01) 1,'NTITLE'

    call DATE_AND_TIME(date, time, zone, date_time)
    write(file,'("*  DATE:",i6,"/",i2"/",i2,i7,":",i2,":",i2)') &
             date_time(2), date_time(3), MOD(date_time(1),100), &
             date_time(5), date_time(6), date_time(7)

    return

  end subroutine write_psf_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_atom
  !> @brief      write atom section to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_atom(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    character(8)          :: cres_no
    integer               :: i, natom

      
    ! write the number of atoms
    !
    if (associated(psf%atom_no)) then
       natom = SIZE(psf%atom_no)
    else
       natom = 0
    end if

    write(file,hdr_info%fmt01) natom, 'NATOM'
    

    ! write atom
    !
    if (psf%type == PsfTypeCHARMM) then

       do i = 1, natom
          write(cres_no, '(i4)') psf%residue_no(i)
          write(file, hdr_info%fmt02) &
               psf%atom_no(i),        &
               psf%segment_name(i),   &
               ADJUSTL(cres_no),      &
               psf%residue_name(i),   &
               psf%atom_name(i),      &
               psf%atom_cls_no(i),    &
               psf%charge(i),         &
               psf%mass(i),           &
               psf%imove(i)
       end do

    else  ! PsfTypeXPLOR

       do i = 1, natom
           write(cres_no, '(i4)') psf%residue_no(i)
          write(file, hdr_info%fmt03) &
               psf%atom_no(i),        &
               psf%segment_name(i),   &
               ADJUSTL(cres_no),      &
               psf%residue_name(i),   &
               psf%atom_name(i),      &
               psf%atom_cls_name(i),  &
               psf%charge(i),         &
               psf%mass(i),           &
               psf%imove(i)
       end do
    end if

    return

  end subroutine write_psf_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_bond
  !> @brief      write bond information to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_bond(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, nbnd


    ! write the number of bonds
    !
    if (associated(psf%bond_list)) then
       nbnd = SIZE(psf%bond_list(1,:))
    else
       nbnd = 0
    end if

    write(file, hdr_info%fmt01) nbnd, 'NBOND: bonds'
    

    ! write bond
    !
    write(file, hdr_info%fmt04) (psf%bond_list(1, i), &
                                 psf%bond_list(2, i), &
                                 i = 1, nbnd)

    return

  end subroutine write_psf_bond

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_angl
  !> @brief      write angle information in PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_angl(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, nang


    ! write the number of angles
    !
    if (associated(psf%angl_list)) then
       nang = SIZE(psf%angl_list(1,:))
    else
       nang = 0
    end if

    write(file, hdr_info%fmt01) nang, 'NTHETA: angles'
    

    ! write angle
    !
    write(file, hdr_info%fmt05) (psf%angl_list(1, i), &
                                 psf%angl_list(2, i), &
                                 psf%angl_list(3, i), &
                                 i = 1, nang)

    return

  end subroutine write_psf_angl

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_dihe
  !> @brief      write dihedral angle information to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_dihe(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, ndih


    ! write the number of angles
    !
    if (associated(psf%dihe_list)) then
       ndih = SIZE(psf%dihe_list(1,:))
    else
       ndih = 0
    end if

    write(file, hdr_info%fmt01) ndih, 'NPHI: dihedrals'
    

    ! write dihe
    !
    write(file, hdr_info%fmt04) (psf%dihe_list(1, i), &
                                 psf%dihe_list(2, i), &
                                 psf%dihe_list(3, i), &
                                 psf%dihe_list(4, i), &
                                 i = 1, ndih)

    return

  end subroutine write_psf_dihe

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_impr
  !> @brief      write improper angle information to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_impr(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, nimp


    ! write the number of impropers
    !
    if (associated(psf%impr_list)) then
       nimp = SIZE(psf%impr_list(1,:))
    else
       nimp = 0
    end if

    write(file, hdr_info%fmt01) nimp, 'NIMPHI: impropers'
    

    ! Write impr
    !
    write(file, hdr_info%fmt04) (psf%impr_list(1, i), &
                                 psf%impr_list(2, i), &
                                 psf%impr_list(3, i), &
                                 psf%impr_list(4, i), &
                                 i = 1, nimp)

    return

  end subroutine write_psf_impr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_donr
  !> @brief      write donor information to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_donr(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, ndon


    ! write the number of donors
    !
    if (associated(psf%donr_list)) then
       ndon = SIZE(psf%donr_list(1,:))
    else
       ndon = 0
    end if

    write(file, hdr_info%fmt01) ndon, 'NDON: donors'

    ! write donors
    write(file, hdr_info%fmt04) (psf%donr_list(1, i), &
                                 psf%donr_list(2, i), &
                                 i = 1, ndon)

    return

  end subroutine write_psf_donr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_acce
  !> @brief      write acceptor information to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_acce(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, nacc


    ! write the number of acceptors
    !
    if (associated(psf%acce_list)) then
       nacc = SIZE(psf%acce_list(1,:))
    else
       nacc = 0
    end if

    write(file, hdr_info%fmt01) nacc, 'NACC: acceptors'

    ! write acceptors
    write(file, hdr_info%fmt04) (psf%acce_list(1, i), &
                                 psf%acce_list(2, i), &
                                 i = 1, nacc)

    return

  end subroutine write_psf_acce

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_nb
  !> @brief      write nb information to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_nb(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, nnb


    nnb = psf%num_atoms

    ! write the number of nb
    !
    write(file, hdr_info%fmt01) psf%num_NB_exclusions, 'NNB'

    ! write nb
    !
    if (associated(psf%nb_list)) then
       write(file, hdr_info%fmt04) (psf%nb_list(i), i = 1, nnb)
    else
       write(file, hdr_info%fmt04) (0, i = 1, nnb)
    end if

    return

  end subroutine write_psf_nb

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_grp
  !> @brief      write group information to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_grp(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, ngrp


    ! write the number of groups
    !
    if (associated(psf%grp_list)) then
       ngrp = SIZE(psf%grp_list(1,:))
    else
       ngrp = 0
    end if

    write(file, hdr_info%fmt06) ngrp, 0, 'NGRP'

    write(file, hdr_info%fmt05) (psf%grp_list(1, i), &
                                 psf%grp_list(2, i), &
                                 psf%grp_list(3, i), &
                                 i = 1, ngrp)

    return

  end subroutine write_psf_grp

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  write_psf_cmap
  !> @brief      write cmap information to PSF file
  !! @authors    NT, YS
  !! @param[in]  file     : unit number of PSF file
  !! @param[in]  hdr_info : header information
  !! @param[in]  psf      : psf data
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_psf_cmap(file, hdr_info, psf)

    ! formal arguments
    integer,              intent(in) :: file
    type(s_psf_hdr_info), intent(in) :: hdr_info
    type(s_psf),          intent(in) :: psf

    ! local variables
    integer               :: i, ncmap


    ! write the number of cross-term 
    !
    if (associated(psf%cmap_list)) then
       ncmap = SIZE(psf%cmap_list(1,:))
    else
       ncmap = 0
    end if

    write(file, hdr_info%fmt01) ncmap, 'NCRTERM: cross-terms'
    

    ! write impr
    !
    write(file, hdr_info%fmt04) (psf%cmap_list(1, i), &
                                 psf%cmap_list(2, i), &
                                 psf%cmap_list(3, i), &
                                 psf%cmap_list(4, i), &
                                 psf%cmap_list(5, i), &
                                 psf%cmap_list(6, i), &
                                 psf%cmap_list(7, i), &
                                 psf%cmap_list(8, i), &
                                 i = 1, ncmap)

    return

  end subroutine write_psf_cmap

end module fileio_psf_mod
