!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   fileio_trj_mod
!> @brief   trajectory information I/O
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM), Yuji Sugita (YS)
!! @date    2012/05/25 
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

module fileio_trj_mod

  use trajectory_str_mod
  use molecules_str_mod
  use selatoms_str_mod
  use fileio_mod
  use messages_mod

  implicit none
  private

  ! gromacs trj handle
  integer, parameter, private :: GroUnitNo1 = 1000
  integer, parameter, private :: GroMaxHandle = 10
  integer(8),         private :: g_gro_handle(GroMaxHandle) = 0

  ! endian
  integer, parameter, private :: EndianNative = 1
  integer, parameter, private :: EndianBig    = 2
  integer, parameter, private :: EndianLittle = 3

  ! subroutines
  public  :: open_trj
  public  :: close_trj
  public  :: read_trj
  public  :: write_trj
  public  :: seek_trj

  private :: open_pdb_trj
  private :: close_pdb_trj
  private :: read_pdb_trj
  private :: write_pdb_trj_sel
  private :: write_pdb_trj

  private :: open_marble_trj
  private :: close_marble_trj
  private :: read_marble_trj
  private :: write_marble_trj_sel
  private :: write_marble_trj
  private :: skip_marble_trj_header
  private :: write_marble_trj_header

  private :: open_amber_trj
  private :: close_amber_trj
  private :: read_amber_trj
  private :: write_amber_trj_sel
  private :: write_amber_trj
  private :: skip_amber_trj_header
  private :: write_amber_trj_header

  private :: open_tinker_trj
  private :: close_tinker_trj
  private :: read_tinker_trj
  private :: write_tinker_trj_sel
  private :: write_tinker_trj

  private :: open_dcd_trj
  private :: close_dcd_trj
  private :: read_dcd_trj
  private :: read_dcd_trj_header
  private :: write_dcd_trj_header
  private :: write_dcd_trj_sel
  private :: write_dcd_trj
  private :: seek_dcd_trj

  private :: open_gromacs_trj
  private :: close_gromacs_trj
  private :: read_gromacs_trj
  private :: write_gromacs_trj_sel
  private :: write_gromacs_trj

  private :: swap_i4
  private :: swap_r4
  private :: swap_r8
  private :: boxToXtlabc
  private :: torad
  private :: todeg
  private :: rotate
  private :: jacobi

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_trj
  !> @brief        Open trajectory file
  !! @authors      NT, YS
  !! @param[inout] file        trajectory file handle
  !! @param[in]    filename    trajectory file name
  !! @param[in]    trj_format  trajectory file format
  !! @param[in]    trj_type    trajectory file type
  !! @param[in]    in_out      information for input or output
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_trj(file, filename, trj_format, trj_type, in_out)

    ! formal arguments
    type(s_trj_file),   intent(inout) :: file
    character(*),       intent(in)    :: filename
    integer,            intent(in)    :: trj_format
    integer,            intent(in)    :: trj_type
    integer,            intent(in)    :: in_out


    file%name       = filename
    file%trj_format = trj_format
    file%trj_type   = trj_type
    file%in_out     = in_out

    select case (file%trj_format)

    case (TrjFormatPDB)

      call open_pdb_trj(file)

    case (TrjFormatMarble)

      call open_marble_trj(file)

    case (TrjFormatAmber)

      call open_amber_trj(file)

    case (TrjFormatTinker)

      call open_tinker_trj(file)

    case (TrjFormatDCD)

      call open_dcd_trj(file, EndianNative)

    case (TrjFormatGromacs)

      call open_gromacs_trj(file)

    end select

    file%step_no = 1

    return

  end subroutine open_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine     close_trj
  !> @brief         Close trajectory file
  !! @authors       NT, YS
  !! @param[inout]  file trajectory file handle
  !! @date          2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_trj(file)

    ! formal arguments
    type(s_trj_file),   intent(inout) :: file


    select case (file%trj_format)

    case (TrjFormatPDB)

      call close_pdb_trj(file)

    case (TrjFormatMarble)

      call close_marble_trj(file)

    case (TrjFormatAmber)

      call close_amber_trj(file)

    case (TrjFormatTinker)

      call close_tinker_trj(file)

    case (TrjFormatDCD)

      call close_dcd_trj(file)
      call update_dcd_trj_header(file)

    case (TrjFormatGromacs)

      call close_gromacs_trj(file)

    end select

    return

  end subroutine close_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_trj
  !> @brief        Read trajectory file
  !! @authors      NT, YS
  !! @param[inout] file        trajectory file handle
  !! @param[inout] trajectory  information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),   intent(inout) :: file
    type(s_trajectory), intent(inout) :: trajectory


    select case (file%trj_format)

    case (TrjFormatPDB)

      call read_pdb_trj(file, trajectory)

    case (TrjFormatMarble)

      call skip_marble_trj_header(file)
      call read_marble_trj(file, trajectory)

    case (TrjFormatAmber)

      call skip_amber_trj_header(file)
      call read_amber_trj(file, trajectory)

    case (TrjFormatTinker)

      call read_tinker_trj(file, trajectory)

    case (TrjFormatDCD)

      call read_dcd_trj_header(file)
      call read_dcd_trj(file, trajectory)

    case (TrjFormatGromacs)

       call read_gromacs_trj(file, trajectory)

    end select

    file%step_no = file%step_no + 1

    return

  end subroutine read_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_trj
  !> @brief        Write trajectory file
  !! @authors      NT, YS
  !! @param[inout] file        trajectory file handle
  !! @param[in]    trajectory  information for trajectory
  !! @param[in]    selatoms    information for atom selection
  !! @param[in]    molecule    information for molecule
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_trj(file, trajectory, selatoms, molecule)

    ! formal arguments
    type(s_trj_file),             intent(inout) :: file
    type(s_trajectory),           intent(in)    :: trajectory
    type(s_selatoms),   optional, intent(in)    :: selatoms
    type(s_molecule),   optional, intent(in)    :: molecule


    select case (file%trj_format)

    case (TrjFormatPDB)

      if (.not. PRESENT(molecule)) then
        call error_msg('Write_Trj> Molecular information is necessary '//&
                       'for the output in the PDB format')
      end if

      if (PRESENT(selatoms)) then
        call write_pdb_trj_sel(file, trajectory, selatoms, molecule)
      else
        call write_pdb_trj(file, trajectory, molecule)
      end if

    case (TrjFormatMarble)

      if (PRESENT(selatoms)) then
        call write_marble_trj_header(file, SIZE(selatoms%idx))
        call write_marble_trj_sel(file, trajectory, selatoms)
      else
        call write_marble_trj_header(file, SIZE(trajectory%coord(1,:)))
        call write_marble_trj(file, trajectory)
      end if

    case (TrjFormatAmber)

      call write_amber_trj_header(file)

      if (PRESENT(selatoms)) then
        call write_amber_trj_sel(file, trajectory, selatoms)
      else
        call write_amber_trj(file, trajectory)
      end if

    case (TrjFormatTinker)

      if (.not. PRESENT(molecule)) then
        call error_msg('Write_Trj> Molecular information is necessary '//&
                       'for the output in the TINKER format')
      end if

      if (file%trj_type == TrjTypeCoorBox) then
        call error_msg('Write_Trj> TINKER format does not support '//&
                       'PBC box information.')
      end if

      if (PRESENT(selatoms)) then
        call write_tinker_trj_sel(file, trajectory, selatoms, molecule)
      else
        call write_tinker_trj(file, trajectory, molecule)
      end if

    case (TrjFormatDCD)

      if (PRESENT(selatoms)) then
        call write_dcd_trj_header(file, SIZE(selatoms%idx))
        call write_dcd_trj_sel(file, trajectory, selatoms)
      else
        call write_dcd_trj_header(file, SIZE(trajectory%coord(1,:)))
        call write_dcd_trj(file, trajectory)
      end if

    case (TrjFormatGromacs)

      if (PRESENT(selatoms)) then
        call write_gromacs_trj_sel(file, trajectory, selatoms)
      else
        call write_gromacs_trj(file, trajectory)
      end if

    case (TrjFormatCharmmRst)

      call error_msg('Write_Trj> Charmm restart is not supported.')

    case (TrjFormatNamdRst)

      call error_msg('Write_Trj> NAMD restart is not supported.')

    end select

    file%step_no = file%step_no + 1

    return

  end subroutine write_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine     seek_trj
  !> @brief         Seek trajectory file
  !! @authors       NT, YS
  !! @param[inout]  file       trajectory file handle
  !! @param[in]     trajectory information of trajectory 
  !! @param[in]     step_no    step number 
  !! @date          2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine seek_trj(file, trajectory, step_no)

    ! formal arguments
    type(s_trj_file),   intent(inout) :: file
    type(s_trajectory), intent(inout) :: trajectory
    integer,            intent(in)    :: step_no


    select case (file%trj_format)

    case (TrjFormatPDB)

      call error_msg('Seek_Trj> Seek is not supported. [PDB]')

    case (TrjFormatMarble)

      call error_msg('Seek_Trj> Seek is not supported. [MARBLE]')

    case (TrjFormatAmber)

      call error_msg('Seek_Trj> Seek is not supported. [AMBER]')

    case (TrjFormatTinker)

      call error_msg('Seek_Trj> Seek is not supported. [TINKER]')

    case (TrjFormatDCD)

      call read_dcd_trj_header(file)
      call seek_dcd_trj(file, trajectory, step_no)

    case (TrjFormatGromacs)

      call error_msg('Seek_Trj> Seek is not supported. [GROMACS]')

    end select
    return

  end subroutine seek_trj


  !*********************************************************************
  ! pdb

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_pdb_trj
  !> @brief        Open pdb trajectory file
  !! @authors      NT, YS
  !! @param[inout] file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_pdb_trj(file)

    ! formal arugments
    type(s_trj_file),           intent(inout) :: file
    

    call open_file(file%unit_no, file%name, file%in_out)

    return 

  end subroutine open_pdb_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_pdb_trj
  !> @brief        Close pdb trajectory file
  !! @authors      NT, YS
  !! @param[in]    file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_pdb_trj(file)

    ! formal arugments
    type(s_trj_file),           intent(in)    :: file


    call close_file(file%unit_no)

    return

  end subroutine close_pdb_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_pdb_trj
  !> @brief        Read pdb trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[inout] trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_pdb_trj(file, trajectory)
    
    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(inout) :: trajectory

    ! local variables
    character(80)               :: line
    integer                     :: i, j, natm, unit
    double precision, pointer   :: crd(:,:)


    unit =  file%unit_no
    crd  => trajectory%coord
    natm =  SIZE(crd(1,:))

    i = 0

100 read(unit, fmt='(a80)', end=900, err=900) line

    if ((line(1:4) == 'ATOM') .or. &
        (line(1:4) == 'HETA')) then

      i = i + 1 

      if (i <= natm) then

        ! atom coordinates
        read(line, fmt='(30x,3f8.3)', err=900) (crd(j,i), j=1,3)
        if (i == natm) then
          goto 200
        else
          goto 100
        end if

      end if

    else if (line(1:4) == 'CRYS') then

      ! crystal
      read(line, fmt='(6x,3f9.3)', err=900) (trajectory%pbc_box(j,j), j=1,3)
      goto 100

    else
      goto 100
    end if

200 continue

    return

900 call error_msg('Read_Pdb_Trj> read error')

  end subroutine read_pdb_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_pdb_trj_sel
  !> @brief        Write pdb trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @param[in]    selatoms   information for atom selection
  !! @param[in]    molecule   information for molecule
  !! @date         2012/05/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_pdb_trj_sel(file, trajectory, selatoms, molecule)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory
    type(s_selatoms),   target, intent(in)    :: selatoms
    type(s_molecule),   target, intent(in)    :: molecule

    ! local varialbles
    character(4),     pointer   :: nama(:), namr(:)
    integer,          pointer   :: numr(:), numm(:), aidx(:)
    integer                     :: i, j, iatm, nsel, unit, molec, molep
    double precision, pointer   :: crd(:,:)


    unit =  file%unit_no
    nama => molecule%atom_name
    namr => molecule%residue_name
    numr => molecule%residue_no
    numm => molecule%molecule_no
    crd  => trajectory%coord


    if (file%trj_type == TrjTypeCoorBox) then
      write(unit, fmt='(a6,3(f9.3),a37)') 'CRYST1', &
           (trajectory%pbc_box(j,j), j=1,3), &
           '  90.00  90.00  90.00 P 1           1'
    end if

    write(unit, fmt='(a5,5x,i4)') 'MODEL', file%step_no


    aidx  => selatoms%idx
    nsel  = SIZE(selatoms%idx)

    molep = -1

    do i = 1, nsel
      iatm = aidx(i)

      molec = numm(iatm)
      if (molep /= -1 .and. molep /= molec) then
        write(unit,fmt='(a3)') 'TER'
      end if
      molep = molec

      if (numr(iatm) <= 9999) then

        write(unit, fmt='(a4,i7,1x,a4,1x,a4,1x,i4,4x,3f8.3)') &
             'ATOM', iatm, nama(iatm), namr(iatm), numr(iatm), (crd(j,iatm),j=1,3)

      else if (numr(iatm) > 999999) then 

        write(unit, fmt='(a4,i7,1x,a4,1x,a4,i6,3x,3f8.3)') &
             'ATOM', iatm, nama(iatm), namr(iatm), numr(iatm), (crd(j,iatm),j=1,3)

      else

        write(unit, fmt='(a4,i7,1x,a4,1x,a4,i5,4x,3f8.3)') &
             'ATOM', iatm, nama(iatm), namr(iatm), numr(iatm), (crd(j,iatm),j=1,3)

      end if
    end do

    write(unit,fmt='(a3)') 'TER'
    write(unit,fmt='(a6)') 'ENDMDL'

    return

  end subroutine write_pdb_trj_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_pdb_trj
  !> @brief        Write pdb trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @param[in]    molecule   information for molecule
  !! @date         2012/05/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_pdb_trj(file, trajectory, molecule)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory
    type(s_molecule),   target, intent(in)    :: molecule

    ! local varialbles
    character(4),       pointer :: nama(:), namr(:)
    integer,            pointer :: numr(:), numm(:)
    integer                     :: i, j, natm, unit, molec, molep
    double precision,   pointer :: crd(:,:)


    unit =  file%unit_no
    nama => molecule%atom_name
    namr => molecule%residue_name
    numr => molecule%residue_no
    crd  => trajectory%coord


    if (file%trj_type == TrjTypeCoorBox) then
      write(unit, fmt='(a6,3(f9.3),a37)') 'CRYST1', &
           (trajectory%pbc_box(j,j), j=1,3), &
           '  90.00  90.00  90.00 P 1           1'
    end if

    write(unit, fmt='(a5,5x,i4)') 'MODEL', file%step_no


    natm = SIZE(crd(1,:))

    molep = -1

    do i = 1, natm

      molec = numm(i)
      if (molep /= -1 .and. molep /= molec) then
        write(unit,fmt='(a3)') 'TER'
      end if
      molep = molec

      if (numr(i) <= 9999) then

        write(unit, fmt='(a4,i7,1x,a4,1x,a4,1x,i4,4x,3f8.3)') &
             'ATOM', i, nama(i), namr(i), numr(i), (crd(j,i),j=1,3)

      else if (numr(i) > 999999) then 

        write(unit, fmt='(a4,i7,1x,a4,1x,a4,i6,3x,3f8.3)') &
             'ATOM', i, nama(i), namr(i), numr(i), (crd(j,i),j=1,3)

      else

        write(unit, fmt='(a4,i7,1x,a4,1x,a4,i5,4x,3f8.3)') &
             'ATOM', i, nama(i), namr(i), numr(i), (crd(j,i),j=1,3)

      end if
    end do

    write(unit,fmt='(a3)') 'TER'
    write(unit,fmt='(a6)') 'ENDMDL'

    return

  end subroutine write_pdb_trj

  !*********************************************************************
  ! marble

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_marble_trj
  !> @brief        Open marble trajectory file
  !! @authors      NT, YS
  !! @param[inout] file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_marble_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file


    call open_file(file%unit_no, file%name, file%in_out)

    file%rw_header = .true.

    return

  end subroutine open_marble_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_marble_trj
  !> @brief        Close marble trajectory file
  !! @authors      NT, YS
  !! @param[in]    file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_marble_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(in)    :: file


    call close_file(file%unit_no)

    return

  end subroutine close_marble_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_marble_trj
  !> @brief        Read marble trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[inout] trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_marble_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(inout) :: trajectory

    ! local variables
    integer                     :: i, j, natm, unit
    double precision, pointer   :: crd(:,:)


    unit =  file%unit_no
    natm = SIZE(trajectory%coord(1,:))
    crd => trajectory%coord

    !  read coordinate
    !
    do i = 1, natm
      read(unit, *, err=900, end=900) (crd(j,i),j=1,3)
    end do

    !  read boundary
    !     
    if (file%trj_type == TrjTypeCoorBox) then
      do i = 1, 3
        read(unit, *, err=900) (trajectory%pbc_box(j,i), j=1,3)
      end do
    end if

    return

900 call error_msg('Read_Marble_Trj> read error')

  end subroutine read_marble_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_marble_trj_sel
  !> @brief        Write marble trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @param[in]    selatoms   information for atom selection
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_marble_trj_sel(file, trajectory, selatoms)
    
    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory
    type(s_selatoms),   target, intent(in)    :: selatoms

    ! local variables
    integer                     :: i, j, k, natm, unit
    integer,            pointer :: idx(:)
    double precision,   pointer :: crd(:,:)


    unit =  file%unit_no
    crd  => trajectory%coord

    natm =  SIZE(selatoms%idx)
    idx  => selatoms%idx

    do i = 1, natm
      j = idx(i)
      write(unit, '(3(f8.3,1x))') (crd(k, j), k=1,3)
    end do

    if (file%trj_type == TrjTypeCoorBox) then
      do i = 1, 3
        write(unit, '(3(f8.3,1x))') (trajectory%pbc_box(i,k), k=1,3)
      end do
    end if

    return

  end subroutine write_marble_trj_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_marble_trj
  !> @brief        Write marble trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_marble_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory

    ! local variables
    integer                     :: i, j, k, natm, unit
    double precision,   pointer :: crd(:,:)


    unit =  file%unit_no
    crd  => trajectory%coord
    natm =  SIZE(crd(1,:))

    do i = 1, natm
      write(unit, '(3(f8.3,1x))') (crd(j,i), j=1,3)
    end do

    if (file%trj_type == TrjTypeCoorBox) then
      do i = 1, 3
        write(unit, '(3(f8.3,1x))') (trajectory%pbc_box(i,k), k=1,3)
      end do
    end if

    return

  end subroutine write_marble_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    skip_marble_trj_header
  !> @brief        Skip marble trajectory file header
  !! @authors      NT, YS
  !! @param[inout] file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine skip_marble_trj_header(file)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file

    ! local variables
    character(80)               :: line


    if (.not. file%rw_header) &
      return 
    file%rw_header = .false.

    read(file%unit_no, '(a80)', err=900) line

    return 

900 call error_msg('Skip_Marble_Trj_Header> error.')

  end subroutine skip_marble_trj_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_marble_trj_header
  !> @brief        Write marble trajectory file header
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    num_atoms  number of atom
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_marble_trj_header(file, num_atoms)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    integer,                    intent(in)    :: num_atoms

    
    if (.not. file%rw_header) &
      return
    file%rw_header = .false.

    write(file%unit_no, &
         '("# TRAJECTORY BY MARBLE n_atom: ",i10,",output: XB")') &
         num_atoms

    return 

  end subroutine write_marble_trj_header

  !*********************************************************************
  ! amber

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_amber_trj
  !> @brief        Open AMBER trajectory file
  !! @authors      NT, YS
  !! @param[inout] file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_amber_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file


    call open_file(file%unit_no, file%name, file%in_out)

    file%rw_header = .true.

    return 
    
  end subroutine open_amber_trj
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_amber_trj
  !> @brief        Close AMBER trajectory file
  !! @authors      NT, YS
  !! @param[in]    file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_amber_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(in)    :: file


    call close_file(file%unit_no)

    return

  end subroutine close_amber_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_amber_trj
  !> @brief        Read AMBER trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[inout] trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_amber_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(inout) :: trajectory

    ! local variables
    integer                     :: i, j, natm, unit
    double precision, pointer   :: crd(:,:)


    unit =  file%unit_no
    crd  => trajectory%coord
    natm =  SIZE(crd(1,:))

    read(unit,*,err=900,end=900) ((crd(j,i), j=1,3), i=1,natm)

    if (file%trj_type == TrjTypeCoorBox) then
      read(unit, *, err=900,end=900) &
           (trajectory%pbc_box(j,j), j=1,3)
    end if

    return

900 call error_msg('Read_Amber_Trj> read error ')

  end subroutine read_amber_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_amber_trj_sel
  !> @brief        Write AMBER trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @param[in]    selatoms   information for selected atoms
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_amber_trj_sel(file, trajectory, selatoms)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory
    type(s_selatoms),   target, intent(in)    :: selatoms

    ! local variables
    integer                     :: i, j, natm, unit
    integer,            pointer :: idx(:)
    double precision,   pointer :: crd(:,:)


    unit =  file%unit_no
    crd  => trajectory%coord

    natm =  SIZE(selatoms%idx)
    idx  => selatoms%idx

    write(unit, fmt='(10(f8.3))', err=900) ((crd(j, idx(i)), j=1,3), i=1,natm)

    if (file%trj_type == TrjTypeCoorBox) then
      write(unit, fmt='(3(f8.3))', err=900) (trajectory%pbc_box(j,j), j=1,3)
    end if

    return

900 call error_msg('Write_Amber_Trj_Sel> write error.')

  end subroutine write_amber_trj_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_amber_trj
  !> @brief        Write AMBER trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_amber_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory

    ! local variables
    integer                     :: i, j, natm, unit
    double precision,   pointer :: crd(:,:)


    unit =  file%unit_no
    crd  => trajectory%coord
    natm =  SIZE(crd(1,:))

    write(unit, fmt='(10(f8.3))', err=900) ((crd(j, i), j=1,3), i=1,natm)

    if (file%trj_type == TrjTypeCoorBox) then
      write(unit, fmt='(3(f8.3))', err=900) (trajectory%pbc_box(j,j), j=1,3)
    end if

    return

900 call error_msg('Write_Amber_Trj> write error.')

  end subroutine write_amber_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    skip_amber_trj_header
  !> @brief        Skip AMBER trajectory file header
  !! @authors      NT, YS
  !! @param[inout] file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine skip_amber_trj_header(file)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file

    ! local variables
    character(80)     :: line


    if (.not. file%rw_header) &
      return
    file%rw_header = .false.

    read(file%unit_no,'(a80)', err=900) line

    return 

900 call error_msg('Skip_Amber_Trj_Header> error.')

  end subroutine skip_amber_trj_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_amber_trj_header
  !> @brief        Write AMBER trajectory file header
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_amber_trj_header(file)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file


    if (.not. file%rw_header) &
      return
    file%rw_header = .false.

    write(file%unit_no,'(a)') ' CREATED BY tbc_toolset 3.0'
      
    return

  end subroutine write_amber_trj_header

  !*********************************************************************
  ! tinker

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_tinker_trj
  !> @brief        Open TINKER trajectory file
  !! @authors      NT, YS
  !! @param[inout] file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_tinker_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file


    call open_file(file%unit_no, file%name, file%in_out)

    return

  end subroutine open_tinker_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_tinker_trj
  !> @brief        Close TINKER trajectory file
  !! @authors      NT, YS
  !! @param[in]    file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_tinker_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(in)    :: file


    call close_file(file%unit_no)

    return

  end subroutine close_tinker_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_tinker_trj
  !> @brief        Read TINKER trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[inout] trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_tinker_trj(file, trajectory)
    
    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(inout) :: trajectory

    ! local variables
    integer                     :: i, j, natm, unit
    character(80)               :: line
    double precision,   pointer :: crd(:,:)


    unit =  file%unit_no
    crd  => trajectory%coord
    natm =  SIZE(crd(1,:))

    ! read Frame Header
    !
    read(unit, fmt='(a80)', end=900, err=900)  line


    ! read coordinate
    !
    do i = 1, natm

      read(unit, fmt='(a80)', end=200, err=900) line
      read(line, fmt='(11x,3f12.6)', err=900) (crd(j,i), j=1,3)

    end do

200 continue

    return

900 call error_msg('Read_Tinker_Trj> read error ')
    
  end subroutine read_tinker_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_tinker_trj_sel
  !> @brief        Write TINKER trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @param[in]    selatoms   information for selected atoms
  !! @param[in]    molecule   information for molecule
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_tinker_trj_sel(file, trajectory, selatoms, molecule)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory
    type(s_selatoms),   target, intent(in)    :: selatoms
    type(s_molecule),   target, intent(in)    :: molecule

    ! local variables
    character(4),       pointer :: nam(:)
    integer                     :: i, j, k, natm, unit
    integer,            pointer :: idx(:)
    double precision,   pointer :: crd(:,:)


    unit = file%unit_no

    nam  => molecule%atom_name
    crd  => trajectory%coord

    natm =  SIZE(selatoms%idx)
    idx  => selatoms%idx

    ! write Frame Header
    !
    write(unit, fmt='(i6)', err=900) natm


    ! write coordinate
    !
    do i = 1, natm
      
      j = idx(i)

      write(unit, fmt='(i6,1x,a4,3(1x,f11.6))', err=900) &
           i, nam(j), (crd(k, j), k=1,3)

    end do

    return

900 call error_msg('Write_Tinker_Trj_Sel> write error.')

  end subroutine write_tinker_trj_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_tinker_trj
  !> @brief        Write TINKER trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @param[in]    molecule   information for molecule
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_tinker_trj(file, trajectory, molecule)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory
    type(s_molecule),   target, intent(in)    :: molecule

    ! local variables
    character(4),       pointer :: nam(:)
    integer                     :: i, j, natm, unit
    double precision,   pointer :: crd(:,:)


    unit =  file%unit_no

    nam  => molecule%atom_name
    crd  => trajectory%coord

    natm =  SIZE(crd(1,:))

    ! write Frame Header
    !
    write(unit, fmt='(i6)', err=900) natm


    ! write coordinate
    !
    do i = 1, natm

      write(unit, fmt='(i6,1x,a4,3(1x,f11.6))', err=900) &
            i, nam(i), (crd(j, i), j=1,3)

    end do

    return

900 call error_msg('Write_Tinker_Trj> write error.')

  end subroutine write_tinker_trj

  !*********************************************************************
  ! dcd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_dcd_trj
  !> @brief        Open DCD trajectory file
  !! @authors      NT, YS
  !! @param[inout] file trajectory file handle
  !! @date         2012/09/11 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_dcd_trj(file, endian)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    integer,                    intent(in)    :: endian

    ! local variable 
    integer           :: unit
    character(50)     :: cnv_str


    select case(endian)
    case (EndianNative)
      cnv_str = 'native'
    case (EndianBig)
      cnv_str = 'big_endian'
    case (EndianLittle)
      cnv_str = 'little_endian'
    end select

    unit = get_unit_no()
    if (file%in_out == IOFileInput) then
      open(unit, file=file%name, status='old', form='unformatted', err=900, &
           convert=cnv_str)
    else
      open(unit, file=file%name, status='new', form='unformatted', err=900, &
           convert=cnv_str)
    end if

    file%unit_no  = unit
    file%rw_header = .true.

    return

900 write(ErrOut,*) 'Open_DCD_trj> ', TRIM(file%name)
    call error_msg('Open_Dcd_Trj> I/O Error.')

    return

  end subroutine open_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_dcd_trj
  !> @brief        Close DCD trajectory file
  !! @authors      NT, YS
  !! @param[in]    file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_dcd_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(in)    :: file

    ! local variables
    integer                     :: unit


    unit = file%unit_no

    close(unit)
    call free_unit_no(unit)

    return

  end subroutine close_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_dcd_trj_header
  !> @brief        Read DCD trajectory file header
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @date         2012/09/11 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_dcd_trj_header(file)

    ! parameter
    character(*), parameter :: EndianStr(3) = (/ &
                              'Native', &
                              'Big   ', &
                              'Little'/)

    ! formal arguments
    type(s_trj_file),     intent(inout) :: file

    ! local variables
    integer               :: unit

    real(4)               :: delta
    integer(4)            :: nfile, npriv, nsavc, nstep, ipbc, chm24, natom
    integer(4)            :: ntitle, dummy(10)
    character(80)         :: title(2)
    character(4)          :: hdr

    integer               :: endian


    if (.not. file%rw_header) &
      return
    file%rw_header = .false.

    endian = EndianNative

10  unit = file%unit_no

    ! section HDR
    !
    read(unit,err=900,end=900)&
               hdr,        &
               nfile,      &
               npriv,      &
               nsavc,      &
               nstep,      &
               dummy(1:5), &
               delta,      &
               ipbc,       &
               dummy(1:8), &
               chm24
               
    if (hdr /= 'CORD') &
      call error_msg('Read_Dcd_Trj_Header> unknown file format: '//&
           TRIM(file%name))

    ! seciton TITLE
    !
    read(unit,err=900,end=900) ntitle, title

    ! section NATOM
    !
    read(unit) natom

    !
    if (ipbc >= 1) then
      file%trj_type = TrjTypeCoorBox
    else
      file%trj_type = TrjTypeCoor
    end if

    !write(MsgOut,'(A)') 'Read_Dcd_Trj_Header> '
    !write(MsgOut,'(A,A)')    '    HDR     : ', hdr
    !write(MsgOut,'(A,I0)')   '    NFILE   : ', nfile
    !write(MsgOut,'(A,I0)')   '    NPRIV   : ', npriv
    !write(MsgOut,'(A,I0)')   '    NSAVC   : ', nsavc
    !write(MsgOut,'(A,I0)')   '    NSTEP   : ', nstep
    !write(MsgOut,'(A,F0.3)') '    DELTA   : ', delta
    !write(MsgOut,'(A,I0)')   '    IPBC    : ', ipbc
    !write(MsgOut,'(A,I0)')   '    CHM24   : ', chm24
    !write(MsgOut,'(A)')      ' '
    !write(MsgOut,'(A,I0)')   '    NTITLE  : ', ntitle
    !write(MsgOut,'(A,A)')    '    TITLE1  : ', TRIM(title(1))
    !write(MsgOut,'(A,A)')    '    TITLE2  : ', TRIM(title(2))
    !write(MsgOut,'(A)')      ' '
    !write(MsgOut,'(A,I0)')   '    NATOM   : ', natom

    return

900 endian = endian + 1
    if (endian > EndianLittle) &
      call error_msg('Read_Dcd_Trj_Header> ERROR unknown file format.')

    !write(MsgOut,'(a)') 'Read_Dcd_Trj_Header> Re-open as '// &
    !                    trim(EndianStr(endian))// &
    !                    ' endian.'

    !$omp critical

    call close_dcd_trj(file)
    call open_dcd_trj(file, endian)

    !$omp end critical

    goto 10

  end subroutine read_dcd_trj_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_dcd_trj_header
  !> @brief        Write DCD trajectory file header
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    num_atoms  number of atom
  !! @date         2012/06/18 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_dcd_trj_header(file, num_atoms)

    ! formal arguments
    type(s_trj_file),     intent(inout) :: file
    integer,              intent(in)    :: num_atoms

    ! local variables
    integer               :: unit

    real(4)               :: delta
    integer(4)            :: nfile, npriv, nsavc, nstep, ipbc, chm24, natom
    integer(4)            :: ntitle, dummy(10)
    character(80)         :: title(2)
    character(4)          :: hdr


    if (.not. file%rw_header) &
      return
    file%rw_header = .false.

    unit = file%unit_no

    ! section HEADER
    !
    hdr      = 'CORD'
    nfile    = 0
    npriv    = 1
    nsavc    = 1
    nstep    = 0
    delta    = 1.0
    if (file%trj_type == TrjTypeCoorBox) then
      ipbc   = 1
    else
      ipbc   = 0
    end if
    chm24    = 24
    dummy(1:5) = 0

    write(unit) hdr,        &
                nfile,      &
                npriv,      &
                nsavc,      &
                nstep,      &
                dummy(1:5), &
                delta,      &
                ipbc,       &
                dummy(1:8), &
                chm24

    ! section TITLE
    !
    ntitle   = 2
    title(1) = 'CREATED BY REMD CONVERTER'
    title(2) = ''
    write(unit) ntitle, title

    ! section NATOM
    !
    natom = num_atoms
    write(unit) num_atoms

    return

  end subroutine write_dcd_trj_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_dcd_trj_header
  !> @brief        Update DCD trajectory file header
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_dcd_trj_header(file)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file

    ! local variables
    integer                     :: unit

    if (file%in_out /= IOFileInput) then

      unit = get_unit_no()

#if defined(__GNUC__)
      open(unit,               &
           file   = file%name, &
           status = 'old',     &
           form   = 'unformatted', &
           access = 'direct',  &
           recl   = 4)
#else
      open(unit,               &
           file   = file%name, &
           status = 'old',     &
           form   = 'binary',  &
           access = 'direct',  &
           recl   = 4)
#endif
      write(unit,rec=3) file%step_no - 1
      write(unit,rec=6) file%step_no - 1

      close(unit)
      call free_unit_no(unit)

    end if

    return

  end subroutine update_dcd_trj_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_dcd_trj
  !> @brief        Read DCD trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[inout] trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_dcd_trj(file, trajectory)
    
    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory),         intent(inout) :: trajectory

    ! local variables
    integer                     :: i, natm, unit
    double precision            :: xtl(6), box(6)

    real(4), allocatable        :: crd4f(:)


    unit     = file%unit_no

    ! read PBC coordinates
    !
    if (file%trj_type == TrjTypeCoorBox) then

      read(unit) xtl

      call xtlabcToBox(xtl, box)

      trajectory%pbc_box(1,1) = box(1)
      trajectory%pbc_box(2,2) = box(2)
      trajectory%pbc_box(3,3) = box(3)

    end if

    ! read atom coordinates
    !
    natm = SIZE(trajectory%coord(1,:))

    allocate(crd4f(natm))

    read(unit) crd4f
    do i = 1, natm
      trajectory%coord(1,i) = crd4f(i)
    end do

    read(unit) crd4f
    do i = 1, natm
      trajectory%coord(2,i) = crd4f(i)
    end do

    read(unit) crd4f
    do i = 1, natm
      trajectory%coord(3,i) = crd4f(i)
    end do

    deallocate(crd4f)

    return

  end subroutine read_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_dcd_trj_sel
  !> @brief        Write DCD trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @param[in]    selatoms   information for atom selection
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_dcd_trj_sel(file, trajectory, selatoms)
    
    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory),         intent(in)    :: trajectory
    type(s_selatoms),           intent(in)    :: selatoms

    ! local variables
    integer                     :: i, natm, unit
    double precision            :: box(8), xtl(6)

    real(4), allocatable        :: crd4f(:)


    unit = file%unit_no

    ! write PBC coordinates
    !
    if (file%trj_type == TrjTypeCoorBox) then

      box(:) = 0.d0
      box(1) = trajectory%pbc_box(1,1)
      box(2) = trajectory%pbc_box(2,2)
      box(3) = trajectory%pbc_box(3,3)

      call boxToXtlabc(box, xtl)
      write(unit) xtl

    end if

    ! write atom coordinates
    !
    natm = SIZE(selatoms%idx)

    allocate(crd4f(natm))

    do i = 1, natm
      crd4f(i) = trajectory%coord(1, selatoms%idx(i))
    end do
    write(unit) crd4f

    do i = 1, natm
      crd4f(i) = trajectory%coord(2, selatoms%idx(i))
    end do
    write(unit) crd4f

    do i = 1, natm
      crd4f(i) = trajectory%coord(3, selatoms%idx(i))
    end do
    write(unit) crd4f

    deallocate(crd4f)

    return

  end subroutine write_dcd_trj_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_dcd_trj
  !> @brief        Write DCD trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @date         2012/06/18 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_dcd_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory),         intent(in)    :: trajectory

    ! local variables
    integer                     :: i, natm, unit
    double precision            :: box(8), xtl(6)

    real(4), allocatable        :: crd4f(:)


    unit = file%unit_no

    ! write PBC coordinates
    !
    if (file%trj_type == TrjTypeCoorBox) then

      box(:) = 0.d0
      box(1) = trajectory%pbc_box(1,1)
      box(2) = trajectory%pbc_box(2,2)
      box(3) = trajectory%pbc_box(3,3)

      call boxToXtlabc(box, xtl)
      write(unit) xtl

    end if

    ! write atom coordinates
    !
    natm = SIZE(trajectory%coord(1,:))

    allocate(crd4f(natm))

    do i = 1, natm
      crd4f(i) = trajectory%coord(1, i)
    end do
    write(unit) crd4f

    do i = 1, natm
      crd4f(i) = trajectory%coord(2, i)
    end do
    write(unit) crd4f

    do i = 1, natm
      crd4f(i) = trajectory%coord(3, i)
    end do
    write(unit) crd4f

    deallocate(crd4f)

    return

  end subroutine write_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    seek_dcd_trj
  !> @brief        Seek DCD trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[inout] trajectory information of trajectory
  !! @param[in]    step_no    step number
  !! @date         2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine seek_dcd_trj(file, trajectory, step_no)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory),         intent(inout) :: trajectory
    integer,                    intent(in)    :: step_no

    ! local variables
    integer                     :: unit, natm, alloc_stat
    integer(4)                  :: i4
    double precision            :: xtl(6)

    real(4),   allocatable      :: crd4f(:)


    unit = file%unit_no
    natm = SIZE(trajectory%coord(1,:))

    if (file%step_no > step_no) then
      call error_msg('Seek_Dcd_Trj> Seeking to the back is not supported.')
    else if (file%step_no == step_no) then
      return 
    end if

    alloc_stat = 0
    allocate(crd4f(natm), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    do while (file%step_no /= step_no)

      if (file%trj_type == TrjTypeCoorBox) then
        read(unit) xtl(:)
      end if

      read(unit) crd4f(:)
      read(unit) crd4f(:)
      read(unit) crd4f(:)

      file%step_no = file%step_no + 1

    end do
    
    !write(MsgOut,'(A,A,A,I7)') 'Seek_Dcd_Trj> file:', TRIM(file%name), &
    !                           ' current:', file%step_no

    deallocate(crd4f, stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine seek_dcd_trj

  !*********************************************************************
  ! gromacs

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_gromacs_trj
  !> @brief        Open Gromacs trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_gromacs_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file

    ! local variables
    character(100)              :: str
    integer                     :: il, i, unit
    integer                     :: c_in_out
    integer(8)                  :: c_handle


    str = file%name
    il = LEN_TRIM(str)
    str(il+1:il+2) = CHAR(0)

    if (file%in_out == IOFileInput) then
      c_in_out = 2
    else
      c_in_out = 1
    end if

    unit = 0

    call copengromacstrj(c_handle, c_in_out, str)

    if (c_handle /= 0) then
      do i = 1, GroMaxHandle
        if (g_gro_handle(i) == 0) then
          g_gro_handle(i) = c_handle
          unit = i + GroUnitNo1

          file%unit_no = unit

          return
        end if
      end do
    end if

    call error_msg('Open_Gromacs_Trj> open error :'//TRIM(file%name))

    return

  end subroutine open_gromacs_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_gromacs_trj
  !> @brief        Close Gromacs trajectory file
  !! @authors      NT, YS
  !! @param[in]    file trajectory file handle
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_gromacs_trj(file)

    ! formal arguments
    type(s_trj_file),           intent(in)    :: file

    ! local variables
    integer(8)                  :: c_handle
    integer                     :: ofile


    ofile = file%unit_no - GroUnitNo1

    if (ofile < 1 .or. ofile > GroMaxHandle) &
         return 
    c_handle = g_gro_handle(ofile)

    call cclosegromacstrj(c_handle)

    g_gro_handle(ofile) = 0

    return

  end subroutine close_gromacs_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_gromacs_trj
  !> @brief        Read Gromacs trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[inout] trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_gromacs_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory),         intent(inout) :: trajectory

    ! local variables
    integer                     :: ofile, ret, natom
    integer(8)                  :: c_handle


    ofile = file%unit_no - GroUnitNo1

    if (ofile < 1 .or. ofile > GroMaxHandle) &
      return 
    c_handle = g_gro_handle(ofile)

    natom = SIZE(trajectory%coord(1,:))

    call cinpgromacstrj(c_handle, &
                        natom, &
                        trajectory%coord, &
                        trajectory%pbc_box, &
                        ret)

    if (ret == -1) then
      call error_msg('Read_Gromacs_Trj> read error.')
    end if

    return

  end subroutine read_gromacs_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_gromacs_trj_sel
  !> @brief        Write Gromacs trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @param[in]    selatoms   information for atom selection
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_gromacs_trj_sel(file, trajectory, selatoms)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory),         intent(in)    :: trajectory
    type(s_selatoms),           intent(in)    :: selatoms

    ! local variables
    integer                     :: ofile
    integer                     :: c_out_pbc_box, ret
    integer(8)                  :: c_handle


    ofile = file%unit_no - GroUnitNo1

    if (ofile < 1 .or. ofile > GroMaxHandle) &
      return 
    c_handle = g_gro_handle(ofile)

    if (file%trj_type == TrjTypeCoorBox) then
      c_out_pbc_box = 2
    else
      c_out_pbc_box = 1
    end if

    call coutgromacstrjsel(c_handle, &
                           SIZE(selatoms%idx), &
                           selatoms%idx, &
                           trajectory%coord, &
                           trajectory%pbc_box, &
                           c_out_pbc_box, &
                           ret)

    if (ret == -1) then
      call error_msg('Write_Gromacs_Trj> write error.')
    end if

    return

  end subroutine write_gromacs_trj_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_gromacs_trj
  !> @brief        Write Gromacs trajectory file
  !! @authors      NT, YS
  !! @param[inout] file       trajectory file handle
  !! @param[in]    trajectory information for trajectory
  !! @date         2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_gromacs_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory),         intent(in)    :: trajectory

    ! local variables
    integer                     :: ofile, natom
    integer                     :: i, c_out_pbc_box, ret
    integer(8)                  :: c_handle
    type(s_selatoms)            :: selatoms


    ofile = file%unit_no - GroUnitNo1

    if (ofile < 1 .or. ofile > GroMaxHandle) &
      return 
    c_handle = g_gro_handle(ofile)

    if (file%trj_type == TrjTypeCoorBox) then
      c_out_pbc_box = 2
    else
      c_out_pbc_box = 1
    end if

    natom = SIZE(trajectory%coord(1,:))

    call alloc_selatoms(selatoms, natom)
    do i = 1, natom
      selatoms%idx(i) = i
    end do

    call coutgromacstrjsel(c_handle, &
                           SIZE(selatoms%idx), &
                           selatoms%idx, &
                           trajectory%coord, &
                           trajectory%pbc_box, &
                           c_out_pbc_box, &
                           ret)

    call dealloc_selatoms(selatoms)

    if (ret == -1) then
      call error_msg('Write_Gromacs_Trj> write error.')
    end if

    return

  end subroutine write_gromacs_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine swap_i4
  !> @brief     swap_i4
  !! @authors   NT, YS
  !! @param[inout] ival : target integer value
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine swap_i4(ival)

    ! formal arguments
    integer, intent(inout) :: ival

    ! local variables
    integer  :: ival2

    ival2 = 0
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF'), 24))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF00'), 8))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF0000'), -8))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF000000'), -24))
    ival = ival2;

    return

  end subroutine swap_i4

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine swap_r4
  !> @brief     swap_r4
  !! @authors   NT, YS
  !! @param[inout] rval : target real value
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine swap_r4(rval)

    ! formal arguments
    real(4), intent(inout) :: rval

    ! local variables
    integer  :: ival, ival2

    ival = transfer(rval, ival)
    ival2 = 0
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF'), 24))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF00'), 8))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF0000'), -8))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF000000'), -24))
    rval = transfer(ival2, rval)

    return

  end subroutine swap_r4

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine swap_r8
  !> @brief     swap_r8
  !! @authors   NT, YS
  !! @param[inout] rval : target real value
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine swap_r8(rval)

    ! formal arguments
    real(8),   intent(inout) :: rval

    ! local variables
    integer(8) :: ival, ival2

    ival = transfer(rval, ival)
    ival2 = 0
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF'), 56))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF00'), 40))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF0000'), 24))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF000000'), 8))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF00000000'), -8))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF0000000000'), -24))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF000000000000'), -40))
    ival2 = IOR(ival2, ISHFT(IAND(ival, x'FF00000000000000'), -56))
    rval = transfer(ival2, rval)

    return

  end subroutine swap_r8

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine boxToXtlabc
  !> @brief     boxToXtlabc
  !! @authors   NT, YS
  !! @param[in]    box : box
  !! @param[inout] xtl : xtl
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine boxToXtlabc(box, xtl)
    
    ! formal arguments 
    double precision, intent(in)    :: box(:)
    double precision, intent(inout) :: xtl(:)

    ! local variables
    double precision  :: a, b, c
    double precision  :: hth(3,3)
    double precision  :: evalue(3), ev(3,3)

    hth(1,1) = box(1)*box(1)

    a = box(6) - 90.0
    if (ABS(a) > 0.000000001) then
      hth(2,1) = box(1)*box(2)*cos(TORAD(box(6)))
    else 
      hth(2,1) = 0.0
    end if
    hth(1,2) = hth(2,1)

    hth(2,2) = box(2)*box(2)

    a = box(5) - 90.0
    if (ABS(a) > 0.000000001) then
      hth(3,1) = box(1)*box(3)*cos(TORAD(box(5)))
    else
      hth(3,1) = 0.0
    end if
    hth(1,3) = hth(3,1)

    a = box(4) - 90.0
    if (ABS(a) > 0.000000001) then
      hth(2,3) = box(2)*box(3)*cos(TORAD(box(4)))
    else
      hth(2,3) = 0.0
    end if
    hth(3,2) = 0.0

    hth(3,3) = box(3)*box(3)

    call jacobi(hth, 3, evalue, ev)

    if (evalue(1) < 0.000000001 .or. &
        evalue(2) < 0.000000001 .or. &
        evalue(3) < 0.000000001) then
      xtl(:) = 0.d0
      return 
    end if

    a = sqrt(evalue(1))
    b = sqrt(evalue(2))
    c = sqrt(evalue(3))

    xtl(1) = a*ev(1,1)*ev(1,1) + b*ev(1,2)*ev(1,2) + c*ev(1,3)*ev(1,3)
    xtl(3) = a*ev(2,1)*ev(2,1) + b*ev(2,2)*ev(2,2) + c*ev(2,3)*ev(2,3)
    xtl(6) = a*ev(3,1)*ev(3,2) + b*ev(3,2)*ev(3,2) + c*ev(3,3)*ev(3,3)
    xtl(2) = a*ev(1,1)*ev(2,1) + b*ev(1,2)*ev(2,2) + c*ev(1,3)*ev(2,3)
    xtl(4) = a*ev(1,1)*ev(3,1) + b*ev(1,2)*ev(3,2) + c*ev(1,3)*ev(3,3)
    xtl(5) = a*ev(2,1)*ev(3,1) + b*ev(2,2)*ev(3,2) + c*ev(2,3)*ev(3,3)

  end subroutine boxToXtlabc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine xtlabcToBox
  !> @brief     xtlabcToBox
  !! @authors   NT, YS
  !! @param[in]    xtl : xtl
  !! @param[inout] box : box
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine xtlabcToBox(xtl, box)

    ! formal arguments 
    double precision, intent(in)    :: xtl(:)
    double precision, intent(inout) :: box(:)

    ! local variables
    double precision  :: ab, bc, ca

    box(1) = sqrt(xtl(1)*xtl(1) + xtl(2)*xtl(2) + xtl(4)*xtl(4))
    box(2) = sqrt(xtl(2)*xtl(2) + xtl(3)*xtl(3) + xtl(5)*xtl(5))
    box(3) = sqrt(xtl(4)*xtl(4) + xtl(5)*xtl(5) + xtl(6)*xtl(6))

    ab = xtl(2)*(xtl(1) + xtl(3)) + xtl(4)*xtl(5)
    bc = xtl(5)*(xtl(3) + xtl(6)) + xtl(2)*xtl(4)
    ca = xtl(4)*(xtl(1) + xtl(6)) + xtl(2)*xtl(5)

    box(4) = TODEG(acos(bc / (box(2)*box(3))))
    box(5) = TODEG(acos(ca / (box(3)*box(1))))
    box(6) = TODEG(acos(ab / (box(1)*box(2))))

    return

  end subroutine xtlabcToBox

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function   torad
  !> @brief     torad
  !! @authors   NT, YS
  !! @param[in] degree : degree
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function torad(degree)

    double precision :: torad

    ! formal arguments
    double precision, intent(in) :: degree

    torad = (degree/180.0*3.141592653589793)
    return

  end function torad

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function   todeg
  !> @brief     todeg
  !! @authors   NT, YS
  !! @param[in] radian : radian
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function todeg(radian)

    double precision :: todeg

    ! formal arguments
    double precision, intent(in) :: radian

    todeg = (radian/3.141592653589793*180.0)
    return

  end function todeg

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine rotate
  !> @brief     rotate
  !! @authors   NT, YS
  !! @param[inout] a : a
  !! @param[in]    i : i
  !! @param[in]    j : j
  !! @param[in]    k : k
  !! @param[in]    l : l
  !! @param[in]    s : s
  !! @param[in]    tau : tau
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine rotate(a, i, j, k, l, s, tau)
    
    ! formal arguments
    double precision,    intent(inout) :: a(3,3)
    integer,             intent(in)    :: i
    integer,             intent(in)    :: j
    integer,             intent(in)    :: k
    integer,             intent(in)    :: l
    double precision,    intent(in)    :: s
    double precision,    intent(in)    :: tau

    ! local variables
    double precision     :: g, h


    g = a(i,j)
    h = a(k,l)
    a(i,j) = g - s * (h + g * tau)
    a(k,l) = h + s * (g - h * tau)

    return

  end subroutine rotate

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine jacobi
  !> @brief     jacobi
  !! @authors   NT, YS
  !! @param[inout] a : a
  !! @param[in]    n : n
  !! @param[inout] d : d
  !! @param[inout] v : v
  !! @date      2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine jacobi(a, n, d, v)
    
    integer, parameter :: MAX_ITERATIONS = 50

    ! formal arguments
    double precision, intent(inout) :: a(3,3)
    integer,          intent(in)    :: n
    double precision, intent(inout) :: d(3)
    double precision, intent(inout) :: v(3,3)

    ! local variables
    integer           :: j, iq, ip, i, nrot
    double precision  :: tresh, theta, tau, t, sm, s, h, g, c
    double precision, allocatable :: b(:), z(:)


    allocate(b(n), z(n))

    do ip = 1, n
      do iq = 1, n
        v(ip,iq) = 0.0
      end do
      v(ip,ip) = 1.0
    end do

    do ip = 1, n
      d(ip) = a(ip,ip)
      b(ip) = d(ip)
      z(ip) = 0.0
    end do

    nrot = 0

    do i = 1,MAX_ITERATIONS
      sm = 0.0
      do ip = 1, n-1
        do iq = ip + 1, n
          sm = sm + abs(a(ip,iq))
        end do
      end do

      if (sm == 0) then
        deallocate(b, z)
        return 
      end if

      if (i < 4) then
        tresh = 0.2 * sm / (n**2)
      else
        tresh = 0.0
      end if

      do ip = 1, n-1
        do iq = ip + 1, n
          g = 100.0 * abs(a(ip,iq))
          if (i > 4 .and. &
               abs(d(ip)) + g == abs(d(ip)) .and. &
               abs(d(iq)) + g == abs(d(iq))) then
            a(ip,iq) = 0.0
          else if (abs(a(ip,iq)) > tresh) then
            h = d(iq) - d(ip)
            if (abs(h) + g == abs(h)) then
              t = (a(ip,iq)) / h
            else 
              theta = 0.5 * h / (a(ip,iq))
              t = 1.0 / (abs(theta) + sqrt(1.0+theta*theta))
              if (theta < 0.0) t = -t
            end if
            c=1.0/sqrt(1+t*t)
            s=t*c
            tau=s/(1.0+c)
            h=t*a(ip,iq)
            z(ip) = z(ip) - h
            z(iq) = z(iq) + h
            d(ip) = d(ip) - h
            d(iq) = d(iq) + h
            a(ip,iq)=0.0

            do j = 1, ip-1
              call rotate(a,j,ip,j,iq, s, tau)
            end do

            do j = ip+1,iq-1
              call rotate(a,ip,j,j,iq, s, tau)
            end do

            do j = iq+1,n
              call rotate(a,ip,j,iq,j, s, tau)
            end do

            do j = 1,n
              call rotate(v,j,ip,j,iq, s, tau)
            end do
            nrot = nrot + 1
          end if
        end do
      end do

      do ip=1,n
        b(ip) = b(ip) + z(ip)
        d(ip) = b(ip)
        z(ip) = 0.0
      end do

    end do

    call error_msg('Jacobi> Too many iterations in function jacobi()')

    return

  end subroutine jacobi

end module fileio_trj_mod
