!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : molecules_mod
!> @brief  define molecules
!! @author Norio Takase(NT), Naoyuki Miyashita(NM), Yuji Sugita (YS)
!! @date   2012/05/28
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module molecules_mod

  use fileio_pdb_mod
  use fileio_psf_mod
  use selatoms_str_mod
  use molecules_str_mod
  use messages_mod

  implicit none
  private

  ! OutputMolecule type
  integer, public, parameter :: MolFileTypePDB = 1

  ! subroutines
  public  :: define_molecules
  private :: setup_molecule_psfcrd
  private :: setup_molecule_crd
  private :: define_mole_atom

  public  :: export_molecules
  private :: export_mole_sel
  private :: export_mole

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  define_molecules
  !> @brief      a driver subroutine for defining moleucles
  !! @authors    NT, YS
  !! @param[in]  molecule : structure of molecule
  !! @param[in]  pdb      : pdb file (Optional)
  !! @param[out] psf      : psf file (Optional)
  !! @date       2012/05/28 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_molecules(molecule, pdb, psf)

    ! formal arguments
    type(s_molecule), intent(inout)           :: molecule
    type(s_pdb),      intent(in)    ,optional :: pdb
    type(s_psf),      intent(inout) ,optional :: psf

    ! local variables
    integer :: num_atoms_pdb, num_atoms_psf, num_atoms_crd, num_atoms_ref
    logical :: present_pdb, present_psf


    present_pdb = .false.
    present_psf = .false.

    if (present(pdb)) then
      if (associated(pdb%atom_no)) present_pdb = .true.
    end if
    if (present(psf)) then
      if (associated(psf%atom_no)) present_psf = .true.
    end if


    ! Check about the number of atoms in PSF, PDB, CRD, and REF
    !
    if (present_pdb) num_atoms_pdb = SIZE(pdb%atom_no)
    if (present_psf) num_atoms_psf = SIZE(psf%atom_no)

    if (present_psf .and. present_pdb) then
      if (num_atoms_psf /= num_atoms_pdb) then
        write(MsgOut,*) 'Define_Molecules> ATOM in PSF = ', num_atoms_psf
        write(MsgOut,*) 'Define_Molecules> ATOM in PDB = ', num_atoms_pdb
        call error_msg('Define_Molecules> Number of atoms differs.')
      end if
    end if


    ! Setup molecule
    !
    if (present_psf) then  !< CHARMM

      if (.not. present_pdb) then
        call error_msg('Define_Molecules> coordinates are not present.')
      end if

      call setup_molecule_psfcrd(molecule, psf, pdb)

    else

      if (.not. present_pdb) then
        call error_msg('Define_Molecules> coordinates are not present.')
      end if

      call setup_molecule_crd(molecule, pdb)

    end if

    return

  end subroutine define_molecules

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  setup_molecule_psfcrd
  !> @brief      copy information of PDB/CRD and PSF into molecule
  !! @authors    NT, YS
  !! @param[in]  psf  : PSF information
  !! @param[in]  pdb  : PDB information
  !! @param[out] mole : molecule information
  !! @date       2011/05/28 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_molecule_psfcrd(mole, psf, pdb)

    ! formal arguments
    type(s_molecule),           intent(inout) :: mole
    type(s_psf),                intent(in)    :: psf
    type(s_pdb),                intent(in)    :: pdb

    ! local variables
    integer                     :: natom
    integer                     :: nbond
    integer                     :: nangl
    integer                     :: ndihe
    integer                     :: nimpr
    integer                     :: ncmap
    integer                     :: nmole

    ! initialize
    !
    call init_molecules(mole)

    ! allocation
    !
    natom = psf%num_atoms
    nbond = psf%num_bonds
    nangl = psf%num_angles
    ndihe = psf%num_dihedrals
    nimpr = psf%num_impropers
    ncmap = psf%num_cross_terms
    nmole = psf%num_molecules

    call alloc_molecules(mole, MoleculeAtom, natom)
    call alloc_molecules(mole, MoleculeBond, nbond)
    call alloc_molecules(mole, MoleculeAngl, nangl)
    call alloc_molecules(mole, MoleculeDihe, ndihe)
    call alloc_molecules(mole, MoleculeImpr, nimpr)
    call alloc_molecules(mole, MoleculeCmap, ncmap)
    call alloc_molecules(mole, MoleculeMole, nmole)

    ! copy information from PSF
    !
    mole%atom_no(1:natom)       = psf%atom_no(1:natom)
    mole%segment_name(1:natom)  = psf%segment_name(1:natom)
    mole%segment_no(1:natom)    = psf%segment_no(1:natom)
    mole%residue_no(1:natom)    = psf%residue_no(1:natom)
    mole%residue_c_no(1:natom)  = psf%residue_c_no(1:natom)
    mole%residue_name(1:natom)  = psf%residue_name(1:natom)
    mole%molecule_no(1:natom)   = psf%molecule_no(1:natom)
    mole%atom_name(1:natom)     = psf%atom_name(1:natom)
    mole%atom_cls_name(1:natom) = psf%atom_cls_name(1:natom)
!    mole%atom_cls_no(1:natom)   = psf%atom_cls_no(1:natom)
    mole%charge(1:natom)        = psf%charge(1:natom)
    mole%mass(1:natom)          = psf%mass(1:natom)
    mole%inv_mass(1:natom)      = 1.0d0/psf%mass(1:natom)
    mole%imove(1:natom)         = psf%imove(1:natom)

    mole%atom_connect_no(1:natom)  = psf%atom_connect_no(1:natom) 
    mole%atom_connect(1:4,1:natom) = psf%atom_connect(1:4,1:natom) 

    mole%bond_list(1:2,1:nbond) = psf%bond_list(1:2,1:nbond)
    mole%angl_list(1:3,1:nangl) = psf%angl_list(1:3,1:nangl)
    mole%dihe_list(1:4,1:ndihe) = psf%dihe_list(1:4,1:ndihe)
    mole%impr_list(1:4,1:nimpr) = psf%impr_list(1:4,1:nimpr)
    mole%cmap_list(1:8,1:ncmap) = psf%cmap_list(1:8,1:ncmap)

    ! copy information from PDB
    !
    mole%chain_id(1:natom)          = pdb%chain_id(1:natom)
    mole%atom_coord(1:3,1:natom)    = pdb%atom_coord(1:3,1:natom)
    mole%atom_occupancy(1:natom)    = pdb%atom_occupancy(1:natom)
    mole%atom_temp_factor(1:natom)  = pdb%atom_temp_factor(1:natom)
       
    ! number of variables
    !
    mole%num_atoms     = natom
    mole%num_bonds     = nbond
    mole%num_angles    = nangl
    mole%num_dihedrals = ndihe
    mole%num_impropers = nimpr
    mole%num_cmaps     = ncmap
    mole%num_deg_freedom = 3*natom

    mole%num_residues   = psf%num_residues
    mole%num_molecules  = psf%num_molecules
    mole%num_segments   = psf%num_segments

    ! redefine atom_cls_no
    !
    mole%atom_cls_no(1:natom) = 0

    ! set molecule_atom_no and mole_mass
    !
    call define_mole_atom(mole)

    return

  end subroutine setup_molecule_psfcrd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  setup_molecule_crd
  !> @brief      copy information of PDB/CRD into molecule
  !! @authors    NT, YS
  !! @param[in]  pdb  : PDB information
  !! @param[out] mole : molecule information
  !! @date       2011/05/28 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_molecule_crd(mole, pdb)

    ! formal arguments
    type(s_molecule),           intent(inout) :: mole
    type(s_pdb),                intent(in)    :: pdb

    ! local variables
    integer                     :: natom


    ! initialize
    !
    call init_molecules(mole)

    ! copy information from PDB
    !

    natom = pdb%num_atoms
    call alloc_molecules(mole, MoleculeAtom, natom)

    mole%atom_no(1:natom)           = pdb%atom_no(1:natom)
    mole%segment_name(1:natom)      = pdb%segment_name(1:natom)
    mole%segment_no(1:natom)        = 0
    mole%residue_no(1:natom)        = pdb%residue_no(1:natom)
    mole%residue_name(1:natom)      = pdb%residue_name(1:natom)
    mole%atom_name(1:natom)         = pdb%atom_name(1:natom)
    mole%atom_cls_name(1:natom)     = ''
    mole%atom_cls_no(1:natom)       = 0
    mole%charge(1:natom)            = 0.d0
    mole%mass(1:natom)              = 0.d0
    mole%inv_mass(1:natom)          = 0.d0
    mole%imove(1:natom)             = 0
    mole%chain_id(1:natom)          = pdb%chain_id(1:natom)
    mole%atom_coord(1:3,1:natom)    = pdb%atom_coord(1:3,1:natom)
    mole%atom_occupancy(1:natom)    = pdb%atom_occupancy(1:natom)
    mole%atom_temp_factor(1:natom)  = pdb%atom_temp_factor(1:natom)

    ! number of variables
    !
    mole%num_atoms     = natom
    mole%num_deg_freedom = 3*natom


    return

  end subroutine setup_molecule_crd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  define_mole_atom
  !> @brief      molecular number is assigned to each atom
  !! @authors    NT, YS
  !! @param[out] mole : molecule information
  !! @date       2012/05/28 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_mole_atom(mole)

    ! formal arguments
    !
    type(s_molecule), intent(inout) :: mole

    ! local variables
    !
    integer          :: i, j
    integer          :: imole, imole_pre, initial, final
    double precision :: mass

    do i = 1, mole%num_atoms
      if (i == 1) then
        imole = mole%molecule_no(i)
        mole%molecule_atom_no(imole) = i
      else
        imole_pre = mole%molecule_no(i-1)  
        imole     = mole%molecule_no(i)
        if (imole /= imole_pre) then
          mole%molecule_atom_no(imole) = i
        end if
      end if
      mole%molecule_name(imole) = mole%segment_name(i)
    end do


    do i = 1, mole%num_molecules
      initial  = mole%molecule_atom_no(i)
      if (i /= mole%num_molecules) then
        final = mole%molecule_atom_no(i+1) - 1
      else
        final = mole%num_atoms
      end if

      mass = 0.0d0
      do j = initial, final
        mass = mass + mole%mass(j)
      end do
      mole%molecule_mass(i) = mass

    end do

    return

  end subroutine define_mole_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  export_molecules
  !> @brief      a driver subroutine for undefining moleucles
  !! @authors    NT, YS
  !! @param[in]  molecule : structure of molecule
  !! @param[in]  selatoms : information of atom selection (Optional)
  !! @param[out] pdb      : pdb file (Optional)
  !! @param[out] psf      : psf file (Optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine export_molecules(molecule, selatoms, pdb, psf)

    ! formal arguments
    type(s_molecule), intent(in)              :: molecule
    type(s_selatoms), intent(in)    ,optional :: selatoms
    type(s_pdb),      intent(inout) ,optional :: pdb
    type(s_psf),      intent(inout) ,optional :: psf


    if (present(pdb)) then

      if (present(selatoms)) then
        call export_mole_sel(molecule, selatoms, pdb)
      else
        call export_mole    (molecule, pdb)
      end if

    end if

    return

  end subroutine export_molecules

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  export_mole_sel
  !> @brief      export molecule to pdb
  !! @authors    NT, YS
  !! @param[in]  molecule : structure of molecule
  !! @param[in]  selatoms : information of atom selection (Optional)
  !! @param[out] pdb      : pdb file (Optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine export_mole_sel(molecule, selatoms, pdb)

    ! formal arguments
    type(s_molecule),    intent(in)    :: molecule
    type(s_selatoms),    intent(in)    :: selatoms
    type(s_pdb),         intent(inout) :: pdb

    ! local variables
    integer              :: i, idx, natom, nresi, nter, mole1, mole2
    integer, allocatable :: ter_no(:)


    natom = SIZE(selatoms%idx)
    nresi = 0
    nter  = 0

    allocate(ter_no(natom + 1))

    ! init pdb
    call init_pdb(pdb)

    pdb%hetatm_rec = .false.
    pdb%cryst_rec  = .false.
    pdb%model_rec  = .true.
    pdb%ter_rec    = .true.
    pdb%end_rec    = .true.
    pdb%segment    = .true.

    ! setup atom data
    call alloc_pdb(pdb, PdbAtom, natom)

    do i = 1, natom

      idx = selatoms%idx(i)

      pdb%hetatm(i)       = .false.
      pdb%atom_name(i)    = molecule%atom_name(idx)
      pdb%residue_name(i) = molecule%residue_name(idx)
      pdb%segment_name(i) = molecule%segment_name(idx)
      pdb%chain_id(i)     = molecule%chain_id(idx)
      pdb%atom_no(i)      = i
      pdb%residue_no(i)   = molecule%residue_no(idx)
      pdb%atom_coord(:,i) = molecule%atom_coord(:,idx)
      pdb%atom_occupancy(i)   = 1.d0
      pdb%atom_temp_factor(i) = 0.d0

      nresi = MAX(nresi, pdb%residue_no(i))

      if (i /= natom) then
        mole1 = molecule%molecule_no(idx)
        mole2 = molecule%molecule_no(selatoms%idx(i + 1))
        if (mole1 /= mole2) then
          nter = nter + 1
          ter_no(nter) = i + 1
        end if
      end if

    end do

    nter = nter + 1
    ter_no(nter) = natom + 1

    ! setup ter data
    call alloc_pdb(pdb, PdbTerLine, nter)

    pdb%ter_line_no(1:nter) = ter_no(1:nter)


    if (natom > 999999) &
       pdb%atom_col7 = .true.

    if (nresi > 9999) then
       pdb%res_col5 = .true.
       if (nresi > 99999) &
          pdb%res_col6 = .true.
    end if

    deallocate(ter_no)

    return

  end subroutine export_mole_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  export_mole
  !> @brief      export molecule to pdb
  !! @authors    NT, YS
  !! @param[in]  molecule : structure of molecule
  !! @param[out] pdb      : pdb file (Optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine export_mole(molecule, pdb)

    ! formal arguments
    type(s_molecule),    intent(in)    :: molecule
    type(s_pdb),         intent(inout) :: pdb

    ! local variables
    integer              :: i, natom, nresi, nter, mole1, mole2
    integer, allocatable :: ter_no(:)


    natom = SIZE(molecule%atom_name)
    nresi = 0
    nter  = 0

    allocate(ter_no(natom + 1))

    ! init pdb
    call init_pdb(pdb)

    pdb%hetatm_rec = .false.
    pdb%cryst_rec  = .false.
    pdb%model_rec  = .true.
    pdb%ter_rec    = .true.
    pdb%end_rec    = .true.
    pdb%segment    = .true.

    ! setup atom data
    call alloc_pdb(pdb, PdbAtom, natom)

    do i = 1, natom

      pdb%hetatm(i)       = .false.
      pdb%atom_name(i)    = molecule%atom_name(i)
      pdb%residue_name(i) = molecule%residue_name(i)
      pdb%segment_name(i) = molecule%segment_name(i)
      pdb%chain_id(i)     = molecule%chain_id(i)
      pdb%atom_no(i)      = i
      pdb%residue_no(i)   = molecule%residue_no(i)
      pdb%atom_coord(:,i) = molecule%atom_coord(:,i)
      pdb%atom_occupancy(i)   = 1.d0
      pdb%atom_temp_factor(i) = 0.d0

      nresi = MAX(nresi, pdb%residue_no(i))

      if (i /= natom) then
        mole1 = molecule%molecule_no(i)
        mole2 = molecule%molecule_no(i + 1)
        if (mole1 /= mole2) then
          nter = nter + 1
          ter_no(nter) = i + 1
        end if
      end if

    end do

    nter = nter + 1
    ter_no(nter) = natom + 1

    ! setup ter data
    call alloc_pdb(pdb, PdbTerLine, nter)

    pdb%ter_line_no(1:nter) = ter_no(1:nter)


    if (natom > 999999) &
       pdb%atom_col7 = .true.

    if (nresi > 9999) then
       pdb%res_col5 = .true.
       if (nresi > 99999) &
          pdb%res_col6 = .true.
    end if

    deallocate(ter_no)

    return

  end subroutine export_mole

end module molecules_mod
