!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : molecules_str_mod
!> @brief  molecule structure
!! @author Norio Takase (NT), Naoyuki Miyashita (NM), Yuji Sugita (YS)
!! @date   2012/05/24
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module molecules_str_mod

  use messages_mod

  implicit none
  private

  ! structures
  !
  type, public :: s_molecule

    integer          :: num_deg_freedom = 0

    integer          :: num_atoms     = 0
    integer          :: num_bonds     = 0
    integer          :: num_angles    = 0
    integer          :: num_dihedrals = 0
    integer          :: num_impropers = 0
    integer          :: num_cmaps     = 0

    integer          :: num_residues  = 0
    integer          :: num_molecules = 0
    integer          :: num_segments  = 0

    double precision :: pbc_box(3,3)

    ! atom
    integer,          pointer :: atom_no(:)          => null()
    character(4),     pointer :: segment_name(:)     => null()
    integer,          pointer :: segment_no(:)       => null()
    integer,          pointer :: residue_no(:)       => null()
    integer,          pointer :: residue_c_no(:)     => null()
    character(4),     pointer :: residue_name(:)     => null()
    character(4),     pointer :: atom_name(:)        => null()
    character(4),     pointer :: atom_cls_name(:)    => null()
    integer,          pointer :: atom_cls_no(:)      => null()
    double precision, pointer :: charge(:)           => null()
    double precision, pointer :: mass(:)             => null()
    double precision, pointer :: inv_mass(:)         => null()
    integer,          pointer :: imove(:)            => null()

    integer,          pointer :: atom_connect_no(:)  => null()
    integer,          pointer :: atom_connect(:,:)   => null()
    integer,          pointer :: molecule_no(:)      => null()

    character(1),     pointer :: chain_id(:)         => null()
    double precision, pointer :: atom_coord(:,:)     => null()
    double precision, pointer :: atom_occupancy(:)   => null()
    double precision, pointer :: atom_temp_factor(:) => null()
    double precision, pointer :: atom_refcoord(:,:)  => null()

    ! bond
    integer,          pointer :: bond_list(:,:)      => null()

    ! angle
    integer,          pointer :: angl_list(:,:)      => null()

    ! dihedral
    integer,          pointer :: dihe_list(:,:)      => null()

    ! improper
    integer,          pointer :: impr_list(:,:)      => null()

    ! cmap
    integer,          pointer :: cmap_list(:,:)      => null()

    ! molecule
    integer,          pointer :: molecule_atom_no(:) => null()
    double precision, pointer :: molecule_mass(:)    => null()
    character(10),    pointer :: molecule_name(:)    => null()

  end type s_molecule

  ! parameters for allocatable variables
  integer,  public, parameter :: MoleculeAtom = 1
  integer,  public, parameter :: MoleculeBond = 2
  integer,  public, parameter :: MoleculeAngl = 3
  integer,  public, parameter :: MoleculeDihe = 4
  integer,  public, parameter :: MoleculeImpr = 5
  integer,  public, parameter :: MoleculeCmap = 6
  integer,  public, parameter :: MoleculeMole = 7
  integer,  public, parameter :: MoleculeRefc = 8

  ! subroutines
  public :: init_molecules
  public :: alloc_molecules
  public :: dealloc_molecules
  public :: dealloc_molecules_all

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  init_molecules
  !> @brief      initialize molecule information
  !! @authors    NT, YS
  !! @param[out] molecule : structure of molecule
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine init_molecules(molecule)

    ! formal arguments
    type(s_molecule), intent(inout) :: molecule

    molecule%num_deg_freedom  = 0
    molecule%num_atoms        = 0
    molecule%num_bonds        = 0
    molecule%num_angles       = 0
    molecule%num_dihedrals    = 0
    molecule%num_impropers    = 0
    molecule%num_cmaps        = 0
    molecule%num_residues     = 0
    molecule%num_segments     = 0
    molecule%num_molecules    = 0
    molecule%pbc_box(1:3,1:3) = 0.0d0

    return

  end subroutine init_molecules

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  alloc_molecules
  !> @brief      allocate molecule information
  !! @authors    NT, YS
  !! @param[in]  variable : selected variable
  !! @param[in]  var_size : size of the selected variable
  !! @param[out] molecule : structure of molecule information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine alloc_molecules(molecule, variable, var_size)

    ! formal arguments
    type(s_molecule),     intent(inout) :: molecule
    integer,              intent(in)    :: variable
    integer,              intent(in)    :: var_size

    ! local variables
    integer                             :: alloc_stat
    integer                             :: dealloc_stat


    ! initialize
    !
    alloc_stat   = 0
    dealloc_stat = 0

    ! allocate selected variables

    select case (variable)

    case(MoleculeAtom)

       if (associated(molecule%atom_no)) then
          if (SIZE(molecule%atom_no) == var_size) return
          deallocate(molecule%atom_no,          &
                     molecule%segment_name,     &
                     molecule%segment_no,       &
                     molecule%residue_no,       &
                     molecule%residue_c_no,     &
                     molecule%residue_name,     &
                     molecule%atom_name,        &
                     molecule%atom_cls_name,    &
                     molecule%atom_cls_no,      &
                     molecule%charge,           &
                     molecule%mass,             &
                     molecule%inv_mass,         &
                     molecule%imove,            &
                     molecule%chain_id,         &
                     molecule%atom_coord,       &
                     molecule%atom_occupancy,   &
                     molecule%atom_temp_factor, &
                     molecule%atom_connect_no,  &
                     molecule%atom_connect,     &
                     molecule%molecule_no,      &
                     stat = dealloc_stat)
       end if

       allocate(molecule%atom_no(var_size),           &
                molecule%segment_name(var_size),      &
                molecule%segment_no(var_size),        &
                molecule%residue_no(var_size),        &
                molecule%residue_c_no(var_size),      &
                molecule%residue_name(var_size),      &
                molecule%atom_name(var_size),         &
                molecule%atom_cls_name(var_size),     &
                molecule%atom_cls_no(var_size),       &
                molecule%charge(var_size),            &
                molecule%mass(var_size),              &
                molecule%inv_mass(var_size),          &
                molecule%imove(var_size),             &
                molecule%chain_id(var_size),          &
                molecule%atom_coord(3,var_size),      &
                molecule%atom_occupancy(var_size),    &
                molecule%atom_temp_factor(var_size),  &
                molecule%atom_connect_no(var_size),   &
                molecule%atom_connect(4,var_size),    &
                molecule%molecule_no(var_size),       &
                stat = alloc_stat)

    case (MoleculeBond)

       if (associated(molecule%bond_list)) then
          if (SIZE(molecule%bond_list) == var_size) &
             return
          deallocate(molecule%bond_list, &
                     stat = dealloc_stat)
       end if

       allocate(molecule%bond_list(2, var_size), &
                stat = alloc_stat)

    case (MoleculeAngl)

       if (associated(molecule%angl_list)) then
          if (SIZE(molecule%angl_list) == var_size) &
             return
          deallocate(molecule%angl_list, &
                     stat = dealloc_stat)
       end if

       allocate(molecule%angl_list(3, var_size), &
                stat = alloc_stat)

    case (MoleculeDihe)

       if (associated(molecule%Dihe_list)) then
          if (SIZE(molecule%dihe_list) == var_size) &
             return
          deallocate(molecule%dihe_list, &
                     stat = dealloc_stat)
       end if

       allocate(molecule%dihe_list(4, var_size), &
                stat = alloc_stat)

    case (MoleculeImpr)

       if (associated(molecule%impr_list)) then
          if (SIZE(molecule%impr_list) == var_size) &
             return
          deallocate(molecule%impr_list, &
                     stat = dealloc_stat)
       end if

       allocate(molecule%impr_list(4, var_size), &
                stat = alloc_stat)

    case (MoleculeCmap)

       if (associated(molecule%cmap_list)) then
          if (SIZE(molecule%cmap_list(1,:)) == var_size) &
             return
          deallocate(molecule%cmap_list, &
                     stat = dealloc_stat)
       end if

       allocate(molecule%cmap_list(8, var_size), &
                stat = alloc_stat)

    case (MoleculeMole)

       if (associated(molecule%molecule_atom_no)) then
          if (SIZE(molecule%molecule_atom_no) == var_size) &
             return
          deallocate(molecule%molecule_atom_no, &
                     molecule%molecule_mass,    &
                     molecule%molecule_name,    &
                     stat = dealloc_stat)
       end if

       allocate(molecule%molecule_atom_no(var_size), &
                molecule%molecule_mass(var_size),    &
                molecule%molecule_name(var_size),    &
                stat = alloc_stat)
   
    case (MoleculeRefc)

       if (associated(molecule%atom_refcoord)) then
          if (SIZE(molecule%atom_refcoord(1,:)) == var_size) &
             return
          deallocate(molecule%atom_refcoord, &
                     stat = dealloc_stat)
       end if

       allocate(molecule%atom_refcoord(3,var_size), &
                stat = alloc_stat)
   

    case default

       call error_msg('Alloc_Molecule> bad variable')

    end select

    if (alloc_stat /= 0)   call error_msg_alloc
    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine alloc_molecules

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  dealloc_molecules
  !> @brief      deallocate molecule information
  !! @authors    NT, YS
  !! @param[in]  variable : selected variable
  !! @param[out] molecule : structure of molecule information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_molecules(molecule, variable)

    ! formal arguments
    type(s_molecule),     intent(inout) :: molecule
    integer,              intent(in)    :: variable

    ! local variables
    integer                             :: dealloc_stat


    ! initialize
    !
    dealloc_stat = 0

    ! deallocate selected variable
    !
    select case (variable)

    case(MoleculeAtom)

       if (associated(molecule%atom_no)) then
          deallocate(molecule%atom_no,          &
                     molecule%segment_name,     &
                     molecule%segment_no,       &
                     molecule%residue_no,       &
                     molecule%residue_c_no,     &
                     molecule%residue_name,     &
                     molecule%atom_name,        &
                     molecule%atom_cls_name,    &
                     molecule%atom_cls_no,      &
                     molecule%charge,           &
                     molecule%mass,             &
                     molecule%inv_mass,         &
                     molecule%imove,            &
                     molecule%chain_id,         &
                     molecule%atom_coord,       &
                     molecule%atom_occupancy,   &
                     molecule%atom_temp_factor, &
                     molecule%atom_connect_no,  &
                     molecule%atom_connect,     &
                     molecule%molecule_no,      &
                     stat = dealloc_stat)
          molecule%atom_no          => null()
          molecule%segment_name     => null()
          molecule%segment_no       => null()
          molecule%residue_no       => null()
          molecule%residue_c_no     => null()
          molecule%residue_name     => null()
          molecule%atom_name        => null()
          molecule%atom_cls_name    => null()
          molecule%atom_cls_no      => null()
          molecule%charge           => null()
          molecule%mass             => null()
          molecule%inv_mass         => null()
          molecule%imove            => null()
          molecule%chain_id         => null()
          molecule%atom_coord       => null()
          molecule%atom_occupancy   => null()
          molecule%atom_temp_factor => null()
          molecule%atom_connect_no  => null()
          molecule%atom_connect     => null()
          molecule%molecule_no      => null()
       end if

    case (MoleculeBond)

       if (associated(molecule%bond_list)) then
          deallocate(molecule%bond_list, &
                     stat = dealloc_stat)
          molecule%bond_list => null()
       end if

    case (MoleculeAngl)

       if (associated(molecule%angl_list)) then
          deallocate(molecule%angl_list, &
                     stat = dealloc_stat)
          molecule%angl_list => null()
       end if

    case (MoleculeDihe)

       if (associated(molecule%Dihe_list)) then
          deallocate(molecule%dihe_list, &
                     stat = dealloc_stat)
          molecule%Dihe_list => null()
       end if


    case (MoleculeImpr)

       if (associated(molecule%impr_list)) then
          deallocate(molecule%impr_list, &
                     stat = dealloc_stat)
          molecule%impr_list => null()
       end if

    case (MoleculeCmap)

       if (associated(molecule%cmap_list)) then
          deallocate(molecule%cmap_list, &
                     stat = dealloc_stat)
          molecule%cmap_list => null()
       end if

    case (MoleculeMole)

       if (associated(molecule%molecule_atom_no)) then
          deallocate(molecule%molecule_atom_no, &
                     molecule%molecule_mass,    &
                     molecule%molecule_name,    &
                     stat = dealloc_stat)
          molecule%molecule_atom_no => null()
       end if

    case (MoleculeRefc)

       if (associated(molecule%atom_refcoord)) then
          deallocate(molecule%atom_refcoord, &
                     stat = dealloc_stat)
          molecule%atom_refcoord => null()
       end if

    case default

       call error_msg('Dealloc_Molecule> bad variable')

    end select

    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine dealloc_molecules

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  dealloc_molecules_all
  !> @brief      deallocate molecule information
  !! @authors    NT, YS
  !! @param[out] molecule : structure of molecule information
  !! @date       2012/05/24 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_molecules_all(molecule)

    ! formal arguments
    type(s_molecule),     intent(inout) :: molecule


    call dealloc_molecules(molecule, MoleculeAtom)
    call dealloc_molecules(molecule, MoleculeBond)
    call dealloc_molecules(molecule, MoleculeAngl)
    call dealloc_molecules(molecule, MoleculeDihe)
    call dealloc_molecules(molecule, MoleculeImpr)
    call dealloc_molecules(molecule, MoleculeCmap)
    call dealloc_molecules(molecule, MoleculeMole)
    call dealloc_molecules(molecule, MoleculeRefc)

    return

  end subroutine dealloc_molecules_all

end module molecules_str_mod
