!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   remdcnv_ctrl_mod
!> @brief   read control parameters
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/05/25 
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

module remdcnv_ctrl_mod

  use fileio_control_mod
  use messages_mod

  implicit none
  private

  ! structures

  ! section : INPUT
  type, public :: s_inp_info
    character(200)               :: psffile        = 'initial.psf'
    character(200)               :: reffile        = 'initial.pdb'
    character(10)                :: trj_format     = 'DCD'
    character(100)               :: energy_column  = ''
    character(100)               :: rein_version   = 'latest'

    character(200),      pointer :: directory(:)
    character(200),      pointer :: trjfile(:)
    character(200),      pointer :: rnkfile(:)
    character(200),      pointer :: enefile(:)
    character(100),      pointer :: first_step(:)
    character(100),      pointer :: last_step(:)
    integer,             pointer :: step_increment(:)
  end type s_inp_info

  ! section : OUTPUT
  type, public :: s_out_info
    character(200)               :: trjfile
    character(10)                :: trj_format     = 'DCD'
    character(200)               :: pdbfile        = ''
  end type s_out_info

  ! section : CONVERT
  type, public :: s_cnv_info
    character(10)                :: type           = 'REPLICA'
    character(200)               :: list           = ''
  end type s_cnv_info

  ! section : SELECTION
  type, public :: s_sel_info
    character(1000)              :: select_atom    = 'all'
    character(1000),     pointer :: mole_names(:)
  end type s_sel_info

  ! section : PBC
  type, public :: s_pbc_info
    character(10)                :: pbc_correct    = 'no'
  end type s_pbc_info

  type, public :: s_ctrl_data
    type(s_inp_info)             :: inp_info
    type(s_out_info)             :: out_info
    type(s_cnv_info)             :: cnv_info
    type(s_sel_info)             :: sel_info
    type(s_pbc_info)             :: pbc_info
  end type s_ctrl_data

  ! subroutines
  public  :: control
  private :: read_input
  private :: read_output
  private :: read_convert
  private :: read_selection
  private :: read_pbc

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  control
  !> @brief      a driver subroutine for reading control parmeters
  !! @authors    NT, NM
  !! @param[in]  filename    filename of control file
  !! @param[out] ctrl_data   control parameters in the program
  !! @date       2012/05/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine control(filename, ctrl_data)

    ! formal arguments
    character(*),      intent(in)    :: filename
    type(s_ctrl_data), intent(inout) :: ctrl_data

    ! local variables
    integer                          :: handle

    ! open control file
    !
    call open_ctrlfile(filename, handle)

    if (handle == 0) then
      call error_msg('Remdcnv_Ctrl> File IO Error')
    end if

    ! read section : INPUT
    !
    call read_input(handle, ctrl_data%inp_info)

    ! read section : OUTPUT
    !
    call read_output(handle, ctrl_data%out_info)

    ! read section : CONVERT
    !
    call read_convert(handle, ctrl_data%cnv_info)

    ! read section : SELECTION
    !
    call read_selection(handle, ctrl_data%sel_info)

    ! read section : PBC
    !
    call read_pbc(handle, ctrl_data%pbc_info)

    ! close control file
    !
    call close_ctrlfile(handle)

    return

  end subroutine control


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_input
  !> @brief      read control parameters in input section
  !! @authors    NT, NM
  !! @date       2012/05/25 (NT)
  !! @date       2012/10/05 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_input(handle, inp_info)

    character(*), parameter :: Section = 'Input'

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_inp_info),       intent(inout) :: inp_info

    ! local variables
    character(200)          :: dirval, trjval, rnkval, eneval
    character(100)          :: fsval, lsval
    character(20)           :: dirkey
    character(3)            :: numstr
    integer                 :: i, j, k, ndir, repeat, sival


    ! read control parameters
    !
    call read_ctrlfile_string(handle, Section, &
         'psffile', inp_info%psffile)
    call read_ctrlfile_string(handle, Section, &
         'reffile', inp_info%reffile)
    call read_ctrlfile_string(handle, Section, &
         'trj_format', inp_info%trj_format)
    call read_ctrlfile_string(handle, Section, &
         'energy_column', inp_info%energy_column)
    call read_ctrlfile_string(handle, Section, &
         'rein_version', inp_info%rein_version)


    ! check readable trajectory count
    ndir = 0
    do while (.true.)
      dirval = ''
      write(dirkey, fmt='(a9,i3.3)') 'directory', ndir + 1
      call read_ctrlfile_string(handle, Section, dirkey, dirval)
      if (dirval == '') then
        exit
      end if
      ndir = ndir + 1
    end do

    if (ndir == 0) &
      call error_msg('Read_Input> There is no data to convert.')

    allocate(inp_info%directory(ndir), &
             inp_info%trjfile(ndir),   &
             inp_info%rnkfile(ndir),   &
             inp_info%enefile(ndir),   &
             inp_info%first_step(ndir),&
             inp_info%last_step(ndir), &
             inp_info%step_increment(ndir))

    ! read trajectory file name
    do i = 1, ndir
      write(dirkey,fmt='(a9,i3.3)') "directory", i
      call read_ctrlfile_string(handle, Section, dirkey, &
                                inp_info%directory(i))
    end do


    i = 1
    j = 1
    do while(ndir > 0)
      
      write(numstr, '(I3.3)') j
      j = j + 1

      ! read trj file
      trjval = ''
      call read_ctrlfile_string(handle, Section, &
                                'trj_filename'//numstr, trjval)
      if (trjval == '') then
        write(dirkey, fmt='(a9,i3.3)') 'directory', i
        call error_msg('Read_Input> trajectory filename is empty. [' &
        //TRIM(dirkey)//']')
      end if


      ! read energy file
      eneval = ''
      call read_ctrlfile_string(handle, Section, &
                                'energy_trj'//numstr, eneval)
      if (eneval == '') &
        call error_msg('Read_Input> energy filename is empty.')


      ! read rank file
      rnkval = eneval
      call read_ctrlfile_string(handle, Section, &
                                'rank_trj'//numstr, rnkval)

      ! read first step
      fsval = ''
      call read_ctrlfile_string(handle, Section, &
                                'first_step'//numstr, fsval)

      ! read last step
      lsval = ''
      call read_ctrlfile_string(handle, Section, &
                                'last_step'//numstr, lsval)

      ! read step increment
      sival = 1
      call read_ctrlfile_integer(handle, Section, &
                                 'step_increment'//numstr, sival)
      if (sival == 0) &
        call error_msg('Read_Input> step increment is not set.')

      ! read repeat
      call read_ctrlfile_integer(handle, Section, &
                                 'repeat'//numstr, repeat)

      do k = 1, repeat
        if (i > ndir) then
          call error_msg('Read_Input> directories and trjectory files'// &
                         ' settings are mismatch.')
        end if
        inp_info%trjfile(i)        = trjval
        inp_info%rnkfile(i)        = rnkval
        inp_info%enefile(i)        = eneval
        inp_info%first_step(i)     = fsval
        inp_info%last_step(i)      = lsval
        inp_info%step_increment(i) = sival
        i = i + 1
      end do
      if (i > ndir) then
        exit
      end if
    end do


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Input> Parameters of Inputs'
    write(MsgOut,'(A20,A)') '        psffile   = ', &
         TRIM(inp_info%psffile)
    write(MsgOut,'(A20,A)') '        reffile   = ', &
         TRIM(inp_info%reffile)
    write(MsgOut,'(A20,A)') '     trj format   = ', &
         TRIM(inp_info%trj_format)
    write(MsgOut,'(A20,A)') '  energy column   = ', &
         TRIM(inp_info%energy_column)
    write(MsgOut,'(A20,I10)') ' # of directories = ', ndir
    do i = 1, ndir
      write(MsgOut,'(A20,I3.3,A3,A)') &
           ' directory          ', i, ' = ', TRIM(inp_info%directory(i))
      write(MsgOut,'(A20,A)') &
           '      trj file    : ', TRIM(inp_info%trjfile(i))
      write(MsgOut,'(A20,A)') &
           '      rank file   : ', TRIM(inp_info%rnkfile(i))
      write(MsgOut,'(A20,A)') &
           '      energy file : ', TRIM(inp_info%enefile(i))
      write(MsgOut,'(A20,A)') &
           '       first step : ', TRIM(inp_info%first_step(i))
      write(MsgOut,'(A20,A)') &
           '        last step : ', TRIM(inp_info%last_step(i))
      write(MsgOut,'(A20,I5)') &
           '   step increment : ', inp_info%step_increment(i)
    end do
    write(MsgOut,*) ''

    return

  end subroutine read_input


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_output
  !> @brief      read control parameters in output section
  !! @authors    NT, NM
  !! @date       2012/05/25 (NT)
  !! @date       2012/10/05 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_output(handle, out_info)

    character(*), parameter :: Section = 'Output'

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_out_info),       intent(inout) :: out_info


    ! read control parameters
    !
    call read_ctrlfile_string(handle, Section, &
         'trj_filename', out_info%trjfile)
    if (out_info%trjfile == '') &
         call error_msg('Read_Output> trajectory filename is empty.')

    call read_ctrlfile_string(handle, Section, &
         'trj_format', out_info%trj_format)

    call read_ctrlfile_string(handle, Section, &
         'out_pdbfile', out_info%pdbfile)
    if (out_info%pdbfile == '') &
         call error_msg('Read_Output> PDB filename is empty.')


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Output> Parameters of Outputs'
    write(MsgOut,'(A20,A)') '    trjfilename   = ', &
         TRIM(out_info%trjfile)
    write(MsgOut,'(A20,A)') '     trj format   = ', &
         TRIM(out_info%trj_format)
    write(MsgOut,'(A20,A)') ' output pdbfile   = ', &
         TRIM(out_info%pdbfile)
    write(MsgOut,*) ''

    return

  end subroutine read_output


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_convert
  !> @brief      read control parameters in convert section
  !! @authors    NT, NM
  !! @date       2012/05/25 (NT)
  !! @date       2012/10/05 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_convert(handle, cnv_info)

    character(*), parameter :: Section = 'Convert'

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_cnv_info),       intent(inout) :: cnv_info


    ! read control parameters
    !
    call read_ctrlfile_string(handle, Section, &
         'type', cnv_info%type)
    call read_ctrlfile_string(handle, Section, &
         'rank', cnv_info%list)


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Convert> Parameters of Convert'
    write(MsgOut,'(A20,A)') '           type   = ', &
         TRIM(cnv_info%type)
    write(MsgOut,'(A20,A)') '           rank   = ', &
         TRIM(cnv_info%list)
    write(MsgOut,*) ''

    return 

  end subroutine read_convert


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_selection
  !> @brief      read control parameters in SELECTION section
  !! @authors    NT, NM
  !! @param[in]  handle   unit number of control file [int] 
  !! @param[out] sel_info information of selection section [str] 
  !! @date       2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_selection(handle, sel_info)
 
    character(*), parameter :: Section = 'Selection'

    ! formal arguments
    integer,                intent(in)    :: handle
    type(s_sel_info),       intent(inout) :: sel_info

    ! local variables
    character(1000)         :: value
    character(12)           :: mole_name
    integer                 :: i, nmole


    call read_ctrlfile_string(handle, Section, &
         'select_atom', sel_info%select_atom)

    ! check readable molecule name
    nmole = 0
    do while (.true.)
      value = ''
      write(mole_name, fmt='(A9,I3.3)') 'mole_name', nmole + 1
      call read_ctrlfile_string(handle, Section, mole_name, value)
      if (value == '') then
        exit
      end if
      nmole = nmole + 1
    end do

    ! allocate for molecule names
    allocate(sel_info%mole_names(nmole))
 
    ! read molecule names
    do i = 1, nmole
      write(mole_name,fmt='(A9,I3.3)') 'mole_name', i
      call read_ctrlfile_string(handle, Section, mole_name, &
                                sel_info%mole_names(i))
    end do


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Selection> Parameters of Selection'
    write(MsgOut,'(A20,A)') '    select atom   = ', &
         TRIM(sel_info%select_atom)
    write(MsgOut,'(A20,I10)') '  # of mole names = ', nmole
    do i = 1, nmole
      write(MsgOut,'(A10,I3.3,A3,A)') &
           '    name  ', i, ' = ', TRIM(sel_info%mole_names(i))
    end do
    write(MsgOut,'(A)') ' '

    return

  end subroutine read_selection


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_pbc
  !> @brief      read control parameters in pbc section
  !! @authors    NT, NM
  !! @date       2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_pbc(handle, pbc_info)
 
    character(*), parameter :: Section = 'PBC'

    ! formal arguments
    integer,                  intent(in)    :: handle
    type(s_pbc_info),         intent(inout) :: pbc_info


    call read_ctrlfile_string(handle, Section, &
         'pbc_correct', pbc_info%pbc_correct)


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_PBC> Parameters of PBC'
    write(MsgOut,'(A20,A)') '    pbc correct   = ', &
         TRIM(pbc_info%pbc_correct)
    write(MsgOut,'(A)') ' '

    return

  end subroutine read_pbc

end module remdcnv_ctrl_mod
