!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   remdcnv_run_mod
!> @brief   run REMD converter
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/05/29 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module remdcnv_run_mod

  use cnv_data_str_mod
  use trajectory_str_mod
  use molecules_str_mod
  use selatoms_str_mod
  use replace_string_mod
  use fileio_control_mod
  use fileio_trj_mod
  use fileio_mod
  use messages_mod
  use omp_lib

  implicit none
  private

  integer, private, parameter       :: MaxEneColumn = 100

  ! global variables
  type(s_trajectory),   allocatable :: g_trajectory(:)

  ! subroutines
  public  :: run
  private :: run_dir
  private :: run_pbc_correct

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run
  !> @brief        run convert
  !! @authors      NT, NM
  !! @param[in]    cnv_data    convert data information
  !! @param[in]    molecule    molecule information
  !! @param[in]    trjout_atom atom selection information
  !! @date         2012/06/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine run(cnv_data, molecule, trjout_atom)

    ! formal arguments
    type(s_cnv_data),  intent(in)    :: cnv_data
    type(s_molecule),  intent(in)    :: molecule
    type(s_selatoms),  intent(in)    :: trjout_atom

    ! local variables
    type(s_trj_file)   :: crd_out
    integer            :: nthreads
    integer            :: i, inum, idat, idir, ene_out
    character(200)     :: filecrd, fileene



    ! allocate trajectory information for each threads
    !
    nthreads = omp_get_max_threads()

    allocate(g_trajectory(nthreads))
    do i = 1, nthreads
      call alloc_trajectory(g_trajectory(i), molecule%num_atoms)
    end do


    !$omp parallel private(inum, idat, idir, filecrd, fileene, crd_out, ene_out)
    !

    do i = 1, SIZE(cnv_data%cnv_list), omp_get_num_threads()

      inum = i + omp_get_thread_num()
      if (inum > SIZE(cnv_data%cnv_list)) &
        cycle

      idat = cnv_data%cnv_list(inum)      

      !$omp critical

      call make_outname(cnv_data%trjout_file, &
                        cnv_data%cnv_type,    &
                        idat, 'crdtrj', filecrd)
       
      call make_outname(cnv_data%trjout_file, &
                        cnv_data%cnv_type,    &
                        idat, 'enetrj', fileene)
      
      call open_trj (crd_out, filecrd,       &
                     cnv_data%trjout_format, &
                     cnv_data%trjout_type,   &
                     IOFileOutputReplace)

      call open_file(ene_out, fileene, IOFileOutputReplace)

      !$omp end critical

      write(MsgOut,'(A,A)') '  Crd trj: ', TRIM(filecrd)
      write(MsgOut,'(A,A)') '  Ene trj: ', TRIM(fileene)


      ! convert process
      !

      if (cnv_data%cnv_type == CnvTypeReplica) then
        write(MsgOut,'(A,I0)') 'Replica: ',   idat
      else
        write(MsgOut,'(A,I0)') 'Condition: ',   idat
      end if

      do idir = 1, SIZE(cnv_data%cnv_in)
        call run_dir(crd_out, ene_out, &
                     idat, idir, &
                     cnv_data, molecule, trjout_atom)
      end do


      !$omp critical
      call close_file(ene_out)
      call close_trj (crd_out)
      !$omp end critical

    end do

    !
    !$omp end parallel


    ! deallocate trajectory information for each threads
    !
    do i = 1, nthreads
      call dealloc_trajectory(g_trajectory(i))
    end do
    deallocate(g_trajectory)


    write(MsgOut,'(A)') ' '

    return

  end subroutine run


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_dir_rep
  !> @brief        convert a directory by replica
  !! @authors      NT, NM
  !! @param[in]    crd_out     file unit number for coordinate trajectory
  !! @param[in]    ene_out     file unit number for coordinate trajectory
  !! @param[in]    idat        index of replica or condition data
  !! @param[in]    idir        index of directory
  !! @param[in]    cnv_data    convert data information
  !! @param[in]    molecule    molecule information
  !! @param[in]    trjout_atom atom selection information
  !! @date         2012/06/12 (NT)
  !! @date         2012/10/05 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine run_dir(crd_out, &
                     ene_out, &
                     idat,    &
                     idir,    &
                     cnv_data, molecule, trjout_atom)

    ! formal arguments
    type(s_trj_file),  intent(inout) :: crd_out
    integer,           intent(in)    :: ene_out
    integer,           intent(in)    :: idat
    integer,           intent(in)    :: idir
    type(s_cnv_data),  intent(in)    :: cnv_data
    type(s_molecule),  intent(in)    :: molecule
    type(s_selatoms),  intent(in)    :: trjout_atom

    ! local variables
    type(s_trj_file)   :: trj_in
    double precision   :: col(MaxEneColumn)
    integer            :: tid
    integer            :: rnk_in, ene_in, nstp, nrep, rnk
    integer            :: irep, istp, ibegin, ifirst, ilast, istinc, trjin
    character(200)     :: rnkfile, enefile, trjfile, inpname, line
    character(10)      :: str

    integer            :: replica_number, condition_number
    character(200)     :: line_data

    integer,           allocatable :: rep_no(:)
   
    integer            :: itmp


    ! create replica number list
    !

    nstp = cnv_data%cnv_in(idir)%num_steps
    nrep = cnv_data%num_replicas

    rnkfile = TRIM(cnv_data%cnv_in(idir)%directory)//'/'//&
              TRIM(cnv_data%cnv_in(idir)%rnkfile)

    allocate(rep_no(nstp))

    if (cnv_data%cnv_type == CnvTypeReplica) then

      !rep_no(:) = idat
      !$omp critical
      call open_file(rnk_in, rnkfile, IOFileInput)
      
      do istp = 1, nstp
        read(rnk_in,'(A)') line
        do irep = 1, nrep
          if (cnv_data%ene_column_old) then
            read(rnk_in,*) rnk
          else if (cnv_data%ene_column_09) then
            read(rnk_in,*) rnk
          else
            itmp = cnv_data%num_ene_column
            read(rnk_in,*) rnk, condition_number, col(1:itmp)
          end if
          if (rnk == idat) &
            rep_no(istp) = irep
        end do
      end do

      call close_file(rnk_in)
      !$omp end critical

    else

      !$omp critical
      call open_file(rnk_in, rnkfile, IOFileInput)
      
      do istp = 1, nstp
        read(rnk_in,'(A)') line
        do irep = 1, nrep
          if (cnv_data%ene_column_old) then
            read(rnk_in,*) rnk
          else if (cnv_data%ene_column_09) then
            read(rnk_in,*) rnk
          else
            itmp = cnv_data%num_ene_column 
            read(rnk_in,*) replica_number, rnk, col(1:itmp)
          end if
          if (rnk == idat) &
            rep_no(istp) = irep
        end do
      end do

      call close_file(rnk_in)
      !$omp end critical

    end if


    ! output coordinate trajectory
    !

    tid    = omp_get_thread_num()
    ibegin = cnv_data%cnv_in(idir)%begin_step
    ifirst = cnv_data%cnv_in(idir)%first_step
    ilast  = cnv_data%cnv_in(idir)%last_step
    istinc = cnv_data%cnv_in(idir)%step_increment

    write(MsgOut,'(A,I3,A,I3,A,I4,A,I4,A,I0)') &
                      ' tid: ',  tid, &
                      '  dat: ', idat, &
                      '  dir: ', idir, &
                      '   fs: ', ifirst, &
                      '   ls: ', ilast

    do istp = ifirst, ilast, istinc

      call make_inpname(cnv_data%cnv_in(idir)%trjfile, &
                        rep_no(istp), istp + ibegin, inpname)

      trjfile = TRIM(cnv_data%cnv_in(idir)%directory)//'/'//inpname

      !$omp critical
      call open_trj (trj_in, trjfile, cnv_data%trj_format, &
                     TrjTypeCoorBox, IOFileInput)
      !$omp end critical


      ! read trajectory
      !
      call read_trj(trj_in, g_trajectory(tid+1))


      ! run pbc-correct
      !
      call run_pbc_correct(cnv_data%pbc_correct, &
                           molecule, g_trajectory(tid+1))


      ! write trajectory
      !
      call write_trj(crd_out, g_trajectory(tid+1), &
                     trjout_atom, molecule)


      !$omp critical
      call close_trj(trj_in)
      !$omp end critical

    end do


    ! output energy trajectory
    !

    enefile = TRIM(cnv_data%cnv_in(idir)%directory)//'/'//&
              TRIM(cnv_data%cnv_in(idir)%enefile)

    !$omp critical
    call open_file(ene_in, enefile, IOFileInput)

    ! skip first REMD step
    do istp = 1, ifirst - 1
      read(ene_in,*) line
      do irep = 1, nrep
        read(ene_in,*) line
      end do
    end do

    do istp = ifirst, ilast

      read(ene_in,*) line

      ! skip first replica
      do irep = 1, rep_no(istp) -1
        read(ene_in,*) line
      end do

      ! read energy
      if (MOD(istp - ifirst, istinc) == 0) then
        if (cnv_data%ene_column_old) then
          read(ene_in,*) col(1:5), str, col(6)
          write(ene_out,'(I7,I7,100F21.5)') istp + ibegin, &
                                          rep_no(istp),  &
                                         col(cnv_data%ene_column(:))
        else if (cnv_data%ene_column_09) then
          read(ene_in,*) col(1:cnv_data%num_ene_column)
          write(ene_out,'(I7,I7,100F21.5)') istp + ibegin, &
                                          rep_no(istp),  &
                                         col(cnv_data%ene_column(:))
        else
          itmp = cnv_data%num_ene_column
          read(ene_in,*) replica_number, condition_number, col(1: itmp)
          write(ene_out,'(I8,2I8,100F21.5)') istp + ibegin, &
                                          replica_number, condition_number, &
                                          col(cnv_data%ene_column(:))
        end if
      else
        read(ene_in,*) line
      end if

      ! skip last
      do irep = rep_no(istp), nrep -1
        read(ene_in,*) line
      end do

    end do

    call close_file(ene_in)
    !$omp end critical


    ! delete replica number list
    !
    deallocate(rep_no)

    return

  end subroutine run_dir


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_inpname
  !> @brief        make output filename
  !! @authors      NT, NM
  !! @date         2012/06/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine make_inpname(format_str, &
                          rep_no,     &
                          step,       &
                          filename)

    ! formal arguments
    character(*),     intent(in)     :: format_str
    integer,          intent(in)     :: rep_no
    integer,          intent(in)     :: step
    character(*),     intent(inout)  :: filename

    ! local variables
    character(20)     :: keyword, numstr
    integer           :: kw_index


    filename = format_str

    do while (check_replace(filename, kw_index, keyword, "{}"))

      select case(keyword)

      case ('rep_no')
        write(numstr,'(I0)') rep_no
        call replace(filename, kw_index, keyword, numstr)

      case ('step')
        write(numstr,'(I0)') step
        call replace(filename, kw_index, keyword, numstr)

      case default
        call error_msg('Make_Inpname> Unknown keyword:"'//TRIM(keyword)//&
             '" in File:'//TRIM(format_str))

      end select

    end do

    return

  end subroutine make_inpname


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_outname
  !> @brief        make output filename
  !! @authors      NT, NM
  !! @date         2012/06/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine make_outname(format_str, &
                          cnv_type,   &
                          number,     &
                          ext_str,    &
                          filename)

    ! formal arguments
    character(*),     intent(in)     :: format_str
    integer,          intent(in)     :: cnv_type
    integer,          intent(in)     :: number
    character(*),     intent(in)     :: ext_str
    character(*),     intent(inout)  :: filename

    ! local variables
    character(20)     :: keyword, numstr
    integer           :: kw_index


    filename = format_str

    do while (check_replace(filename, kw_index, keyword, "{}"))

      select case(keyword)

      case ('type')

        select case (cnv_type)
          case (CnvTypeReplica)
            call replace(filename, kw_index, keyword, 'replica')

          case (CnvTypeCondition)
            call replace(filename, kw_index, keyword, 'condition')

        end select

      case ('number')
        write(numstr,'(I0)') number
        call replace(filename, kw_index, keyword, numstr)

      case ('extension')
        call replace(filename, kw_index, keyword, ext_str)

      case default
        call error_msg('Make_Outname> Unknown keyword:"'//TRIM(keyword)//&
             '"  in File:'//TRIM(format_str))

      end select

    end do

    return

  end subroutine make_outname


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_pbc_correct
  !> @brief        run pbc correct
  !! @authors      NT, NM
  !! @param[in]    pbcc       PBC-correct mode [character(*)]
  !! @param[in]    molecule   information of molecule [s_molecule]
  !! @param[inout] trajectory trajectory for pbc correction [s_trajectory]
  !! @date         2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine run_pbc_correct(pbcc, molecule, trajectory)

    ! formal argments
    logical,                    intent(in)    :: pbcc
    type(s_molecule),           intent(in)    :: molecule
    type(s_trajectory), target, intent(inout) :: trajectory

    ! local variables
    integer             :: i, j, initial, final
    double precision    :: com(3)
    double precision    :: boxx, boxy, boxz
    double precision    :: iboxx, iboxy, iboxz
    double precision, pointer :: coord(:,:)


    if (.not. pbcc) &
      return

    if (molecule%num_molecules < 1) then
      call error_msg('Run_Pbc_Correct> molecule is not defined.')
    end if

    boxx  = trajectory%pbc_box(1,1)
    boxy  = trajectory%pbc_box(2,2)
    boxz  = trajectory%pbc_box(3,3)
    iboxx = 1.0d0/boxx
    iboxy = 1.0d0/boxy
    iboxz = 1.0d0/boxz

    coord => trajectory%coord

    do i = 1, molecule%num_molecules

      initial  = molecule%molecule_atom_no(i)
      if (i /= molecule%num_molecules) then
        final = molecule%molecule_atom_no(i+1) - 1
      else
        final = molecule%num_atoms
      end if

      ! compute center of mass of a molecule
      !
      com(1) = 0.0d0
      com(2) = 0.0d0
      com(3) = 0.0d0
      do j = initial, final
        com(1) = com(1) + coord(1,j) * molecule%mass(j)
        com(2) = com(2) + coord(2,j) * molecule%mass(j)
        com(3) = com(3) + coord(3,j) * molecule%mass(j)
      end do
      com(1) = com(1)/molecule%molecule_mass(i)
      com(2) = com(2)/molecule%molecule_mass(i)
      com(3) = com(3)/molecule%molecule_mass(i)

      ! move molecule into the unit cell
      !
      do j = initial, final
        coord(1,j) = coord(1,j) - boxx * nint(com(1)*iboxx)
        coord(2,j) = coord(2,j) - boxy * nint(com(2)*iboxy)
        coord(3,j) = coord(3,j) - boxz * nint(com(3)*iboxz)
      end do

    end do

    return

  end subroutine run_pbc_correct

end module remdcnv_run_mod
