!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   remdcnv_setup_mod
!> @brief   setup REMD converter
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/05/29 
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

module remdcnv_setup_mod

  use remdcnv_ctrl_mod
  use fileio_pdb_mod
  use fileio_psf_mod
  use fileio_mod
  use molecules_mod
  use select_mod
  use cnv_data_str_mod
  use trajectory_str_mod
  use molecules_str_mod
  use selatoms_str_mod
  use fileio_control_mod
  use messages_mod

  implicit none
  private

  ! constant parameters
  integer, private, parameter :: MaxRangeRes     = 50
  integer, private, parameter :: MaxGroupRes     = 50
  integer, private, parameter :: MaxGroupResName = 10

  integer, private, parameter :: TblType         = 1
  integer, private, parameter :: TblMolNo        = 2
  integer, private, parameter :: TblMolID        = 3

  ! structures
  type s_range_res
    character(20)     :: name
    character(4)      :: seg_name_b
    character(4)      :: seg_name_e
    integer           :: res_no_b
    integer           :: res_no_e
    character(4)      :: res_name_b
    character(4)      :: res_name_e
  end type s_range_res

  type s_group_res
    character(20)     :: name
    integer           :: num_res
    character(4)      :: res_name(MaxGroupResName)
  end type s_group_res

  type s_mole_info
    integer           :: num_range
    integer           :: num_group
    type(s_range_res) :: range_res(MaxRangeRes)
    type(s_group_res) :: group_res(MaxGroupRes)
  end type s_mole_info
  
  ! subroutines
  public  :: setup

  private :: setup_selection
  private :: setup_pbc_correct
  private :: setup_cnv_data
  private :: parse_mole_names
  private :: setup_mole_no
  private :: make_number
  private :: get_no

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup
  !> @brief        a subroutine for setup convert
  !! @authors      NT, NM
  !! @param[in]    ctrl_data   control parameters in the program
  !! @param[inout] cnv_data    convert data information
  !! @param[inout] molecule    molecule information
  !! @param[inout] trjout_atom atom selection information
  !! @date         2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup(ctrl_data, cnv_data, molecule, trjout_atom)

    ! formal arguments
    type(s_ctrl_data), intent(in)    :: ctrl_data
    type(s_cnv_data),  intent(inout) :: cnv_data
    type(s_molecule),  intent(inout) :: molecule
    type(s_selatoms),  intent(inout) :: trjout_atom

    ! local variables
    type(s_pdb)        :: ref, ref_out
    type(s_psf)        :: psf
    logical            :: use_psf


    ! read reference coordinate file
    !
    use_psf = LEN_TRIM(ctrl_data%inp_info%psffile) > 0

    if (use_psf) &
      call input_psf(ctrl_data%inp_info%psffile, psf)

    call input_pdb(ctrl_data%inp_info%reffile, ref)


    ! define molecule
    !
    if (use_psf) then
      call define_molecules(molecule, ref, psf=psf)
    else
      call define_molecules(molecule, ref)
    end if


    ! setup selection
    !
    call setup_selection(ctrl_data%sel_info, molecule, trjout_atom)


    ! setup pbc correction
    !
    call setup_pbc_correct(ctrl_data%pbc_info%pbc_correct, molecule)


    ! setup convert data
    !
    call setup_cnv_data(ctrl_data%inp_info, &
                        ctrl_data%cnv_info, &
                        ctrl_data%out_info, &
                        ctrl_data%pbc_info, &
                        cnv_data)


    ! export reference molecules
    !
    call export_molecules(molecule, trjout_atom, ref_out)
    call output_pdb(ctrl_data%out_info%pdbfile, ref_out)

    call dealloc_pdb_all(ref_out)

    write(MsgOut,'(A)') ' '

    return

  end subroutine setup


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_selection
  !> @brief        setup selection
  !! @authors      NT, NM
  !! @param[in]    sel_info  information of selection section
  !! @param[inout] molecule  information of molecule
  !! @param[inout] selatoms  information of select atom
  !! @date         2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_selection(sel_info, molecule, selatoms)
  
    ! formal argments
    type(s_sel_info),   intent(in)    :: sel_info
    type(s_molecule),   intent(inout) :: molecule
    type(s_selatoms),   intent(inout) :: selatoms

    ! local variables
    type(s_mole_info)   :: mole_info
    integer             :: nmole


    nmole  = SIZE(sel_info%mole_names)

    ! setup molecule number
    !
    if (nmole > 0) then

      if (associated(molecule%molecule_atom_no)) then
        write(MsgOut,'(A)') &
             'Setup_Selection> Original molecule information is overwrote.'
      end if

      call parse_mole_names(sel_info%mole_names, mole_info)
      call setup_mole_no   (mole_info, molecule)

      if (molecule%num_molecules > 0) then
        write(MsgOut,'(A)') &
             'Setup_Selection> Molecule information is defined newly.'
        write(MsgOut,'(A,I7)') &
             '                   number of molecules : ', &
             molecule%num_molecules
        write(MsgOut,'(A)') ' '
      end if

    end if

    ! atom selection
    !
    call select_atom(molecule, sel_info%select_atom, selatoms)

    return

  end subroutine setup_selection


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_pbc_correct
  !> @brief        setup pbc-correction
  !! @authors      NT, NM
  !! @param[in]    pbcc       PBC-correct mode [character(*)]
  !! @param[inout] molecule   information of molecule [s_molecule]
  !! @date         2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_pbc_correct(pbcc, molecule)

    ! formal argments
    character(*),       intent(in)    :: pbcc
    type(s_molecule),   intent(inout) :: molecule

    ! local variables
    character(100)      :: str
    integer             :: i, j, natom, nmole, first, last
    logical             :: set_mass


    ! setup for PBC-correct mode "molecule"
    str = pbcc;
    call tolower(str)
    if (str /= 'molecule') &
      return

    ! check molecules information
    if (.not. associated(molecule%molecule_atom_no)) &
      call error_msg('Setup_Pbc_Correct> Molecule information is not defined.')

    set_mass = .true.

    natom = SIZE(molecule%atom_coord(1,:))
    nmole = SIZE(molecule%molecule_atom_no)

    do i = 1, natom
      if (molecule%mass(i) > 0) then
        set_mass = .false.
        exit
      end if
    end do

    if (set_mass) then

      write(MsgOut,'(A)') &
           'Setup_PBC_Correct> Setup molecule mass information.'

      molecule%mass(:) = 1.d0

      do i = 1, nmole

        first = molecule%molecule_atom_no(i)
        if (i /= nmole) then
          last = molecule%molecule_atom_no(i + 1) - 1
        else
          last = natom
        end if

        molecule%molecule_mass(i) = 0.d0

        do j = first, last
          molecule%molecule_mass(i) = molecule%molecule_mass(i) + &
               molecule%mass(j)
        end do
      end do

    end if

    return

  end subroutine setup_pbc_correct


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_cnv_data
  !> @brief        setup convert data
  !! @authors      NT, NM
  !! @param[in]    inp_info  information of input section
  !! @param[in]    cnv_info  information of convert section
  !! @param[in]    out_info  information of output section
  !! @param[in]    pbc_info  information of pbc section
  !! @param[inout] cnv_data  information of convert data
  !! @date         2012/06/12 (NT)
  !! @date         2012/10/05 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_cnv_data(inp_info, cnv_info, out_info, pbc_info, cnv_data)
  
    ! formal argments
    type(s_inp_info),   intent(in)    :: inp_info
    type(s_cnv_info),   intent(in)    :: cnv_info
    type(s_out_info),   intent(in)    :: out_info
    type(s_pbc_info),   intent(in)    :: pbc_info
    type(s_cnv_data),   intent(inout) :: cnv_data

    ! local variables
    double precision    :: ene(100)
    integer             :: i, rnk_in, ene_in, nrep, nstp, istp, idx
    character(200)      :: line
    character(100)      :: str, rnk_file, ene_file
    integer             :: irank_cond, irank_rep
    character(200)      :: ntmp1, ntmp2, ntmp3


    ! setup trj_format
    !
    str = inp_info%trj_format;
    call tolower(str)

    select case(str)
    case ('pdb')
      cnv_data%trj_format = TrjFormatPDB
    case ('marble')
      cnv_data%trj_format = TrjFormatMarble
    case ('amber')
      cnv_data%trj_format = TrjFormatAmber
    case ('tinker')
      cnv_data%trj_format = TrjFormatTinker
    case ('dcd')
      cnv_data%trj_format = TrjFormatDCD
    case ('gromacs')
      cnv_data%trj_format = TrjFormatGromacs
    case default
      call error_msg('Setup_Cnv_Data> Unknown trajectory format '//&
           TRIM(inp_info%trj_format))
    end select

    ! setup rein version
    select case ( TRIM ( inp_info%rein_version ) )
    case ('old')
      cnv_data%ene_column_old = .TRUE.
      cnv_data%ene_column_09  = .FALSE.
    case ('09' )
      cnv_data%ene_column_old = .FALSE.
      cnv_data%ene_column_09  = .TRUE.
    case ('latest')
      cnv_data%ene_column_old = .FALSE.
      cnv_data%ene_column_09  = .FALSE.
    case default
    end select

    ! setup cnv in(s)
    !
    allocate(cnv_data%cnv_in(SIZE(inp_info%directory)))

    do i = 1, SIZE(inp_info%directory)

      ! directory
      cnv_data%cnv_in(i)%directory  = inp_info%directory(i)

      ! input trajectory file
      cnv_data%cnv_in(i)%trjfile    = inp_info%trjfile(i)

      ! input rank file
      cnv_data%cnv_in(i)%rnkfile    = inp_info%rnkfile(i)

      ! input energy file
      cnv_data%cnv_in(i)%enefile    = inp_info%enefile(i)

      ! number of steps
      rnk_file = TRIM(cnv_data%cnv_in(i)%directory)//'/'//&
                 TRIM(cnv_data%cnv_in(i)%rnkfile)
      call open_file(rnk_in, rnk_file, IOFileInput)

      nstp = 0
      do while(.true.)
        read(rnk_in,'(a)',err=920,end=30) str
        if (str(1:11) == ' #=========') then
          nstp = nstp + 1
        end if
      end do

30    cnv_data%cnv_in(i)%num_steps = nstp
      write(MsgOut,'(A,I4,A,I0)') &
           '  Directory:', i, '   # of REMD steps: ', nstp

      ! begin step
      REWIND(rnk_in)

      read(rnk_in,'(a)',err=920,end=920) str
      
      idx = SCAN(str, '=', .true.)
      if (idx == 0) goto 920
      read(str(idx+1:),*) istp

      call close_file(rnk_in)

      cnv_data%cnv_in(i)%begin_step = istp - 1
      write(MsgOut,'(A,I0)') &
           '    begin step: ', istp

      ! first step
      istp = 0
      if (LEN_TRIM(inp_info%first_step(i)) > 0) &
        read(inp_info%first_step,*,err=900,end=900) istp
      if (istp <= 0) &
        istp = 1

      cnv_data%cnv_in(i)%first_step = istp
      write(MsgOut,'(A,I0)') &
           '    first step: ', istp

      ! last step
      istp = 0
      if (LEN_TRIM(inp_info%last_step(i)) > 0) &
        read(inp_info%last_step,*,err=910,end=910) istp
      if (istp <= 0) &
        istp = nstp

      cnv_data%cnv_in(i)%last_step = istp
      write(MsgOut,'(A,I0)') &
           '     last step: ', istp

      ! step increment
      cnv_data%cnv_in(i)%step_increment = inp_info%step_increment(i)

    end do
    write(MsgOut,'(A)') ' '


    ! setup num_ene_column
    !
    ene_file = TRIM(cnv_data%cnv_in(1)%directory)//'/'//&
               TRIM(cnv_data%cnv_in(1)%enefile)
    call open_file(ene_in, ene_file, IOFileInput)

    read(ene_in,*)     line
    read(ene_in,'(A)') line
    read(line,*,err=930,end=10) ene(1:5), str, ene(6:6)

10  if (SCAN(str,'=>') /= 0) then
      cnv_data%num_ene_column = 6
      cnv_data%ene_column_old = .true.
    else
      ene(:) = -100000000
      if (cnv_data%ene_column_09) then
        read(line,*,err=20,end=20) ene
      else
        read(line,*,err=20,end=20) irank_rep, irank_cond, ene
      end if
20    do i = 1, SIZE(ene)
        if (ene(i) == -100000000) &
          exit
      end do
      cnv_data%num_ene_column = i - 1
      cnv_data%ene_column_old = .false.

    end if

    call close_file(ene_in)

    write(MsgOut,'(A,I0)') &
         ' # of columns of energy file: ', cnv_data%num_ene_column
    if (cnv_data%ene_column_old) &
      write(MsgOut,'(A)') '           energy file is old version.'


    ! setup ene_column
    !
    
    ntmp3 = inp_info%energy_column
    if (LEN_TRIM(inp_info%energy_column) == 0) then
      if (cnv_data%ene_column_09) then
        write(ntmp1,*) cnv_data%num_ene_column
      else if ( cnv_data%ene_column_old ) then
        write(ntmp1,*) cnv_data%num_ene_column
      else
        write(ntmp1,*) cnv_data%num_ene_column
      end if
      ntmp2 = "1-" // TRIM ( ntmp1 ) 
      ntmp3 = TRIM ( ntmp2 )
    end if
    write(MsgOut,'(A,I0)') &
         ' # of columns of energy file: ', cnv_data%num_ene_column
    if (cnv_data%ene_column_old) &
      write(MsgOut,'(A)') '           energy file is old version.'

    call make_number(ntmp3,                   &
                     cnv_data%num_ene_column, &
                     cnv_data%ene_column)


    ! setup convert type
    !
    str = cnv_info%type
    call tolower(str)

    select case(str)
    case ('replica')
      cnv_data%cnv_type = CnvTypeReplica
    case ('condition')
      cnv_data%cnv_type = CnvTypeCondition  
    case default
      call error_msg('Setup_Cnv_Data> Unknown convert type '//&
           TRIM(cnv_info%type))
    end select


    ! setup convert list
    !
    rnk_file = TRIM(cnv_data%cnv_in(1)%directory)//'/'//&
               TRIM(cnv_data%cnv_in(1)%rnkfile)
    call open_file(rnk_in, rnk_file, IOFileInput)

    i = 0
    nrep = 0
    do while(.true.)
      read(rnk_in,'(a)',err=920,end=40) str
      if (str(1:11) == ' #=========') then
        i = i + 1
        if (i == 2) &
          exit
      else
        nrep = nrep + 1
      end if
    end do

40  call close_file(rnk_in)

    call make_number(cnv_info%list, nrep, cnv_data%cnv_list)


    ! setup number of replicas
    !
    cnv_data%num_replicas = nrep


    ! pbc correct
    !
    str = pbc_info%pbc_correct
    call tolower(str)
    cnv_data%pbc_correct = str == 'molecule'


    ! setup trjout_format, trjout_type
    !
    str = out_info%trj_format
    call tolower(str)

    select case(str)
    case ('pdb')
      cnv_data%trjout_format = TrjFormatPDB
      cnv_data%trjout_type   = TrjTypeCoorBox
    case ('marble')
      cnv_data%trjout_format = TrjFormatMarble
      cnv_data%trjout_type   = TrjTypeCoorBox
    case ('amber')
      cnv_data%trjout_format = TrjFormatAmber
      cnv_data%trjout_type   = TrjTypeCoorBox
    case ('tinker')
      cnv_data%trjout_format = TrjFormatTinker
      cnv_data%trjout_type   = TrjTypeCoor
    case ('dcd')
      cnv_data%trjout_format = TrjFormatDCD
      cnv_data%trjout_type   = TrjTypeCoorBox
    case ('gromacs')
      cnv_data%trjout_format = TrjFormatGromacs
      cnv_data%trjout_type   = TrjTypeCoorBox
    case default
      call error_msg('Setup_Cnv_Data> Unknown trajectory format '//&
           TRIM(out_info%trj_format))
    end select


    ! setup trjout file
    !
    cnv_data%trjout_file = out_info%trjfile

    return

900 call error_msg('Setup_Cnv_Data> First step line is BAD format.')
910 call error_msg('Setup_Cnv_Data> Last step line is BAD format.')
920 call error_msg('Setup_Cnv_Data> Rank file is BAD format.'//TRIM(rnk_file))
930 call error_msg('Setup_Cnv_Data> Ener file is BAD format.'//TRIM(ene_file))

  end subroutine setup_cnv_data

  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    parse_mole_names
  !> @brief        parse molecule name definition
  !! @authors      NT, NM
  !! @param[in]    mole_names molecule names
  !! @param[inout] mole_info  molecule information
  !! @date         2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine parse_mole_names(mole_names, mole_info)

    ! formal arguments
    character(*),      intent(in)    :: mole_names(:)
    type(s_mole_info), intent(inout) :: mole_info

    ! local variables
    character(1000)    :: mole_name
    character(100)     :: s1, s2, s3
    character(4)       :: seg_name, res_name
    integer            :: res_no, i, j, nmole
    

    nmole = SIZE(mole_names)

    mole_info%num_range = 0
    mole_info%num_group = 0

    do j = 1, nmole

      mole_name = mole_names(j)
       
      s1 = ''
      s2 = ''
      s3 = ''
      read(mole_name,*,end=110,err=900) s1, s2, s3

110   if (s1 == '') &
        cycle

      if (s3 /= '') then
        ! residue by range 
        !

        mole_info%num_range = mole_info%num_range + 1
        if (mole_info%num_range > MaxRangeRes) &
          cycle

        do i = 1, LEN(s2)
          if (s2(i:i) == ':') s2(i:i) = ' '
        end do

        ! begin value
        seg_name = ''
        res_no   = -1
        res_name = ''
        read(s2,*,end=120,err=120) seg_name, res_no, res_name

120     if (seg_name == '') &
          call error_msg('Parse_Mole_Range> ERROR : molecule name : '//&
          TRIM(s1))

        if (res_name == '') then
          seg_name = ''
          res_no   = -1
          read(s2,*,end=121,err=121) res_no, res_name
121       if (res_no == -1 .or. res_name == '') &
            call error_msg('Parse_Mole_Range> ERROR : molecule name : '//&
            TRIM(s1))
        end if

        mole_info%range_res(mole_info%num_range)%name       = s1
        mole_info%range_res(mole_info%num_range)%seg_name_b = seg_name
        mole_info%range_res(mole_info%num_range)%res_no_b   = res_no
        mole_info%range_res(mole_info%num_range)%res_name_b = res_name

        do i = 1, LEN(s3)
          if (s3(i:i) == ':') s3(i:i) = ' '
        end do

        ! end value
        seg_name = ''
        res_no   = -1
        res_name = ''
        read(s3,*,end=130,err=130) seg_name, res_no, res_name

130     if (seg_name == '') &
           call error_msg('Parse_Mole_Range> ERROR : molecule name : '//&
           TRIM(s1))

        if (res_name == '') then
          seg_name = ''
          res_no   = -1
          read(s3,*,end=131,err=131) res_no, res_name
131       if (res_no == -1 .or. res_name == '') &
            call error_msg('Parse_Mole_Range> ERROR : molecule name : '//&
            TRIM(s1))
        end if

        mole_info%range_res(mole_info%num_range)%name       = s1
        mole_info%range_res(mole_info%num_range)%seg_name_e = seg_name
        mole_info%range_res(mole_info%num_range)%res_no_e   = res_no
        mole_info%range_res(mole_info%num_range)%res_name_e = res_name

      else
        ! residue by group
        !

        mole_info%num_group = mole_info%num_group + 1
        if (mole_info%num_group > MaxGroupRes) &
             cycle

        do i = 1, LEN(s2)
          if (s2(i:i) == ':') s2(i:i) = ' '
        end do

        mole_info%group_res(mole_info%num_group)%name = s1

        mole_info%group_res(mole_info%num_group)%res_name(:) = ''
        read(s2,*,end=140,err=900) &
          (mole_info%group_res(mole_info%num_group)%res_name(i), &
          i=1,MaxGroupResName)

140     continue

        do i = 1, MaxGroupResName
          if (mole_info%group_res(mole_info%num_group)%res_name(i) == '') &
            exit
        end do

        if (i - 1 > 0) then
          mole_info%group_res(mole_info%num_group)%num_res = i - 1
        else
          mole_info%num_group = mole_info%num_group - 1
        end if

      end if

    end do

100 continue

    return

900 call error_msg_fileio
    return

  end subroutine parse_mole_names


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mole_no
  !> @brief        setup molecule number
  !! @authors      NT, NM
  !! @param[in]    mole_info  molecule information
  !! @param[inout] molecule   molecule information
  !! @date         2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_mole_no(mole_info, molecule)

    ! formal argments
    type(s_mole_info),             intent(in)    :: mole_info
    type(s_molecule),  target,     intent(inout) :: molecule

    ! local variables
    character(4)                   :: snb, sne, rnab, rnae
    character(4)                   :: gr(MaxGroupResName)
    integer                        :: alloc_stat
    integer                        :: i, j, k, natom, nresi, nmole
    integer                        :: rnob, rnoe, si1, si2
    integer                        :: mid, ngr
    integer                        :: ty, rn

    character(4),      allocatable :: segname(:)
    character(4),      allocatable :: resname(:)
    integer,           allocatable :: resno(:), atomidx(:,:), mol_tbl(:,:)

    character(4),      pointer     :: psegname(:), presname(:)
    integer,           pointer     :: presno(:)


    natom = molecule%num_atoms
    if (natom == 0) then
      call alloc_molecules(molecule, MoleculeMole, 0)
      return
    end if

    psegname => molecule%segment_name
    presname => molecule%residue_name
    presno   => molecule%residue_no

    ! allocate memory
    nresi = 1
    do i = 2, natom
      if (psegname(i-1) /= psegname(i) .or. &
          presno  (i-1) /= presno  (i)) &
          nresi = nresi + 1
    end do

    allocate(segname(nresi),   &
             resno  (nresi),   &
             resname(nresi),   &
             atomidx(2,nresi), &
             mol_tbl(3,natom), &
             stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    ! prepare residue lists
    nresi = 1
    segname(nresi)   = psegname(1)
    resno  (nresi)   = presno  (1)
    resname(nresi)   = presname(1)
    atomidx(1,nresi) = 1

    do i = 2, natom
      if (psegname(i-1) /= psegname(i) .or. &
          presno  (i-1) /= presno  (i)) then
        atomidx(2,nresi) = i - 1
        nresi = nresi + 1
        segname(nresi)   = psegname(i)
        resno  (nresi)   = presno  (i)
        resname(nresi)   = presname(i)
        atomidx(1,nresi) = i
      end if
    end do
    atomidx(2,nresi) = natom

    mid = 1
    mol_tbl(:,:) = 0

    ! prepare range residue
    do i = 1, mole_info%num_range
      snb  = mole_info%range_res(i)%seg_name_b
      sne  = mole_info%range_res(i)%seg_name_e
      rnob = mole_info%range_res(i)%res_no_b
      rnoe = mole_info%range_res(i)%res_no_e
      rnab = mole_info%range_res(i)%res_name_b
      rnae = mole_info%range_res(i)%res_name_e

      si1 = -1
      do j = 1, nresi

        if (('' == snb .or. segname(j) == snb) .and. &
             resno(j) == rnob .and. &
             resname(j) == rnab) then
          si1 = atomidx(1,j)
        end if

        if (si1 == -1) &
             cycle

        if (('' == sne .or. segname(j) == sne) .and. &
             resno(j) == rnoe .and. &
             resname(j) == rnae) then
          si2 = atomidx(2,j)

          mid = mid + 1
          do k = si1, si2
            mol_tbl(:,k) = (/1,i,mid/)
          end do

          si1 = -1
        end if
      end do
    end do

    ! prepare group residue
    do i = 1, mole_info%num_group
      ngr      = mole_info%group_res(i)%num_res
      gr(:ngr) = mole_info%group_res(i)%res_name(:ngr)

      j = 1
      do while(j <= nresi)
        do k = 1, ngr
          if (nresi < j+k-1) &
            exit
          if (resname(j+k-1) /= gr(k)) &
            exit
        end do
        if (k > ngr) then
          si1 = atomidx(1,j)
          si2 = atomidx(2,j+ngr-1)

          mid = mid + 1
          do k = si1, si2
            if (mol_tbl(TblType,k) == 0) &
              mol_tbl(:,k) = (/2,i,mid/)
          end do
          j = j + ngr
        else
          j = j + 1
        end if

      end do
    end do

    ! setup molecule number
    nmole = 1
    do i = 2, natom
      if (mol_tbl(TblMolID,i-1) /= mol_tbl(TblMolID,i)) &
        nmole = nmole + 1
    end do

    call alloc_molecules(molecule, MoleculeMole, nmole)
    
    nmole = 1
    ty = mol_tbl(TblType, 1)
    rn = mol_tbl(TblMolNo, 1)
    select case(ty)
    case (0)
      molecule%molecule_name(nmole) = '@UNDEFINED@'
    case (1)
      molecule%molecule_name(nmole) = mole_info%range_res(rn)%name
    case (2)
      molecule%molecule_name(nmole) = mole_info%group_res(rn)%name
    end select
    molecule%molecule_atom_no(nmole) = 1
    molecule%molecule_no(1) = nmole

    do i = 2, natom
      if (mol_tbl(TblMolID,i-1) /= mol_tbl(TblMolID,i)) then

        nmole = nmole + 1
        ty = mol_tbl(TblType, i)
        rn = mol_tbl(TblMolNo, i)
        select case(ty)
        case (0)
          molecule%molecule_name(nmole) = '@UNDEFINED@'
        case (1)
          molecule%molecule_name(nmole) = mole_info%range_res(rn)%name
        case (2)
          molecule%molecule_name(nmole) = mole_info%group_res(rn)%name
        end select
        molecule%molecule_atom_no(nmole) = i

      end if
      molecule%molecule_no(i) = nmole
    end do

    molecule%num_molecules = nmole

    ! deallocate memory
    deallocate(segname, resno, resname, atomidx, mol_tbl, &
               stat = alloc_stat)

    return

  end subroutine setup_mole_no


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_number
  !> @brief        make number list from line string
  !! @authors      NT, NM
  !! @date         2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_number(str, max_val, no_list)

    ! formal arguments
    character(*),              intent(in)    :: str
    integer,                   intent(in)    :: max_val
    integer,      pointer                    :: no_list(:)
    
    ! local variables
    logical                    :: range_on
    integer                    :: alloc_stat
    integer                    :: i, cnt, no, str_ptr
    integer                    :: nof, not, prv_no
    integer,      allocatable  :: no_list_tmp(:)  


    if (str == '') then
       allocate(no_list(max_val), stat = alloc_stat)
       if (alloc_stat /= 0) &
          call error_msg_alloc
       do i = 1, max_val
          no_list(i) = i
       end do
       return
    end if

    allocate(no_list_tmp(max_val), stat = alloc_stat)
    if (alloc_stat /= 0) &
       call error_msg_alloc

    range_on = .false.
    str_ptr = 1
    cnt = 0
    do while(get_no(str, str_ptr, no))
       if (no /= -1) then
          if (range_on) then
             nof = prv_no+1
             not = no
             range_on = .false.
          else
             nof = no
             not = no
             prv_no = no
          end if
       else
          range_on = .true.
          cycle
       end if
       if (nof < 1 .or. nof > not .or. nof > max_val .or. not > max_val) &
          call error_msg('Make_Number> out of range: '//TRIM(str))
       do i = nof, not
          cnt = cnt + 1
          no_list_tmp(cnt) = i
       end do
    end do

    allocate(no_list(cnt), stat = alloc_stat)
    if (alloc_stat /= 0) &
       call error_msg_alloc
    no_list(1:cnt) = no_list_tmp(1:cnt)

    deallocate(no_list_tmp, stat = alloc_stat)
    if (alloc_stat /= 0) &
       call error_msg_alloc

    return

  end subroutine make_number


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    get_number
  !> @brief        get next number from line string
  !! @authors      NT, NM
  !! @date         2012/05/29 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function get_no(str, str_ptr, no)

    ! return value
    logical :: get_no

    ! formal arguments
    character(*),     intent(in)    :: str
    integer,          intent(inout) :: str_ptr
    integer,          intent(out)   :: no

    ! local variables
    logical           :: isdigit
    character         :: c
    character(10)     :: dig_str
    integer           :: i, len, dig_cnt


    len = LEN_TRIM(str) + 1

    isdigit = .false.
    dig_str = ''

    do i = str_ptr, len
       c = str(i:i)

       if (c >= '0' .and. c <= '9') then

          if (isdigit) then
             dig_cnt = dig_cnt + 1
          else
             isdigit = .true.
             dig_cnt = 1
          end if
          dig_str(dig_cnt:dig_cnt) = c

       else if (c == '-') then

          if (isdigit) then
             read(dig_str,*,err=900,end=900) no
             str_ptr = i
          else
             no = -1
             str_ptr = i + 1
          end if
          get_no = .true.
          return

       else if (c == ' ') then

          if (isdigit) then
             read(dig_str,*,err=900,end=900) no
             str_ptr = i + 1
             get_no = .true.
             return
          end if

       else
          goto 900
       end if
    end do

    get_no = .false.
    return 

900 call error_msg('RemdReaderInit> read error: '//TRIM(str))

  end function get_no

end module remdcnv_setup_mod
