!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   select_mod
!> @brief   define selection modules
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM), Yuji Sugita (YS)
!! @date    2012/05/24
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module select_mod

  use select_contact_mod
  use select_parser_mod
  use molecules_str_mod
  use messages_mod

  implicit none
  private

  ! module variables
  double precision, private, save              :: g_around_scale = 1.d0

  ! subroutines
  public  :: select_atom
  public  :: reselect_atom
  private :: make_selatoms
  private :: make_bracket
  private :: make_and
  private :: make_or
  private :: make_not
  private :: make_around_atom
  private :: make_around_resi
  private :: make_around_mole
  private :: make_atom_name
  private :: make_atom_idx
  private :: make_atom_no
  private :: make_res_name
  private :: make_res_no
  private :: make_mol_name
  private :: make_seg_id
  private :: make_hydrogen
  private :: make_backbone
  private :: make_heavy
  private :: make_all
  private :: make_sel_hyd
  private :: calc_selatoms_union
  private :: calc_selatoms_isect
  private :: calc_selatoms_not
  private :: calc_selatoms_not_isect
  private :: quicksort
  private :: reselect_atom_
  private :: adjust_selection

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    select_atom
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[in]    expression : characters for atom selection
  !! @param[inout] selatoms   : structure of selatoms
  !! @date         2012/05/24
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine select_atom(molecule, expression, selatoms)

    ! formal arguments
    type(s_molecule),  intent(in)    :: molecule
    character(*),      intent(in)    :: expression
    type(s_selatoms),  intent(inout) :: selatoms

    ! local variables
    logical                  :: ret
    type(s_sp_node), pointer :: parse_tree

    ! selection
    !
    call parser_build_tree(expression, ret)

    if (.not. ret) &
      call error_msg('Select_atom> Syntax error. '//TRIM(parser_get_error()))

    parse_tree => parser_get_tree()

    g_around_scale = 1.d0

    if (associated(parse_tree)) then
      call make_selatoms(molecule, parse_tree)
      call alloc_selatoms(selatoms, SIZE(parse_tree%selatoms%idx))
      selatoms%idx(:) = parse_tree%selatoms%idx(:)
    else
      call alloc_selatoms(selatoms, 0)
    end if

    call parser_dealloc_tree

    return

  end subroutine select_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    reselect_atom
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule      structure of molecule information
  !! @param[in]    expression    characters for atom selection
  !! @param[in]    coord_new     new atom coordinates
  !! @param[in]    selatoms_org  structure of selatoms
  !! @param[inout] selatoms_new  structure of selatoms
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reselect_atom(molecule,     &
                           expression,   &
                           coord_new,    &
                           selatoms_org, &
                           selatoms_new)

    ! parameters
    double precision,  parameter     :: DecScale = 0.90d0
    double precision,  parameter     :: IncScale = 1.10d0

    ! formal arguments
    type(s_molecule),  intent(in)    :: molecule
    character(*),      intent(in)    :: expression
    double precision,  intent(in)    :: coord_new(:,:)
    type(s_selatoms),  intent(in)    :: selatoms_org
    type(s_selatoms),  intent(inout) :: selatoms_new

    ! local variables
    type(s_sp_node),   pointer       :: parse_tree

    logical            :: ret
    character(1000)    :: expr0
    integer            :: i, norg, nnew, nnew0
    type(s_molecule)   :: molecule_new
    type(s_selatoms)   :: selatoms_new0


    ! check whether re-selection is necessary.
    !
    expr0 = expression
    do i = 1, LEN(expr0)
      if (expr0(i:i) >= 'A' .and. expr0(i:i) <= 'Z') then
        expr0(i:i) = CHAR(ICHAR(expr0(i:i)) + 32)
      end if
    end do
    if (INDEX(expr0, "around") == 0) then
      call alloc_selatoms(selatoms_new, SIZE(selatoms_org%idx))
      selatoms_new%idx(:) = selatoms_org%idx(:)
      return
    end if

    norg = SIZE(selatoms_org%idx)
    write(MsgOut,'(A,I7)') &
         'Reselect_Atom> number of original seleciton : ', norg

    ! define molecule by new coordinates
    !
    call alloc_molecules(molecule_new, MoleculeAtom, molecule%num_atoms)
    call alloc_molecules(molecule_new, MoleculeMole, molecule%num_molecules)

    molecule_new%num_atoms       = molecule%num_atoms
    molecule_new%atom_no(:)      = molecule%atom_no(:)
    molecule_new%segment_name(:) = molecule%segment_name(:)
    molecule_new%segment_no(:)   = molecule%segment_no(:)
    molecule_new%residue_no(:)   = molecule%residue_no(:)
    molecule_new%residue_name(:) = molecule%residue_name(:)
    molecule_new%atom_name(:)    = molecule%atom_name(:)
    molecule_new%molecule_no(:)  = molecule%molecule_no(:)
    molecule_new%chain_id(:)     = molecule%chain_id(:)
    molecule_new%atom_coord(:,:) = coord_new(:,:)
    molecule_new%num_molecules       = molecule%num_molecules
    molecule_new%molecule_atom_no(:) = molecule%molecule_atom_no(:)
    molecule_new%molecule_mass(:)    = molecule%molecule_mass(:)
    molecule_new%molecule_name(:)    = molecule%molecule_name(:)
    
    ! selection
    !
    call parser_build_tree(expression, ret)

    if (.not. ret) &
      call error_msg('Reselect_Atom> Syntax error. '// &
                     TRIM(parser_get_error()))

    parse_tree => parser_get_tree()

    ! select (1)
    !
    g_around_scale = 1.d0

    write(MsgOut,'(A)') 'Reselect_Atom> normal selection: '

    call reselect_atom_(parse_tree, molecule_new, selatoms_new)

    nnew = SIZE(selatoms_new%idx)

    write(MsgOut,'(A,I7)') &
         'Reselect_Atom> number of new selection : ', nnew
    write(MsgOut,*) ''

    ! select (2)
    !
    if (norg < nnew) then

      do while (.true.)

        g_around_scale = g_around_scale * DecScale

        write(MsgOut,'(A,F6.2)') &
             'Reselect_Atom> around scale:', g_around_scale

        call reselect_atom_(parse_tree, molecule_new, selatoms_new0)

        nnew0 = SIZE(selatoms_new0%idx)

        write(MsgOut,'(A,I7)') &
             'Reselect_Atom> number of new selection : ', nnew0
        write(MsgOut,*) ''

        if (norg == nnew0) then

          call alloc_selatoms(selatoms_new, nnew0)
          selatoms_new%idx(:) = selatoms_new0%idx(:)
          exit

        else if (norg > nnew0) then

          call adjust_selection(norg, &
                                molecule_new, &
                                selatoms_new0, &
                                selatoms_new)
          exit

        end if
      end do

    else if (norg > nnew) then

      do while (.true.)

        g_around_scale = g_around_scale * IncScale

        write(MsgOut,'(A,F6.2)') &
             'Reselect_Atom> around scale:', g_around_scale

        call reselect_atom_(parse_tree, molecule_new, selatoms_new0)

        nnew0 = SIZE(selatoms_new0%idx)

        write(MsgOut,'(A,I7)') &
             'Reselect_Atom> number of new selection : ', nnew0
        write(MsgOut,*) ''

        if (norg == nnew0) then

          call alloc_selatoms(selatoms_new, nnew0)
          selatoms_new%idx(:) = selatoms_new0%idx(:)
          exit

        else if (norg < nnew0) then

          call adjust_selection(norg, &
                                molecule_new, &
                                selatoms_new, &
                                selatoms_new0)
          call alloc_selatoms(selatoms_new, SIZE(selatoms_new0%idx))
          selatoms_new%idx(:) = selatoms_new0%idx(:)
          exit

        end if
      end do

    end if

    ! deallocate
    !
    call parser_dealloc_tree
    call dealloc_selatoms(selatoms_new0)
    call dealloc_molecules(molecule_new, MoleculeAtom)
    call dealloc_molecules(molecule_new, MoleculeMole)

    write(MsgOut,'(a,i7)') 'Reselect_Atom> number of new selection '//&
         '(final): ', SIZE(selatoms_new%idx)
    write(MsgOut,*) ''

    return

  end subroutine reselect_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_selatoms
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_selatoms(molecule, sp_node)
    
    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    select case(sp_node%type)

    case(NT_Top)

    case(NT_Bracket)
      call make_bracket(molecule, sp_node)

    case(NT_AND)
      call make_and(molecule, sp_node)

    case(NT_OR)
      call make_or(molecule, sp_node)

    case(NT_NOT)
      call make_not(molecule, sp_node)

    case (NT_AroundAtom)
       call make_around_atom(molecule, sp_node)

    case (NT_AroundRes)
       call make_around_resi(molecule, sp_node)

    case (NT_AroundMol)
       call make_around_mole(molecule, sp_node)

    case(NT_AtomName)
      call make_atom_name(molecule, sp_node)

    case(NT_AtomIdx)
      call make_atom_idx(molecule, sp_node)

    case(NT_AtomNo)
      call make_atom_no(molecule, sp_node)

    case(NT_ResName)
      call make_res_name(molecule, sp_node)

    case(NT_ResNo)
      call make_res_no(molecule, sp_node)

    case (NT_MolName)
      call make_mol_name(molecule, sp_node)

    case(NT_SegId)
      call make_seg_id(molecule, sp_node)

    case(NT_Hydrogen)
      call make_hydrogen(molecule, sp_node)

    case(NT_Backbone)
      call make_backbone(molecule, sp_node)

    case(NT_Heavy)
      call make_heavy(molecule, sp_node)

    case(NT_All)
      call make_all(molecule, sp_node)

    end select

    return

  end subroutine make_selatoms

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_bracket
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_bracket(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    integer,         pointer :: idx(:)

    call make_selatoms(molecule, sp_node%child_node1)

    idx => sp_node%child_node1%selatoms%idx

    call alloc_selatoms(sp_node%selatoms, SIZE(idx))
    sp_node%selatoms%idx(:) = idx(:)


    call dealloc_selatoms(sp_node%child_node1%selatoms)

    return

  end subroutine make_bracket

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_and
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_and(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    call make_selatoms(molecule, sp_node%child_node1)
    call make_selatoms(molecule, sp_node%child_node2)

    call calc_selatoms_isect(sp_node%child_node1%selatoms, &
                              sp_node%child_node2%selatoms, &
                              sp_node%selatoms)

    call dealloc_selatoms(sp_node%child_node1%selatoms)
    call dealloc_selatoms(sp_node%child_node2%selatoms)

    return

  end subroutine make_and

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_or
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_or(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    call make_selatoms(molecule, sp_node%child_node1)
    call make_selatoms(molecule, sp_node%child_node2)

    call calc_selatoms_union(sp_node%child_node1%selatoms, &
                              sp_node%child_node2%selatoms, &
                              sp_node%selatoms)

    call dealloc_selatoms(sp_node%child_node1%selatoms)
    call dealloc_selatoms(sp_node%child_node2%selatoms)

    return

  end subroutine make_or

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_not
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_not(molecule, sp_node)
    
    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node


    call make_selatoms(molecule, sp_node%child_node1)

    call calc_selatoms_not(molecule%num_atoms, &
                           sp_node%child_node1%selatoms, &
                           sp_node%selatoms)

    call dealloc_selatoms(sp_node%child_node1%selatoms)

    return

  end subroutine make_not

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_around_atom
  !> @brief        return list of atom indices
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_around_atom(molecule, sp_node)
    
    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    double precision          :: distance
    type(s_selatoms)          :: selatoms_not


    read(sp_node%param_str,*,err=900,end=900) distance

    call make_selatoms(molecule, sp_node%child_node1)

    call calc_selatoms_not(molecule%num_atoms, &
                           sp_node%child_node1%selatoms, &
                           selatoms_not)

    if (SIZE(sp_node%child_node1%selatoms%idx) /= 0 .and. &
        SIZE(selatoms_not%idx) /= 0) then

      call select_contact(SelectModeAtom, &
                          molecule, &
                          molecule, &
                          sp_node%child_node1%selatoms, &
                          selatoms_not, &
                          distance * g_around_scale, &
                          sp_node%selatoms)

    else

      call alloc_selatoms(sp_node%selatoms, 0)

    end if

    call dealloc_selatoms(selatoms_not)
    call dealloc_selatoms(sp_node%child_node1%selatoms)

    return

900 call error_msg('Make_Around_Atom> unexpected error.')

    return

  end subroutine make_around_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_around_resi
  !> @brief        return list of atom indices
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_around_resi(molecule, sp_node)
    
    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    double precision          :: distance
    type(s_selatoms)          :: selatoms_not


    read(sp_node%param_str,*,err=900,end=900) distance

    call make_selatoms(molecule, sp_node%child_node1)

    call calc_selatoms_not(molecule%num_atoms, &
                           sp_node%child_node1%selatoms, &
                           selatoms_not)

    if (SIZE(sp_node%child_node1%selatoms%idx) /= 0 .and. &
        SIZE(selatoms_not%idx) /= 0) then

      call select_contact(SelectModeResi, &
                          molecule, &
                          molecule, &
                          sp_node%child_node1%selatoms, &
                          selatoms_not, &
                          distance * g_around_scale, &
                          sp_node%selatoms)

    else

      call alloc_selatoms(sp_node%selatoms, 0)

    end if

    call dealloc_selatoms(selatoms_not)
    call dealloc_selatoms(sp_node%child_node1%selatoms)

    return

900 call error_msg('Make_Around_Resi> unexpected error.')

  end subroutine make_around_resi

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_around_mole
  !> @brief        return list of atom indices
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_around_mole(molecule, sp_node)
    
    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    double precision          :: distance
    type(s_selatoms)          :: selatoms_not


    read(sp_node%param_str,*,err=900,end=900) distance

    call make_selatoms(molecule, sp_node%child_node1)

    call calc_selatoms_not(molecule%num_atoms, &
                           sp_node%child_node1%selatoms, &
                           selatoms_not)

    if (SIZE(sp_node%child_node1%selatoms%idx) /= 0 .and. &
        SIZE(selatoms_not%idx) /= 0) then

      call select_contact(SelectModeMole, &
                          molecule, &
                          molecule, &
                          sp_node%child_node1%selatoms, &
                          selatoms_not, &
                          distance * g_around_scale, &
                          sp_node%selatoms)

    else

      call alloc_selatoms(sp_node%selatoms, 0)

    end if

    call dealloc_selatoms(selatoms_not)
    call dealloc_selatoms(sp_node%child_node1%selatoms)

    return

900 call error_msg('Make_Around_Mole> unexpected error.')

  end subroutine make_around_mole

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_atom_name
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_atom_name(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    character(4)              :: an
    integer, allocatable      :: idx(:)
    integer                   :: i, num_atm, num_sel, alloc_stat, dealloc_stat
    
    
    num_atm = molecule%num_atoms

    num_sel = 0
    allocate(idx(num_atm), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    do i = 1, num_atm
      read(molecule%atom_name(i), *, err=900) an
      if (an == sp_node%param_str) then
        num_sel = num_sel + 1
        idx(num_sel) = i
      end if
    end do

    call alloc_selatoms(sp_node%selatoms, num_sel)
    sp_node%selatoms%idx(:) = idx(1:num_sel)

    deallocate(idx, stat = dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

    return

900 call error_msg('Make_atom_name> read error.')

  end subroutine make_atom_name

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_atom_idx
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_atom_idx(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    logical          :: fromto
    character(100)   :: str
    integer          :: i, natm, nsel, bidx, eidx


    str = sp_node%param_str

    natm = molecule%num_atoms

    fromto = .false.
    do i = 1, LEN(str)
      if (str(i:i) == '-') then
        str(i:i) = ' '
        fromto = .true.
      end if
    end do

    eidx = natm
    read(str,*,err=100,end=100) bidx, eidx

100 if (.not. fromto .or. bidx > eidx) &
      eidx = bidx

    nsel = MIN(natm, eidx) - MAX(1, bidx) + 1
    if (nsel < 0) &
      nsel = 0

    call alloc_selatoms(sp_node%selatoms, nsel)

    nsel = 0
    do i = bidx, eidx
      if (i > 0 .and. i <= natm) then
        nsel = nsel + 1
        sp_node%selatoms%idx(nsel) = i
      end if
    end do

    return

  end subroutine make_atom_idx

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_atom_no
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_atom_no(molecule, sp_node)
    
    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    logical                   :: fromto
    character(100)            :: str
    integer                   :: i, alloc_stat, dealloc_stat
    integer                   :: natm, nsel, bno, eno, atmno
    integer, allocatable      :: idx(:)


    str = sp_node%param_str

    natm = molecule%num_atoms

    fromto = .false.
    do i = 1, LEN(str)
      if (str(i:i) == '-') then
        str(i:i) = ' '
        fromto = .true.
      end if
    end do

    eno = -1
    read(str,*,err=100,end=100) bno, eno

100 if (eno == -1) then
      if (fromto) then
        do i = 1, natm
          eno = MAX(molecule%atom_no(i), eno)
        end do
      else
        eno = bno
      end if
    end if

    allocate(idx(natm), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    if (eno < bno) &
      eno = bno

    nsel = 0
    do i = 1, natm
      atmno = molecule%atom_no(i)
      if (bno <= atmno .and. atmno <= eno) then
        nsel = nsel + 1
        idx(nsel) = i
      end if
    end do

    call alloc_selatoms(sp_node%selatoms, nsel)
    sp_node%selatoms%idx(:) = idx(1:nsel)

    if (dealloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine make_atom_no

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_res_name
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_res_name(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    character(10)             :: resnam
    integer                   :: i, natm, nsel, alloc_stat, dealloc_stat
    integer, allocatable      :: idx(:)


    resnam = sp_node%param_str

    natm = molecule%num_atoms

    allocate(idx(natm), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    nsel = 0
    do i = 1, natm
      if (resnam == molecule%residue_name(i)) then
        nsel = nsel + 1
        idx(nsel) = i
      end if
    end do

    call alloc_selatoms(sp_node%selatoms, nsel)
    sp_node%selatoms%idx(:) = idx(1:nsel)

    deallocate(idx, stat = dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine make_res_name

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_res_no
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_res_no(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    logical                   :: fromto
    character(100)            :: str
    integer                   :: i, alloc_stat, dealloc_stat
    integer                   :: natm, nsel, bno, eno, resno
    integer, allocatable      :: idx(:)
    
    
    str = sp_node%param_str

    natm = molecule%num_atoms

    fromto = .false.
    do i = 1, LEN(str)
      if (str(i:i) == '-') then
        str(i:i) = ' '
        fromto = .true.
      end if
    end do

    eno = -1
    read(str,*,err=100,end=100) bno, eno

100 if (eno == -1) then
      if (fromto) then
        do i = 1, natm
          eno = MAX(molecule%residue_no(i), eno)
        end do
      else
        eno = bno
      end if
    end if

    allocate(idx(natm), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    if (eno < bno) &
      eno = bno

    nsel = 0
    do i = 1, natm
      resno = molecule%residue_no(i)
      if (bno <= resno .and. resno <= eno) then
        nsel = nsel + 1
        idx(nsel) = i
      end if
    end do

    call alloc_selatoms(sp_node%selatoms, nsel)
    sp_node%selatoms%idx(:) = idx(1:nsel)

    deallocate(idx, stat = dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine make_res_no

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_mol_name
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_mol_name(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    character(20)    :: molname
    integer          :: alloc_stat
    integer          :: i, j, natm, nsel, nmol, first, last

    integer, allocatable :: idx(:)


    molname = sp_node%param_str

    natm = molecule%num_atoms

    allocate(idx(natm), stat = alloc_stat)
    if (alloc_stat /= 0) &
       call error_msg_alloc

    nmol = SIZE(molecule%molecule_name)

    nsel = 0
    do i = 1, nmol
      if (molecule%molecule_name(i) /= molname) &
        cycle

      first = molecule%molecule_atom_no(i)
      if (i /= nmol) then
        last = molecule%molecule_atom_no(i+1) - 1
      else
        last = natm
      end if

      do j = first, last
        nsel = nsel + 1
        idx(nsel) = j
      end do
    end do

    call alloc_selatoms(sp_node%selatoms, nsel)
    sp_node%selatoms%idx(:) = idx(1:nsel)

    deallocate(idx, stat = alloc_stat)
    if (alloc_stat /= 0) &
       call error_msg_dealloc

    return

  end subroutine make_mol_name

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_res_no
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_seg_id(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    character(10)             :: segid
    integer                   :: i, natm, nsel, alloc_stat, dealloc_stat
    integer, allocatable      :: idx(:)
    
    
    segid = sp_node%param_str

    natm = molecule%num_atoms

    allocate(idx(natm), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    nsel = 0
    do i = 1, natm
      if (segid == ADJUSTL(molecule%segment_name(i))) then
        nsel = nsel + 1
        idx(nsel) = i
      end if
    end do

    call alloc_selatoms(sp_node%selatoms, nsel)
    sp_node%selatoms%idx(:) = idx(1:nsel)

    deallocate(idx, stat = dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine make_seg_id

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_hydrogen
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_hydrogen(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node


    call make_sel_hyd(.true., molecule, sp_node)

    return

  end subroutine make_hydrogen

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_backbone
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_backbone(molecule, sp_node)
    
    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    character(4), pointer     :: atmnam(:)
    integer                   :: i, natm, nsel, alloc_stat, dealloc_stat
    integer, allocatable      :: idx(:)


    atmnam => molecule%atom_name

    natm = molecule%num_atoms

    allocate(idx(natm), stat = alloc_stat)
    if (alloc_stat /= 0) &
       call error_msg_alloc

    nsel = 0
    do i = 1, natm
      if (atmnam(i)(1:4) == 'N   ' .or. &
          atmnam(i)(1:4) == 'CA  ' .or. &
          atmnam(i)(1:4) == 'C   ' .or. &
          atmnam(i)(1:4) == 'O   ') then

        nsel = nsel + 1
        idx(nsel) = i
      end if
    end do

    call alloc_selatoms(sp_node%selatoms, nsel)
    sp_node%selatoms%idx(:) = idx(1:nsel)

    deallocate(idx, stat = dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine make_backbone

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_heavy
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_heavy(molecule, sp_node)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node


    call make_sel_hyd(.false., molecule, sp_node)

    return

  end subroutine make_heavy

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_all
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_all(molecule, sp_node)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    integer          :: i


    call alloc_selatoms(sp_node%selatoms, molecule%num_atoms)

    do i = 1, molecule%num_atoms
      sp_node%selatoms%idx(i) = i
    end do

    return

  end subroutine make_all

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_sel_hyd
  !> @brief        return list of atom indexes
  !! @authors      NT
  !! @param[in]    molecule   : structure of molecule information
  !! @param[inout] sp_node    : structure of parse tree node
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine make_sel_hyd(sel_hyd, molecule, sp_node)

    ! formal arguments
    logical,                  intent(in)    :: sel_hyd
    type(s_molecule), target, intent(in)    :: molecule
    type(s_sp_node), pointer                :: sp_node

    ! local variables
    character(4), pointer     :: atmnam(:)
    integer                   :: i, natm, nsel, alloc_stat, dealloc_stat
    integer, allocatable      :: idx(:)


    atmnam => molecule%atom_name

    natm = molecule%num_atoms

    allocate(idx(natm), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    nsel = 0
    do i = 1, natm
      if ((atmnam(i)(1:1) == 'H' .and. sel_hyd .eqv. .true.) .or. &
          (atmnam(i)(1:1) /= 'H' .and. sel_hyd .eqv. .false.)) then
        nsel = nsel + 1
        idx(nsel) = i
      end if
    end do

    call alloc_selatoms(sp_node%selatoms, nsel)
    sp_node%selatoms%idx(:) = idx(1:nsel)

    deallocate(idx, stat = dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine make_sel_hyd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    calc_selatoms_union
  !> @brief        operate : c = a or b
  !! @authors      NT
  !! @param[in]    selatoms_a   : list a of selected atom index
  !! @param[in]    selatoms_b   : list b of selected atom index
  !! @param[inout] selatoms_c   : list c of selected atom index
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine calc_selatoms_union(selatoms_a, selatoms_b, selatoms_c)

    ! formal arguments
    type(s_selatoms),      intent(in)    :: selatoms_a
    type(s_selatoms),      intent(in)    :: selatoms_b
    type(s_selatoms),      intent(inout) :: selatoms_c

    ! local variables
    integer, allocatable   :: aidx(:), bidx(:), cidx(:)
    integer                :: alloc_stat, numc, i, ia, ib, v
    integer                :: numa, numb
    integer                :: dealloc_stat

    numa = SIZE(selatoms_a%idx)
    numb = SIZE(selatoms_b%idx)
    numc = numa + numb

    allocate(aidx(numa), bidx(numb), cidx(numc), stat = alloc_stat)
    if (alloc_stat /= 0) call error_msg_alloc
    aidx(1:numa) = selatoms_a%idx(1:numa)
    bidx(1:numb) = selatoms_b%idx(1:numb)

    if (numa > 1) &
      call quicksort(aidx, 1, numa)
    if (numb > 1) &
      call quicksort(bidx, 1, numb)

    ia = 1
    ib = 1
    numc = 0
    do while (ia <= numa .and. ib <= numb)
      if (aidx(ia) < bidx(ib)) then
        v = aidx(ia)
        ia = ia + 1
      else if(aidx(ia) > bidx(ib)) then
        v = bidx(ib)
        ib = ib + 1
      else
        v = aidx(ia)
        ia = ia + 1
        ib = ib + 1
      end if
      numc = numc + 1
      cidx(numc) = v
    end do
    do i = ia, numa
      numc = numc + 1
      cidx(numc) = aidx(i)
    end do
    do i = ib, numb
      numc = numc + 1
      cidx(numc) = bidx(i)
    end do

    call alloc_selatoms(selatoms_c, numc)
    selatoms_c%idx(1:numc) = cidx(1:numc)

    deallocate(aidx, bidx, cidx, stat = dealloc_stat)
    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine calc_selatoms_union

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    calc_selatoms_isect
  !> @brief        operate : c = a and b
  !! @authors      NT
  !! @param[in]    selatoms_a   : list a of selected atom index
  !! @param[in]    selatoms_b   : list b of selected atom index
  !! @param[inout] selatoms_c   : list c of selected atom index
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine calc_selatoms_isect(selatoms_a, selatoms_b, selatoms_c)

    ! formal arguments
    type(s_selatoms),      intent(in)    :: selatoms_a
    type(s_selatoms),      intent(in)    :: selatoms_b
    type(s_selatoms),      intent(inout) :: selatoms_c

    ! local variables
    integer, allocatable   :: aidx(:), bidx(:), cidx(:)
    integer                :: alloc_stat, dealloc_stat, numc, ia, ib
    integer                :: numa, numb

    numa = SIZE(selatoms_a%idx)
    numb = SIZE(selatoms_b%idx)
    numc = MIN(numa, numb)

    allocate(aidx(numa), bidx(numb), cidx(numc), stat = alloc_stat)

    if (alloc_stat /= 0)   call error_msg_alloc
    aidx(1:numa) = selatoms_a%idx(1:numa)
    bidx(1:numb) = selatoms_b%idx(1:numb)

    if (numa > 1) &
      call quicksort(aidx, 1, numa)
    if (numb > 1) &
      call quicksort(bidx, 1, numb)

    ia = 1
    ib = 1
    numc = 0
    do while (ia <= numa .and. ib <= numb)
      if (aidx(ia) < bidx(ib)) then
        ia = ia + 1
      else if(aidx(ia) > bidx(ib)) then
        ib = ib + 1
      else
        numc = numc + 1
        cidx(numc) = aidx(ia)
        ia = ia + 1
        ib = ib + 1
      end if
    end do

    call alloc_selatoms(selatoms_c, numc)

    selatoms_c%idx(1:numc) = cidx(1:numc)
    deallocate(aidx, bidx, cidx, stat = dealloc_stat)

    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine calc_selatoms_isect

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    calc_selatoms_not
  !> @brief        operate : b = not a
  !! @authors      NT
  !! @param[in]    num_atoms     : number of atoms
  !! @param[in]    selatoms_a   : list a of selected atom index
  !! @param[inout] selatoms_b   : list b of selected atom index
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine calc_selatoms_not(num_atoms, selatoms_a, selatoms_b)

    ! formal arguments
    integer,                   intent(in)    :: num_atoms
    type(s_selatoms),          intent(in)    :: selatoms_a
    type(s_selatoms),  target, intent(inout) :: selatoms_b

    ! local variables
    integer, allocatable       :: aidx(:)
    integer, pointer           :: bidx(:)
    integer                    :: i, alloc_stat, dealloc_stat
    integer                    :: numb, iatm, natm
    integer                    :: numa

    numa = SIZE(selatoms_a%idx)

    if (numa /= 0) then
      do i = 1, numa
        natm = MAX(selatoms_a%idx(i), num_atoms)
      end do
    else
      natm = num_atoms
    end if

    numb = natm - numa

    allocate(aidx(numa), stat = alloc_stat)
    if (alloc_stat /= 0)   call error_msg_alloc
    aidx(1:numa) = selatoms_a%idx(1:numa)

    if (numa > 1) &
      call quicksort(aidx, 1, numa)

    call alloc_selatoms(selatoms_b, numb)
    bidx => selatoms_b%idx

    iatm = 1
    numb = 0
    do i = 1, natm
      if (iatm <= numa) then
        if (i == aidx(iatm)) then
          iatm = iatm + 1
        else
          numb = numb + 1
          bidx(numb) = i
        end if
      else
        numb = numb + 1
        bidx(numb) = i
      end if
    end do

    deallocate(aidx, stat = dealloc_stat)
    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine calc_selatoms_not

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    calc_selatoms_not_isect
  !> @brief        operate : c = a not b
  !! @authors      NT
  !! @param[in]    selatoms_a   : list a of selected atom index
  !! @param[in]    selatoms_b   : list b of selected atom index
  !! @param[inout] selatoms_c   : list c of selected atom index
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine calc_selatoms_not_isect(selatoms_a, selatoms_b, selatoms_c)

    ! formal arguments
    type(s_selatoms),      intent(in)    :: selatoms_a
    type(s_selatoms),      intent(in)    :: selatoms_b
    type(s_selatoms),      intent(inout) :: selatoms_c

    ! local variables
    integer, allocatable   :: aidx(:), bidx(:), cidx(:)
    integer                :: alloc_stat, dealloc_stat, numc, i, ia, ib
    integer                :: numa, numb

    numa = SIZE(selatoms_a%idx)
    numb = SIZE(selatoms_b%idx)
    numc = numa

    allocate(aidx(numa), bidx(numb), cidx(numc), stat = alloc_stat)
    if (alloc_stat /= 0)   call error_msg_alloc
    aidx(1:numa) = selatoms_a%idx(1:numa)
    bidx(1:numb) = selatoms_b%idx(1:numb)

    if (numa > 1) &
      call quicksort(aidx, 1, numa)
    if (numb > 1) &
      call quicksort(bidx, 1, numb)

    ia = 1
    ib = 1
    numc = 0
    do while (ia <= numa .and. ib <= numb)
      if (aidx(ia) < bidx(ib)) then
        numc = numc + 1
        cidx(numc) = aidx(ia)
        ia = ia + 1
      else if (aidx(ia) > bidx(ib)) then
        ib = ib + 1
      else
        ia = ia + 1
        ib = ib + 1
      end if
    end do
    do i = ia, numa
      numc = numc + 1
      cidx(numc) = aidx(i)
    end do

    call alloc_selatoms(selatoms_c, numc)
    selatoms_c%idx(1:numc) = cidx(1:numc)

    deallocate(aidx, bidx, cidx, stat = dealloc_stat)
    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine calc_selatoms_not_isect

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    quicksort
  !> @brief        sorting function
  !! @authors      NT
  !! @param[inout] a   : list of integers
  !! @param[in]    numa: number of array a
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine quicksort(a, start, end)

    ! formal arguments
    integer, intent(inout) ::  a(*)
    integer, intent(in)    :: start, end

    ! local variables
    integer                :: i, j, t, x

    x = a((start + end) / 2)
    i = start
    j = end
    do
      do while (a(i) < x)
        i = i + 1
      end do
      do while (x < a(j))
        j = j - 1
      end do
      if (i >= j) exit
      t = a(i)
      a(i) = a(j)
      a(j) = t
      i = i + 1
      j = j - 1
    end do

    if (start < i - 1) call quicksort(a, start, i - 1)
    if (j + 1 < end)  call quicksort(a, j + 1, end)

    return

  end subroutine quicksort

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    reselect_atom_
  !> @brief        reselect_atom_
  !! @authors      NT
  !! @param[in]    parse_tree
  !! @param[in]    molecule_new
  !! @param[inout] selatoms_new
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reselect_atom_(parse_tree,   &
                            molecule_new, &
                            selatoms_new)

    ! formal arguments
    type(s_sp_node),  pointer                :: parse_tree
    type(s_molecule),          intent(in)    :: molecule_new
    type(s_selatoms),          intent(inout) :: selatoms_new
    

    if (associated(parse_tree)) then
      call make_selatoms(molecule_new, parse_tree)
      call alloc_selatoms(selatoms_new, SIZE(parse_tree%selatoms%idx))
      selatoms_new%idx(:) = parse_tree%selatoms%idx(:)
    else
      call alloc_selatoms(selatoms_new, 0)
    end if

    return

  end subroutine reselect_atom_

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    adjust_selection
  !> @brief        adjust_selection
  !! @authors      NT
  !! @param[in]    target_count
  !! @param[in]    molecule
  !! @param[in]    sa_base
  !! @param[inout] sa_target
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine adjust_selection(target_count, &
                              molecule, &
                              sa_base, &
                              sa_target)

    ! parameter
    double precision,  parameter     :: AdjustDist = 10.d0

    ! formal arguments
    integer,           intent(in)    :: target_count
    type(s_molecule),  intent(in)    :: molecule
    type(s_selatoms),  intent(in)    :: sa_base
    type(s_selatoms),  intent(inout) :: sa_target

    ! local variables
    type(s_selatoms)  :: sa_adjust1, sa_adjust2



    ! calculate adjust = target - base
    !
    call calc_selatoms_not_isect(sa_target, sa_base, sa_adjust1)


    ! return value : sa_target
    !
    call alloc_selatoms(sa_target, SIZE(sa_base%idx))
    sa_target%idx(:) = sa_base%idx(:)


    ! by molecule
    !
    write(MsgOut,'(A)') 'Adjust_Selection> adjust selection by molecule '

    call select_contact(SelectModeMole, &
                        molecule,   &
                        molecule,   &
                        sa_base,    &
                        sa_adjust1, &
                        AdjustDist, &
                        sa_adjust2, &
                        target_count - SIZE(sa_target%idx), &
                        .true.)

    write(MsgOut,'(A,I7)') 'Adjust_Selection> adjusted : ', SIZE(sa_adjust2%idx)

    call calc_selatoms_union(sa_target, sa_adjust2, sa_target)

    if (target_count == SIZE(sa_target%idx)) &
      return
       
    call calc_selatoms_not_isect(sa_adjust1, sa_adjust2, sa_adjust1)


    ! by residue
    !
    write(MsgOut,'(A)') 'Adjust_Selection> adjust selection by residue : '

    call select_contact(SelectModeResi, &
                        molecule,   &
                        molecule,   &
                        sa_base,    &
                        sa_adjust1, &
                        AdjustDist, &
                        sa_adjust2, &
                        target_count - SIZE(sa_target%idx), &
                        .true.)

    write(MsgOut,'(A,I7)') 'Adjust_Selection> adjusted : ', SIZE(sa_adjust2%idx)

    call calc_selatoms_union(sa_target, sa_adjust2, sa_target)

    if (target_count == SIZE(sa_target%idx)) &
      return

    call calc_selatoms_not_isect(sa_adjust1, sa_adjust2, sa_adjust1)


    ! by atom
    !
    write(MsgOut,'(A)') 'Adjust_Selection> adjust selection by atom    : '

    call select_contact(SelectModeAtom, &
                        molecule,   &
                        molecule,   &
                        sa_base,    &
                        sa_adjust1, &
                        AdjustDist, &
                        sa_adjust2, &
                        target_count - SIZE(sa_target%idx), &
                        .true.)

    write(MsgOut,'(A,I7)') 'Adjust_Selection> adjusted : ', SIZE(sa_adjust2%idx)

    call calc_selatoms_union(sa_target, sa_adjust2, sa_target)

    if (target_count /= SIZE(sa_target%idx)) &
      call error_msg('Adjust_Selection> ERROR: cannot adjust selection.')

    return

  end subroutine adjust_selection

end module select_mod
