!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   select_contact_mod
!> @brief   select contact atom
!! @authors Norio Takase (NT)
!! @date    2012/05/24 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module select_contact_mod

  use molecules_str_mod
  use selatoms_str_mod
  use messages_mod

  implicit none
  private

  ! constant parametes
  integer,    public, parameter :: SelectModeAtom = 1
  integer,    public, parameter :: SelectModeResi = 2
  integer,    public, parameter :: SelectModeMole = 3

  integer,            parameter :: Empty = 0

  double precision,   parameter :: EPS   = 1.0d-10


  ! structures
  type s_cell_linked_list

     ! origin      : simulation box origin
     ! region      : simulation box region
     ! cell_count  : cell count for each axis
     ! cell_length : edge length of each cell
     !
     ! head        : head[c] holds the index of the first atom in the c-th cell
     ! lscl        : An array implementation of the linked lists

     double precision          :: origin(3)
     double precision          :: region(3)
     integer                   :: cell_count(3)
     double precision          :: cell_length(3)

     integer,          pointer :: head(:) => null()
     integer,          pointer :: lscl(:) => null()

     double precision, pointer :: coords(:,:)

  end type s_cell_linked_list

  type s_sdata_atom
     double precision          :: sq_dis
     integer                   :: idx
  end type s_sdata_atom

  type s_sdata_mole
     double precision          :: sq_dis
     integer                   :: idx1
     integer                   :: idx2
  end type s_sdata_mole

  ! subroutines
  public  :: select_contact
  private :: init_select
  private :: select_atom_ctct
  private :: select_mole_ctct
  private :: quicksort_atom
  private :: quicksort_mole
  private :: build_cell_linked_list
  private :: dealloc_cell_linked_list
  private :: search_neighbour_coord

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    select_contact
  !> @brief        select atoms in less than distance of atoms-A from atoms-B
  !! @authors      NT
  !! @param[in]    mode         selection mode
  !! @param[in]    molecule_a   array of atom coordinate A
  !! @param[in]    molecule_b   array of atom coordinate B
  !! @param[in]    selatoms_a   atom seleciton A
  !! @param[in]    selatoms_b   atom seleciton B
  !! @param[in]    distance     contact distance 
  !! @param[inout] selatoms     result atom selection
  !! @param[in]    o_max_atom   maximum atom count (optional)
  !! @param[in]    o_sort_list  sort list by distance (optional)
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine select_contact(mode,       &
                            molecule_a, &
                            molecule_b, &
                            selatoms_a, &
                            selatoms_b, &
                            distance,   &
                            selatoms,   &
                            o_max_atom, &
                            o_sort_list)

    ! formal arguments
    integer,                     intent(in)    :: mode
    type(s_molecule),            intent(in)    :: molecule_a
    type(s_molecule),            intent(in)    :: molecule_b
    type(s_selatoms),            intent(in)    :: selatoms_a
    type(s_selatoms),            intent(in)    :: selatoms_b
    double precision,            intent(in)    :: distance
    type(s_selatoms),            intent(inout) :: selatoms
    integer,           optional, intent(in)    :: o_max_atom
    logical,           optional, intent(in)    :: o_sort_list
                                 
    ! local variables
    type(s_cell_linked_list)     :: list_a, list_b
    integer                      :: max_atom, alloc_stat
    logical                      :: sort_list
    
    integer, pointer             :: resi_mole_no(:)      => null()
    integer, pointer             :: resi_mole_atom_no(:) => null()

    character(*), parameter      :: ModeStr(3) =   (/'Atom','Resi','Mole'/)

    
    if (present(o_max_atom)) then
      max_atom = o_max_atom
    else
      max_atom = 0
    end if

    if (present(o_sort_list)) then
      sort_list = o_sort_list
    else
      sort_list = .false.
    end if

    write(MsgOut, '(A,A4)') &
         'Select_Contact> mode        : ', ModeStr(mode)
    write(MsgOut, '(A,I7)') &
         '                sel atoms A : ', SIZE(selatoms_a%idx)
    write(MsgOut, '(A,I7)') &
         '                sel atoms B : ', SIZE(selatoms_b%idx)
    write(MsgOut, '(A,F7.2)') &
         '                distance    : ', distance


    ! initialize selection
    !
    call init_select(molecule_a, &
                     molecule_b, &
                     selatoms_a, &
                     selatoms_b, &
                     distance,   &
                     list_a,     &
                     list_b)

    select case(mode) 

    case (SelectModeAtom)

      call select_atom_ctct(distance,   &
                            max_atom,   &
                            sort_list,  &
                            list_a,     &
                            list_b,     &
                            selatoms_b, &
                            selatoms)

    case (SelectModeResi)

      call make_resi_mole_no(molecule_b,   &
                             resi_mole_no, &
                             resi_mole_atom_no)

      call select_mole_ctct(distance,   &
                            max_atom,   &
                            sort_list,  &
                            list_a,     &
                            list_b,     &
                            selatoms_b, &
                            resi_mole_no,      &
                            resi_mole_atom_no, &
                            selatoms)

      deallocate(resi_mole_no, &
                 resi_mole_atom_no, stat = alloc_stat)
      if (alloc_stat /= 0) &
        call error_msg_dealloc

    case (SelectModeMole)

      call select_mole_ctct(distance,   &
                            max_atom,   &
                            sort_list,  &
                            list_a,     &
                            list_b,     &
                            selatoms_b, &
                            molecule_b%molecule_no,      &
                            molecule_b%molecule_atom_no, &
                            selatoms)

    end select

    write(MsgOut, '(A,I7)') &
         '                contact atom list : ', SIZE(selatoms%idx)
    write(MsgOut, *) ''


    ! deallocate
    !
    deallocate(list_a%coords, list_b%coords, stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_dealloc

    call dealloc_cell_linked_list(list_a)
    call dealloc_cell_linked_list(list_b)

    return

  end subroutine select_contact

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    init_select
  !> @brief        initialize cell-linked list
  !! @authors      NT
  !! @param[in]    molecule_a   array of atom coordinate A
  !! @param[in]    molecule_b   array of atom coordinate B
  !! @param[in]    selatoms_a   atom seleciton A
  !! @param[in]    selatoms_b   atom seleciton B
  !! @param[in]    distance     contact distance 
  !! @param[inout] list_a       cell-linked list A
  !! @param[inout] list_b       cell-linked list B
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine init_select(molecule_a, &
                         molecule_b, &
                         selatoms_a, &
                         selatoms_b, &
                         distance,   &
                         list_a,     &
                         list_b)

    ! formal arguments
    type(s_molecule),         intent(in)    :: molecule_a
    type(s_molecule),         intent(in)    :: molecule_b
    type(s_selatoms),         intent(in)    :: selatoms_a
    type(s_selatoms),         intent(in)    :: selatoms_b
    double precision,         intent(in)    :: distance
    type(s_cell_linked_list), intent(inout) :: list_a
    type(s_cell_linked_list), intent(inout) :: list_b

    ! local variables
    integer                   :: i, alloc_stat
    integer                   :: num_a, num_b
    double precision          :: minbou(3), maxbou(3), mincl(3), region(3)

    integer                   :: cell_count(3)
    double precision          :: cell_length(3)

    double precision, pointer :: coord_a(:,:)
    double precision, pointer :: coord_b(:,:)


    ! setup coordinates
    !
    num_a = SIZE(selatoms_a%idx)
    num_b = SIZE(selatoms_b%idx)

    allocate(coord_a(3, num_a), &
             coord_b(3, num_b), &
             stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc


    minbou(1:3) =  1000000.0
    maxbou(1:3) = -1000000.0

    do i = 1, num_a
      coord_a(1:3, i) = molecule_a%atom_coord(1:3, selatoms_a%idx(i))
      minbou(1:3) = MIN(minbou(1:3), coord_a(1:3, i))
      maxbou(1:3) = MAX(maxbou(1:3), coord_a(1:3, i))
    end do

    do i = 1, num_b
      coord_b(1:3, i) = molecule_b%atom_coord(1:3, selatoms_b%idx(i))
      minbou(1:3) = MIN(minbou(1:3), coord_b(1:3, i))
      maxbou(1:3) = MAX(maxbou(1:3), coord_b(1:3, i))
    end do

    minbou(1:3) = minbou(1:3) - EPS
    maxbou(1:3) = maxbou(1:3) + EPS
    

    ! compute cell parameter
    !
    mincl(1:3) = distance
    region(1:3) = maxbou(1:3) - minbou(1:3)

    call compute_cell_by_region(region,     &
                                mincl,      &
                                cell_count, &
                                cell_length)


    ! build cell linked-list
    !
    call build_cell_linked_list(minbou(1:3), &
                                region,      &
                                cell_count,  &
                                coord_a,     &
                                list_a)

    call build_cell_linked_list(minbou(1:3), &
                                region,      &
                                cell_count,  &
                                coord_b,     &
                                list_b)

    return

  end subroutine init_select

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    select_atom_ctct
  !> @brief        select atoms in less than distance of atoms-A from atoms-B
  !! @authors      NT
  !! @param[in]    distance     contact distance 
  !! @param[in]    max_atom     maximum selection atom count
  !! @param[in]    sort_list    sort list
  !! @param[in]    list_a       cell-linked list A
  !! @param[in]    list_b       cell-linked list B
  !! @param[in]    selatoms_b   atom seleciton B
  !! @param[inout] selatoms     contact atom seleciton
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine select_atom_ctct(distance,   &
                              max_atom,   &
                              sort_list,  &
                              list_a,     &
                              list_b,     &
                              selatoms_b, &
                              selatoms)

    ! formal arguments
    double precision,         intent(in)    :: distance
    integer,                  intent(in)    :: max_atom
    logical,                  intent(in)    :: sort_list
    type(s_cell_linked_list), intent(in)    :: list_a
    type(s_cell_linked_list), intent(in)    :: list_b
    type(s_selatoms),         intent(in)    :: selatoms_b
    type(s_selatoms),         intent(inout) :: selatoms

    ! local variables
    integer                   :: ccx, ccy, ccz, ccyz
    integer                   :: cxa, cya, cza, cxb, cyb, czb, ca, cb
    integer                   :: i, natom, natom_b, nsel, a, b, alloc_stat
    double precision          :: sqr_d, dx, dy, dz, dxyz

    logical,            allocatable :: atom_flg(:)
    double precision,   allocatable :: atom_dis(:)
    type(s_sdata_atom), allocatable :: sort_array(:)

    double precision, pointer      :: crda(:,:), crdb(:,:)


    crda => list_a%coords
    crdb => list_b%coords
    sqr_d = distance * distance

    natom_b = SIZE(selatoms_b%idx)


    ! allocate
    !
    allocate(atom_flg(natom_b), &
             atom_dis(natom_b), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    atom_flg(1:natom_b) = .false.


    ! search
    !
    ccx  = list_a%cell_count(1)
    ccy  = list_a%cell_count(2)
    ccz  = list_a%cell_count(3)
    ccyz = ccy * ccz

    do cxa = 0, ccx - 1
    do cya = 0, ccy - 1
    do cza = 0, ccz - 1
      
      ca = cxa * ccyz + &
           cya * ccz +  &
           cza + 1

      do cxb = cxa - 1, cxa + 1
        if (cxb < 0 .or. cxb >= ccx) cycle
      do cyb = cya - 1, cya + 1
        if (cyb < 0 .or. cyb >= ccy) cycle
      do czb = cza - 1, cza + 1
        if (czb < 0 .or. czb >= ccz) cycle

        cb = cxb * ccyz + &
             cyb * ccz +  &
             czb + 1

        a = list_a%head(ca)

        do while(a /= Empty)

          b = list_b%head(cb)

          do while(b /= Empty)

            dx = crda(1, a) - crdb(1, b)
            dy = crda(2, a) - crdb(2, b)
            dz = crda(3, a) - crdb(3, b)
            dxyz = dx*dx + dy*dy + dz*dz
            if (sqr_d > dxyz) then
              atom_flg(b) = .true.
              atom_dis(b) = dxyz
            end if

            b = list_b%lscl(b)

          end do
          a = list_a%lscl(a)

        end do

      end do
      end do
      end do

    end do
    end do
    end do


    ! setup result
    !

    nsel = 0
    do i = 1, natom_b
      if (atom_flg(i)) &
        nsel = nsel + 1
    end do

    if (max_atom /= 0) then
      natom = MIN(max_atom, nsel)
    else
      natom = nsel
    end if

    call alloc_selatoms(selatoms, natom)

    if (nsel /= 0) then

      ! sort
      if (sort_list) then

        allocate(sort_array(nsel), stat = alloc_stat)
        if (alloc_stat /= 0) &
             call error_msg_alloc

        nsel = 0
        do i = 1, natom_b
          if (atom_flg(i)) then
            nsel = nsel + 1
            sort_array(nsel)%sq_dis = atom_dis(i)
            sort_array(nsel)%idx    = selatoms_b%idx(i)
          end if
        end do

        call quicksort_atom(sort_array, 1, nsel)

        do i = 1, natom
          selatoms%idx(i) = sort_array(i)%idx
        end do

        deallocate(sort_array, stat = alloc_stat)

        ! not sort
      else

        nsel = 0
        do i = 1, natom_b
          if (atom_flg(i)) then
            nsel = nsel + 1
            if (nsel > natom) &
              exit
            selatoms%idx(nsel) = selatoms_b%idx(i)
          end if
        end do

      end if

    end if


    ! dealloc
    !
    deallocate(atom_flg, atom_dis, stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine select_atom_ctct

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    select_mole_ctct
  !> @brief        select molecule in less than distance of atoms A from atoms B
  !! @authors      NT
  !! @param[in]    distance       contact distance 
  !! @param[in]    max_atom       maximum selection atom count
  !! @param[in]    sort_list      sort list
  !! @param[in]    list_a         cell-linked list A
  !! @param[in]    list_b         cell-linked list B
  !! @param[in]    selatoms_b     atom seleciton B
  !! @param[in]    mole_no_b      molecule number list
  !! @param[in]    mole_atom_no_b atom number list
  !! @param[inout] selatoms       contact atom seleciton
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine select_mole_ctct(distance,     &
                              max_atom,     &
                              sort_list,    &
                              list_a,       &
                              list_b,       &
                              selatoms_b,   &
                              mole_no_b,    &
                              mole_atom_no_b, &
                              selatoms)

    ! formal arguments
    double precision,         intent(in)    :: distance
    integer,                  intent(in)    :: max_atom
    logical,                  intent(in)    :: sort_list
    type(s_cell_linked_list), intent(in)    :: list_a
    type(s_cell_linked_list), intent(in)    :: list_b
    type(s_selatoms),         intent(in)    :: selatoms_b
    integer,                  intent(in)    :: mole_no_b(:)
    integer,                  intent(in)    :: mole_atom_no_b(:)
    type(s_selatoms),         intent(inout) :: selatoms

    ! local variables
    integer                   :: ccx, ccy, ccz, ccyz
    integer                   :: cxa, cya, cza, cxb, cyb, czb, ca, cb
    integer                   :: natom, natom0, nmole, nmole_b, midx, a, b
    integer                   :: i, j, alloc_stat
    integer                   :: first, last
    double precision          :: sqr_d, dx, dy, dz, dxyz

    logical,            allocatable :: mole_flg(:)
    double precision,   allocatable :: mole_dis(:)
    type(s_sdata_mole), allocatable :: sort_array(:)
    double precision,   pointer     :: crda(:,:), crdb(:,:)


    crda => list_a%coords
    crdb => list_b%coords
    sqr_d = distance * distance

    nmole_b = SIZE(mole_atom_no_b)


    ! check molecule atom count
    !
    if (nmole_b == 0) then
      call alloc_selatoms(selatoms, 0)
      return 
    end if

    ! allocate
    !
    allocate(mole_flg(nmole_b), &
             mole_dis(nmole_b), stat = alloc_stat)
    if (alloc_stat /= 0) &
       call error_msg_alloc


    mole_flg(1:nmole_b) = .false.
    mole_dis(1:nmole_b) = sqr_d

    ccx  = list_a%cell_count(1)
    ccy  = list_a%cell_count(2)
    ccz  = list_a%cell_count(3)
    ccyz = ccy * ccz

    do cxa = 0, ccx - 1
    do cya = 0, ccy - 1
    do cza = 0, ccz - 1

      ca = cxa * ccyz + &
           cya * ccz +  &
           cza + 1

      do cxb = cxa - 1, cxa + 1
        if (cxb < 0 .or. cxb >= ccx) cycle
      do cyb = cya - 1, cya + 1
        if (cyb < 0 .or. cyb >= ccy) cycle
      do czb = cza - 1, cza + 1
        if (czb < 0 .or. czb >= ccz) cycle

        cb = cxb * ccyz + &
             cyb * ccz +  &
             czb + 1

        a = list_a%head(ca)

        do while(a /= Empty)

          b = list_b%head(cb)

          do while(b /= Empty)

            dx = crda(1, a) - crdb(1, b)
            dy = crda(2, a) - crdb(2, b)
            dz = crda(3, a) - crdb(3, b)
            dxyz = dx*dx + dy*dy + dz*dz
            if (sqr_d > dxyz) then

              midx = mole_no_b(selatoms_b%idx(b))
              mole_flg(midx) = .true.
              mole_dis(midx) = MIN(mole_dis(midx), dxyz)

            end if

            b = list_b%lscl(b)

          end do
          a = list_a%lscl(a)

        end do

      end do
      end do
      end do

    end do
    end do
    end do


    ! setup result
    !

    ! sort
    if (sort_list) then

      nmole = 0
      do i = 1, nmole_b
        if (mole_flg(i)) &
          nmole = nmole + 1
      end do

      allocate(sort_array(nmole), stat = alloc_stat)
      if (alloc_stat /= 0) &
        call error_msg_alloc

      nmole = 0
      do i = 1, nmole_b
        if (mole_flg(i)) then

          nmole = nmole + 1

          first = mole_atom_no_b(i)
          if (i /= nmole_b) then
            last = mole_atom_no_b(i+1) - 1
          else
            last = SIZE(mole_no_b)
          end if

          sort_array(nmole)%sq_dis = mole_dis(i)
          sort_array(nmole)%idx1 = first
          sort_array(nmole)%idx2 = last

        end if

      end do

      if (nmole /= 0) &
        call quicksort_mole(sort_array, 1, nmole)
        
      natom = 0
      do i = 1, nmole
        natom0 = sort_array(i)%idx2 - sort_array(i)%idx1 + 1
        if (max_atom /= 0) then
          if (natom == max_atom) &
               exit
          if (natom + natom0 > max_atom) &
               cycle
        end if
        !write(MsgOut,*) i, '  len:', natom0, '  dis:', sort_array(i)%sq_dis
        natom = natom + natom0
      end do

      call alloc_selatoms(selatoms, natom)

      natom = 0

      do i = 1, nmole
        natom0 = sort_array(i)%idx2 - sort_array(i)%idx1 + 1
        if (max_atom /= 0) then
          if (natom == max_atom) &
               exit
          if (natom + natom0 > max_atom) &
               cycle
        end if

        do j = sort_array(i)%idx1, sort_array(i)%idx2
          natom = natom + 1
          selatoms%idx(natom) = j
        end do
      end do

      deallocate(sort_array, stat = alloc_stat)

    ! not sort
    else

      natom = 0

      do i = 1, nmole_b
        if (mole_flg(i)) then

          first = mole_atom_no_b(i)
          if (i /= nmole_b) then
            last = mole_atom_no_b(i+1) - 1
          else
            last = SIZE(mole_no_b)
          end if

          natom0 = last - first + 1
          if (max_atom /= 0 .and. natom + natom0 > max_atom) &
            exit
          natom = natom + natom0

        end if
      end do

      call alloc_selatoms(selatoms, natom)

      natom = 0

      do i = 1, nmole_b
        if (mole_flg(i)) then

          first = mole_atom_no_b(i)
          if (i /= nmole_b) then
            last = mole_atom_no_b(i+1) - 1
          else
            last = SIZE(mole_no_b)
          end if

          natom0 = last - first + 1
          if (max_atom /= 0 .and. natom + natom0 > max_atom) &
            exit
          do j = first, last
            natom = natom + 1
            selatoms%idx(natom) = j
          end do

        end if
      end do

    end if


    ! dealloc
    !
    deallocate(mole_dis, mole_flg, stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_dealloc

    return

  end subroutine select_mole_ctct

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_resi_mole_no
  !> @brief        make_resi_mole_no
  !! @authors      NT
  !! @param[in]    molecule          information of molecule
  !! @param[inout] resi_mole_no      array of residue molecule number
  !! @param[inout] resi_mole_atom_no array of residue molecule atom number
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_resi_mole_no(molecule,     &
                               resi_mole_no, &
                               resi_mole_atom_no)

    ! formal arguments
    type(s_molecule), target,  intent(in)    :: molecule
    integer,          pointer                :: resi_mole_no(:)
    integer,          pointer                :: resi_mole_atom_no(:)

    ! local variables
    integer                    :: i, natom, nresi
    integer                    :: alloc_stat

    character(4), pointer      :: psegname(:)
    integer,      pointer      :: presno(:)


    psegname => molecule%segment_name
    presno   => molecule%residue_no

    natom = molecule%num_atoms
    if (natom == 0) &
       return

    nresi = 1
    do i = 2, natom
      if (psegname(i-1) /= psegname(i) .or. &
          presno  (i-1) /= presno  (i)) &
        nresi = nresi + 1
    end do

    allocate(resi_mole_no(natom), &
             resi_mole_atom_no(nresi), &
             stat = alloc_stat)

    nresi = 1
    resi_mole_atom_no(nresi) = 1
    resi_mole_no(1) = nresi

    do i = 2, natom
      if (psegname(i-1) /= psegname(i) .or. &
          presno  (i-1) /= presno  (i)) then

        nresi = nresi + 1
        resi_mole_atom_no(nresi) = i
      end if
      resi_mole_no(i) = nresi
    end do

    return

  end subroutine make_resi_mole_no

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    quicksort_atom
  !> @brief        sorting function
  !! @authors      NT
  !! @param[inout] a       array
  !! @param[in]    start   start index
  !! @param[in]    end     end index
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine quicksort_atom(a, start, end)

    ! formal arguments
    type(s_sdata_atom), intent(inout) :: a(:)
    integer,            intent(in)    :: start
    integer,            intent(in)    :: end

    ! local variables
    type(s_sdata_atom)  :: t, x
    integer             :: i, j

    x = a((start + end) / 2)
    i = start
    j = end
    do
      do while (a(i)%sq_dis < x%sq_dis)
        i = i + 1
      end do
      do while (x%sq_dis < a(j)%sq_dis)
        j = j - 1
      end do
      if (i >= j) exit
      t = a(i)
      a(i) = a(j)
      a(j) = t
      i = i + 1
      j = j - 1
    end do

    if (start < i - 1) call quicksort_atom(a, start, i - 1)
    if (j + 1 < end)   call quicksort_atom(a, j + 1, end)

    return

  end subroutine quicksort_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    quicksort_mole
  !> @brief        sorting function
  !! @authors      NT
  !! @param[inout] a       array
  !! @param[in]    start   start index
  !! @param[in]    end     end index
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine quicksort_mole(a, start, end)

    ! formal arguments
    type(s_sdata_mole), intent(inout) :: a(:)
    integer,            intent(in)    :: start
    integer,            intent(in)    :: end

    ! local variables
    type(s_sdata_mole)  :: t, x
    integer             :: i, j

    x = a((start + end) / 2)
    i = start
    j = end
    do
      do while (a(i)%sq_dis < x%sq_dis)
        i = i + 1
      end do
      do while (x%sq_dis < a(j)%sq_dis)
        j = j - 1
      end do
      if (i >= j) exit
      t = a(i)
      a(i) = a(j)
      a(j) = t
      i = i + 1
      j = j - 1
    end do

    if (start < i - 1) call quicksort_mole(a, start, i - 1)
    if (j + 1 < end)   call quicksort_mole(a, j + 1, end)

    return

  end subroutine quicksort_mole

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    build_cell_linked_list
  !> @brief        build cell-linked list
  !! @authors      NT
  !! @param[in]    origin     simulation box origin
  !! @param[in]    region     simulation box region
  !! @param[in]    cell_count cell count for each axis
  !! @param[in]    coords     array of 3d-coordinate
  !! @param[out]   list       cell linked-list
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_cell_linked_list(origin,      &
                                    region,      &
                                    cell_count,  &
                                    coords,      &
                                    list)

    ! formal arguments
    double precision,         intent(in)          :: origin(3)
    double precision,         intent(in)          :: region(3)
    integer,                  intent(in)          :: cell_count(3)
    double precision,         intent(in), target  :: coords(:,:)
    type(s_cell_linked_list), intent(out)         :: list

    ! local variables
    integer                   :: total_cell, total_cell_yz, total_coord
    integer                   :: i, alloc_stat, dealloc_stat
    integer                   :: ci, cix, ciy, ciz


    ! check parameter
    !
    if (region(1) <= 0.0 .or. &
        region(2) <= 0.0 .or. &
        region(3) <= 0.0 .or. &
        cell_count(1) <= 0 .or. &
        cell_count(2) <= 0 .or. &
        cell_count(3) <= 0) then
      call error_msg('build_cell_linked_list> ERROR : cell linked-list'//&
                     ' parameter.')
    end if
       
    total_cell_yz = cell_count(2) * cell_count(3)
    total_cell    = cell_count(1) * total_cell_yz
    total_coord   = SIZE(coords(1,:))

    if (total_coord == 0) then
      call error_msg('build_cell_linked_list> ERROR : There are no'//&
                     ' coordinates.')
    end if

    list%origin(1:3)      = origin(1:3)
    list%region(1:3)      = region(1:3)
    list%cell_count(1:3)  = cell_count(1:3)
    list%cell_length(1:3) = region(1:3) / DBLE(cell_count(1:3))

    list%coords => coords

    !write(MsgOut, *) ' '
    !write(MsgOut, fmt='(" build_cell_linked_list> origin      : '//&
    !                  '(",3f8.3,")")'), list%origin(1:3)
    !write(MsgOut, fmt='("                         region      : '//&
    !                  '(",3f8.3,")")'), list%region(1:3)
    !write(MsgOut, fmt='("                         cell count  : '//&
    !                  '(",3i3,")")'), list%cell_count(1:3)
    !write(MsgOut, fmt='("                         cell length : '//&
    !                  '(",3f8.3,")")'), list%cell_length(1:3)
    !write(MsgOut, fmt='("                         total cell  : ",i8)'), &
    !     total_cell
    !write(MsgOut, fmt='("                         total coord : ",i8)'), &
    !     total_coord


    ! allocate list
    !
    if (associated(list%head)) then
      deallocate(list%head, stat = dealloc_stat)
      if (dealloc_stat /= 0) &
        call error_msg_dealloc
      list%head => null()
    end if

    if (associated(list%lscl)) then
      deallocate(list%lscl, stat = dealloc_stat)
      if (dealloc_stat /= 0) &
        call error_msg_dealloc
      list%lscl => null()
    end if

    allocate(list%head(total_cell),  &
             list%lscl(total_coord), &
             stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc


    ! build
    !

    do i = 1, total_cell
       list%head(i) = EMPTY
    end do

    do i = 1, total_coord
      cix = FLOOR((coords(1, i) - list%origin(1)) / list%cell_length(1))
      ciy = FLOOR((coords(2, i) - list%origin(2)) / list%cell_length(2))
      ciz = FLOOR((coords(3, i) - list%origin(3)) / list%cell_length(3))

      ci = cix * total_cell_yz +      &
           ciy * list%cell_count(3) + &
           ciz + 1

      if (ci < 1 .or. ci > total_cell) &
        call error_msg('build_cell_linked_list> coordinate is out of region.')

      list%lscl(i) = list%head(ci)

      list%head(ci) = i

    end do

    return

  end subroutine build_cell_linked_list

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    dealloc_cell_linked_list
  !> @brief        dealloc cell-linked list
  !! @authors      NT
  !! @param[inout] list       cell linked-list
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_cell_linked_list(list)

    ! formal arguments
    type(s_cell_linked_list), intent(inout) :: list

    ! local variables
    integer                   dealloc_stat


    if (associated(list%head)) then
      deallocate(list%head, stat = dealloc_stat)
      if (dealloc_stat /= 0) &
        call error_msg_dealloc
      list%head => null()
    end if

    if (associated(list%lscl)) then
      deallocate(list%lscl, stat = dealloc_stat)
      if (dealloc_stat /= 0) &
        call error_msg_dealloc
      list%lscl => null()
    end if

    nullify(list%coords)

    return

  end subroutine dealloc_cell_linked_list

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    search_neighbour_coord
  !> @brief        search neighbour coordinates of target coordinate
  !! @authors      NT
  !! @param[in]    list       cell linked-list
  !! @param[in]    distance   distance to judge as a neighbour coord
  !! @param[in]    coord      target coordinate
  !! @param[inout] coord_list coordinate index list
  !! @param[inout] num_list   number of index list
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine search_neighbour_coord(list,       &
                                    distance,   &
                                    coord,      &
                                    coord_list, &
                                    num_list)

    ! formal arguments
    type(s_cell_linked_list), intent(in)    :: list
    double precision,         intent(in)    :: distance
    double precision,         intent(in)    :: coord(3)
    integer,                  intent(inout) :: coord_list(:)
    integer,                  intent(inout) :: num_list

    ! local variables
    integer                   :: i, cx, cy, cz, dcx, dcy, dcz, dc
    integer                   :: ccx, ccy, ccz, ccyz
    double precision          :: dis2, d2
    double precision, pointer :: coords(:,:)


    num_list = 0

    ccx  = list%cell_count(1)
    ccy  = list%cell_count(2)
    ccz  = list%cell_count(3)
    ccyz = ccy * ccz

    cx = FLOOR((coord(1) - list%origin(1)) / list%cell_length(1))
    cy = FLOOR((coord(2) - list%origin(2)) / list%cell_length(2))
    cz = FLOOR((coord(3) - list%origin(3)) / list%cell_length(3))

    dis2 = distance**2

    coords => list%coords

    if (cx < 0 .or. cx >= ccx .or. &
        cy < 0 .or. cy >= ccy .or. &
        cz < 0 .or. cz >= ccz)     &
      call error_msg('search_cell_linked_list> coordinate is out of region.')

    do dcx = cx - 1, cx + 1
      if (dcx < 0 .or. dcx >= ccx) cycle
    do dcy = cy - 1, cy + 1
      if (dcy < 0 .or. dcy >= ccy) cycle
    do dcz = cz - 1, cz + 1
      if (dcz < 0 .or. dcz >= ccz) cycle

       dc = dcx * ccyz + &
            dcy * ccz +  &
            dcz + 1
             
       i = list%head(dc)

       do while(i /= EMPTY)
          d2 = (coord(1) - coords(1,i))**2 + &
               (coord(2) - coords(2,i))**2 + &
               (coord(3) - coords(3,i))**2
          if (d2 > EPS .and. d2 <= dis2) then
            num_list = num_list + 1
            coord_list(num_list) = i
          end if
          i = list%lscl(i)
       end do

    end do
    end do
    end do

    return

  end subroutine search_neighbour_coord

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_cell_by_coords
  !> @brief        compute the most suitable cell parameter 
  !! @authors      NT
  !! @param[in]    coords       array of 3d-coordinate
  !! @param[in]    min_cell_len minimum cell length
  !! @param[out]   origin       simulation box origin
  !! @param[out]   region       simulation box region
  !! @param[out]   cell_count   computed cell count
  !! @param[out]   cell_length  computed cell length
  !! @date         2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_cell_by_coords(coords,       &
                                    min_cell_len, &
                                    origin,       &
                                    region,       &
                                    cell_count,   &
                                    cell_length)

    ! formal arguments
    double precision, intent(in)  :: coords(:,:)
    double precision, intent(in)  :: min_cell_len(3)
    double precision, intent(out) :: origin(3)
    double precision, intent(out) :: region(3)
    integer,          intent(out) :: cell_count(3) 
    double precision, intent(out) :: cell_length(3)

    ! local variables
    integer           :: i, num
    double precision  :: minb(3), maxb(3)


    ! compute boundary
    num = SIZE(coords(1,:))
    if (num <= 0) &
      call error_msg('compute_cell_by_coords')

    minb(1:3) = coords(1:3, 1)
    maxb(1:3) = coords(1:3, 1)
    
    do i = 2, num
      minb(1:3) = MIN(minb(1:3), coords(1:3, i))
      maxb(1:3) = MAX(maxb(1:3), coords(1:3, i))
    end do
    
    minb(1:3) = minb(1:3) - EPS
    maxb(1:3) = maxb(1:3) + EPS

    origin(1:3) = minb(1:3)
    region(1:3) = maxb(1:3) - minb(1:3)

    !write(MsgOut, fmt='(" compute_cell_by_coords> origin : '//&
    !                  '(",3f8.3,")")'), origin(1:3)
    !write(MsgOut, fmt='("                         region : '//&
    !                  '(",3f8.3,")")'), region(1:3)

    call compute_cell_by_region(region,       &
                                min_cell_len, &
                                cell_count,   &
                                cell_length)

    return

  end subroutine compute_cell_by_coords

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_cell_by_region
  !> @brief        compute the most suitable cell parameter 
  !! @authors      NT
  !! @param[in]    region       simulation box region
  !! @param[in]    min_cell_len minimum cell length
  !! @param[out]   cell_count   computed cell count
  !! @param[out]   cell_length  computed cell length
  !! @date         2012/05/24
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_cell_by_region(region,       &
                                    min_cell_len, &
                                    cell_count,   &
                                    cell_length)

    ! formal arguments
    double precision, intent(in)  :: region(3)
    double precision, intent(in)  :: min_cell_len(3)
    integer,          intent(out) :: cell_count(3) 
    double precision, intent(out) :: cell_length(3)

    ! local variables
    double precision  :: r_cell_count(3)


    if (min_cell_len(1) == 0.0 .or. &
        min_cell_len(2) == 0.0 .or. &
        min_cell_len(3) == 0.0) &
      call error_msg('compute_cell_by_region> ERROR : min_cell_len is zero.')

    r_cell_count(1:3) = region(1:3) / min_cell_len(1:3)

    cell_count(1:3) = INT(FLOOR(r_cell_count(1:3)))
    if (cell_count(1) < 1) cell_count(1) = 1
    if (cell_count(2) < 1) cell_count(2) = 1
    if (cell_count(3) < 1) cell_count(3) = 1
    cell_length(1:3) = region(1:3) / DBLE(cell_count(1:3))

    !write(MsgOut, fmt='(" compute_cell_by_region> cell count  : '//&
    !                  '(",3i3,")")'), cell_count(1:3)
    !write(MsgOut, fmt='("                         cell length : '//&
    !                  '(",3f8.3,")")'), cell_length(1:3)

    return

  end subroutine compute_cell_by_region

end module select_contact_mod
