!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   select_lexer_mod
!> @brief   lexical analyzer for select_mod
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM), Yuji Sugita (YS)
!! @date    2012/05/24 
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module select_lexer_mod

  implicit none

  private :: check_space
  private :: check_token

  ! token type
  !     SLTTString : string token
  !     SLTTOther  : other token
  integer,                  parameter :: SLTTString = 1
  integer,                  parameter :: SLTTOther  = 2

  ! string token
  integer,         private, parameter :: NumToken = 7
  character(*),    private, parameter :: Token(NumToken) = &
                                        (/'(',')','&','|','!',':','-'/)

  ! space (delimiter)
  integer,         private, parameter :: NumSpace = 2
  character(1),    private, parameter :: Space(NumSpace) = (/' ',CHAR(9)/)

  ! global variables
  integer,         private            :: g_lex_ptr
  integer,         private            :: g_lex_str_len
  character(1000), private            :: g_lex_str

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  lexer_set
  !> @brief      set string for lexical analysis
  !! @authors    NT
  !! @param[in]  str : string for atoms election
  !! @date       2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine lexer_set(str)

    ! formal arguments
    character(*)    :: str


    g_lex_ptr = 1
    g_lex_str_len = LEN_TRIM(str)
    g_lex_str = str
    
    return

  end subroutine lexer_set

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function       lexer_next_token
  !> @brief         set the next token
  !! @authors       NT
  !! @param[inout]  token_str  : next token
  !! @param[out]    token_type : type of token (string/others)
  !! @date          2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function lexer_next_token(token_str, token_type)

    ! return value
    logical       :: lexer_next_token

    ! formal arguments
    character(*), intent(inout) :: token_str
    integer,      intent(out)   :: token_type

    ! local variables
    integer       :: i, token_len


    lexer_next_token = .false.
    token_str = ''

    if (g_lex_ptr > g_lex_str_len) &
      return

    do i = g_lex_ptr, g_lex_str_len

      if (.not. check_space(i)) &
        exit

    end do

    g_lex_ptr = i

    if (check_token(g_lex_ptr, token_len)) then

      token_str = g_lex_str(g_lex_ptr:g_lex_ptr + token_len - 1)
      token_type = SLTTString
      g_lex_ptr = g_lex_ptr + token_len

    else

      do i = g_lex_ptr, g_lex_str_len
        if (check_space(i) .or. check_token(i, token_len)) &
          exit
      end do

      token_str = g_lex_str(g_lex_ptr:i - 1)
      token_type = SLTTOther
      g_lex_ptr = i

    end if

    lexer_next_token = .true.
    return

  end function lexer_next_token

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine     lexer_back
  !> @brief         set the lexical pointer back
  !! @authors       NT
  !! @param[in]     token_str  : token
  !! @date          2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine lexer_back(token_str)

    ! formal arguments
    character(*), intent(in) :: token_str


    g_lex_ptr = g_lex_ptr - LEN_TRIM(token_str)

    return

  end subroutine lexer_back

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function       check_space
  !> @brief         check space
  !! @authors       NT
  !! @param[in]     iptr  : pointer
  !! @date          2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function check_space(iptr)

    ! return value
    logical    :: check_space

    ! formal arguments
    integer,   intent(in) :: iptr

    ! local variable
    character  :: c
    integer    :: i


    c = g_lex_str(iptr:iptr)
    do i = 1, NumSpace
      if (c == Space(i)) then
        check_space = .true.
        return 
      end if
    end do

    check_space = .false.
    return

  end function check_space

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function       check_token
  !> @brief         check token
  !! @authors       NT
  !! @param[in]     iptr      : pointer
  !! @param[out]    token_len : length of token
  !! @date          2012/05/24 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function check_token(iptr, token_len)

    ! return value
    logical   :: check_token

    ! formal arguments
    integer,  intent(in)  :: iptr
    integer,  intent(out) :: token_len

    ! local variables
    integer   :: i
    

    do i = 1, NumToken
      token_len = LEN_TRIM(Token(i))
      if (g_lex_str(iptr:iptr + token_len - 1) == Token(i)) then
        check_token = .true.
        return
      end if
    end do

    check_token = .false.
    return

  end function check_token

end module select_lexer_mod
